%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Purpose: Main Module for ASN.1 compile time functions
%% Author: Kenneth Lundin

-module(asn1ct_check).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/london/14 $ ').
%-compile(export_all).
-export([check/2,storeindb/1]).
-include("asn1_records.hrl").

check(S,{Types,Values}) ->
    Terror= checkt(S,Types,[]),
    Verror= checkv(S,Values,[]),
    case {Terror,Verror} of
	{[],[]} -> ok;
	L ->{error,L}
    end.

checkt(S,[Name|T],Acc) ->
    %io:format("check_typedef:~p~n",[Name]),
    Result = case asn1_db:dbget(S#state.mname,Name) of
	undefined ->
	    error({type,{internal_error,'???'},S});
	Type when record(Type,typedef) ->
	    NewS = S#state{type=Type,tname=Name},
	    case catch(check_type(NewS,Type,Type#typedef.typespec)) of
		{error,Reason} ->
		    error({type,Reason,NewS});
		{'EXIT',Reason} ->
		    error({type,{internal_error,Reason},NewS});
		Ts ->
		    NewType = Type#typedef{typespec = Ts},
		    asn1_db:dbput(NewS#state.mname,Name,NewType), % update the type
		    ok
	    end
	     end,
    case Result of
	ok ->
	    checkt(S,T,Acc);
	_ ->
	    checkt(S,T,[Result|Acc])
    end;
checkt(S,[],Acc) ->
    lists:reverse(Acc).

checkv(S,[Name|T],Acc) ->
						%io:format("check_valuedef:~p~n",[Name]),
    Result = case asn1_db:dbget(S#state.mname,Name) of
		 undefined -> error({value,{internal_error,'???'},S});
		 Value when record(Value,valuedef) ->
		     NewS = S#state{value=Value},
		     case catch(check_value(NewS,Value)) of
			 {error,Reason} ->
			     error({type,Reason,NewS});
			 {'EXIT',Reason} ->
			     error({type,{internal_error,Reason},NewS});
			 V ->
			     ok % must do more here
		     end
	     end,
    case Result of
	ok ->
	    checkv(S,T,Acc);
	_ ->
	    checkv(S,T,[Result|Acc])
    end;
checkv(Module,[],Acc) ->
    lists:reverse(Acc).

check_value(OldS,V) ->
    #valuedef{name=Name,type=Vtype,value=Value} = V,
    Def = Vtype#type.def,
    Constr = Vtype#type.constraint,
    S = OldS#state{type=Vtype,tname=Def,value=V,vname=Name},
    put(state,S),
    case Def of
	{typereference,Pos,TypeName} ->
	    case asn1_db:dbget(S#state.mname,TypeName) of
		undefined -> io:format("Type ~s is not defined~n",[Name]);
		Type -> check_value(S,V#valuedef{type=Type#typedef.typespec})
	    end;
	'ANY' ->
	    throw({error,{asn1,{'cant check value of type',Def}}});
	'INTEGER' ->
	    validate_integer(S,Value,[],Constr);
	{'INTEGER',NamedNumberList} ->
	    validate_integer(S,Value,NamedNumberList,Constr);
	{'BIT STRING',NamedNumberList} ->
	    validate_bitstring(S,Value,NamedNumberList,Constr);
	'NULL' ->
	    validate_null(S,Value,Constr);
	'OBJECT IDENTIFIER' ->
	    validate_objectidentifier(S,Value,Constr);
	'ObjectDescriptor' ->
	    validate_objectdescriptor(S,Value,Constr);
	{'ENUMERATED',NamedNumberList} ->
	    validate_enumerated(S,Value,NamedNumberList,Constr);
	'BOOLEAN'->
	    validate_boolean(S,Value,Constr);
	'OCTET STRING' ->
	    validate_octetstring(S,Value,Constr);
	'NumericString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'TeletexString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'VideotexString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'UTCTime' ->
	    exit({'cant check value of type' ,Def});
	'GeneralizedTime' ->
	    exit({'cant check value of type' ,Def});
	'GraphicString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'VisibleString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'GeneralString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'PrintableString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'IA5String' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	'BMPString' ->
	    validate_restrictedstring(S,Value,Def,Constr);
	{'SEQUENCE',Components} ->
	    validate_sequence(S,Value,Components,Constr);
	{'SEQUENCE OF',Components} ->
	    validate_sequenceof(S,Value,Components,Constr);
	{'CHOICE',Components} ->
	    validate_choice(S,Value,Components,Constr);
	{'SET',Components} ->
	    validate_set(S,Value,Components,Constr);
	{'SET OF',Components} ->
	    validate_setof(S,Value,Components,Constr);
	Other ->
	    exit({'cant check value of type' ,Other})
    end.

validate_integer(S,{identifier,Pos,Id},NamedNumberList,Constr) ->
    case lists:keysearch(Id,1,NamedNumberList) of
	{value,_} -> ok;
	false -> error({value,"unknown NamedNumber",S})
    end;
validate_integer(S,Value,NamedNumberList,Constr) when integer(Value) ->
    check_integer_range(Value,Constr).

check_integer_range(Int,Constr) when list(Constr) ->
    NewConstr = [X || #constraint{c=X} <- Constr],
    check_constr(Int,NewConstr);

check_integer_range(Int,Constr) ->
						%io:format("~p~n",[Constr]),
    ok.

check_constr(Int,[{'ValueRange',Lb,Ub}|T]) when Int >= Lb, Int =< Ub ->
    check_constr(Int,T);
check_constr(Int,[]) ->
    ok.

validate_bitstring(S,Value,NamedNumberList,Constr) ->
    ok.

validate_null(S,'NULL',Constr) ->
    ok.

validate_objectidentifier(S,Value,Constr) ->
    ok.

validate_objectdescriptor(S,Value,Constr) ->
    ok.

validate_enumerated(S,{identifier,Pos,Id},NamedNumberList,Constr) ->
    case lists:keysearch(Id,1,NamedNumberList) of
	{value,_} -> ok;
	false -> error({value,"unknown ENUMERATED",S})
    end.

validate_boolean(S,Value,Constr) ->
    ok.

validate_octetstring(S,Value,Constr) ->
    ok.

validate_restrictedstring(S,Value,Def,Constr) ->
    ok.

validate_sequence(S,Value,Components,Constr) ->
    ok.

validate_sequenceof(S,Value,Components,Constr) ->
    ok.

validate_choice(S,Value,Components,Constr) ->
    ok.

validate_set(S,Value,Components,Constr) ->
    ok.

validate_setof(S,Value,Components,Constr) ->
    ok.


check_type(S,Type,Ts) when record(Ts,type) ->
    Tag = Ts#type.tag,
    Def = Ts#type.def,
    Constr = Ts#type.constraint,
    NewDef = 
	case Def of 
	    Tref when record(Tref,typereference) ->
		MaybeChoice = get_referenced_type(S,Tref),
		case {Tag,(MaybeChoice#typedef.typespec)#type.def} of
		    {#tag{type='IMPLICIT'},{'CHOICE',_}} ->
			throw({error,{asn1,{implicit_tag_before_choice}}});
		    _ ->
			ok
		end,
		check_typereference(S,Tref);
	    Ext when record(Ext,'Externaltypereference') ->
		MaybeChoice = get_referenced_type(S,Ext),
		case {Tag,(MaybeChoice#typedef.typespec)#type.def} of
		    {#tag{type='IMPLICIT'},{'CHOICE',_}} ->
			throw({error,{asn1,{implicit_tag_before_choice}}});
		    _ ->
			ok
		end,
		check_externaltypereference(S,Ext);
	    'ANY' ->
		exit({'cant check',Def});
	    'INTEGER' ->
		check_integer(S,[],Constr),
		unchanged;
	    {'INTEGER',NamedNumberList} ->
		{'INTEGER',check_integer(S,NamedNumberList,Constr)};
	    {'BIT STRING',NamedNumberList} ->
		{'BIT STRING',check_bitstring(S,NamedNumberList,Constr)};
	    'NULL' ->
		unchanged;
	    'OBJECT IDENTIFIER' ->
		check_objectidentifier(S,Constr),
		unchanged;
	    'ObjectDescriptor' ->
		unchanged;
	    {'ENUMERATED',NamedNumberList} ->
		{'ENUMERATED',check_enumerated(S,NamedNumberList,Constr)};
	    'BOOLEAN'->
		check_boolean(S,Constr),
		unchanged;
	    'OCTET STRING' ->
		check_octetstring(S,Constr),
		unchanged;
	    'NumericString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'TeletexString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'VideotexString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'UTCTime' ->
		unchanged;
	    'GeneralizedTime' ->
		unchanged;
	    'GraphicString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'VisibleString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'GeneralString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'PrintableString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'IA5String' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'BMPString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    'UniversalString' ->
		check_restrictedstring(S,Def,Constr),
		unchanged;
	    {'SEQUENCE',Components} ->
		{'SEQUENCE',check_sequence(S,Type,Components)};
	    {'SEQUENCE OF',Components} ->
		{'SEQUENCE OF',check_sequenceof(S,Type,Components)};
	    {'CHOICE',Components} ->
		case Tag of
		    #tag{type='IMPLICIT'} ->
			throw({error,{asn1,{implicit_tag_before_choice}}});
		    _ -> 
			ok
		end,
		{'CHOICE',check_choice(S,Type,Components)};
	    {'SET',Components} ->
		{'SET',check_set(S,Type,Components)};
	    {'SET OF',Components} ->
		{'SET OF',check_setof(S,Type,Components)};
	    Other ->
		exit({'cant check' ,Other})
	end,
    case NewDef of
	unchanged ->
	    Ts;
	_ ->
	    Ts#type{def=NewDef}
    end.

check_imported(S,Imodule,Name) ->
    case asn1_db:dbget(Imodule,'MODULE') of
	undefined ->
	    io:format("~s.asn1db not found~n",[Imodule]),
	    io:format("Type ~s imported from non existing module ~s~n",[Name,Imodule]);
	Im when record(Im,module) ->
	    case is_exported(Im,Name) of
		false ->
		    io:format("Imported type ~s not exported from module ~s~n",[Name,Imodule]);
		_ ->
		    ok
	    end
    end,
    ok.



is_exported(Module,Name) when record(Module,module) ->
    {exports,Exports} = Module#module.exports,
    case Exports of
	all ->
	    true;
	[] ->
	    false;
	L when list(L) -> 
	    case lists:keysearch(Name,3,Exports) of
		false -> false;
		_ -> true
	    end
    end.
    
    
check_externaltypereference(S,Etref) when record(Etref,'Externaltypereference') ->
    Emod = Etref#'Externaltypereference'.module,
    Etype = Etref#'Externaltypereference'.type,
    Pos = Etref#'Externaltypereference'.pos,
    Currmod = S#state.mname,
    case Emod of
	Currmod ->
	    %% reference to current module
	    check_typereference(S,#typereference{val=Etype,pos=Pos});
	 _ ->
	    io:format("Type ~s IMPORTED FROM ~s~n",[Etype,Emod]),
	    Etref
    end.

check_typereference(S,Tref) when record(Tref,typereference) ->
    Name = Tref#typereference.val,
    case asn1_db:dbget(S#state.mname,Name) of
	undefined ->
	    case imported(S,Name) of
		{ok,Imodule} ->
		    io:format("Type ~s IMPORTED FROM ~s~n",[Name,Imodule]),
		    check_imported(S,Imodule,Name),
		    #'Externaltypereference'{module=Imodule,type=Name};
		_ ->
		    io:format("Type ~s is not defined~n",[Name]),
		    Tref
	    end;
	_ ->
	    Tref    
    end.

get_referenced_type(S,#'Externaltypereference'{module=Emod,type=Etype}) ->
    case asn1_db:dbget(Emod,Etype) of
	undefined ->
	    throw({error,{asn1,{undefined_type,{Emod,Etype}}}});
	T ->
	    Spec = T#typedef.typespec,
	    case Spec#type.def of
		Tref when record(Tref,typereference) ->
		    Def = #'Externaltypereference'{module=Emod,
					     type=Tref#typereference.val,
					     pos=Tref#typereference.pos},
		    
		    
		    T#typedef{typespec=Spec#type{def=Def}};
		_ ->
		    T % should add check that T is exported here
	    end
    end;
get_referenced_type(S,#typereference{val=Name,pos=Pos}) ->
    case asn1_db:dbget(S#state.mname,Name) of
	undefined ->
	    case imported(S,Name) of
		{ok,Imodule} ->
		    case asn1_db:dbget(Imodule,'MODULE') of
			undefined ->
			    throw({error,{asn1,{module_not_found,Imodule}}});
			Im when record(Im,module) ->
			    case is_exported(Im,Name) of
				false ->
				    throw({error,{asn1,{not_exported,{Im,Name}}}});
				_ ->
				    get_referenced_type(S,
							#'Externaltypereference'{module=Imodule,
										 type=Name,pos=Pos})
			    end
		    end;
		_ ->
		    throw({error,{asn1,{undefined_type,Name}}})
	    end;
	T ->
	    T
    end;
get_referenced_type(S,Type) ->
    Type.
     
imported(S,Name) ->
    {imports,Ilist} = (S#state.module)#module.imports,
    imported1(Name,Ilist).

imported1(Name,
	  [{'SymbolsFromModule',Symlist,
	    {typereference,_,ModuleName}}|T]) ->
    case lists:keysearch(Name,3,Symlist) of
	{value,V} ->
	    {ok,ModuleName};
	_ ->
	    imported1(Name,T)
    end;
imported1(Name,[]) ->
    false.
	

check_integer(S,[],C) ->
    ok;
check_integer(S,NamedNumberList,C) ->
    case check_unique(NamedNumberList,2) of
	[] -> 
	    check_int(S,NamedNumberList,[]);
	L when list(L) ->
	    error({type,{duplicates,L},S}),
	    unchanged
		   
    end.

check_int(S,[{'NamedNumber',Id,Num}|T],Acc) when integer(Num) ->
    check_int(S,T,[{Id,Num}|Acc]);
check_int(S,[{'NamedNumber',Id,{identifier,_,Name}}|T],Acc) ->
    Val = dbget_ex(S,S#state.mname,Name),
    check_int(S,[{'NamedNumber',Id,Val#valuedef.value}|T],Acc);
check_int(S,[],Acc) ->
    lists:keysort(2,Acc).


check_bitstring(S,NamedNumberList,Constr) ->
    NamedNumberList.
    


%% Check ENUMERATED
%% ****************************************
%% Check that all values are unique
%% assign values to un-numbered identifiers
%% check that the constraints are allowed and correct
%% put the updated info back into database
check_enumerated(S,NamedNumberList,Constr) ->
    NewList = check_enum(S,NamedNumberList,[],[]),
    NewList.

%% identifiers are put in Acc2
%% returns either [{Name,Number}] or {[{Name,Number}],[{ExtName,ExtNumber}]}
%% the latter is returned if the ENUMERATION contains EXTENSIONMARK
check_enum(S,[{'NamedNumber',Id,Num}|T],Acc1,Acc2) when integer(Num) ->
    check_enum(S,T,[{Id,Num}|Acc1],Acc2);
check_enum(S,[{'NamedNumber',Id,{identifier,_,Name}}|T],Acc1,Acc2) ->
    Val = dbget_ex(S,S#state.mname,Name),
    check_enum(S,[{'NamedNumber',Id,Val#valuedef.value}|T],Acc1,Acc2);
check_enum(S,['EXTENSIONMARK'|T],Acc1,Acc2) ->
    NewAcc2 = lists:keysort(2,Acc1),
    NewList = enum_number(Acc2,NewAcc2,0,[]),
    { NewList, check_enum(S,T,[],[])};
check_enum(S,[Id|T],Acc1,Acc2) when atom(Id) ->
    check_enum(S,T,Acc1,[Id|Acc2]);
check_enum(S,[],Acc1,Acc2) ->
    NewAcc2 = lists:keysort(2,Acc1),
    enum_number(Acc2,NewAcc2,0,[]).


% assign numbers to identifiers , numbers from 0 ... but must not
% be the same as already assigned to NamedNumbers
enum_number([H|T],[{Id,Num}|T2],Cnt,Acc) when Num > Cnt ->
    enum_number(T,[{Id,Num}|T2],Cnt+1,[{H,Cnt}|Acc]);
enum_number([H|T],[{Id,Num}|T2],Cnt,Acc) when Num < Cnt -> % negative Num
    enum_number(T,T2,Cnt+1,[{H,Cnt},{Id,Num}|Acc]);
enum_number([],L2,Cnt,Acc) ->
    lists:concat([lists:reverse(Acc),L2]);
enum_number(L,[{Id,Num}|T2],Cnt,Acc) -> % Num == Cnt
    enum_number(L,T2,Cnt+1,[{Id,Num}|Acc]);
enum_number([H|T],[],Cnt,Acc) ->
    enum_number(T,[],Cnt+1,[{H,Cnt}|Acc]).
    

check_boolean(S,Constr) ->
    ok.

check_octetstring(S,Constr) ->
    ok.

% check all aspects of a SEQUENCE
% - that all component names are unique
% - that all TAGS are ok (when TAG default is applied)
% - that each component is of a valid type
% - that the extension marks are valid
check_sequence(S,Type,Components) when list(Components) ->
    case check_unique([C||C <- Components,record(C,'ComponentType')]
		      ,#'ComponentType'.name) of
	[] ->
    %%    sort_canonical(Components),
    	NewComponents = check_each_component(S,Type,Components,[]),
    	check_unique_sequence_tags(S,NewComponents),
    	NewComponents;
	Dupl ->
		throw({error,{asn1,{duplicate_sequence_components,Dupl}}})
    end.

check_unique_sequence_tags(S,[#'ComponentType'{prop=mandatory}|Rest]) ->
    check_unique_sequence_tags(S,Rest);
check_unique_sequence_tags(S,[C|Rest]) when record(C,'ComponentType') ->
    check_unique_sequence_tags1(S,Rest,[C]);% optional or default
check_unique_sequence_tags(S,[ExtensionMarker|Rest]) ->
    check_unique_sequence_tags(S,Rest);
check_unique_sequence_tags(S,[]) ->
    true.

check_unique_sequence_tags1(S,[C|Rest],Acc) when record(C,'ComponentType') ->
    case C#'ComponentType'.prop of
	mandatory ->
	    check_unique_tags(S,lists:reverse([C|Acc])),
	    check_unique_sequence_tags(S,Rest);
	_  ->
	    check_unique_sequence_tags1(S,Rest,[C|Acc]) % default or optional
    end;
check_unique_sequence_tags1(S,[H|Rest],Acc) ->
    check_unique_sequence_tags1(S,Rest,[H|Acc]);
check_unique_sequence_tags1(S,[],Acc) ->
    check_unique_tags(S,lists:reverse(Acc)).

check_sequenceof(S,Type,Component) when record(Component,type) ->
    check_type(S,Type,Component).

check_set(S,Type,Components) ->
    case check_unique([C||C <- Components,record(C,'ComponentType')],
		      #'ComponentType'.name) of
	[] ->
    %%    sort_canonical(Components),
    	NewComponents = check_each_component(S,Type,Components,[]),
    	check_unique_tags(S,NewComponents),
    	NewComponents;
	Dupl ->
		throw({error,{asn1,{duplicate_set_components,Dupl}}})
    end.

check_setof(S,Type,Component) when record(Component,type) ->
    check_type(S,Type,Component).

check_restrictedstring(S,Def,Constr) ->
    ok.

check_objectidentifier(S,Constr) ->
    ok.

% check all aspects of a CHOICE
% - that all alternative names are unique
% - that all TAGS are ok (when TAG default is applied)
% - that each alternative is of a valid type
% - that the extension marks are valid
check_choice(S,Type,Components) when list(Components) ->
    case check_unique([C||C <- Components,
			  record(C,'ComponentType')],#'ComponentType'.name) of
	[] -> 
    %%    sort_canonical(Components),
	    NewComponents = check_each_alternative(S,Type,Components,[]),
	    check_unique_tags(S,NewComponents),
	    NewComponents;
	Dupl ->
	    throw({error,{asn1,{duplicate_choice_alternatives,Dupl}}})
    end;
check_choice(S,Type,[]) -> 
    [].

check_unique_tags(S,C) ->
    case (S#state.module)#module.tagdefault of
	'AUTOMATIC' ->
	    true;
	_ ->
	    collect_and_sort_tags(C,[])
    end.

collect_and_sort_tags([C|Rest],Acc) when record(C,'ComponentType') ->
    collect_and_sort_tags(Rest,C#'ComponentType'.tags ++ Acc);
collect_and_sort_tags([H|Rest],Acc) ->
    collect_and_sort_tags(Rest,Acc);
collect_and_sort_tags([],Acc) ->
    {Dupl,_}= lists:mapfoldl(fun(El,El)->{{dup,El},El};(El,Prev)-> {El,El} end,notag,lists:sort(Acc)),
    Dupl2 = [Dup|| {dup,Dup} <- Dupl],
    if 
	length(Dupl2) > 0 ->
	    throw({error,{asn1,{duplicates_of_the_tags,Dupl2}}});
	true ->
	    true
    end.

check_unique(L,Pos) ->
    Slist = lists:keysort(Pos,L),
    check_unique2(Slist,Pos,[]).

check_unique2([A,B|T],Pos,Acc) when element(Pos,A) == element(Pos,B) ->
    check_unique2([B|T],Pos,[element(Pos,B)|Acc]);
check_unique2([H|T],Pos,Acc) ->
    check_unique2(T,Pos,Acc);
check_unique2([],Pos,Acc) ->
    lists:reverse(Acc).

check_each_component(S,Type,[C|Ct],Acc) when record(C,'ComponentType') ->
    #'ComponentType'{name=Cname,typespec=Ts,prop=Prop} = C,
    NewTags = get_taglist(S,Ts),
    NewC = C#'ComponentType'{typespec=check_type(S,Type,Ts),tags=NewTags},
    check_each_component(S,Type,Ct,[NewC|Acc]);
check_each_component(S,Type,[C|Ct],Acc) -> % skip 'EXTENSIONMARK'
    check_each_component(S,Type,Ct,[C|Acc]);
check_each_component(S,Type,[],Acc) ->
    lists:reverse(Acc).

check_each_alternative(S,Type,[C|Ct],Acc) when record(C,'ComponentType') ->
    #'ComponentType'{name=Cname,typespec=Ts,prop=Prop} = C,
    NewTags = get_taglist(S,Ts),
    NewC = C#'ComponentType'{typespec=check_type(S,Type,Ts),tags=NewTags},
    check_each_alternative(S,Type,Ct,[NewC|Acc]);
check_each_alternative(S,Type,[C|Ct],Acc) -> % skip 'EXTENSIONMARK'
    check_each_alternative(S,Type,Ct,[C|Acc]);
check_each_alternative(S,Type,[],Acc) ->
    lists:reverse(Acc).

get_taglist(S,Ext) when record(Ext,'Externaltypereference') ->
    T = get_referenced_type(S,Ext),
    get_taglist(S,T#typedef.typespec);
get_taglist(S,Tref) when record(Tref,typereference) ->
    T = get_referenced_type(S,Tref),
    get_taglist(S,T#typedef.typespec);
get_taglist(S,Type) when record(Type,type) ->
    case Type#type.tag of
	undefined ->
	    get_taglist(S,Type#type.def);
	Tag  ->
	    [asn1ct_gen:def_to_tag(Tag)]
    end;
get_taglist(S,{'CHOICE',Components}) ->
    get_taglist1(S,Components);
get_taglist(S,Def) ->
    [asn1ct_gen:def_to_tag(Def)].

get_taglist1(S,[#'ComponentType'{name=Cname,typespec=Ts,tags=TagL}|Rest]) when list(TagL) -> 
    %% tag_list has been here , just return TagL and continue with next alternative
    TagL ++ get_taglist1(S,Rest);
get_taglist1(S,[#'ComponentType'{name=Cname,typespec=Ts,tags=undefined}|Rest]) ->
    get_taglist(S,Ts) ++ get_taglist1(S,Rest);
get_taglist1(S,[H|Rest]) -> % skip EXTENSIONMARK
    get_taglist1(S,Rest);
get_taglist1(S,[]) ->
    [].


dbget_ex(S,Module,Key) ->
    case asn1_db:dbget(Module,Key) of
	undefined ->
	    
	    throw({error,{asn1,{undefined,{Module,Key}}}}); % this is catched on toplevel type or value
	T -> T
    end.


storeindb(M) when record(M,module) ->
    TVlist = M#module.typeorval,
    NewM = M#module{typeorval=findtypes_and_values(TVlist,[],[])},
    asn1_db:dbnew(NewM#module.name),
    asn1_db:dbput(NewM#module.name,'MODULE',NewM),
    storeindb(NewM#module.name,TVlist).

storeindb(Module,[H|T]) when record(H,typedef) ->
    asn1_db:dbput(Module,H#typedef.name,H),
    storeindb(Module,T);
storeindb(Module,[H|T]) when record(H,valuedef) ->
    asn1_db:dbput(Module,H#valuedef.name,H),
    storeindb(Module,T);
storeindb(_,[]) -> true.

findtypes_and_values([H|T],Tacc,Vacc) when record(H,typedef) ->
    findtypes_and_values(T,[H#typedef.name|Tacc],Vacc);
findtypes_and_values([H|T],Tacc,Vacc) when record(H,valuedef) ->
    findtypes_and_values(T,Tacc,[H#valuedef.name|Vacc]);
findtypes_and_values([],Tacc,Vacc) ->
    {lists:reverse(Tacc),lists:reverse(Vacc)}.
    

error({type,Msg,#state{mname=Mname,type=Type,tname=Typename}}) ->
    io:format("asn1error:~p:~p:~p ~p~n",[Type#typedef.pos,Mname,Typename,Msg]),
    {error,type,Type#typedef.pos,Mname,Typename,Msg};
error({value,Msg,#state{mname=Mname,value=Value,vname=Valuename}}) ->
    io:format("asn1error:~p:~p:~p ~p~n",[Value#valuedef.pos,Mname,Valuename,Msg]),
    {error,value,Value#valuedef.pos,Mname,Valuename,Msg}.
	    











