%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : asn1ct_gen_ber.erl
%%% Author  : Kenneth Lundin <kenneth@erix.ericsson.se>
%%% Purpose : Generate erlang module which handles 
%%%           (PER) encode and decode for all types in an ASN.1 module
%%% Created : 25 Feb 1997 by Kenneth Lundin <kenneth@erix.ericsson.se>

-module(asn1ct_gen_ber).
-author('kenneth@erix.ericsson.se').
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-include("asn1_records.hrl").
%%-compile(export_all).
%%-import(asn1_misc,[get_prop/2]).

-export([pgen/4]).
-export([decode_type/1]).
-export([dec_tag_prim/2]).
-export([decode_length/0]).
-export([decode_length_temp/0]).
-export([add_removed_bytes/0]).
-export([gen_encode_prim/4]).
-export([gen_dec_prim/6]).
-export([mk_var/1]).
-export([check_tag/2]).
-import(asn1ct_gen, [emit/1,demit/1]).

						% the encoding of class of tag bits 8 and 7
-define(UNIVERSAL,   0).
-define(APPLICATION, 16#40).
-define(CONTEXT,     16#80).
-define(PRIVATE,     16#C0).

						% primitive or constructed encoding % bit 6
-define(PRIMITIVE,   0).
-define(CONSTRUCTED, 2#00100000).

						% restricted character string types
-define(T_NumericString,    ?UNIVERSAL bor ?PRIMITIVE bor 18). %can be constructed
-define(T_PrintableString,  ?UNIVERSAL bor ?PRIMITIVE bor 19). %can be constructed
-define(T_TeletexString,    ?UNIVERSAL bor ?PRIMITIVE bor 20). %can be constructed
-define(T_VideotexString,   ?UNIVERSAL bor ?PRIMITIVE bor 21). %can be constructed
-define(T_IA5String,        ?UNIVERSAL bor ?PRIMITIVE bor 22). %can be constructed
-define(T_GraphicString,    ?UNIVERSAL bor ?PRIMITIVE bor 25). %can be constructed
-define(T_VisibleString,    ?UNIVERSAL bor ?PRIMITIVE bor 26). %can be constructed
-define(T_GeneralString,    ?UNIVERSAL bor ?PRIMITIVE bor 27). %can be constructed

%% pgen(Erules, Module, TypeOrVal)
%% Generate Erlang module (.erl) and (.hrl) file corresponding to an ASN.1 module
%% .hrl file is only generated if necessary
%% Erules = per | ber
%% Module = atom()
%% TypeOrVal = {TypeList,ValueList}
%% TypeList = ValueList = [atom()]

pgen(OutFile,Erules,Module,TypeOrVal) ->
    pgen_module(OutFile,Erules,Module,TypeOrVal,true).

pgen_module(OutFile,Erules,Module,TypeOrVal,Indent) ->
    put(outfile,OutFile),
    HrlGenerated = asn1ct_gen:pgen_hrl(Erules,Module,TypeOrVal,Indent),
    asn1ct_name:start(),
    ErlFile = lists:concat([OutFile,".erl"]),
    Fid = asn1ct_gen:fopen(ErlFile,write),
    put(gen_file_out,Fid),
    asn1ct_gen:gen_head(Erules,Module,HrlGenerated),
    asn1ct_gen:pgen_exports(Erules,Module,TypeOrVal),
    pgen_typeorval(Erules,Module,TypeOrVal),
    %% gen_vars(asn1_db:mod_to_vars(Module)),
    %% gen_tag_table(AllTypes),
    file:close(Fid),
    io:format("--~p--~n",[{generated,ErlFile}]).

pgen_typeorval(Erules,Module,{Types,Values}) ->
    pgen_types(Erules,Module,Types),
    pgen_values(Erules,Module,Values).

pgen_values(_,_,[]) ->
    true;
pgen_values(Erules,Module,[H|T]) ->
    Valuedef = asn1_db:dbget(Module,H),
    gen_value(Valuedef),
    pgen_values(Erules,Module,T).

pgen_types(_,_,[]) ->
    true;

pgen_types(Erules,Module,[H|T]) ->
    asn1ct_name:clear(),
    Typedef = asn1_db:dbget(Module,H),
    gen_encode(Erules,Module,Typedef),
    asn1ct_name:clear(),
    gen_decode(Erules,Module,Typedef),
    pgen_types(Erules,Module,T).

gen_types(Erules,Module,Tname,[{'EXTENSIONMARK',_,_}|Rest]) ->
    gen_types(Erules,Module,Tname,Rest);
gen_types(Erules,Module,Tname,[ComponentType|Rest]) ->
    asn1ct_name:clear(),
    gen_encode(Erules,Module,Tname,ComponentType),
    asn1ct_name:clear(),
    gen_decode(Erules,Module,Tname,ComponentType),
    gen_types(Erules,Module,Tname,Rest);
gen_types(_,_,Tname,[]) ->
    true;
gen_types(Erules,Module,Tname,Type) when record(Type,type) ->
    asn1ct_name:clear(),
    gen_encode(Erules,Module,Tname,Type),
    asn1ct_name:clear(),
    gen_decode(Erules,Module,Tname,Type).

%% VARIOUS GENERATOR STUFF 
%% *************************************************
%%**************************************************

mk_var(X) when atom(X) ->
    list_to_atom(mk_var(atom_to_list(X)));

mk_var([H|T]) ->
    [H-32|T].

%% Generate value functions ***************
%% ****************************************
%% Generates a function 'V'/0 for each Value V defined in the ASN.1 module
%% the function returns the value in an Erlang representation which can be
%% used as  input to the runtime encode functions

gen_value(Value) when record(Value,valuedef) ->
    io:format(" ~w ",[Value#valuedef.name]),
    emit({"'",Value#valuedef.name,"'() ->",nl}),
    emit({"{'",asn1ct_gen:get_inner((Value#valuedef.type)#type.def),
	  "',",{asis,Value#valuedef.value},"}.",nl,nl}).


%%===============================================================================
%%===============================================================================
%%===============================================================================
%% Generate ENCODING
%%===============================================================================
%%===============================================================================
%%===============================================================================

%%===============================================================================
%% encode #{typedef, {pos, name, typespec}}
%%===============================================================================

gen_encode(Erules,Module,Type) when record(Type,typedef) ->
    put(ctype,Type#typedef.name),
    %%    io:format(" encode-~w ",[Type#typedef.name]),
    %%    gen_encode_tag(...),
    gen_encode_user(Erules,Module,Type).

%%===============================================================================
%% encode #{type, {tag, def, constraint}}
%%===============================================================================

gen_encode(Erules,Module,Tname,Type) when record(Type,type) ->
    Typename = Tname,
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    case asn1ct_gen:type(InnerType) of
	Ext when record(Ext,'Externaltypereference') ->
	    true;
	{_,user} ->
	    true;
	{notype,_} ->
	    true;
	{primitive,bif} ->
	    true;
	{constructed,bif} ->
	    %%	    emit({nl,"'enc_",Typename,"'({'",Typename,
	    %%		  "',Val}) ->",nl}),
	    %%	    emit({"'enc_",Typename,"'(Val);",nl,nl}),
	    emit({nl,nl,nl,"%%================================"}),
	    emit({nl,"%%  ",Typename}),
	    emit({nl,"%%================================",nl}),
	    case InnerType of
		'SET' -> true;
		'SEQUENCE' -> true;
		_ ->
		    emit({nl,"'enc_",Typename,"'({'",Typename,"',Val}) ->",nl}),
		    emit({"   'enc_",Typename,"'(Val);",nl,nl})
	    end,
	    emit({"'enc_",Typename,"'(Val) ->",nl,"   "}),
	    gen_encode_constructed(Erules,Module,Typename,InnerType,Type)
    end;

%%===============================================================================
%% encode ComponentType
%%===============================================================================

gen_encode(Erules,Module,Tname,{'ComponentType',_Pos,Cname,Type,Prop,Tags}) ->
    NewTname = list_to_atom(lists:concat([Tname,"_",Cname])),
    gen_encode(Erules,Module,NewTname,Type).

gen_encode_user(Erules,Module,D) when record(D,typedef) ->
    Typename = D#typedef.name,
    Type = D#typedef.typespec,
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    emit({nl,nl,nl,"%%================================"}),
    emit({nl,"%%  ",Typename}),
    emit({nl,"%%================================",nl}),
    case InnerType of
	'SET' -> true;
	'SEQUENCE' -> true;
	_ ->
	    emit({nl,"'enc_",Typename,"'({'",Typename,"',Val}) ->",nl}),
	    emit({"   'enc_",Typename,"'(Val);",nl,nl})
    end,
    emit({"'enc_",Typename,"'(Val) ->",nl}),
    case asn1ct_gen:type(InnerType) of
	{primitive,bif} ->
	    gen_encode_prim(Erules,Type,tag),
	    emit({".",nl});
	{constructed,bif} ->
	    gen_encode_constructed(Erules,Module,Typename,InnerType,D);
	{_,user} ->
	    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
	    case TopEnc_type of
		universal ->
		    emit({"   'enc_",InnerType,"'(Val).",nl,nl});
		implicit ->
		    emit({"   ",{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'enc_",InnerType,"'(Val),",nl}),
		    emit({"   asn1rt_ber:change_tag({",
			  {asis,Top_class},",",
			  {asis,Top_tagNr},"}, ",
			  {var,asn1ct_name:curr(encTmp)},")."}),
		    asn1ct_name:new(encTmp);
		explicit ->
		    emit({"   ",{var,asn1ct_name:curr(tag)}," = "}),
		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}), ",nl}),
		    emit({"   ",{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'enc_",InnerType,"'(Val),",nl}),
		    emit({"   ",{var,asn1ct_name:curr(enc)}," = "}),
		    emit({"[asn1rt_ber:encode_length("}),
		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
		    emit({{var,asn1ct_name:curr(encTmp)},com,nl}),
		    emit({"   ",{var,asn1ct_name:curr(tag)}," ++ ",
			  {var,asn1ct_name:curr(enc)},".",nl,nl})
	    end;

	{notype,_} ->
	    emit({"   'enc_",InnerType,"'(Val).",nl,nl});

	#'Externaltypereference'{module=Emod,type=Etype} ->
	    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(Type#type.tag),
	    case TopEnc_type of
		universal ->
		    emit({"   '",Emod,"':'enc_",Etype,"'(Val).",nl});
		implicit ->
		    emit({"   ",{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'",Emod,"':'enc_",Etype,"'(Val),",nl}),
		    emit({"   asn1rt_ber:change_tag({",
			  {asis,Top_class},",",
			  {asis,Top_tagNr},"}, ",
			  {var,asn1ct_name:curr(encTmp)},")."}),
		    asn1ct_name:new(encTmp);
		explicit ->
		    emit({"   ",{var,asn1ct_name:curr(tag)}," = "}),
		    emit({"asn1rt_ber:encode_tag_val({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}), ",nl}),
		    emit({"   ",{var,asn1ct_name:curr(encTmp)}," = "}),
		    emit({"'",Emod,"':'enc_",Etype,"'(Val),",nl}),
		    emit({"   ",{var,asn1ct_name:curr(enc)}," = "}),
		    emit({"[asn1rt_ber:encode_length("}),
		    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
		    emit({{var,asn1ct_name:curr(encTmp)},com,nl}),
		    emit({"   ",{var,asn1ct_name:curr(tag)}," ++ ",
			  {var,asn1ct_name:curr(enc)},"."})
	    end
    end.

gen_encode_constructed(Erules,Module,Typename,InnerType,D) when record(D,type) ->
    
    %%    Rtmod = list_to_atom(lists:concat(["asn1ct_constructed_",Erules])),
    case InnerType of
	'SEQUENCE' ->
	    asn1ct_constructed_ber:gen_encode_sequence(Typename,D),
	    {_,Components} = D#type.def,
	    gen_types(Erules,Module,Typename,Components);
	'CHOICE' ->
	    asn1ct_constructed_ber:gen_encode_choice(Typename,D),
	    {_,Components} = D#type.def,
	    gen_types(Erules,Module,Typename,Components);
	'SEQUENCE OF' ->
	    asn1ct_constructed_ber:gen_encode_sof(Typename,InnerType,D),
	    {_,Type} = D#type.def,
	    gen_types(Erules,Module,list_to_atom(lists:concat([Typename,"_SEQOF"])),Type);
	'SET' ->
	    asn1ct_constructed_ber:gen_encode_set(Typename,D),
	    {_,Components} = D#type.def,
	    gen_types(Erules,Module,Typename,Components);
	'SET OF' ->
	    asn1ct_constructed_ber:gen_encode_sof(Typename,InnerType,D),
	    {_,Type} = D#type.def,
	    gen_types(Erules,Module,list_to_atom(lists:concat([Typename,"_SETOF"])),Type);
	Other ->
	    exit({nyi,InnerType})
    end;

gen_encode_constructed(Erules,Module,Typename,InnerType,D) when record(D,typedef) ->
    gen_encode_constructed(Erules,Module,Typename,InnerType,D#typedef.typespec).

gen_encode_prim(Erules,D,DoTag) ->
    Value = case asn1ct_name:active(val) of
		true ->
		    mk_var(asn1ct_name:curr(val));
		false ->
		    "Val"
	    end,
    
    {Enc_type, Class, Number} = 
	check_tag(D#type.tag, asn1ct_gen:get_inner(D#type.def)),
    case Enc_type of
	universal ->
	    emit({"   "}),
	    gen_encode_prim(Erules,D,tag,Value);
	implicit ->
	    emit({"   ",{var,asn1ct_name:curr(tag)}," = "}),
	    emit({"asn1rt_ber:encode_tag_val({",
		  {asis,Class},",",
		  {asis,?PRIMITIVE},",",
		  {asis,Number},"}), ",nl}),
	    emit({"   ",{var,asn1ct_name:curr(enc)}," = "}),
	    gen_encode_prim(Erules,D,notag,Value),
	    emit({com,nl,"   "}),
	    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}});
	explicit ->
	    emit({"   ",{var,asn1ct_name:curr(tag)}," = "}),
	    emit({"asn1rt_ber:encode_tag_val({",
		  {asis,Class},",",
		  {asis,?CONSTRUCTED},",",
		  {asis,Number},"}), ",nl}),
	    emit({"   ",{var,asn1ct_name:curr(encTmp)}," = "}),
	    gen_encode_prim(Erules,D,tag,Value),
	    emit({com,nl,"   ",{var,asn1ct_name:curr(enc)}," = "}),
	    emit({"[asn1rt_ber:encode_length("}),
	    emit({"length(lists:flatten(",{var,asn1ct_name:curr(encTmp)},")))] ++ "}),
	    emit({var,asn1ct_name:curr(encTmp)}),
	    emit({com,nl,"   "}),
	    emit({{var,asn1ct_name:curr(tag)}," ++ ",{var,asn1ct_name:curr(enc)}})
    end.
						%    gen_encode_prim(Erules,D,DoTag,Value).

gen_encode_prim(Erules,D,DoTag,Value) when record(D,type) ->
    
    Constraint = D#type.constraint,
    
    case D#type.def of
	'BOOLEAN' ->
	    emit({"asn1rt_ber:encode_boolean(",Value,",",DoTag,")"});
	'INTEGER' ->
	    emit({"asn1rt_ber:encode_integer(", 
		  {asis,Constraint},",",Value,",",DoTag,")"});
	{'INTEGER',NamedNumberList} ->
						%	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:encode_integer(",
		  {asis,Constraint},",",Value,",",
		  {asis,NamedNumberList},",",DoTag,")"});
	{'ENUMERATED',NamedNumberList} ->
	    
						%	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:encode_enumerated(",
		  {asis,Constraint},",",Value,",",
		  {asis,NamedNumberList},",",DoTag,")"});
						%		  {asis,NewList},",",DoTag,")"});
	{'BIT STRING',NamedNumberList} ->
	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:encode_bit_string(",{asis,Constraint},",",
		  Value,",",
		  {asis,NewList},",",
		  DoTag,")"});
	'ANY' ->
	    exit({'can not encode' ,'ANY'});
	'NULL' ->
	    emit({"asn1rt_ber:encode_null(",Value,",",DoTag,")"});
	'OBJECT IDENTIFIER' ->
	    emit({"asn1rt_ber:encode_object_identifier(",Value,",",DoTag,")"});
	'ObjectDescriptor' ->
	    emit({"asn1_encode:e_object_descriptor(",Value,",",false,")"});
	'OCTET STRING' ->
	    emit({"asn1rt_ber:encode_octet_string(",
		  {asis,Constraint},",",Value,",",DoTag,")"});
	'NumericString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_NumericString},",",DoTag,")"});
	'TeletexString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_TeletexString},",",DoTag,")"});
	'VideotexString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_VideotexString},",",DoTag,")"});
	'GraphicString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_GraphicString},",",DoTag,")"});
	'VisibleString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_VisibleString},",",DoTag,")"});
	'GeneralString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_GeneralString},",",DoTag,")"});
	'PrintableString' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_PrintableString},",",DoTag,")"});
	'IA5String' ->
	    emit({"asn1rt_ber:encode_restricted_string(",
		  {asis,Constraint},",",Value,",",{asis,?T_IA5String},",",DoTag,")"});
	'UniversalString' ->
	    emit({"asn1rt_ber:encode_universal_string(",
		  {asis,Constraint},",",Value,",",DoTag,")"});
	'BMPString' ->
	    emit({"asn1rt_ber:encode_BMP_string(",{asis,Constraint},",",Value,",",DoTag,")"});
	'UTCTime' ->
	    emit({"asn1rt_ber:encode_utc_time(",{asis,Constraint},",",Value,",",DoTag,")"});
	'GeneralizedTime' ->
	    emit({"asn1rt_ber:encode_generalized_time(",
		  {asis,Constraint},",",Value,",",DoTag,")"});
	XX ->
	    exit({'can not encode' ,XX})
    end.




%%===============================================================================
%%===============================================================================
%%===============================================================================
%% Generate DECODING
%%===============================================================================
%%===============================================================================
%%===============================================================================

%%===============================================================================
%% decode #{typedef, {pos, name, typespec}}
%%===============================================================================

gen_decode(Erules,Module,Type) when record(Type,typedef) ->
    D = Type,
    put(ctype,Type#typedef.name),
    %%    io:format(" decode-~w ",[Type#typedef.name]),
    emit({nl,nl}),
    emit({"'dec_",Type#typedef.name,"'(Bytes,Telltype) ->",nl}),
    emit({"   'dec_",Type#typedef.name,"'(Bytes,Telltype,{no_impl_class,no_tag_no}).",nl,nl}),
    emit({"'dec_",Type#typedef.name,"'(Bytes,Telltype,{ImplTagClass,ImplTagNo}) ->",nl}),
    dbdec(Type#typedef.name),
    gen_decode_user(Erules,Module,D).


%%===============================================================================
%% decode #{type, {tag, def, constraint}}
%%===============================================================================

gen_decode(Erules,Module,Tname,Type) when record(Type,type) ->
    Typename = Tname,
    InnerType = asn1ct_gen:get_inner(Type#type.def),
    case asn1ct_gen:type(InnerType) of
	Ext when record(Ext,'Externaltypereference') ->
	    true;
	{_,user} ->
	    true;
	{notype,_} ->
	    true;
	{primitive,bif} ->
	    true;
	{constructed,bif} ->
	    emit({nl,"'dec_",Typename,"'(Bytes,Telltype) ->",nl}),
	    emit({"   'dec_",Typename,"'(Bytes,Telltype,{no_impl_class,no_tag_no}).",nl,nl}),
	    emit({"'dec_",Typename,"'(Bytes,Telltype,{ImplTagClass,ImplTagNo}) ->",nl}),
	    dbdec(Typename),
	    gen_decode_constructed(Erules,Module,Typename,InnerType,Type)
    end;


%%===============================================================================
%% decode ComponentType
%%===============================================================================

gen_decode(Erules,Module,Tname,{'ComponentType',_Pos,Cname,Type,Prop,Tags}) ->
    NewTname = list_to_atom(lists:concat([Tname,"_",Cname])),
    gen_decode(Erules,Module,NewTname,Type).


gen_decode_user(Erules,Module,D) when record(D,typedef) ->
    Typename = D#typedef.name,
    Def = D#typedef.typespec,
    InnerType = asn1ct_gen:get_inner(Def#type.def),
    InnerTag = Def#type.tag,
    case asn1ct_gen:type(InnerType) of
	{primitive,bif} ->
	    Enc_type = dec_tag_prim(InnerTag, InnerType),
	    case Enc_type of
		universal ->
		    decode_length(),
		    emit({"   {Dec, Rest, ",{var,asn1ct_name:next(rb)},"} = "}),
		    asn1ct_name:new(rb),
		    BytesVar = mk_var(asn1ct_name:curr(bytes)),
		    asn1ct_name:delete(len),
		    LenVar = mk_var(asn1ct_name:curr(len)),
		    asn1ct_name:new(len),
		    gen_dec_prim(ber,Def,BytesVar,notag,LenVar,?PRIMITIVE),
		    asn1ct_name:new(rb),
		    emit({",",nl,"   {Dec, Rest, "}),
		    add_removed_bytes(),
		    emit("}");
		implicit ->
		    decode_length(),
		    emit({"   {Dec, Rest, ",{var,asn1ct_name:next(rb)},"} = "}),
		    asn1ct_name:new(rb),
		    BytesVar = mk_var(asn1ct_name:curr(bytes)),
		    asn1ct_name:delete(len),
		    LenVar = mk_var(asn1ct_name:curr(len)),
		    asn1ct_name:new(len),
		    gen_dec_prim(ber,Def,BytesVar,notag,LenVar,?PRIMITIVE),
		    asn1ct_name:new(rb),
		    emit({",",nl,"   {Dec, Rest, "}),
		    add_removed_bytes(),
		    emit("}");
		explicit ->
		    decode_length(),
		    emit({"   {Dec, Rest, ",{var,asn1ct_name:next(rb)},"} = "}),
		    asn1ct_name:new(rb),
		    BytesVar = mk_var(asn1ct_name:curr(bytes)),
		    asn1ct_name:delete(len),
		    LenVar = mk_var(asn1ct_name:curr(len)),
		    asn1ct_name:new(len),
		    gen_dec_prim(ber,Def,BytesVar,tag,LenVar,?CONSTRUCTED),
		    asn1ct_name:new(rb),
		    emit({",",nl,"   {Dec, Rest, "}),
		    add_removed_bytes(),
		    emit("}")
	    end,
	    emit({".",nl,nl});

	{constructed,bif} ->
	    gen_decode_constructed(Erules,Module,Typename,InnerType,D);

	{Form,user} ->
	    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(InnerTag),
	    case TopEnc_type of
		universal ->
		    emit({"   'dec_",InnerType,"'(Bytes,Telltype).",nl,nl});
		implicit ->
		    emit({"   'dec_",InnerType,"'(Bytes,Telltype,{",
			  Top_class,",",Top_tagNr,"}).",nl,nl});
		explicit ->
		    emit({"   {Bytes1, Rb1} = asn1rt_ber:check_tag({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}, {ImplTagClass, ImplTagNo}, Bytes), ",nl}),
    
		    emit({"   {{Len, Bytes2},Rb2} = asn1rt_ber:decode_length(Bytes1),",nl}),
		    emit({"   {Dec, Rest, Rb3} = 'dec_",InnerType,"'(Bytes2,Telltype),",nl}),
		    emit({"   {Dec, Rest, Rb1+Rb2+Rb3}.",nl,nl})
	    end;

	#'Externaltypereference'{module=Emod,type=Etype} ->
	    {TopEnc_type, Top_class, Top_tagNr} = get_top_enc_type(InnerTag),
	    case TopEnc_type of
		universal ->
		    emit({"   '",Emod,"':'dec_",Etype,"'(Bytes,Telltype).",nl,nl});
		implicit ->
		    emit({"   '",Emod,"':'dec_",Etype,"'(Bytes,Telltype,{",
			  Top_class,",",Top_tagNr,"}).",nl,nl});
		explicit ->
		    emit({"   {Bytes1, Rb1} = asn1rt_ber:check_tag({",
			  {asis,Top_class},",",
			  {asis,?CONSTRUCTED},",",
			  {asis,Top_tagNr},"}, {ImplTagClass, ImplTagNo}, Bytes), ",nl}),
    
		    emit({"   {{Len, Bytes2},Rb2} = asn1rt_ber:decode_length(Bytes1),",nl}),
		    emit({"   {Dec, Rest, Rb3} = '",Emod,"':'dec_",Etype,"'(Bytes2,Telltype),",nl}),
		    emit({"   {Dec, Rest, Rb1+Rb2+Rb3}.",nl,nl})
	    end;
	Other ->
	    exit({error,34665})
    end.





gen_decode_constructed(Erules,Module,Typename,InnerType,D) when record(D,type) ->
    Rtmod = list_to_atom(lists:concat(["asn1ct_constructed_",Erules])),
    case InnerType of
	'SEQUENCE' ->
	    Rtmod:gen_decode_sequence(Module,Typename,D);
	'CHOICE' ->
	    Rtmod:gen_decode_choice(Module,Typename,D);
	'SEQUENCE OF' ->
	    Rtmod:gen_decode_sof(Typename,InnerType,D);	
 	'SET' ->
	    Rtmod:gen_decode_set(Module,Typename,D);
	'SET OF' ->
	    Rtmod:gen_decode_sof(Typename,InnerType,D);
	Other ->
	    exit({nyi,InnerType})
    end;


gen_decode_constructed(Erules,Module,Typename,InnerType,D) when record(D,typedef) ->
    gen_decode_constructed(Erules,Module,Typename,InnerType,D#typedef.typespec).


gen_dec_prim(Erules,Att,BytesVar,DoTag,Length,Form) ->
    Typename = Att#type.def,
    Constraint = Att#type.constraint,
    
    case Typename of
	'BOOLEAN'->
	    emit({"asn1rt_ber:decode_boolean(",BytesVar,","});
	'INTEGER' ->
	    emit({"asn1rt_ber:decode_integer(",BytesVar,",",
		  {asis,Constraint},","});
	{'INTEGER',NamedNumberList} ->
	    emit({"asn1rt_ber:decode_integer(",BytesVar,",",
		  {asis,Constraint},",",
		  {asis,NamedNumberList},","});
	{'ENUMERATED',{NamedNumberList,_}} -> % Extensionmark (currently ignored for BER)
	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:decode_enumerated(",BytesVar,",",
		  {asis,Constraint},",",
		  {asis,NamedNumberList},","});
	{'ENUMERATED',NamedNumberList} ->
	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:decode_enumerated(",BytesVar,",",
		  {asis,Constraint},",",
		  {asis,NamedNumberList},","});
	{'BIT STRING',NamedNumberList} ->
	    NewList = [{X,Y}||{_,X,Y} <- NamedNumberList],
	    emit({"asn1rt_ber:decode_bit_string(",BytesVar,",",
		  {asis,Constraint},",",
		  {asis,NewList},","});
	'NULL' ->
	    emit({"asn1rt_ber:decode_null(",BytesVar,","});
	'OBJECT IDENTIFIER' ->
	    emit({"asn1rt_ber:decode_object_identifier(",BytesVar,","});
	'ObjectDescriptor' ->
	    emit({"asn1rt_ber:decode_object_descriptor(",
		  BytesVar,")"});
	'OCTET STRING' ->
	    emit({"asn1rt_ber:decode_octet_string(",BytesVar,",",{asis,Constraint},","});
	'NumericString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_NumericString},","});
	'TeletexString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_TeletexString},","});
	'VideotexString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_VideotexString},","});
	'GraphicString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_GraphicString},","});
	'VisibleString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_VisibleString},","});
	'GeneralString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_GeneralString},","});
	'PrintableString' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_PrintableString},","});
	'IA5String' ->
	    emit({"asn1rt_ber:decode_restricted_string(",
		  BytesVar,",",{asis,Constraint},",",{asis,?T_IA5String},","});
	'UniversalString' ->
	    emit({"asn1rt_ber:decode_universal_string(",
		  BytesVar,",",{asis,Constraint},","});
	'BMPString' ->
	    emit({"asn1rt_ber:decode_BMP_string(",
		  BytesVar,",",{asis,Constraint},","});
	'UTCTime' ->
	    emit({"asn1rt_ber:decode_utc_time(",
		  BytesVar,",",{asis,Constraint},","});
	'GeneralizedTime' ->
	    emit({"asn1rt_ber:decode_generalized_time(",
		  BytesVar,",",{asis,Constraint},","});
	'ANY' ->
	    exit({'can not decode' ,'ANY'});
	Other ->
	    exit({'cant decode' ,Other})
    end,
    
    
    case DoTag of
	tag ->
	    emit({" ",DoTag,", ",Length,")"});
	notag ->
	    emit({" {notag, ",{var,asn1ct_name:curr(form)},"}, ",Length, ")" }),
	    asn1ct_name:new(form)
    end.


dbdec(Type) ->
    demit({"io:format(\"decoding: ",{asis,Type},"~w~n\",[Bytes]),",nl}).





decode_type('BOOLEAN') -> 1;
decode_type('INTEGER') -> 2;
decode_type('BIT STRING') -> 3; 
decode_type('OCTET STRING') -> 4; 
decode_type('NULL') -> 5;
decode_type('OBJECT IDENTIFIER') -> 6;
decode_type('OBJECT DESCRIPTOR') -> 7;
decode_type('EXTERNAL') -> 8;
decode_type('REAL') -> 9;
decode_type('ENUMERATED') -> 10;
decode_type('EMBEDDED_PDV') -> 11;
decode_type('SEQUENCE') -> 16;
decode_type('SEQUENCE OF') -> 16;
decode_type('SET') -> 17;
decode_type('SET OF') -> 17;
decode_type('NumericString') -> 18;  
decode_type('PrintableString') -> 19;  
decode_type('TeletexString') -> 20;  
decode_type('VideotexString') -> 21;  
decode_type('IA5String') -> 22;  
decode_type('UTCTime') -> 23;  
decode_type('GeneralizedTime') -> 24;  
decode_type('GraphicString') -> 25;  
decode_type('VisibleString') -> 26;  
decode_type('GeneralString') -> 27;  
decode_type('UniversalString') -> 28;  
decode_type('BMPString') -> 30;  
decode_type(Else) -> exit({error,{asn1,{unrecognized_type,Else}}}).


decode_class('UNIVERSAL') ->
    ?UNIVERSAL;
decode_class('APPLICATION') ->
    ?APPLICATION;
decode_class('CONTEXT') ->
    ?CONTEXT;
decode_class('PRIVATE') ->
    ?PRIVATE.
 


get_top_enc_type(Tag) ->
    case Tag of
	undefined ->
	    {universal,?UNIVERSAL,undefined};
	{tag, 'UNIVERSAL', TagNr, 'IMPLICIT'} ->
	    {universal,?UNIVERSAL,'IMPLICIT'};
	{tag, 'UNIVERSAL', TagNr, 'EXPLICIT'} ->
	    {universal,?UNIVERSAL,'EXPLICIT'};
	{tag, 'APPLICATION', TagNr, 'IMPLICIT'} ->
	    {implicit,?APPLICATION,TagNr};
	{tag, 'APPLICATION', TagNr, 'EXPLICIT'} ->
	    {explicit,?APPLICATION,TagNr};
	{tag, 'PRIVATE', TagNr, 'IMPLICIT'} ->
	    {implicit,?PRIVATE,TagNr};
	{tag, 'PRIVATE', TagNr, 'EXPLICIT'} ->
	    {explicit,?PRIVATE,TagNr};
	{tag, 'CONTEXT', TagNr, 'IMPLICIT'} ->
	    {implicit,?CONTEXT,TagNr};
	{tag, 'CONTEXT', TagNr, 'EXPLICIT'} ->
	    {explicit,?CONTEXT,TagNr};
	UnknownTag ->
	    exit({error,{asn1,{not_recognized_tag,UnknownTag}}})
    end.




check_tag(TagIn,TypeIn) ->
    Type = case TypeIn of
	       X when atom(X) ->
		   decode_type(X);
	       {X,List} when atom(X) ->
		   decode_type(X);
	       X ->
		   exit({error,{asn1,{unrecognized_type,X}}})
	   end,
    
    {Tag, Class, Number} = case TagIn of
			       undefined ->
				   {universal,?UNIVERSAL,Type};
			       
			       {tag,'UNIVERSAL',NumberIn,_TagType} ->
				   {universal,?UNIVERSAL,Type};
			       
			       {tag,'APPLICATION',NumberIn,'IMPLICIT'} ->
				   {implicit,?APPLICATION,NumberIn};
			       
			       {tag,'APPLICATION',NumberIn,'EXPLICIT'} ->
				   {explicit,?APPLICATION,NumberIn};
			       
			       {tag,'PRIVATE',NumberIn,'IMPLICIT'} ->
				   {implicit,?PRIVATE,NumberIn};
			       
			       {tag,'PRIVATE',NumberIn,'EXPLICIT'} ->
				   {explicit,?PRIVATE,NumberIn};
			       
			       {tag,'CONTEXT',NumberIn,'IMPLICIT'} ->
				   {implicit,?CONTEXT,NumberIn};
			       
			       {tag,'CONTEXT',NumberIn,'EXPLICIT'} ->
				   {explicit,?CONTEXT,NumberIn}
			   end.


dec_tag_prim(TagIn,TypeIn) ->
    {EncTag, Class, Number} = check_tag(TagIn, TypeIn),
    emit("   "),
    case EncTag of
	universal ->
	    emit({"{",{var,asn1ct_name:curr(tag)},",",
		  {var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rb)}}),
	    emit({"} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"), ",nl}),
	    asn1ct_name:new(rb),
	    asn1ct_name:new(bytes),
	    emit({"   ",{var,asn1ct_name:next(form)}," = ",nl}),
	    emit({"      case ",{var,asn1ct_name:curr(tag)}," of",nl}),
	    emit({"         {",?UNIVERSAL,", ",{var,asn1ct_name:curr(form)},", ",
		  decode_type(asn1ct_gen:get_inner(TypeIn)),"} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         {ImplTagClass, ",{var,asn1ct_name:curr(form)},
		  ", ImplTagNo} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         ",{var,asn1ct_name:next(else)},
		  " -> exit({error,{asn1,{invalid_tag,",{var,asn1ct_name:next(else)},"}}})",nl}),
	    emit({"      end,",nl}),
	    asn1ct_name:new(tag),
	    asn1ct_name:new(else),
	    asn1ct_name:new(form);
	implicit ->
	    emit({"{",{var,asn1ct_name:curr(tag)},",",
		  {var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rb)}}),
	    emit({"} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"), ",nl}),
	    asn1ct_name:new(rb),
	    asn1ct_name:new(bytes),
	    emit({"   ",{var,asn1ct_name:next(form)}," = ",nl}),
	    emit({"      case ",{var,asn1ct_name:curr(tag)}," of",nl}),
	    emit({"         {",{asis,Class},", ",{var,asn1ct_name:curr(form)},
		  ", ",{asis,Number},"} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         {ImplTagClass, ",{var,asn1ct_name:curr(form)},
		  ", ImplTagNo} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         ",{var,asn1ct_name:next(else)},
		  " -> exit({error,{asn1,{invalid_tag,",{var,asn1ct_name:next(else)},"}}})",nl}),
	    emit({"      end,",nl}),
	    asn1ct_name:new(tag),
	    asn1ct_name:new(else),
	    asn1ct_name:new(form);
	
	explicit ->
	    emit({"{",{var,asn1ct_name:next(tag)},",",
		  {var,asn1ct_name:next(bytes)},",",
		  {var,asn1ct_name:next(rb)}}),
	    emit({"} = asn1rt_ber:decode_tag(",{var,asn1ct_name:curr(bytes)},"), ",nl}),
	    asn1ct_name:new(rb),
	    asn1ct_name:new(bytes),
	    emit({"   ",{var,asn1ct_name:next(form)}," = ",nl}),
	    emit({"      case ",{var,asn1ct_name:next(tag)}," of",nl}),
	    emit({"         {",{asis,Class},", ",{var,asn1ct_name:curr(form)},
		  ", ",{asis,Number},"} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         {ImplTagClass, ",{var,asn1ct_name:curr(form)},
		  ", ImplTagNo} -> ",{var,asn1ct_name:curr(form)},";",nl}),
	    emit({"         ",{var,asn1ct_name:next(else)},
		  " -> exit({error,{asn1,{invalid_tag,",{var,asn1ct_name:next(else)},"}}})",nl}),
	    emit({"      end,",nl}),
	    asn1ct_name:new(tag),
	    asn1ct_name:new(else),
	    asn1ct_name:new(form)
    end,
    EncTag.





decode_length() ->
    emit({"   {{",{var,asn1ct_name:curr(len)},",",
	  {var,asn1ct_name:next(bytes)},"},",
	  {var,asn1ct_name:next(rb)},"} = asn1rt_ber:decode_length(",
	  {var,asn1ct_name:curr(bytes)},"),",nl}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(len),
    asn1ct_name:new(rb).


decode_length_temp() ->
    emit({"   {{",{var,asn1ct_name:curr(len)},",",
	  {var,asn1ct_name:next(bytes)},"},",
	  {var,asn1ct_name:next(rbTmp)},"} = asn1rt_ber:decode_length(",
	  {var,asn1ct_name:curr(bytes)},"),",nl}),
    asn1ct_name:new(bytes),
    asn1ct_name:new(len),
    asn1ct_name:new(rbTmp).


remove_length() ->
    emit({"   ",{var,asn1ct_name:next(bytes)},
	  " = asn1rt_ber:remove_length(",
	  {var,asn1ct_name:curr(bytes)},"), ",nl}).

add_removed_bytes() ->
    asn1ct_name:delete(rb),
    add_removed_bytes(asn1ct_name:all(rb)).

add_removed_bytes([H,T1|T]) ->
    emit({{var,H},"+"}),
    add_removed_bytes([T1|T]);
add_removed_bytes([H|T]) ->
    emit({{var,H}}),
    add_removed_bytes(T);
add_removed_bytes([]) ->
    true.



