%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994,1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created: 94-10-27 by Tobbe (tobbe@erix.ericsson.se)
%% 
%% Function: Interface Generator
%%-----------------------------------------------------------------------
-module(ig).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: ig.erl,v 1.12 1996/04/10 13:43:18 tobbe Exp $.').

-export([compile/3,all/1,gen/1,gen/2,init/3,get_token/1,get_option/2,adr_of/2,sizeof/2,
	 nice_print/1,get_options/1,stdio/1,stdin/2,stdout/2,leex/1,yecc/1]).

-import(lists,[append/2]).

-include("ig.hrl").
-include("erl_compile.hrl").

%% -CSASL-
%%
%% For documentation about CSASL: see iggenerate.erl
%%

gen(Fname) -> 
    gen(Fname,[]).

gen(Fname,Options) -> 
    FileName = add_dot_h(Fname),
    Pid = spawn_link(ig,init,[self(),FileName,Options]),
    receive {Pid,Res} -> Res end.

%% This should be used when calling IG from the erl script,
%% preferably at install time.

all([H|T]) -> 
    Arg = str2term(H),
    io:format("~n<--- apply(ig,gen,~p) --->~n",[Arg]),
    catch apply(ig,gen,Arg),
    all(T);
all([]) -> ok.

%% compile(AbsFileName, Outfile, Options)
%%   Compile entry point for erl_compile.

compile(File, _OutFile, Options) ->
    case gen(File, Options#options.specific) of
	ok -> ok;
	Other -> Other
    end.

str2term(X) -> catch str2term_1(X).

str2term_1(Atom) when atom(Atom) -> 
    str2term_1(atom_to_list(Atom));
str2term_1(String) -> 
    element(2,erl_parse:parse_term(element(2,io_lib:scan(append(String,". "))))).

%% Add the '.h' suffix if necessary
add_dot_h(Fname) when atom(Fname) ->
    add_dot_h(atom_to_list(Fname));
add_dot_h(Fname) ->
    case string:index(Fname,".h") of
	0 -> ?lc([string:strip(Fname,right,$ ),".h"]);
	_ -> Fname
    end.

adr_of(S,Name) -> igserver:adr_of(S,Name).
sizeof(S,Name) -> igserver:sizeof(S,Name).

%% To be called at release time
leex([A,B]) -> leex:gen(A,B).
yecc([A,B])   -> yecc:yecc(A,B);
yecc([A,B,C]) -> yecc:yecc(A,B,C);
yecc([A,B,C,D]) -> yecc:yecc(A,B,C,D).


init(From,Fname,Options) ->
    case catch init_(From,Fname,Options) of
	{'EXIT',Reason} ->
	    From ! {self(),{error,Reason}};
	_ -> ok
    end.

init_(From,Fname,Options) ->
    S = igsym:new(),
    set_default_options(S,Fname),
    store_options(S,From,Options),
    ?TELL("Calling parser", []),
    {ok,Xforms} = parse_extra_files(From,S),
    case catch do_parse(From,Fname) of
	{ok,Forms,0} ->
	    debug(S,Forms),
	    push_symbols(S,Forms,Fname),
	    case catch igsemantic:check(S,Fname,append(Forms,Xforms)) of
		{error,Reason} ->
		    From ! {self(),{error,Reason}};
		SForms ->
		    debug(S,SForms),
		    Gen = iggenerator:gen(S,Fname,SForms),
		    catch check_for_warnings(S,SForms,Fname),
		    From ! {self(),Gen}
	    end;
	{ok,Forms,I} when I>0 ->
	    From ! {self(),{error,when_parsing}};
	{error,Reason} ->
	    From ! {self(),{error,Reason}};
	{error,Reason,WhatLine} ->
	    report_error(Fname,WhatLine,Reason),
	    From ! {self(),{error,when_parsing}};
	{'EXIT',Reason} ->
	    io:format("Internal error, aborting !"),
	    From ! {self(),{error,Reason}}
    end.

parse_extra_files(From,S) ->
    case catch igsym:out(S,{option,files}) of
	Files when list(Files) ->
	    parse_extra_files(From,S,Files,[]);
	_ -> 
	    {ok,[]}
    end.

parse_extra_files(From,S,[Fname|Fs],Xforms) ->
    case catch do_parse(From,Fname) of
	{ok,Forms,0} ->
	    push_symbols(S,Forms,Fname),
	    parse_extra_files(From,S,Fs,append(Forms,Xforms));
	{ok,Forms,I} when I>0 ->
	    From ! {self(),{error,when_parsing}},
	    exit(normal);
	{error,Reason} ->
	    From ! {self(),{error,Reason}},
	    exit(normal);
	{error,Reason,WhatLine} ->
	    report_error(Fname,WhatLine,Reason),
	    From ! {self(),{error,when_parsing}},
	    exit(normal);
	{'EXIT',Reason} ->
	    io:format("Internal error, aborting !"),
	    From ! {self(),{error,Reason}},
	    exit(normal)
    end;
parse_extra_files(_,_,[],Xforms) -> 
    {ok,Xforms}.

%% Store the symbol-file relation.
push_symbols(S,[H|T],Fname) when tuple(H),size(H)>3 ->
    case element(3,H) of
	Id when tuple(Id),size(Id)>3 ->
	    SymName = element(3,Id),
	    igsym:in(S,{sym_file,SymName},Fname),
	    push_symbols(S,T,Fname);
	_ ->
	    push_symbols(S,T,Fname)
    end;
push_symbols(S,[_|T],Fname) ->
    push_symbols(S,T,Fname);
push_symbols(_,[],_) -> ok.

%% Parse statement for statement
do_parse(From,Fname) ->
    Fd = open_file(From,Fname),
    Res = do_parse(Fname,Fd,[],0,1),
    file:close(Fd),
    Res.

do_parse(Fname,Fd,Forms,Ecount,Line) ->
    case igparser:parse_and_scan({ig,get_token,[Fd]}) of
	{error,{WhatLine,Mod,Emsg}} ->
	    ?andthen(Ecount==0, io:nl()),
	    report_error(Fname,WhatLine,Emsg),
	    NewLine = skip_to_eos(WhatLine,Fd),
	    do_parse(Fname,Fd,Forms,Ecount+1,NewLine);
	{error,Reason,WhatLine} ->
	    report_error(Fname,WhatLine,"hejsan"),
	    NewLine = skip_to_eos(WhatLine,Fd),
	    do_parse(Fname,Fd,Forms,Ecount+1,NewLine);
	{ok,[]} -> 
	    {ok,Forms,Ecount};
	{ok,eof} -> 
	    {ok,Forms,Ecount};
	{ok,Form} ->
%	    io:format("do_parse ~p~n", [Form]),
	    NewLine = element(2,lists:last(Form)),
%%	    io:format("~n<~p>~p~n",[NewLine,Form]),
	    do_parse(Fname,Fd,append(Form,Forms),Ecount,NewLine);
	XX ->
	    io:format("~n<MEGA ERROR> do_parse got: ~p~n",[XX]),
	    exit(hejda)
    end.

%% This function will be called by the parser whenever 
%% a new token shall be produced
get_token(Fd) ->
    get_token(get_line(),Fd).

get_token(Line,Fd) ->
    case catch io:requests(Fd,[{get_until,'',igscanner,token,[Line]}]) of
	{ok,{end_of_statement,NewLine},What} ->
	    put_line(NewLine),
	    {ok,[{';',NewLine},{'$end',NewLine}],NewLine};
	{ok,Token,NewLine} ->
	    put_line(NewLine),
	    {ok,[Token],NewLine};
	eof -> 
%%	    {ok,[{'$end',Line}],Line};
	    {eof,Line};
	{error,Reason,WhatLine} ->
	    {error,Reason,WhatLine};
	XX -> 
	    exit(XX)
    end.

get_line() ->
    case get(ig_secret_line_number) of
	N when integer(N) -> N;
	_ -> 1
    end.
put_line(Line) ->
    put(ig_secret_line_number, Line).

%% Drive the tokenizer through the erroneous statement
skip_to_eos(Line,Fd) ->
    case catch io:requests(Fd,[{get_until,'',igscanner,token,[Line]}]) of
	{ok,{end_of_statement,NewLine},What} -> 
	    put_line(NewLine),
	    NewLine;
	{ok,X,NewLine} ->
	    skip_to_eos(NewLine,Fd);
	eof -> 
	    0;
%	    throw({error,end_of_file_in_scanner,Line});
	{error,Reason,WhatLine} ->
	    throw({error,Reason,WhatLine});
	XX -> 
	    throw({error,XX})
    end.


open_file(From,Fname) ->
    case file:open(Fname,read) of
	{error,Reason} -> 
	    From ! {self(),{error,Reason}},
	    exit(Reason);
	{ok,Fd} -> Fd
    end.

check_for_warnings(S,[{ig_call,Line,_,_,_}|_],Fname) ->
    case get_option(S,cback_mod) of
	'_' -> warn({Fname,Line,no_cback_mod});
	_ -> ok
    end;
check_for_warnings(S,[{ig_cast,Line,_,_,_}|_],Fname) ->
    case get_option(S,cback_mod) of
	'_' -> warn({Fname,Line,no_cback_mod});
	_ -> ok
    end;
check_for_warnings(S,[_|T],Fname) ->
    check_for_warnings(S,T,Fname);
check_for_warnings(_,[],_) -> ok.

set_default_options(S,Fname) ->
    igsym:in(S,{option,nouse_stdio},false),
    CbackMod = string:sub_word(Fname,1,$.),
    igsym:in(S,{option,cback_mod},'_'),
    igsym:in(S,{option,debug},false),
    igsym:in(S,{option,files},false).
    
store_options(S,From,Options) ->
    igsym:in(S,options,Options),
    store_options_0(S,From,Options).

store_options_0(S,From,[{cback_mod,Mod}|T]) ->
    igsym:in(S,{option,cback_mod},Mod),
    store_options_0(S,From,T);
store_options_0(S,From,[debug|T]) ->
    igsym:in(S,{option,debug},true),
    store_options_0(S,From,T);
store_options_0(S,From,[nouse_stdio|T]) ->
    igsym:in(S,{option,nouse_stdio},true),
    store_options_0(S,From,T);
store_options_0(S,From,[{file,Fname}|T]) ->
    FileName = add_dot_h(Fname),
    case igsym:out(S,{option,files}) of
	false -> igsym:in(S,{option,files},[FileName]);
	Files -> igsym:in(S,{option,files},[FileName|Files])
    end,
    store_options_0(S,From,T);
store_options_0(S,From,[Option|T]) ->
    io:format("<ERROR> wrong option: ~p , aborting...~n",[Option]),
    From ! {self(),{error,wrong_option}},
    exit(normal);
store_options_0(S,_,[]) -> 
    set_fd(S);
store_options_0(S,From,XX) ->
    io:format("<ERROR> wrong option argument: ~p , aborting...~n",[XX]),
    From ! {self(),{error,wrong_option}},
    exit(normal).

get_option(S,Option) -> igsym:out(S,{option,Option}).
get_options(S) -> igsym:out(S,options).

set_fd(S) ->
    case get_option(S,nouse_stdio) of
	true ->
	    igsym:in(S,stdio,nouse_stdio),
	    igsym:in(S,{stdin,'fun'},3),
	    igsym:in(S,{stdin,cback},3),
	    igsym:in(S,{stdout,'fun'},4),
	    igsym:in(S,{stdout,cback},4);
	false ->
	    igsym:in(S,stdio,use_stdio),
	    igsym:in(S,{stdin,'fun'},0),
	    igsym:in(S,{stdin,cback},0),
	    igsym:in(S,{stdout,'fun'},1),
	    igsym:in(S,{stdout,cback},1)
    end. 

%    case lists:member(use_sockets,Options) of
%	true ->
%	    in(S,stdio,sockets),
%	    in(S,{stdin,fun},?IG_FUN_FD),
%	    in(S,{stdin,cback},?IG_CBACK_FD),
%	    in(S,{stdout,fun},?IG_FUN_FD),
%	    in(S,{stdout,cback},?IG_CBACK_FD);
%	false ->



%%-------------------------------------------------------
%% Channel may be either of: 
%%  fun   - Function calls from Erlang to C
%%  cback - Function calls from C to Erlang
%%-------------------------------------------------------
stdin(S,Channel)  -> igsym:out(S,{stdin,Channel}).
stdout(S,Channel) -> igsym:out(S,{stdout,Channel}).
stdio(S)          -> igsym:out(S,stdio).



debug(S,Forms) ->
    case get_option(S,debug) of
	true -> nice_print(Forms);
	_ -> ok
    end.

nice_print([H|T]) ->
    io:format("~p~n",[H]),
    nice_print(T);
nice_print([]) -> true.

%report_error(Fname,Line,end_of_file_in_scanner) ->
%    io:format("~s~n",
%	      [?lc([Fname,":",Line,
%		    ": scanner reached end of file !"])]);
%report_error(Source_file, Line, end_of_file_in_scanner) ->
%    io:format("~n~s~n", [lists:concat([Source_file, ": ", Line, ": ",
%				     "end of file in scanner"])]);
report_error(Source_file, Line,end_of_file_in_scanner ) ->
    io:format("~n~s~n", [lists:concat([Source_file, ": ", Line, ": ",
				       format_error(end_of_file_in_scanner)])]);
report_error(Source_file, Line, Message) ->
    io:format("~s~n", [lists:concat([Source_file, ": ", Line, ": ",
			     format_error(Message)])]).

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

%%report_error(Fname,Line,end_of_file_in_scanner) ->
%%    io:format("~s~n",
%%	      [?lc([Fname,":",Line,
%%		    ": scanner reached end of file !"])]).

warn({Fname,Line,no_cback_mod}) ->
    io:format("~s~n",
	      [?lc([Fname,":",Line,
		    ": Warning : callback declared, but no ",
		    "callback module specified !"])]);
warn(_) -> ok.
