%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
%%-----------------------------------------------------------------------
%% Created: 95-08-09 by Tobbe (tobbe@erix.ericsson.se)
%% 
%% Function: Does some semantic checks. It also works as an interface
%%           between the front/back end of IG, transforming the parsed
%%           form into a form that suit the backend.
%%-----------------------------------------------------------------------
-module(igsemantic).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: igsemantic.erl,v 1.5 1996/04/24 10:58:55 tobbe Exp $.').
-export([check/3]).
-include("ig.hrl").
-include("igabsform.hrl").

-ifdef(debug).
-define( io(X,Y) , io:format(X,Y) ).
-else.
-define( io(X,Y) , ok ).
-endif.

format_error(obsolete_binary) ->
    "obsolete binary declaration, use IG_binaryPtr instead ";
format_error({tcat_error,{X,Y}}) ->
    ?lc(["interfering type declaration for ",tcat2str(X),
	 " and ",tcat2str(Y)," "]);
format_error(tcat_none) ->
    "no valid type declaration found ";
format_error({sclass,{X,Y}}) ->
    ?lc(["multiple storage classes found <",X,",",Y,"> "]);
format_error({crashed,What}) ->
    ?lc(["crashed , reason : ",What]);
format_error({internal,AtLine}) ->
    ?lc(["internal error in file <",?FILE,"> at line : ",AtLine]).

%%-----------------------------------------------------------------------
%% check(SymbolTable,FileName,ListOfForms)
%%-----------------------------------------------------------------------
check(S,Fname,[Form|Fs]) ->
    case catch check_form(Form) of
	{ok,Sform} ->
	    [Sform|check(S,Fname,Fs)];
	{local_error,What} ->
	    Line = get_line(Form),
	    fatal_error(S,Fname,Line,What);
	{local_error,What,Line} ->
	    fatal_error(S,Fname,Line,What);
	{'EXIT',What} ->
	    Line = get_line(Form),
	    fatal_error(S,Fname,Line,{crashed,What})
    end;
check(_,_,[]) -> [].

check_form(Form) when tuple(Form),ig_var==element(1,Form) -> check_ig_var(Form);
check_form(Form) when tuple(Form),ig_fun==element(1,Form) -> check_fun(Form);
check_form(Form) when tuple(Form),ig_void==element(1,Form) -> check_fun(Form);
check_form(Form) when tuple(Form),ig_call==element(1,Form) -> check_fun(Form);
check_form(Form) when tuple(Form),ig_cast==element(1,Form) -> check_fun(Form);
check_form(Form) when tuple(Form),ig_extern==element(1,Form) -> check_fun(Form);
check_form(Form) when tuple(Form),type==element(1,Form) -> check_ig_type(Form);
check_form(Form) when tuple(Form),hdef==element(1,Form) -> check_hdef(Form);
check_form(Form) when tuple(Form),ifdef==element(1,Form) -> check_ifdef(Form);
check_form(Form) when tuple(Form),ifndef==element(1,Form) -> check_ifndef(Form);
check_form(Form) when tuple(Form),else==element(1,Form) -> check_else(Form);
check_form(Form) when tuple(Form),endif==element(1,Form) -> check_endif(Form);
check_form(Form) -> local_error({internal,?LINE}).

check_ig_var({ig_var,Line,I,D}) when record(I,identifier),record(D,declarations) ->
    ?io("~n<ig_var>: ~p~n",[{ig_var,Line,I,D}]),
    Name = I#identifier.name,
    Rtype = mk_descriptor(I,D),
    Form = {ig_var,Line,Rtype,Name},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_fun({IG_keyw,Line,I,D,Parameters}) when record(I,identifier),record(D,declarations) ->
    ?io("~n<~p>: ~p~n",[IG_keyw,{IG_keyw,Line,I,D,Parameters}]),
    Fun = I#identifier.name,
    Desc = mk_descriptor(I,D),
    Rtype = case D#declarations.ds of
		%% Special case when return type is void and no pointer
		[void] when Desc#descriptor.cat==std -> Desc#descriptor{cat=void};
		_ -> Desc
	    end,
    Params = mk_params(Parameters),
    Form = {IG_keyw,Line,Fun,Rtype,Params},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_ig_type({type,Line,I,D}) when record(I,identifier),record(D,declarations) ->
    ?io("~n<ig_type>: ~p~n",[{type,Line,I,D}]),
    Name = I#identifier.name,
    Desc = mk_descriptor(I,D),
    Form = {type,Line,Name,Desc},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form};
check_ig_type({type,Line,I,M}) when record(I,identifier),list(M) ->
    ?io("~n<ig_type>: ~p~n",[{type,Line,I,M}]),
    Name = I#identifier.name,
    Members = mk_members(M),
    Form = {type,Line,Name,Members},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_hdef({hdef,Line,I,C}) when record(I,identifier) ->
    ?io("~n<hdef>: ~p~n",[{hdef,Line,I,C}]),
    Name = I#identifier.name,
    Form = {hdef,Line,Name,C},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_ifdef({ifdef,Line,I}) when record(I,identifier) ->
    ?io("~n<ifdef>: ~p~n",[{ifdef,Line,I}]),
    Name = I#identifier.name,
    Form = {ifdef,Line,Name},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_ifndef({ifndef,Line,I}) when record(I,identifier) ->
    ?io("~n<ifndef>: ~p~n",[{ifndef,Line,I}]),
    Name = I#identifier.name,
    Form = {ifndef,Line,Name},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_else({else,Line}) ->
    ?io("~n<else>: ~p~n",[{else,Line}]),
    Form = {else,Line},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

check_endif({endif,Line}) ->
    ?io("~n<endif>: ~p~n",[{endif,Line}]),
    Form = {endif,Line},
    ?io("FORM: ~p~n",[Form]),
    {ok,Form}.

mk_members([{D,Line,I}|T]) when record(D,declarations),record(I,identifier) ->
    case catch mk_descriptor(I,D) of
	{local_error,What} ->
	    local_error(What,Line);
	Desc when record(Desc,descriptor) ->
	    Name = I#identifier.name,
	    Member = #member{desc=Desc,name=Name},
	    [Member|mk_members(T)]
    end;
mk_members([{D,Line,{I,C}}|T]) when record(D,declarations),record(I,identifier),integer(C) ->
    case catch mk_descriptor(I,D) of
	{local_error,What} ->
	    local_error(What,Line);
	Desc when record(Desc,descriptor) ->
	    Name = {I#identifier.name,C},
	    Member = #member{desc=Desc,name=Name},
	    [Member|mk_members(T)]
    end;
mk_members([]) -> [];
mk_members(XX) -> 
    ?io("mk_members got: ~p~n",[XX]),
    local_error({internal,?LINE}).


mk_params([{D,I}]) when D#declarations.ds==[void],I#identifier.cat==id ->
    %% Special case, void parameter
    Desc = #descriptor{cat=void,storage_class=none,ansi_string="void"},
    [Desc];
mk_params([{D,I}|T]) when record(D,declarations),record(I,identifier) ->
    Desc = mk_descriptor(I,D),
    [Desc|mk_params(T)];
mk_params([]) -> [];
mk_params(XX) -> 
%%    ?io("mk_params got: ~p~n",[XX]),
    local_error({internal,?LINE}).

%%-------------------------------------------------------
%% Create a type descriptor containing the type category,
%% the storage class, and the (ANSI) type as a string.
%%-------------------------------------------------------
mk_descriptor(I,D) when I#identifier.cat==id,record(D,declarations) ->
    {Cat,Tname} = case type_category(D) of
		      {type,Tname0} -> {type,Tname0};
		      Cat0 -> {Cat0,[]}
		  end,
    {Sclass,Ts,Ds} = mk_desc_data(I,D),
    As = lists:append(Ds,Ts),
    #descriptor{cat=Cat,storage_class=Sclass,ansi_string=As,tname=Tname};
mk_descriptor(I,D) when I#identifier.cat==ptr,record(D,declarations) ->
    %% Just check for interfering types since
    %% we already know it is a pointer.
    type_category(D),
    Cat = ptr,
    {Sclass,Ts,Ds} = mk_desc_data(I,D),
    As = ?lc([Ds," * ",Ts]),
    #descriptor{cat=Cat,storage_class=Sclass,ansi_string=As,tname=[]};
mk_descriptor(I,D) when I#identifier.cat==type_ptr,record(D,declarations) ->
    %% Just check for interfering types since
    %% we already know it is a pointer.
    type_category(D),
    Cat = ptr,
    {Sclass,Ts,Ds} = mk_desc_data(I,D),
    As = ?lc([Ds," * ",Ts]),
    #descriptor{cat=Cat,storage_class=Sclass,ansi_string=As,tname=[]};
mk_descriptor(I,D) when I#identifier.cat==type,record(D,declarations) ->
    {Cat,Tname} = case type_category(D) of
		      {type,Tname0} -> {type,Tname0};
		      Cat0 -> {Cat0,[]}
		  end,
    {Sclass,Ts,Ds} = mk_desc_data(I,D),
    As = lists:append(Ds,Ts),
    #descriptor{cat=Cat,storage_class=Sclass,ansi_string=As,tname=Tname}.

mk_desc_data(I,D) ->
    Sclass = storage_class(D),
    Ts = mk_ansi_string(I#identifier.tq),
    Ds = mk_ansi_string(D),
    {Sclass,Ts,Ds}.
    

%%-------------------------------------------------------
%% Find the storage class. If multiple storage classes
%% are found, then throw an error !!
%%-------------------------------------------------------
storage_class(D) when record(D,declarations) ->
    sclass(D#declarations.ds,none).

sclass([extern|T],none) -> sclass(T,extern);
sclass([static|T],none) -> sclass(T,static);
sclass([register|T],none) -> sclass(T,register);
sclass([auto|T],none) -> sclass(T,auto);
sclass([extern|T],F) when F=/=none -> sclass_error({extern,F});
sclass([static|T],F) when F=/=none -> sclass_error({static,F});
sclass([register|T],F) when F=/=none -> sclass_error({register,F});
sclass([auto|T],F) when F=/=none -> sclass_error({auto,F});
sclass([_|T],F) -> sclass(T,F);
sclass([],none) -> auto;
sclass([],F) -> F.

sclass_error(What) -> local_error({sclass,What}).

%%-------------------------------------------------------
%% Find out what category the type declaration belong to.
%% Valid categories are: 
%%   std, ptr, type, 'IG_string', binaryPtr, assumed_ptr 
%% Since they are exclusive, then throw an error if
%% more than one of them could be deduced.
%%-------------------------------------------------------
type_category(D) -> type_category(D,none).
type_category(D,Cat) when record(D,declarations) ->
    tcat(D#declarations.ds,Cat).

tcat([binaryPtr|T],none)-> tcat(T,binaryPtr);
tcat([binaryPtr|T],F)   -> tcat_error({binaryPtr,F});
tcat([{type,Type}|T],none) -> tcat(T,{type,Type});
tcat([{type,Type}|T],F) -> tcat_error({Type,F});
tcat(['IG_string'|T],none) -> tcat(T,'IG_string');
tcat(['IG_string'|T],F) -> tcat_error({'IG_string',F});
tcat([{assumed_ptr,_}|T],none) -> tcat(T,ptr);
tcat([{assumed_ptr,Ap}|T],F) -> tcat_error({Ap,F});
tcat([const|T],F) -> tcat(T,F);
tcat([volatile|T],F) -> tcat(T,F);
tcat([Type|T],F) -> tcat_0([Type|T],F);
tcat([],none)           -> tcat_error(none);
tcat([],F) -> F.

tcat_0([Type|T],F) ->
    case lists:member(Type,?STD_C_TYPES) of
	true ->
	    case F of
		none -> tcat(T,std);
		std  -> tcat(T,std);
		ptr  -> tcat(T,ptr);
		Else -> tcat_error({std,Else})
	    end;
	false ->
	    case lists:member(Type,?STORAGE_CLASS) of
		true  -> tcat(T,F); %% Ignore
		false -> tcat_error({unexpected_type,Type})
	    end
    end.

tcat_error(none) -> local_error(tcat_none);
tcat_error(What) -> local_error({tcat_error,What}).

tcat2str(ptr)         -> "a pointer";
tcat2str(assumed_ptr) -> "an  assumed pointer";
tcat2str(type)        -> "a user defined type";
tcat2str('IG_string') -> "an IG_string type";
tcat2str(binaryPtr)   -> "a IG_binaryPtr type";
tcat2str(std)         -> "a standard C type";
tcat2str(Type)        -> ?lc(["the ",Type]).


%%-------------------------------------------------------
%% Create a string of the (ANSI) type that can be used
%% when generating C declarations.
%%-------------------------------------------------------
mk_ansi_string(D) when record(D,declarations) ->
    mk_ansi_string(D#declarations.ds);
mk_ansi_string([Atom|T]) when atom(Atom),Atom=/=binaryPtr -> 
    case lists:member(Atom,?STORAGE_CLASS) of
	true  -> mk_ansi_string(T); %% Ignore
	false -> ?lc([Atom," ",mk_ansi_string(T)])
    end;
mk_ansi_string([{type,Name}|T]) -> 
    ?lc([Name," ",mk_ansi_string(T)]);
mk_ansi_string([{assumed_ptr,'IG_binary'}|T]) -> 
    local_error(obsolete_binary);
mk_ansi_string([{assumed_ptr,Name}|T]) -> 
    ?lc([Name," ",mk_ansi_string(T)]);
mk_ansi_string([binaryPtr|T]) -> 
    ?lc(["IG_binaryPtr ",mk_ansi_string(T)]);
mk_ansi_string([_|T]) -> 
    local_error({internal,?LINE});
mk_ansi_string([]) -> "".

%% Throw an error in a "standard" format
fatal_error(S,Fname,Line,What) ->
    Emsg = format_error(What),
    report_error(Fname,Line,Emsg),
    throw({error,semantic}).

%% To be used when to escape deep down...
local_error(What) -> throw({local_error,What}).
local_error(What,Line) -> throw({local_error,What,Line}).

report_error(Fname,Line,Emsg) ->
    io:format("~n~s~n",[?lc([Fname,":",Line," ",Emsg])]).

get_line(Form) when tuple(Form) -> element(2,Form).

