%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
%%-------------------------------------------------------
%% Created:  7 Jun 1995 by tobbe@erix.ericsson.se
%%
%% Function: Symbol table functions for IG
%%-------------------------------------------------------
-module(igsym).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('tobbe@erix.ericsson.se').
-rcs('$Id: igsym.erl,v 1.1 1996/04/09 13:01:10 tobbe Exp tobbe $.').

-export([new/0,delete/1,in/3,out/2]).
-export([mk_sym/3,first_sym/1,next_sym/2,mk_fun_sym/2,
	 first_fun_sym/1,next_fun_sym/2,max_sym/1,get_sym/2]).
-export([alloc_var/2,alloc_var/3,free_var/3]).
-export([store_types/2,get_type_def/2,get_used_vars/1]).

-include("ig.hrl").

%%-------------------------------------------------------
%% First we have some basic operations
%%-------------------------------------------------------
new() -> ets:new(ig,[set,public]).
delete(S) -> ets:delete(S).

in(S,Key,Val) -> ets:insert(S,{Key,Val}).

out(S,Key) ->
    case ets:lookup(S,Key) of
	[{K,Val}] -> Val;
	_ -> throw({error,out})
    end.



%%-------------------------------------------------------
%% Store/Get the defined type(s)
%%-------------------------------------------------------
store_types(S,[{type,Line,Name,Def}|Tail]) ->
    in(S,{type,Name},{type,Line,Name,Def}),
    store_types(S,Tail);
store_types(S,[_|Tail]) -> store_types(S,Tail);
store_types(_,[]) -> ok.

get_type_def(S,Name) ->
    Key = {type,Name},
    case catch out(S,Key) of
	{error,_} -> throw({error,no_type_def});
	Type -> Type
    end.

%%-------------------------------------------------------
%% Get a free variable name for a given type
%%-------------------------------------------------------
alloc_var(S,Type) when list(Type) ->
    alloc_var(S,Type,global).

alloc_var(S,Type,Scope) when list(Type) ->
    case catch out(S,{free_vars,Type}) of
	{Max,[]} ->
	    Var = mk_var(Max+1,Type),
	    in(S,{free_vars,Type},{Max+1,[]}),
	    add_used_var(S,{Var,Type},Scope),
	    Var;
	{Max,[Var|T]} ->
	    in(S,{free_vars,Type},{Max,T}),
	    Var;
	%% First time this routine is called !
	{error,_} ->
	    Var = mk_var(0,Type),
	    in(S,{free_vars,Type},{0,[]}),
	    add_used_var(S,{Var,Type},Scope),
	    Var
    end.

add_used_var(S,VarType,global) ->
    case catch out(S,used_vars) of
	Vs when list(Vs) -> in(S,used_vars,[VarType|Vs]);
	%% First time this routine is called !
	{error,_} -> in(S,used_vars,[VarType])
    end;
add_used_var(_,_,local) -> ok.


get_used_vars(S) -> catch out(S,used_vars).
    

free_var(S,Type,Var) when list(Type),list(Var) ->
    case catch out(S,used_vars) of
	Vs when list(Vs) ->
	    %% The variable might have been local
	    %% and should then be removed from the
	    %% free list. Yes, it's a hack !!
	    case lists:member({Var,Type},Vs) of
		true  -> free_var_0(S,Type,Var);
		false -> ok
	    end;
	_ ->
	    %% We must have at least one used variable
	    %% since we want to free one !!
	    %% But it could have been local (got you Tobbe /Peter)
	    %%throw({error,free_var})
	    ok
    end.

free_var_0(S,Type,Var) ->
    case catch out(S,{free_vars,Type}) of
	{Max,T} -> in(S,{free_vars,Type},{Max,[Var|T]});
	{error,_} -> throw({error,free_var})
    end.
    
%% Create a new variable
mk_var(N,Type) -> ?lc(["IG",rm_sp_star(Type),N]).

%% Remove Space chars and substitute: * with P
rm_sp_star([$ |T]) -> rm_sp_star(T);
rm_sp_star([$*|T]) -> [$P|rm_sp_star(T)];
rm_sp_star([X|T])  -> [X|rm_sp_star(T)];
rm_sp_star([])     -> [].

%%-------------------------------------------------------
%% Routines for handling of generated #DEFINE symbols
%%-------------------------------------------------------
first_sym(S)   -> catch get_sym(S,0).
next_sym(S,No) -> catch get_sym(S,No+1).

mk_fun_sym(S,Name) -> mk_sym(S,Name,?IG_FUN_TAG).
first_fun_sym(S)   -> first_sym(S,?IG_FUN_TAG).
next_fun_sym(S,No) -> next_sym(S,?IG_FUN_TAG,No).

mk_sym(S,Name,Tag) when atom(Name) ->
    mk_sym(S,atom_to_list(Name),Tag);
mk_sym(S,Name,Tag) when list(Name) ->
    add_sym(S,Name,Tag),
    Name.

max_sym(S) -> 
    case catch out(S,def_sym_max) of
	Max when integer(Max) -> Max;
	_ -> -1
    end.

%% Add a #DEFINE symbol and keep a count on the total 
%% number of symbols.
add_sym(S,Name,Tag) ->
    case catch out(S,def_sym_max) of
	Max when integer(Max) -> 
	    case catch out(S,{def_sym,Name}) of
		X when record(X,symbol) -> ok;
		{error,_} ->
		    %% First time !!
		    No = Max+1,
		    in(S,def_sym_max,No),
		    Sym = #symbol{name=Name,no=No,tag=Tag},
		    %% Use both Name and No as keys
		    in(S,{def_sym,Name},#symbol{name=Name,no=No,tag=Tag}),
		    in(S,{def_sym,No},#symbol{name=Name,no=No,tag=Tag})
	    end;
	{error,_} -> 
	    %% The very first time !!
	    in(S,def_sym_max,0),
	    in(S,{def_sym,Name},#symbol{name=Name,no=0,tag=Tag}),
	    in(S,{def_sym,0},#symbol{name=Name,no=0,tag=Tag})
    end.

get_sym(S,Key) when atom(Key) ->
    get_sym(S,atom_to_list(Key));
get_sym(S,Key) ->
    case catch out(S,{def_sym,Key}) of
	X when record(X,symbol) -> {ok,X};
	{error,_} -> throw({error,no_sym})
    end.

first_sym(S,Tag) ->
    case catch get_sym(S,0) of
	{ok,X} when X#symbol.tag == Tag -> {ok,X};
	{ok,X} when X#symbol.tag =/= Tag -> next_sym(S,Tag,0);
	{error,Reason} -> {error,Reason}
    end.

next_sym(S,Tag,No) ->
    case catch get_sym(S,No+1) of
	{ok,X} when X#symbol.tag == Tag -> {ok,X};
	{ok,X} when X#symbol.tag =/= Tag -> next_sym(S,Tag,No+1);
	{error,Reason} -> {error,Reason}
    end.

	
