%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(httpd_socket).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('jocke@erix.ericsson.se').
-export([start/1,listen/2,accept/2,deliver/3,send/3,peername/2,resolve/1,
	 close/2,config/1]).

-include("httpd.hrl").
-include_lib("kernel/include/inet.hrl").

%% start

start(ip_comm) ->
  inet_db:start();
start({ssl,SSLConfigString}) ->
  ssl_socket:start().

%% listen

listen(ip_comm,Port) ->
    case gen_tcp:listen(Port,
			[{packet,0},
			 {backlog,128},
			 {reuseaddr,true},
			 {tcp_nodelay,true}]) of
	{ok,ListenSocket} ->
	    ListenSocket;
	Error ->
	    Error
    end;
listen({ssl,SSLConfigString},Port) ->
    ?DEBUG("SSL: Listening to port ~p", [Port]),
    ssl_socket:listen('STREAM','AF_INET',Port,raw).

%% accept

accept(ip_comm,ListenSocket) ->
  case gen_tcp:accept(ListenSocket) of
    {ok,Socket} ->
      Socket;
    Error ->
      Error
  end;
accept({ssl,SSLConfigString},ListenSocket) ->
    ?DEBUG("SSL: Accept on socket ~p", [ListenSocket]),
    ssl_socket:accept(ListenSocket,SSLConfigString).

%% deliver

deliver(SocketType, Socket, IOListOrBinary)  ->
    send(SocketType, Socket, IOListOrBinary),
    receive
	{tcp_closed,Socket} ->
	    socket_closed;
	{tcp_error,Socket,Reason} ->
	    socket_closed;
	%% Not likely to happen!
	{Socket,{socket_closed,normal}} ->
	    socket_closed;
	%% Not likely to happen!
	{Socket,{socket_closed,error}} ->
	    socket_closed
    after 0 ->
	    ok
    end.

%% send

send(ip_comm,Socket,Data) ->
    gen_tcp:send(Socket,Data);
send({ssl,SSLConfig},Socket,Data) when binary(Data)->
    ?DEBUG("SSL: Send ~p bytes on socket ~p", [size(Data), Socket]),
    Socket ! {self(),{deliver,Data}};
send({ssl,SSLConfig},Socket,Data) ->
    ?DEBUG("SSL: Send ~p bytes on socket ~p", [httpd_util:flatlength(Data), Socket]),
    Socket ! {self(),{deliver,Data}}.

%% peername

peername(ip_comm, Socket) ->
    case inet:peername(Socket) of
	{ok,{{A,B,C,D},Port}} ->
	    {Port,integer_to_list(A)++"."++integer_to_list(B)++"."++
	     integer_to_list(C)++"."++integer_to_list(D)};
	{error,Reason} ->
	    {-1,"unknown"}
    end;
peername({ssl,_SSLConfig},Socket) ->
    case catch ssl_socket:peername(Socket) of
	{'EXIT',_} ->
	    {-1,"unknown"};
	PeerName ->
	    ?DEBUG("SSL: Peername on socket ~p:~p", [Socket, PeerName]),
	    PeerName
    end.

%% resolve

resolve(ip_comm) ->
    {ok,Name} = inet:gethostname(),
    Name;
resolve({ssl,_SSLConfig}) ->
    case ssl_socket:resolve() of
	{resolve, Name} ->          %% New ssl_socket
	    Name;
	Name when list(Name) ->     %% Old ssl_socket
	    Name;
	Other ->
	    Other
    end.

%% close

close(ip_comm,Socket) ->
    gen_tcp:close(Socket);
close({ssl,SSLConfigString},Socket) ->
    ?DEBUG("SSL: Close socket ~p", [Socket]),
    ssl_socket:close(Socket).

%% config (debug: "-cert /var/tmp/server_root/conf/ssl_server.pem")

config(ConfigDB) ->
  case httpd_util:lookup(ConfigDB,com_type,ip_comm) of
    ssl ->
      case ssl_certificate_file(ConfigDB) of
	undefined ->
	  {error,
	   ?NICE("Directive SSLCertificateFile not found in the config file")};
	SSLCertificateFile ->
	  {ssl,SSLCertificateFile++ssl_certificate_key_file(ConfigDB)++
	   ssl_verify_client(ConfigDB)}
      end;
    ip_comm ->
      ip_comm
  end.

ssl_certificate_file(ConfigDB) ->
    case httpd_util:lookup(ConfigDB,ssl_certificate_file) of
      undefined ->
	    undefined;
	SSLCertificateFile ->
	    "-cert "++SSLCertificateFile
    end.

ssl_certificate_key_file(ConfigDB) ->
    case httpd_util:lookup(ConfigDB,ssl_certificate_key_file) of
	undefined ->
	    "";
	SSLCertificateKeyFile ->
	    " -key "++SSLCertificateKeyFile
    end.

ssl_verify_client(ConfigDB) ->
    case httpd_util:lookup(ConfigDB,ssl_verify_client) of
	undefined ->
	    "";
	SSLVerifyClient ->
	    " -verify "++integer_to_list(SSLVerifyClient)
    end.

