%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
%% File    : constraint.erl
%% Author  : Hans Nilsson
%% Purpose : 
%% Created : Mon Jan  9 13:53:19 1995

-module(mnemosyne_constraint).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-export([invert_constraint/1,
	 check/2]).

%%-define(debug,1).

-include("mnemosyne_debug.hrl").
-include("mnemosyne_internal_form.hrl").

%%%----------------------------------------------------------------
%%% 		Exports

%%% called by the unifier when a variable is bound.
%%% Returns  remove_trigger | keep_trigger | throw(fail)
%%% Process dict when_called = compile_time | undefined

check(Bindings, Constraint) ->
    C = mnemosyne_unify:instantiate(Constraint,Bindings),
    check(get(when_called), 
	  C#constraint.exprL, 
	  C#constraint.op, 
	  C#constraint.exprR).
    
invert_constraint(C) ->
    C#constraint{op=inv_op(C#constraint.op)}.

%%%----------------------------------------------------------------
%%% 		Private

check(WhenCalled, ExprL, Op, ExprR) ->
    Result = (catch eval(WhenCalled, {'#expr',
				      eval(WhenCalled, ExprL),
				      Op,
				      eval(WhenCalled, ExprR)})),
    ?debugmsg(1, "~s ~s ~s : ~w\n", [mnemosyne_pp:e(ExprL),
				     Op,
				     mnemosyne_pp:e(ExprR),
				     Result]),
    case Result of
	true ->  remove_trigger;
	cannot_evaluate -> keep_trigger;
	fail -> throw(fail)
    end.
    

%%% eval assumes that all sub expressions are evaluated. That implies
%%% that the arguments are ground and contains no function calls.

eval(WhenCalled, {'#expr',L,Op,R}) ->
    case Op of
	'+' -> L+R;
	'-' -> L-R;
	'*' -> L*R;
	'/' -> L/R;
	'<'   when L <  R -> true;
	'=<'  when L =< R -> true;
	'>'   when L >  R -> true;
	'>='  when L >= R -> true;
	'=='  when L == R -> true;
	'=/=' when L =/= R-> true;
	'='   when L == R -> true;
	'!='  when L =/= R-> true;
	_  -> throw(fail)
    end;

eval(WhenCalled, {'#expr',Op,E}) -> 
    case Op of
	'-' -> - eval(WhenCalled, E);
	'+' -> eval(WhenCalled, E)
    end;

eval(WhenCalled, {'#var',V})  -> throw(cannot_evaluate);

eval(WhenCalled, {'#value',V}) -> V;

eval(compile_time, {'#funcall',M,F,Args}) -> throw(cannot_evaluate);
eval(W, {'#funcall',M,F,Args}) -> apply(eval(W,M),
					eval(W,F),
					eval(W,Args));

eval(WhenCalled, L) when list(L) ->
    lists:map(fun(E) -> eval(WhenCalled,E) end, L);

eval(WhenCalled, T) when tuple(T) ->
    list_to_tuple(eval(WhenCalled,tuple_to_list(T)));

eval(WhenCalled, V) -> V.


%%%----------------------------------------------------------------

inv_op('<')  -> '>=' ;
inv_op('=<') -> '>'  ;
inv_op('>')  -> '=<' ;
inv_op('>=') -> '<'  ;
inv_op('==') -> '=/=';
inv_op('=/=')-> '==' ;
inv_op('=')  -> '!=' ;
inv_op('!=') -> '='  .

%%%----------------------------------------------------------------
regexp_match(String, Regexp) when list(String), list(Regexp) ->
    case string:re_match(String,Regexp) of
	{match,_,_} -> true;
	_ -> throw(fail)
    end;

regexp_match(Atom, Regexp) when atom(Atom), list(Regexp) ->
    regexp_match(atom_to_list(Atom), Regexp);

regexp_match(Integer, Regexp) when integer(Integer), list(Regexp) ->
    regexp_match(integer_to_list(Integer), Regexp);

regexp_match(_, _) ->
    throw(cannot_evaluate).


