%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1996, Ellemtel Telecommunications Systems Laboratories
%% File    : q.erl
%% Author  : Hans Nilsson
%% Purpose : 
%% Created : Thu Jan 11 17:35:27 1996


-module(mnemosyne_exec).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').

-export([eval_collect/3, mk_patterns/1]).

-export([setup_collect/6, setup/2]).

-export([setup_collector_and_query/1, init_query/2,
	 get_answers/3,
	 kill_processes/1]).

-export([start_simple_loop1/4, 
	 start_simple_loop2/3,
	 start_complicated_loop/3]). %% spawned
%%-export([simple_loop1/4, simple_loop2/3, complicated_loop/3]). %% spawned

%%-define(debug,1).

-include("mnemosyne_debug.hrl").
-include("mnemosyne_internal_form.hrl").

%%-define(max_slice_sice, 1).
%%-define(max_slice_sice, 10).
%%-define(max_slice_sice, 100).
-define(max_slice_sice, 1000).
%%-define(max_slice_sice, 10000).
%%-define(max_slice_sice, 100000).

%%%================================================================
%%% 		Exports

%%% eval_collect(Nsols, Query, Fun) where
%%%	Fun = fun(Accu,P) is foldl'd over all bound patterns and the result is
%%%			  returned
%%% Ex1. fun(Accu,P) ->
%%%          [P|Accu]
%%%       end
%%%     collects all solutions.

%%% Ex2. fun(Accu,P) ->
%%%          m:f(P),
%%%          Accu
%%%       end
%%%     applies m:f on every solution but does not return any result



-record(spec, {func,
	       table,
	       pattern
	      }).

eval_collect(N, D, Fun) when record(D,optimizer_result) ->
    Spec = #spec{func=Fun, pattern=D#optimizer_result.pattern},
    case D#optimizer_result.how_to_eval of
	mnesia_match_build ->
	    [P] = D#optimizer_result.code,
	    ?debugmsg(1, "Using mnesia:match_object(~w)\n", 
		      [P#pred_sym.pattern]),

	    %% No need to catch EXIT's from get_from_table
	    Bss = 
		mnemosyne_op:get_from_table(D#optimizer_result.common_bs, 
					    P, []),
	    Result =
		filter_result(Spec#spec{table=
					ets:new(filter1,[set,public])},
			      Bss),
	    ets:delete(Spec#spec.table),
	    Result;
	
	query_eval when D#optimizer_result.code==[] ->
	    ?debugmsg(1, 
		      "Empty goal. pattern= ~s" ++ 
		      case D#optimizer_result.common_bs of
			  [] -> "~i";
			  _ -> " bindings= ~s\n"
		      end,
		      [mnemosyne_pp:e(D#optimizer_result.pattern),
		       mnemosyne_pp:e({'#bindings',
				       D#optimizer_result.common_bs})]),
	    [D#optimizer_result.pattern];
	
	query_eval ->
	    ?debugmsg(1, 
		      "Using the general query evaluator on:\n"
		      "\tpattern= ~s, code= ~s\n" ++ 
		      case D#optimizer_result.common_bs of
			  [] -> "~i";
			  _ -> "\tbindings= ~s\n"
		      end,
		      [mnemosyne_pp:e(D#optimizer_result.pattern),
		       mnemosyne_pp:e(D#optimizer_result.code),
		       mnemosyne_pp:e({'#bindings',
				       D#optimizer_result.common_bs})]),

	    Tid = mnesia:get_activity_id(),
	    Pid = spawn_link(?MODULE, setup_collect, 
			     [D#optimizer_result.code, 
			      D#optimizer_result.common_bs,
			      Spec, N, Tid, self()]),
	    receive
		{fail, Pid, Cause} ->
		    throw({'EXIT', Cause});
		{result, Pid, Res} ->
		    Res;
		{mnesia_down, Node} ->
		    mnesia:abort({node_not_running, Node})
	    end
    end.

%%%----------------------------------------------------------------
%%% Cursor handling
%%% 
%% Supposed to be called as:
%%     H = setup_collector_and_query(Q),
%%         ....
%%     mnesia:transaction(
%%         fun() ->
%%	     Crsr = init_query(H [,Number_of_answers_to_prefetch]),
%%	         ....
%%	     loop over L1 = get_answers(Crsr [,Min,Max]),
%%	         ....
%%	   end),
%%         ....
%%     kill_processes(H)


setup_collector_and_query(OptQuery) when record(OptQuery,optimizer_result) ->
    Spec = #spec{func = fun(Accu,P) -> [P|Accu] end,
		 pattern = OptQuery#optimizer_result.pattern},
    CollectorPid =
	case OptQuery#optimizer_result.how_to_eval of
	    mnesia_match_build -> 
		%% fetch all solutions using mnesia:match_object
		[Code] = OptQuery#optimizer_result.code,
		Bs = OptQuery#optimizer_result.common_bs,
		?debugmsg(1, "Using mnesia:match_object(~w)\n",
			  [Code#pred_sym.pattern]),
		spawn_link(?MODULE, start_simple_loop1, [self(),Bs,Code,Spec]);

	    query_eval when OptQuery#optimizer_result.code==[] ->
		%% has the single answer as the pattern
		?debugmsg(1, "Empty goal. pattern= ~s" ++ 
			  case OptQuery#optimizer_result.common_bs of
			      [] -> "~i";
			      _ -> " bindings= ~s\n"
			  end,
			  [mnemosyne_pp:e(OptQuery#optimizer_result.pattern),
			   mnemosyne_pp:e({'#bindings',
					   OptQuery#optimizer_result.common_bs})]),
		Solutions = [OptQuery#optimizer_result.pattern],
		spawn_link(?MODULE, start_simple_loop2, [self(),Solutions,Spec]);

	    query_eval -> %% Must start all processes
		?debugmsg(1, "Using the general query evaluator on:\n"
			  "\tpattern= ~s, code= ~s\n" ++ 
			  case OptQuery#optimizer_result.common_bs of
			      [] -> "~i";
			      _ -> "\tbindings= ~s\n"
			  end,
			  [mnemosyne_pp:e(OptQuery#optimizer_result.pattern),
			   mnemosyne_pp:e(OptQuery#optimizer_result.code),
			   mnemosyne_pp:e({'#bindings',
					   OptQuery#optimizer_result.common_bs})]),
		spawn_link(?MODULE, start_complicated_loop, [self(),OptQuery,Spec])
	end,
    CursorOwner = self(),
    {query_collector, CollectorPid, CursorOwner}.


%%%----
init_query({query_collector,CollectorPid,CursorOwner}, N) ->
    Tid = mnesia:get_activity_id(),
    legal_pid_tid(CollectorPid, CursorOwner, Tid),
    CollectorPid ! {attention,CursorOwner},
    CollectorPid ! {init_collector,CursorOwner,Tid,N},
    {exec_cursor, CollectorPid, CursorOwner, Tid}.


%%%----
get_answers({exec_cursor,CollectorPid,CursorOwner,Tid}, Nmin, Nmax) ->
    legal_pid_tid(CollectorPid, CursorOwner, Tid),
    CollectorPid ! {get, CursorOwner, Tid, Nmin, Nmax},
    receive
	{answer,CollectorPid,Answers} ->
	    Answers;
	{fail,_,Cause} ->
	    throw({'EXIT',Cause});
	{mnesia_down, Node} ->
	    mnesia:abort({node_not_running, Node})
    end.
    

legal_pid_tid(CollectorPid, CursorOwner, Tid0) ->
    Tid = mnesia:get_activity_id(),
    CollectorProcessInfo = process_info(CollectorPid),
    if 
	CursorOwner =/= self() ->
	    throw({'EXIT',{aborted,not_cursor_owner}});

	Tid == undefined ->
	    throw({'EXIT',{aborted,no_transaction}});

	Tid =/= Tid0 ->
	    throw({'EXIT',{aborted,wrong_transaction}});

	CollectorProcessInfo == undefined ->
	    throw({'EXIT',{aborted,no_collector}});

	true ->
	    ok
    end.
    


%%%----

kill_processes({query_collector,CollectorPid,CursorOwner}) ->
    if 
	CursorOwner =/= self() ->
	    throw({'EXIT',{aborted,not_cursor_owner}});

	true ->
	    unlink(CollectorPid),
	    exit(CollectorPid, deleted)
    end.


%%%----------------

start_simple_loop1(Self,Bs,Code,Spec) ->
    simple_loop1(Self,Bs,Code,
		 Spec#spec{table=ets:new(filter2,[set,public])}).

start_simple_loop2(Self,Solutions,Spec) ->
    simple_loop2(Self,Solutions,
		 Spec#spec{table=ets:new(filter3,[set,public])}).

start_complicated_loop(Self,OptQuery,Spec) ->
    complicated_loop(Self,OptQuery,
		     Spec#spec{table=ets:new(filter4,[set,public])}).

		      


simple_loop1(CursorOwner, Bs, Code, Spec) ->
    ?debugmsg(1,
	      "simple_loop1(CursorOwner=~w, Bs=~w, Code=~w, Spec=~w)\n",
	      [CursorOwner, Bs, Code, Spec]
	     ),
    receive
	{init_collector, CursorOwner, Tid, Nprefetch} ->
	    mnesia:put_activity_id(Tid),
	    case catch mnemosyne_op:get_from_table(Bs,Code,[]) of
		{'EXIT', Cause} ->
		    CursorOwner ! {fail, self(), Cause},
		    simple_loop1(CursorOwner, Bs, Code, Spec);
		Solutions ->
		    simple_loop(filter_result(Spec,Solutions),
				Tid, CursorOwner, Spec),
		    simple_loop1(CursorOwner, Bs, Code, Spec)
	    end;

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    simple_loop1(CursorOwner, Bs, Code, Spec);

	{attention,CursorOwner} ->
	    simple_loop1(CursorOwner, Bs, Code, Spec)
    end.
    

simple_loop2(CursorOwner, Solutions, Spec) ->
    ?debugmsg(1,
	      "simple_loop2(CursorOwner=~w, Solutions=~w, Spec=~w)\n",
	      [CursorOwner, Solutions, Spec]
	     ),
    receive
	{init_collector, CursorOwner, Tid, Nprefetch} ->
	    mnesia:put_activity_id(Tid),
	    simple_loop(Solutions, Tid, CursorOwner, Spec),
	    simple_loop2(CursorOwner, Solutions, Spec);

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    simple_loop2(CursorOwner, Solutions, Spec);

	{attention,CursorOwner} ->
	    simple_loop2(CursorOwner, Solutions, Spec)
    end.



complicated_loop(CursorOwner, OptQuery, Spec) ->
    ?debugmsg(1,
	      "complicated_loop(CursorOwner=~w, OptQuery=~w, Spec=~w)\n",
	      [CursorOwner, OptQuery, Spec]
	     ),
    %% process_flag(trap_exit, true),
    FirstPid = setup_query(OptQuery#optimizer_result.code, self()),
    complicated_loop_cont(FirstPid, CursorOwner, OptQuery, Spec).

complicated_loop_cont(FirstPid, CursorOwner, OptQuery, Spec) ->
    receive
	{init_collector, CursorOwner, Tid, Nprefetch} ->
	    mnesia:put_activity_id(Tid),
	    FirstPid ! {tid,Tid},
	    EndCounters = [mnemosyne_op:new_end_token(1)],
	    FirstPid ! {bss, 
			[OptQuery#optimizer_result.common_bs], 
			Nprefetch, EndCounters, []},
	    Reply = no_pending([], FirstPid, 
				Tid, CursorOwner, Spec, EndCounters),
	    CursorOwner ! Reply,
	    complicated_loop_cont(FirstPid, CursorOwner, OptQuery, Spec);

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    Message = {tid, Activity},
	    FirstPid ! Message,
	    flush_until(Message),
	    ReplyTo ! {activity_ended, Activity, self()},
	    complicated_loop_cont(FirstPid, CursorOwner, OptQuery, Spec);
	
	{attention,CursorOwner} ->
	    complicated_loop_cont(FirstPid, CursorOwner, OptQuery, Spec)
    end.

flush_until(Message) ->
    receive
	Message -> ok;
	Other -> flush_until(Message)
    end.


simple_loop(Answers, Tid, CursorOwner, Spec) ->
    ?debugmsg(1, "simple_loop Answers=~w\n", [Answers]),
    receive
	{get, CursorOwner, Tid, Nmin, Nmax} ->
	    {First, Last} = mnemosyne_op:split_list(Nmax, Answers),
	    CursorOwner ! {answer, self(), First},
	    simple_loop(Last, Tid, CursorOwner, Spec);

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    ok;

	{fail, SomePid, Cause} ->
	    CursorOwner ! {fail, self(), Cause},
	    ok;

	{attention,CursorOwner} ->
	    ok
    end.
	


ask_for_more(Pid, N, EndCounters) when pid(Pid) ->
    NewEndToken = mnemosyne_op:new_end_token(1),
    Pid ! {bss, [], N, [NewEndToken], []},
    [NewEndToken | EndCounters].
    

send_answers(Pid, Acc, N) ->
    {First, Last} = mnemosyne_op:split_list(N, Acc),
    Pid ! {answer, self(), First},
    Last.
    


no_pending(Answers, FirstPid, Tid, CursorOwner, Spec, EndCounters) ->
    ?debugmsg(1, "no_pending Answers=~w, EndCounters=~w\n",
	      [Answers, EndCounters]),
    receive
	{get, CursorOwner, Tid, Nmin, Nmax} ->
	    %% Answers requested
	    Nanswers = length(Answers),
	    if Nmin =< Nanswers ->
		    %% Has enough answers already.
		    no_pending(send_answers(CursorOwner,Answers,Nmax),
			       FirstPid, Tid, CursorOwner, Spec, EndCounters);
	      Nanswers < Nmin ->
		    %% Has NOT enough answers available, ask for more
		    pending(Answers, FirstPid, Tid, 
				    CursorOwner, Spec,
				    ask_for_more(FirstPid,Nmax,EndCounters),
				    Nmin, Nmax)
	    end;

	{bss, Bss, SliceSize, EndCntrl, Stack} ->
	    case decrease_counters(EndCntrl, EndCounters) of
		[] -> %% No more Bss expected. Check by asking for more
		    wait_end_no_pending(filter_result(Spec,Bss,Answers),
					FirstPid, Tid, CursorOwner,
					Spec,
					ask_for_more(FirstPid, SliceSize,[]));
		NewEndCounters -> %% More Bss expected
		    no_pending(filter_result(Spec,Bss,Answers),
			       FirstPid, Tid, CursorOwner, Spec,NewEndCounters)
	    end;

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    ok;

	{fail, SomePid, Cause} ->
	    CursorOwner ! {fail, self(), Cause},
	    ok;

	{tid, _} ->
	    no_pending(Answers, FirstPid, Tid, CursorOwner, Spec, EndCounters);

	{attention,CursorOwner} ->
	    ok;
	
	Others -> 
	    Others
    end.
	    

pending(Answers, FirstPid, Tid, CursorOwner, Spec, EndCounters, Nmin, Nmax) ->
    ?debugmsg(1, "pending Answers=~w, EndCounters=~w, Nmin=~w, Nmax=~w\n",
	      [Answers, EndCounters, Nmin, Nmax]
	     ),

    receive
	{bss, Bss, SliceSize, EndCntrl, Stack} ->
	    NewAnswers = filter_result(Spec,Bss,Answers),
	    Nnewanswers = length(NewAnswers),

	    case decrease_counters(EndCntrl, EndCounters) of
		[] when Nnewanswers < Nmin ->
		    %% No more Bss expected. Ask for more.
		    wait_end_pending(NewAnswers, FirstPid, Tid, CursorOwner,
				     Spec,
				     ask_for_more(FirstPid, SliceSize, []),
				     Nmin, Nmax);
		[] when Nmin =< Nnewanswers ->
		    %% No more Bss expected. Ask for more. Send Answers
		    wait_end_no_pending(send_answers(CursorOwner,
						     NewAnswers,Nmax),
					FirstPid, Tid, CursorOwner,
					Spec,
					ask_for_more(FirstPid, SliceSize,[]));

		NewEndCounters when Nnewanswers < Nmin -> 
		    %% More Bss expected, just wait for them
		    pending(NewAnswers, FirstPid, Tid, CursorOwner, Spec,
			    NewEndCounters, Nmin, Nmax);

		NewEndCounters when  Nmin =< Nnewanswers -> 
		    %% More Bss expected, send Answers, wait for more Bss
		    no_pending(send_answers(CursorOwner,NewAnswers,Nmax),
			       FirstPid, Tid, CursorOwner, Spec,
			       NewEndCounters)
	    end;

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    ok;

	{fail, SomePid, Cause} ->
	    CursorOwner ! {fail, self(), Cause},
	    ok;

	{tid, _} ->
	    pending(Answers, FirstPid, Tid, CursorOwner, 
		    Spec, EndCounters, Nmin, Nmax);

	{attention,CursorOwner} ->
	    ok;
	
	Others -> 
	    Others
    end.
		
    

wait_end_no_pending(Answers, FirstPid, Tid, CursorOwner,Spec,EndCounters) ->
    ?debugmsg(1, "wait_end_no_pending Answers=~w, EndCounters=~w\n",
	      [Answers, EndCounters]),
    receive
	{bss, Bss, SliceSize, EndCntrl, Stack} ->
	    case decrease_counters(EndCntrl, EndCounters) of
		[] -> %% No more Bss expected. Just wait for get-requests
		    simple_loop(filter_result(Spec,Bss,Answers), 
				Tid, CursorOwner, Spec);

		NewEndCounters when Bss == [] -> %% More Bss expected
		    wait_end_no_pending(Answers, FirstPid, Tid, CursorOwner,
					Spec, NewEndCounters);

		NewEndCounters when Bss =/= [] -> %% More Bss expected
		    no_pending(filter_result(Spec,Bss,Answers),
			       FirstPid, Tid, CursorOwner, Spec,
			       NewEndCounters)
	    end;

	{get, CursorOwner, Tid, Nmin, Nmax} ->
	    %% Answers requested
	    Nanswers = length(Answers),
	    if Nmin =< Nanswers ->
		    %% Has enough answers already.
		    wait_end_no_pending(send_answers(CursorOwner,Answers,Nmax),
					FirstPid, Tid, CursorOwner, Spec,
					EndCounters);
	      Nanswers < Nmin ->
		    %% Has NOT enough answers available, wait for more Bss
		    wait_end_pending(Answers, FirstPid, Tid, 
				     CursorOwner, Spec, EndCounters,
				     Nmin, Nmax)
	    end;

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    ok;

	{fail, SomePid, Cause} ->
	    CursorOwner ! {fail, self(), Cause},
	    ok;

	{tid, _} ->
	    wait_end_no_pending(Answers, FirstPid, Tid, 
				CursorOwner,Spec,EndCounters);

	{attention,CursorOwner} ->
	    ok;
	
	Others -> 
	    Others
    end.

wait_end_pending(Answers, FirstPid, Tid, CursorOwner, Spec, EndCounters,
		 Nmin, Nmax) ->
    ?debugmsg(1, "wait_end_pending Answers=~w, EndCounters=~w, Nmin=~w, Nmax=~w)\n",
	      [Answers, EndCounters, Nmin, Nmax]),
    receive
	{bss, Bss, SliceSize, EndCntrl, Stack} ->
	    NewAnswers = filter_result(Spec,Bss,Answers),
	    Nnewanswers = length(NewAnswers),

	    case decrease_counters(EndCntrl, EndCounters) of
		[] -> %% No more Bss expected. Send Answers and wait for death
		    simple_loop(send_answers(CursorOwner,NewAnswers,Nmax),
				Tid, CursorOwner, Spec);

		NewEndCounters when Bss == [] ->
		    %% More Bss expected. Wait for them
		    wait_end_pending(Answers, FirstPid, Tid, CursorOwner, 
				     Spec, NewEndCounters, Nmin, Nmax) ;

		NewEndCounters when Bss =/= [], Nmin =< Nnewanswers ->
		    %% Have enough answers. send them
		    no_pending(send_answers(CursorOwner,NewAnswers,Nmax),
			       FirstPid, Tid, CursorOwner, Spec,
			       NewEndCounters);

		NewEndCounters when Bss =/= [], Nnewanswers < Nmin ->
		    %% Have not enough answers but expecting more Bss.
		    %% Wait for them
		    pending(NewAnswers, FirstPid, Tid, CursorOwner, Spec,
			    NewEndCounters, Nmin, Nmax)
	    end;

	{activity_ended, Activity, ReplyTo} ->
	    mnesia:put_activity_id(Activity),
	    ReplyTo ! {activity_ended, Activity, self()},
	    ok;

	{fail, SomePid, Cause} ->
	    CursorOwner ! {fail, self(), Cause},
	    ok;

	{tid, _} ->
	    wait_end_pending(Answers, FirstPid, Tid, 
			     CursorOwner, Spec, EndCounters, Nmin, Nmax);

	{attention,CursorOwner} ->
	    ok;
	
	Others -> 
	    Others
    end.


	

%%%================================================================
%%% 		Private


-define(start_op(Type,Args),
	spawn_link(mnemosyne_op, start_op, [Type,Args])
       ).

%%%----------------

setup_query(Query, NextPid) ->
    ?debugmsg(3, "~w (CollectorPid)\n", [NextPid]),
    P = make_split( setup(Query,NextPid) ),
    ?debugmsg(3, "~w (first) -> ~w  max_slice_sice=~w\n", 
	      [self(),P,?max_slice_sice]),
    P.



setup([C|Cs], NextPid) ->
    setup(C, setup(Cs,NextPid));

setup({'#or',Alts}, NextPid) ->
    make_split( lists:map({?MODULE,setup},[NextPid],Alts));

setup({'#bindings',Bs}, NextPid) ->
    ?start_op(bindings, [NextPid,Bs]);

setup(P, NextPid) when record(P,pred_sym) ->
    case P#pred_sym.type of
	table ->
	    Ps = mk_patterns(P),
	    ?start_op(table, [NextPid, Ps]);

	rule when P#pred_sym.recursive==recursive ->
	    ?start_op(call_recursive_op, [NextPid,P])
    end;

setup(F, NextPid) when record(F,fn) ->
    ?start_op(funcall, [NextPid,F#fn.alias_var,F#fn.fndef]);

setup(E, NextPid) when record(E,erl_expr) ->
    ?start_op(erl_expr, [NextPid,E]);

setup(C, NextPid) when record(C,disj_alt) ->
    setup(C#disj_alt.conj, NextPid);

setup({'#not',Q}, NextPid) ->
    HelpPid = ?start_op(negation_help, []),
    Pid = ?start_op(negation, [NextPid,HelpPid]),
    HelpPid ! {pids, Pid, setup(Q,HelpPid)},
    Pid;

setup([], NextPid) ->
    NextPid.




make_split([Pid]) when pid(Pid) -> Pid;
make_split(L) when list(L),pid(hd(L)) -> ?start_op(split, [L]);
make_split(Pid) when pid(Pid) -> Pid.

%%%----------------
-record(mkp, {p=[],	%% The Pattern to send to trans:match (ets:match)
	      d=[],	%% Dictionary used when making p
	      evN=0	%% First unused ets variable
	     }).


mk_patterns(P) when record(P,pred_sym) ->
    S = mkp(P#pred_sym.args,
	    P#pred_sym.defvars,
	    P#pred_sym.singelvars,
	    #mkp{}),
    P#pred_sym{pattern=hd(S#mkp.p),%%% one arg only
	       back_pattern=hd(P#pred_sym.args)
	      }.


mkp([H|T], DefVars, SingelVars, S0) ->
    Sh = mkp1(H, DefVars, SingelVars, S0#mkp{p=[]}),
    mkp(T, DefVars, SingelVars, Sh#mkp{p=[Sh#mkp.p|S0#mkp.p]});

mkp([], DefVars, SingelVars, S) ->
    S#mkp{p=lists:reverse(S#mkp.p)}.



mkp1([H|T], DefVars, SingelVars, S0) ->
    Sh = mkp1(H, DefVars, SingelVars, S0#mkp{p=[]}),
    if
	list(T) ->
	    mkp1(T, DefVars, SingelVars, Sh#mkp{p=[Sh#mkp.p|S0#mkp.p]});
	true ->
	    St = mkp1(T, DefVars, SingelVars, Sh#mkp{p=[]}),
	    St#mkp{p = lists:append(lists:reverse([Sh#mkp.p|S0#mkp.p]), 
				    St#mkp.p)}
    end;

mkp1([], DefVars, SingelVars, S) ->
    S#mkp{p=lists:reverse(S#mkp.p)};

mkp1({'#var','_'}, DefVars, SingelVars, S0) ->
    S0#mkp{p='_'};

mkp1({'#var',V}, DefVars, SingelVars, S0) ->
    case ordsets:is_element(V,SingelVars) of
	true ->
	    S0#mkp{p='_'};
	false ->
	    case ordsets:is_element(V,DefVars) of
		true ->
		    S0#mkp{p={'#var',V}};
		false ->
		    case lists:keysearch(V,1,S0#mkp.d) of
			{value, {V,EtsVar}} ->
			    S0#mkp{p=EtsVar};
			false ->
			    EtsVar = ets_var(S0#mkp.evN),
			    S0#mkp{p=EtsVar, 
				   evN=S0#mkp.evN+1,
				   d=[{V,EtsVar}|S0#mkp.d]}
		    end
	    end
    end;
	
mkp1(T, DefVars, SingelVars, S0) when tuple(T) ->
    S1 = mkp1(tuple_to_list(T),DefVars,SingelVars,S0),
    S1#mkp{p=list_to_tuple(S1#mkp.p)};

mkp1(X, DefVars, SingelVars, S0) ->
    S0#mkp{p=X}.


ets_var(N) -> list_to_atom([$$ | integer_to_list(N)]).


%%%----------------------------------------------------------------
%%%
%%% Collection help functions
%%%

setup_collect(OptQuery, InitialBs, Spec, N, Tid, ResultPid) ->
    ?debugmsg(2, "Goal=~s\nInitial Bindings=~s\n",
	      [mnemosyne_pp:body(OptQuery), 
	       mnemosyne_pp:e({'#bindings',InitialBs})]),
    %% process_flag(trap_exit, true),
    FirstPid = setup_query(OptQuery, self()),
    FirstPid ! {tid, Tid}, 
    SliceSize = slice_size(N),
    EndToken = mnemosyne_op:new_end_token(1),
    FirstPid ! {bss, [InitialBs], SliceSize, [EndToken], []},
    ResultPid ! collector(Spec#spec{table=ets:new(filter5,[set,public])},
			  FirstPid, N, SliceSize, [EndToken]),
    unlink(ResultPid),
    exit(ready).
    
collector(Spec, FirstPid, N, SliceSize, EndCounters) ->
    Acc = [],
    foreach(EndCounters, FirstPid, 0, N, Spec, SliceSize, Acc).


foreach([], Pid, N, Max, Spec, SliceSize, Acc) when N<Max ->
    EndToken = mnemosyne_op:new_end_token(1),
    Pid ! {bss, [], SliceSize, [EndToken], []},
    wait_all_end([EndToken], Pid, N, Max, Spec, Acc);

foreach([], Pid, N, Max, Spec, SliceSize, Acc) ->
    {result, self(), Acc};

foreach(EndCounters, Pid, N, Max, Spec, SliceSize0, Acc) ->
    ?debugmsg(9,"EndCounters=~w, length(Acc)=~w\n", [EndCounters,length(Acc)]),
    ?debug_receive(3)
	{bss,[],SliceSize,EndCntrl,_} ->
	    foreach(decrease_counters(EndCntrl,EndCounters),
		    Pid, N, Max, Spec, SliceSize, Acc);
	{bss,Bss,SliceSize,EndCntrl,_} ->
	    handle_bss(EndCounters, Pid, N, Max, Spec, 
		       Bss, SliceSize, EndCntrl, Acc);
	{fail, SomePid, Cause} ->
	    {fail, self(), Cause}
    end.



handle_bss(EndCounters, Pid, N, all, Spec, Bss, SliceSize, EndCntrl, Acc) ->
    foreach(decrease_counters(EndCntrl,EndCounters),
	    Pid, N+length(Bss), all, Spec, SliceSize, 
	    filter_result(Spec, Bss, Acc));

handle_bss(EndCounters, Pid, N, Max, Spec, Bss, SliceSize, EndCntrl, Acc) 
						when (N+length(Bss)) >= Max ->
    {result, self(), filter_result(Spec, lists:sublist(Bss,1,(Max-N)), Acc)};

handle_bss(EndCounters, Pid, N, Max, Spec, Bss, SliceSize, EndCntrl, Acc) ->
    foreach(decrease_counters(EndCntrl,EndCounters),
	    Pid, N+length(Bss), Max, Spec, SliceSize, 
	    filter_result(Spec, Bss, Acc)).


wait_all_end([], Pid, N, Max, Spec, Acc) ->
    {result, self(), Acc};
    
wait_all_end(EndCounters,Pid, N, Max, Spec, Acc) ->
    ?debugmsg(9, "wait_all_end  EndCounters=~w, length(Acc)=~w\n",
	      [EndCounters,length(Acc)]),
    ?debug_receive(3)
	{bss,[],_,EndCntrl,_} ->
	    wait_all_end(decrease_counters(EndCntrl,EndCounters),
			 Pid, N, Max, Spec, Acc);

	{bss,Bss,SliceSize,EndCntrl,_} ->
	    handle_bss(EndCounters, Pid, N, Max, Spec,
		       Bss, SliceSize, EndCntrl, Acc);

	{fail, SomePid, Cause} ->
	    {fail, self(), Cause}
     end.

%%----------------
filter_result(Spec, Bss) ->
    filter_result(Spec, Bss, []).

filter_result(Spec, Bss, Acc) ->
    lists:foldr(
      fun(Bs,Accu) ->
	      Answer = mnemosyne_unify:instantiate(Spec#spec.pattern,Bs),
	      case ets:lookup(Spec#spec.table, Answer) of
		  []  -> 
		      ets:insert(Spec#spec.table, {Answer}),
		      (Spec#spec.func)(Accu,Answer);
		  _ ->
		      Accu
	      end
      end, Acc, Bss).


%%----------------
decrease_counters([{Key,N}|T], Counters) ->
    decrease_counters(T,
		      case lists:keysearch(Key,1,Counters) of
			  {value, {Key,X}} when X>1 ->
			      lists:keyreplace(Key,1,Counters,{Key,X-1});
			  {value, {Key,X}} ->
			      lists:keydelete(Key,1,Counters);
			  false when N>1 ->
			      [{Key,N-1}|Counters];
			  false ->
			      Counters
		      end);
decrease_counters([], Counters) ->
    Counters.

%%----------------
slice_size(all) -> slice_size(?max_slice_sice);
slice_size(N) when N>?max_slice_sice -> ?max_slice_sice;
slice_size(N) -> N.
