%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File    : mnemosyne_lc.erl
%% Author  : Hans Nilsson (From erl_id_trans.erl by Robert Virding)
%% Purpose : A List Comprehension transformer of Mnemosyne

-module(mnemosyne_lc).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').

-export([parse_transform/2, format_error/1, the_query/1,
	 one_lc_to_handle/1]).

%%-export([abstract_keep_vars/2]).

%%-define(debug, 3).

-include("mnemosyne_debug.hrl").
-include("mnemosyne_internal_form.hrl").


%% This module transforms Mnemosyne things in the erlang source
%% module. Queries are transformed into function calls:
%%
%%           query [ P || B ] end
%%           
%%  --->     mnemosyne_lc:the_query({query,1,P',B',Data})
%%
%%           
%%  Rules are transformed as:
%%
%%           P(PV1) :- PB1;
%%           P(PV2) :- PB2;
%%		 ...
%%           P(PVn) :- PBn.
%%           
%%
%%           S(SV1) :- SB1;
%%           S(SV2) :- SB2;
%%		 ...
%%           S(SVm) :- SBm.
%%           
%%  --->     'MNEMOSYNE RULE'(P) -> [{PV1', PB1'},
%%				     {PV2', PB2'},
%%				       ...
%%				     {PVn', PBn'}];
%%
%%           'MNEMOSYNE RULE'(S) -> [{SV1', SB1'},
%%				     {SV2', SB2'},
%%				       ...
%%				     {SVm', SBm'}].
%%




-record(s, {options = [],
	    recdefs = [],
	    argtypes = [],
	    module = ?NO_MODULE,
	    line = none,
	    erl_vars = ordsets:new_set(),
	    db_vars = ordsets:new_set(),
	    allow_erl_vars = false,
	    allow_db_vars = true,
	    allow_fun_calls = false,
	    rules = []
	    }).

-record(rec_def, {name, fields=[]}).  %% Field <- {name, init}

format_error(Msg) ->
    lists:concat(
      case Msg of
	  mnesia_not_running ->
	      ["Mnesia is not running"];

	  {no_erl_var,Name} ->
	      ["no Erlang variable (",Name,") at this place in a database "
	       "list comprehension"];

%	  {unbound_var,Name} ->
%	      ["variable '",Name,"' in list comprehension is unbound"];

	  {illegal_lc_generator, Side, X} ->
	      case lists:flatten(erl_pp:expr(X, fun(_,_,_,_) -> "" end)) of
		  "" ->
		      ["illegal ",Side,
		       " side in database list comprehension generator"];
		  Str ->
		      ["illegal ",Side,
		       " side in database list comprehension generator: ",
		      Str]
	      end;
	  
	  {illegal_lc_body,X} ->
	      case lists:flatten(erl_pp:expr(X, fun(_,_,_,_) -> "" end)) of
		  "" ->
		      ["illegal body in database list comprehension"];
		  Str ->
		      ["illegal body \"",
		       Str, "\" in database list comprehension"]
	      end;
	  
	  {illegal_expr,X} ->
	      case lists:flatten(erl_pp:expr(X, fun(_,_,_,_) -> "" end)) of
		  "" ->
		      ["illegal expression in database list comprehension"];
		  Str ->
		      ["illegal expression \"", Str,
		       "\" in database list comprehension"]
	      end;
	  
	  {illegal_op,Op} ->
	      ["illegal operation (",Op,") in database list comprehension"];
	  
	  {illegal_field,Name} ->
	      ["illegal database record field value for \"",Name, "\""];
	  
	  {illegal_record_init,Name,Field} ->
	      ["database record \"",Name,
	       "\" has illegal initialization of the field \"", Field, "\""];
	  
	  {illegal_rule_arity,Name,Arity} ->
	      ["the rule \"",Name,"\" has arity =/= 1"];

	  {already_defined,What,Name} ->
	      ["the ", What," \"", Name, "\" is already defined"];

	  {undefined,What,Name} ->
	      ["the ", What," \"", Name, "\" is undefined"];

	  Other ->
	      [io_lib:write(Other)]
      end).


parse_transform(Forms, Options) ->
    case catch pass2( pass1(Forms,Options) ) of
	{error, E} -> [{error, E}];
	R -> R
    end.



%%%----------------------------------------------------------------
one_lc_to_handle([Expr]) ->
    TabAttrs =
	lists:map(
	  fun(Tab) ->
		  case catch mnesia:table_info(Tab, attributes) of
		      {'EXIT',{aborted,no_exists}} ->
			  throw({error, atom_to_list(Tab)++" does not exist"});
		      Attrs when list(Attrs) ->
			  {Tab,Attrs}
		  end
	  end, find_tables(Expr)),

    FakedFunction = {function,0,hale_bopp,0,[{clause,0,[],[],[Expr]}]},
    FakedModule = mk_rec_defs(TabAttrs) ++ [FakedFunction, {eof,0}] ,
    TransformedFakedModule = 
	parse_transform(FakedModule, [report_errors,report_warnings]),
    
    the_query(
      hd(lists:foldl(
	   fun({function,0,hale_bopp,0,
		[{clause,0,[],[],
		  [{call,1,{remote,1,{atom,1,mnemosyne_lc},{atom,1,the_query}},
		    [X]}]}]}, Acc) -> [erl_parse:normalise(X)|Acc];
	      (_, Acc) ->
		   Acc
	   end, [], TransformedFakedModule))).



mk_rec_defs(TabAttrs) ->
    lists:map(fun({Tab,Attrs}) ->
		      {attribute,0,record,
		       {Tab, lists:map(fun(Attr) ->
					       {record_field,0,{atom,0,Attr}}
				       end, Attrs)}}
	      end, TabAttrs).

find_tables(E) -> find_tables(E,[]).


find_tables({generate,_,{var,_,_},{call,_,{atom,_,table},[{atom,_,Table}]}},
	    Acc) ->  [Table | Acc];
find_tables({generate,_,{var,_,_},_}, Acc) ->
    throw({error, "Illegal generator"});
find_tables(T, Acc) when tuple(T) ->  find_tables(tuple_to_list(T), Acc);
find_tables([H|T], Acc) ->  find_tables(T, find_tables(H,Acc));
find_tables(_, Acc) ->   Acc.
    




%%%----------------------------------------------------------------
%%%----------------
pass1(Forms, Options) ->
    lists:foldl(
      fun
	  ({attribute,Line,record,{Name,Defs0}}, {Acc,S}) ->
	      case catch record_defs(Defs0, Name) of
		  {error, E} ->
		      {[{error,E}|Acc], S};
		  {'EXIT', Cause} ->
		      throw({'EXIT', Cause});
		  {ok, Def} ->
		      {[{attribute,Line,record,{Name,Defs0}}|Acc], 
		       S#s{recdefs=[Def|S#s.recdefs]}}
	      end;
      
	  ({attribute,Line,module,Name}, {Acc,S}) ->
	      {[{attribute,Line,module,Name}|Acc], S#s{module=Name}};
      
	  ({attribute,Line,argtype,{RuleName,RecordName}}, {Acc,S}) ->
	      case lists:keysearch(RuleName,1,S#s.argtypes) of
		  {value, _} ->
		      {[{error, 
			 {Line,?MODULE,
			  {already_defined,"argument type of",RuleName}}} |
			Acc], S};
		  false ->
		      {Acc, S#s{argtypes=[{RuleName,RecordName}|S#s.argtypes]}}
	      end;
      
	  ({rule,Line,Name,Arity,Clauses}, {Acc,S}) when Arity == 1 ->
	      case lists:keysearch(Name,1,S#s.rules) of
		  {value,_} ->
		      {[{error, {Line,?MODULE,{already_defined,rule,Name}}} |
			Acc], S};
		  false -> 
		      {Acc, S#s{rules=[{Name,Line,Clauses}|S#s.rules]}}
	      end;
      
	  ({rule,Line,Name,Arity,_}, {Acc,S}) when Arity =/= 1 ->
	      {[{error, 
		 {Line,?MODULE,{illegal_rule_arity,Name,Arity}}}|Acc], S};
      
	  (F, {Acc,S}) ->
	      {[F|Acc], S}
	       
      end, {[],#s{options=Options}}, Forms).




pass2({[{eof,EOFline}|Forms],S0}) -> %% Forms are the forms in reverse order
    {RulePart, S1} =
	case mk_rules_def_fkn(S0) of
	    [] ->
		{[], S0};
	    {ok, FnDefL} ->
		{FnDefL, S0};
	    {error, ErrLst} ->
		{ErrLst, S0#s{rules=nope}} %% prevent export decl of 'MNEMOSYNE RULE'
	end,
    
    {NewForms,_} = 
	lists:foldl(    
	  fun
	      ({attribute,Line,'export',L}, {Acc,S}) when list(S#s.rules) -> 
		  {[{attribute,Line,'export',[{'MNEMOSYNE RULE',1}|L]} | Acc],
		   S#s{rules=nope}}; %% prevent further export decls.
			  
	      ({function,Line,Name,Arity,Clauses}, {Acc,S}) -> 
		  {[case catch fn_clauses(Clauses,S) of
			{error, E} ->
			    {error,E};
			{'EXIT',Cause} ->
			    throw({'EXIT',Cause});
			TClauses ->
			    {function,Line,Name,Arity,TClauses}
		    end |Acc], S};

	      (F, {Acc,S}) ->
		  {[F|Acc], S}
	  end, {[],S1}, Forms),

    NewForms ++ RulePart ++ [{eof,EOFline}];

pass2({L, S}) ->
    L.


%%  N.B. Field names are full expressions here but only atoms are allowed
%%  by the *parser*!

record_defs(Defs, Name) ->
    R = record_defs(Defs, Name, [], []),
    ?debugmsg(3, "RecDef = ~w\n", [R]),
    {ok,R}.

record_defs([{record_field,Line,{atom,La,A},Val0}|Defs], Name, Fs, Is) ->
    case valid_record_def_inits(Val0) of
	true ->
	    record_defs(Defs, Name, [A|Fs], [Val0|Is]);
	false ->
	    throw({error, {La,?MODULE,{illegal_record_init,Name,A}}})
    end;
record_defs([{record_field,Line,{atom,La,A}}|Defs], Name, Fs, Is) ->
    record_defs(Defs, Name, [A|Fs], [{'#var',{Name,A}}|Is]);
record_defs([], Name, Fs, Is) ->
    {Name, lists:reverse(Fs), list_to_tuple([Name|lists:reverse(Is)])}.

valid_record_def_inits(V) -> valid_record_def_inits(V,true).


valid_record_def_inits({atom,_,_}, X) -> X;

valid_record_def_inits({integer,_,_}, X) -> X;

valid_record_def_inits({float,_,_}, X) -> X;

valid_record_def_inits({string,_,_}, X) -> X;

valid_record_def_inits({cons,_,H,T}, X) -> 
    valid_record_def_inits(T, valid_record_def_inits(H,X));

valid_record_def_inits({tuple,_,ElemsL}, X) ->
    lists:foldl(fun valid_record_def_inits/2, X, ElemsL);

valid_record_def_inits({nil,_}, X) -> X;

valid_record_def_inits({record,_,_,FieldsInit}, X) ->
    lists:foldl(fun({record_field,_,_,Init},X1) ->
			valid_record_def_inits(Init,X1);
		   (_,_) -> 
			false
		end, X, FieldsInit);

%%valid_record_def_inits(A,X) -> io:format('~w, ~w\n', [A,X]), false;

valid_record_def_inits(_,_) -> false.

%%%----------------------------------------------------------------
mk_rules_def_fkn(S) ->
    Rules0 = lists:foldl(fun one_rule/3, [S], [], S#s.rules),
    case mnemosyne_compiler:pass2_rules(Rules0, 
					S#s.module, 
					S#s.recdefs) of
	{ok, CompiledRules} ->
	    {ok, [{function,0,'MNEMOSYNE RULE',1,
		   mk_fn_clauses(CompiledRules)++mk_data_clauses(S)}]};

	{error, L} ->
	    {error, L}
    end.


one_rule({Name,Line,Clauses}, Accu0, S0) ->
    lists:foldl(
      fun({clause,CLine,[Arg0],[],LCBody0}, Accu) ->
	      S = S0#s{db_vars=db_vars(LCBody0)},
	      Body0 = tr_body(LCBody0, S),
	      {Arg,Body} =
		  case tr_pattern(Arg0,S) of
		      {'#var',V} ->
			  {{'#var',V}, Body0};
		      NonVar ->
			  NewArg = mnemosyne_lib:unique_var(rule_arg),
			  {NewArg, [#constraint{exprL = NewArg, 
						op    = '=', 
						exprR = NonVar, 
						line  = CLine} | Body0]}
		  end,
	      OrigVars = mnemosyne_unify:variables_and_annonymous(Arg),
	      Head = #pred_sym{module       = S#s.module,
			       functor      = Name,
			       line	    = CLine,
			       type	    = rule,
			       record_def   = arg_record_def(rule,Name,S),
			       args	    = [Arg],
			       original_args_vars = OrigVars
			      },
	      [{rule,Head,Body} | Accu]
      end, Accu0, Clauses).



arg_record_def(What, Name, S) when record(S,s) ->
    arg_record_def(What, Name, S#s.argtypes, S#s.recdefs, S#s.line).

arg_record_def(rule, Name, ArgTypes, RecDefs, Line) ->
    arg_record_def(table,
		   case lists:keysearch(Name,1,ArgTypes) of
		       {Value,{Name,R_name}} -> R_name;
		       false -> Name
		   end, ArgTypes, RecDefs, Line);

arg_record_def(table, Name, ArgTypes, RecDefs, Line) ->
    arg_record_def(get_def, Name, ArgTypes, RecDefs, Line);

arg_record_def(get_def, Name, ArgTypes, RecDefs, Line) ->
    case lists:keysearch(Name,1,RecDefs) of
	{Value,Def} -> Def;
	false -> throw({error, {Line,?MODULE,{undefined,record,Name}}})
    end.



mk_data_clauses(S) ->
    lists:map(fun({Name,Val}) ->
		      {clause, 
		       0, 
		       [abstract_keep_vars({data,Name})], 
		       [], 
		       [abstract_keep_vars(Val)]}
	      end, [{record_defs,  S#s.recdefs},
		    {argtypes, S#s.argtypes}]).



mk_fn_clauses(Rs) -> mk_fn_clauses(Rs, '', [], []).


mk_fn_clauses([{rule,Head,Body}|Rs], Doing, Acc, BigAcc) 
					when Head#pred_sym.functor == Doing ->
    mk_fn_clauses(Rs, Doing, [{rule,Head,Body}|Acc], BigAcc);

mk_fn_clauses([{rule,Head,Body}|Rs], _, Acc, BigAcc) -> %% New functor
    mk_fn_clauses([{rule,Head,Body}|Rs], Head#pred_sym.functor, [], 
	       new_bigacc(Acc,BigAcc));

mk_fn_clauses([], _, Acc, BigAcc) ->
    new_bigacc(Acc,BigAcc).
    


new_bigacc([], BigAcc) -> 
    BigAcc;

new_bigacc(Acc,BigAcc) ->
    [{rule,Head,_}|_] = Acc,
    [{clause, 
      Head#pred_sym.line, 
      [abstract_keep_vars(Head#pred_sym.functor)], 
      [],
      [abstract_keep_vars(Acc)]} 
     | BigAcc].

%%%----------------------------------------------------------------

fn_clauses(Cs, S) ->
    lists:map(fun({clause,Line,H,G,B0}) ->
		      {Form, _} = 
			  replace_lcs(B0,S#s{line=Line, erl_vars=vars(H)}),
		      {clause,Line,H,G,Form}
	      end, Cs).


the_query({first_half, Data}) ->
    case catch the_query1(Data) of
	{'EXIT', Cause} ->
	    case Cause of
		{undef,{Module,'MNEMOSYNE RULE',[_]}} ->
		    throw({'EXIT', {mnemosyne, 
				    {no_rules_in_module,Module}}});
	
		{function_clause,{Module,'MNEMOSYNE RULE',[RuleName]}} ->
		    throw({'EXIT', {mnemosyne, 
				    {no_such_rule,Module,RuleName}}});

		Others ->
		    throw({'EXIT',Others})
	    end;
	Others ->
	    Others
    end;

the_query({second_half,Opt}) ->
    ?debugmsg(1, "Optimized call: ~s\n", [mnemosyne_pp:body(Opt)]),
    mnemosyne_exec:eval_collect(all, Opt,
				fun(Acc,P) ->
					[P|Acc]
				end).


the_query1({Pattern,Goal0,S}) ->
    ?debugmsg(2, 
	      "Calling mnemosyne_compiler:lc_query1 line ~w\n"
	      "   Pattern : ~s\n   Goal    : ~s\n", 
	      [S#s.line,
	       mnemosyne_pp:e(Pattern),
	       mnemosyne_pp:body(Goal0)]),
    Goal = runtime_resolve_tables(Goal0, S),
    Call = mnemosyne_compiler:lc_query1(Pattern, Goal, S#s.recdefs),
    ?debugmsg(1, "Call line ~w = ~s\n", [S#s.line,mnemosyne_pp:body(Call)]),
    Opt = mnemosyne_optimizer:phase2(Call),
    ?debugmsg(1, "Optimized call: ~s\n", [mnemosyne_pp:body(Opt)]),
    {call,Call,Opt}.


vars(X) -> vars(X,ordsets:new_set()).

vars({var,_,V}, S) -> ordsets:add_element(V,S);
vars({cons,_,H,T}, S) -> vars(T, vars(H,S));
vars({tuple,_,As}, S) -> vars(As, S);
vars(L, S) when list(L) -> lists:foldr(fun vars/2, S, L);
vars(T, S) when tuple(T) -> vars(tuple_to_list(T), S);
vars(_, S) -> S.


runtime_resolve_tables(Goal, S) ->
    lists:map(
      fun
	  (P) when record(P,pred_sym), P#pred_sym.record_def == ?UNKNOWN ->
	      P#pred_sym{record_def = 
			 arg_record_def(P#pred_sym.type,
					P#pred_sym.functor,
					mnemosyne_lib:db_data(
					  argtypes,  P#pred_sym.module),
					mnemosyne_lib:db_data(
					  record_defs, P#pred_sym.module),
					P#pred_sym.line)};
	  (Other) ->
	      Other
      end, Goal).




replace_lcs({'query',_,{lc,Line,E0,Qs0}}, S0) ->
    %% dynamically compiled query
    DbVars = db_vars(Qs0),
    S = S0#s{erl_vars = ordsets:subtract(S0#s.erl_vars, DbVars),
	     db_vars = DbVars},
    Pattern0 = tr_pattern(E0, S#s{line=Line}),
    Goal0 = tr_body(Qs0, S#s{line=Line}),
    {mk_call(?MODULE, the_query, [{first_half,{Pattern0,Goal0,S}}], Line), 
     S0};

replace_lcs({var,L,V}, S) ->
    {{var,L,V}, S#s{erl_vars=ordsets:add_element(V,S#s.erl_vars)}};

replace_lcs([H|T], S0) -> %% mapfoldl doesn't handle non-proper lists
    {H1,S1} = replace_lcs(H, S0),
    {T1,S2} = replace_lcs(T, S1),
    {[H1|T1], S2};

replace_lcs(T, S0) when tuple(T) -> 
    {F, S} = replace_lcs(tuple_to_list(T), S0),
    {list_to_tuple(F), S};

replace_lcs(F, S) ->
    {F, S}.


mk_call(Mod, Fun, Args, Line) ->
    Call = {call,Line,{remote,Line,
		       {atom,Line,Mod},
		       {atom,Line,Fun}},
	    lists:map(fun abstract_keep_vars/1, Args)},
    ?debugmsg(3, "mk_call = ~p\n", [Call]),
    Call.
    


abstract_keep_vars(X) ->
    element(1, abstract_keep_vars(X, {[],0})).


abstract_keep_vars({'#erl', X}, L) -> 
    {X, L};

abstract_keep_vars([H0|T0], L0) ->
    {H, L1} = abstract_keep_vars(H0, L0),
    {T, L} = abstract_keep_vars(T0, L1),
    {{cons,0,H,T}, L};

abstract_keep_vars(Tuple, L0) when tuple(Tuple) ->
    {T,L} = lists:mapfoldl(
	      fun(A,B)->
		      abstract_keep_vars(A,B) 
	      end, L0, tuple_to_list(Tuple)),
    {{tuple,0,T}, L};

abstract_keep_vars(R0, {L0,N0}) when reference(R0) ->
  case lists:keysearch(R0,1,L0) of
	{value,{R0,R}} ->
	    {erl_parse:abstract(R), {L0,N0}};
	false ->
	    {erl_parse:abstract(N0), {[{R0,N0}|L0], N0+1}}
    end;

abstract_keep_vars(X, L0) ->
    {erl_parse:abstract(X), L0}.



tr_body(L, S) when list(L) ->
    lists:map(
      fun
          %% DbVar <- Expr
	  ({generate,Line,Left, Right}) ->
	      generate_generator(Left, Right, S#s{line=Line});

	  %% Expr Op Expr
	  ({op,Line,Op,Left,Right}) ->
	      generate_constraint(Left, Right, Op, S#s{line=Line});
	  
	  %% Expr = Expr
	  ({match,Line,Left,Right}) ->
	      generate_constraint(Left, Right, '=', S#s{line=Line});
      
	  %% Fn(Args)
	  ({call,Line,B,Args}) ->
	      generate_constraint({atom,Line,true}, {call,Line,B,Args},'=',
				  S#s{line=Line});
	  (Others) ->
	      throw({error, {S#s.line,?MODULE,{illegal_lc_body,Others}}})
      end, L).


db_vars(L) ->
    lists:foldl(
      fun
	  ({generate,_,{var,_,V},_}, VarSet) -> ordsets:add_element(V, VarSet);
	  (_, VarSet) -> VarSet
      end, ordsets:new_set(), L).
      

generate_constraint(Left, Right, Op, S0) ->
    S = S0#s{allow_erl_vars=true,
	     allow_fun_calls=true},
    #constraint{exprL = expr(Left,S),
		exprR = expr(Right,S),
		op    = constr_op(Op,S#s.line),
		line  = S#s.line}.
    


generate_generator({var,VLL,VL}, {var,VRL,VR}, S) ->
    #erl_expr{alias_var = expr_var({var,VLL,VL}, S#s{allow_erl_vars=true,
						     line=VLL}),
	      expr = expr_var({var,VRL,VR}, S#s{allow_erl_vars=true,
						allow_db_vars=false,
						line=VRL})};

generate_generator({var,VLL,VL}, Right, S) ->
    Var = expr_var({var,VLL,VL}, S#s{line=VLL}),
    case Right of
	%% Var <- table(Name)
	{call,_,{atom,_,table},[{atom,LineP,PredName}]} ->
	    Args = [Var],
	    #pred_sym{module       = S#s.module,
		      functor	 = PredName,
		      line	 = LineP,
		      type	 = table,
		      record_def = arg_record_def(table,PredName,S),
		      recursive	 = non_recursive,
		      args	 = Args,
		      original_args_vars = 
		             mnemosyne_unify:variables_and_annonymous(Args)
		     };
	
	%% Var <- table(Expr)
	{call,_,{atom,_,table},[{var,VTL,VT}]} ->
	    Args = [Var],
	    #pred_sym{module       = S#s.module,
		      functor = {'#erl', {var,VTL,VT}},
		      line = VTL,
		      type = table,
		      recursive = non_recursive,
		      args = Args,
		      original_args_vars = 
		             mnemosyne_unify:variables_and_annonymous(Args)
		     };
	
	%% Var <- table(Others)  ** Illegal
	{call,Line,{atom,_,table}, _} ->
	    throw({error, {Line,?MODULE,{illegal_lc_generator,right,Right}}});

	%% Var <- rule(Name)  (i.e. locally defined rule)
	{call,_,{atom,_,rule}, [{atom,LineP,PredName}]}->
	    Args = [Var],
	    #pred_sym{module = S#s.module,
		      functor = PredName,
		      line = LineP,
		      type = rule,
		      record_def = arg_record_def(rule,PredName,S),
		      recursive = non_recursive, % Hypothesis
		      args = Args,
		      original_args_vars = 
		             mnemosyne_unify:variables_and_annonymous(Args)
		     };

	%% Var <- rule(Mod:Name)
	{call,_,{atom,_,rule}, [{remote,LineP,{atom,_,Mod},{atom,_,PredName}}]}->
	    Args = [Var],
	    #pred_sym{module = Mod,
		      functor = PredName,
		      line = LineP,
		      type = rule,
		      record_def = if 
				       Mod == S#s.module -> 
					   arg_record_def(rule,PredName,S);
				       true ->
					   ?UNKNOWN
				   end,
		      recursive = non_recursiv, % Hypothesis
		      args = Args,
		      original_args_vars = 
		             mnemosyne_unify:variables_and_annonymous(Args)
		     };

	%% Var <- rule(Others)  ** Illegal
	{call,Line,{atom,_,rule}, _}->
	    throw({error, {Line,?MODULE,{illegal_lc_generator,right,Right}}});

	%% Var <- Name(Args)
	{call,FL,F,Args} ->
	    #erl_expr{alias_var = Var,
		      expr = expr(Right,
				  S#s{allow_erl_vars = true,
				      allow_fun_calls = true})};
	%% Var <- List
	{cons,Line,H,T} ->
	    #erl_expr{alias_var = Var,
		      expr = expr(Right,
				  S#s{allow_erl_vars = true,
				      allow_fun_calls=true})};
	X ->
	    throw({error, 
		   {S#s.line,?MODULE, {illegal_lc_generator,right,X}}})
    end;

generate_generator({op,NL,'not',{var,VLL,VL}}, Right, S) ->
    {'#not', [generate_generator({var,VLL,VL}, Right, S#s{line=NL})]};

generate_generator(Left, _, S) ->
    throw({error, {S#s.line,?MODULE, {illegal_lc_generator,left,Left}}}).
    


%%% The <pattern> part in a list comprehension
tr_pattern(P, S) ->
    expr(P, S#s{allow_erl_vars=true,
		allow_fun_calls=false}).

%%% data without function calls like {a,X,[#r{},1]}. Not: NOT 1+4.

expr({record_field,Line,{var,L1,Var},Name,{atom,L3,Field}},
     S)->
    %% X#r.a
    case expr_var(L1,Var,S) of
	{'#erl', V} ->
	    {'#erl', 
	     {record_field,Line,V,Name,{atom,L3,Field}}};
	V ->
	    #rec_f{var = V,
		   name = Name,
		   line = Line,
		   field = Field}
    end;	    

expr({record_field,Line,{var,L1,Var},{atom,L2,Field}}, S) ->
    %% X.a
    case expr_var(L1,Var,S) of
	{'#erl', V} ->
	    {'#erl', {record_field,Line,V,{atom,L2,Field}}};

	V ->
	    #rec_f{var = V,
		   line = Line,
		   field = Field}
    end;	    

expr({record,Line,Name,Fields}, S) ->
    %% #r{a=..., ...}
    Fs = lists:map(fun({record_field,Lf,{atom,_,Nf},Vf}) ->
			   {Nf,Lf,expr(Vf,S#s{line=Lf})};
		       (Other) ->
			   throw({error,
				  {Line,?MODULE,{illegal_field,Name}}})
		   end, Fields),
    #rec_c{line=Line,
	   name=Name,
	   fields=Fs};

expr({var,Line,Name}, S) -> expr_var(Line, Name, S);
expr({integer,_,I}, S) -> I;
expr({float,_,F}, S) -> F;
expr({atom,_,A}, S) -> 
    case atom_to_list(A) of
	[$#|T] -> list_to_atom([$#,$#|T]);
	_ -> A
    end;
expr({string,_,Str}, S) -> Str;
expr({nil,_}, S) -> [];

expr({cons,Line,H0,T0}, S0) -> 
    S = S0#s{line=Line},
    [expr(H0,S) | expr(T0,S)];

expr({tuple,Line,Es0}, S0) -> 
    S = S0#s{line=Line},
    list_to_tuple( lists:map(fun (E) ->
				     expr(E,S)
			     end,
			     Es0) );
expr({call,Line,B,Args}, S0) when S0#s.allow_fun_calls==true ->
    Exp = {call,Line,B,Args},
    case B of
	{remote, _, M, F} -> f_call(M,F,Args,S0#s{line=Line});
	{atom,_,F} -> f_call({atom,0,S0#s.module},B,Args,S0#s{line=Line});
	{var,_,F} -> f_call({atom,0,S0#s.module},B,Args,S0#s{line=Line});
	_ ->
	    ?debugmsg(2, "B=~w\n", [B]),
	    throw({error, {S0#s.line,?MODULE,{illegal_expr,Exp}}})
    end;

expr(X, S) ->
    ?debugmsg(2, "~p\n~p\n", [X,S]),
    throw({error, {S#s.line,?MODULE,{illegal_expr,X}}}).



expr_var({var,Line,Name}, S) -> expr_var(Line,Name,S).

expr_var(Line, Name, S) ->
    case ordsets:is_element(Name,S#s.erl_vars) of
	true when S#s.allow_erl_vars==true ->
	    {'#erl',{var,Line,Name}};
	true when S#s.allow_erl_vars==false ->
	    throw({error, {Line,?MODULE,{no_erl_var,Name}}});
	false ->
{'#var',Name}
%	    case ordsets:is_element(Name,S#s.db_vars) of
%		true -> {'#var',Name};
%		false -> throw({error, {Line,?MODULE,{unbound_var,Name}}})
%	    end
    end.


f_call(M, F, Args0, S0) ->
    C = {'#funcall', 
	 expr(M,S0),
	 expr(F,S0),
	 lists:map(fun (A) ->
			   expr(A, S0)
		   end, Args0)
	},
    ?debugmsg(2, "~p\n", [C]),
    C.



constr_op('=',_) -> '=';
constr_op('/=',_) -> '!=';
constr_op('<',_) -> '<';
constr_op('>',_) -> '>';
constr_op('=<',_) -> '=<';
constr_op('>=',_) -> '>=';
constr_op(Op, Line) -> throw({error, {Line,?MODULE,{illegal_op,Op}}}).
