%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1996, Ellemtel Telecommunications Systems Laboratories
%% File    : op.erl
%% Author  : Hans Nilsson
%% Purpose : 
%% Created : Thu Jan 18 17:43:18 1996


-module(mnemosyne_op).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-export([bindings/2, funcall/3, erl_expr/2, split/1, table/2, negation/2,
	 negation_help/0, call_recursive_op/2,
	 split_list/2, get_from_table/3, new_end_token/1, start_op/2
	]).

%%-define(debug,3).
-define(no_not_yet, yes).

-include("mnemosyne_debug.hrl").
-include("mnemosyne_internal_form.hrl").

-record(acc, {sols=[], bss=[]}).

-ifdef(debug).
-define(trace_printout(Type,Args,Where),
	case Type of
	    negation_help when Where==1 ->
		wait_a_bit;
	    _ ->
		?debugmsg(1, "~w (~w) -> ~w. Args: ~s\n", 
			  [self(), Type, hd(Args),
			   type_dependent_args(Type, Args)])
	end).

type_dependent_args(bindings, [_,Bs]) ->
    [" ",mnemosyne_pp:e({'#bindings',Bs})];

type_dependent_args(table, [_,P]) when ?debug==2 ->
    [" line=",mnemosyne_pp:e({'#line',P#pred_sym.line}),
     " p=",mnemosyne_pp:e(P#pred_sym.pattern),
     " bp=",mnemosyne_pp:e(P#pred_sym.back_pattern),
     " idx=",io_lib:write(P#pred_sym.idx_method)
    ];

type_dependent_args(table, [_,P]) when ?debug>2 ->
    [" line=",mnemosyne_pp:e({'#line',P#pred_sym.line}),
     " goal=",mnemosyne_pp:e(P),
     " p=",mnemosyne_pp:e(P#pred_sym.pattern),
     " bp=",mnemosyne_pp:e(P#pred_sym.back_pattern),
     " idx=",io_lib:write(P#pred_sym.idx_method),
     " defvars=",io_lib:write(P#pred_sym.defvars)
    ];

type_dependent_args(negation, [_,HelpPid]) ->
    [" help=",io_lib:write(HelpPid)
    ];

type_dependent_args(negation_help, [Qpid,MainPid]) ->
    [" QueryPid=",io_lib:write(Qpid),
     " MainPid=",io_lib:write(MainPid)
    ];

type_dependent_args(Type, Args) ->
    [" ",mnemosyne_pp:e(Args)].

-else.
-define(trace_printout(Type,Args,Where),ok).
-endif.

%%%================================================================
%%% 		Exports

start_op(Type, Args) ->
    ?trace_printout(Type, Args, 1),
    %% The Type(Args) call never returns unless there is an exception:
    case catch apply(?MODULE, Type, Args) of
	{'EXIT',Cause} -> exit(Cause);
	Others -> exit({throw,Others})
    end.


new_end_token(N) ->
    {{make_ref(),erlang:now()}, N}.


%%%----------------------------------------------------------------
%%%---- Adds bindings to a set of bindings

bindings(NextPid, Bs) ->
    bindings(NextPid, Bs, Bs).

bindings(NextPid, Bs, BsOrig) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} when Bss =/= [] ->
	    NextPid ! {bss, bss_union(Bss,Bs), Max, EndCntrl, Stack},
	    bindings(NextPid, [], BsOrig);

	{bss, _, Max, EndCntrl, Stack} ->
	    NextPid ! {bss, [], Max, EndCntrl, Stack},
	    bindings(NextPid, [], BsOrig);
	
	{fail, SomePid, Cause} ->
	    NextPid ! {fail, self(), Cause},
	    bindings(NextPid, BsOrig);

	{tid, Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid},
	    bindings(NextPid, BsOrig)
    end.



%%%----------------------------------------------------------------
%%%---- Call an Erlang function

funcall(NextPid, LeftSide, FunCall) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} ->
	    case eval_funcall(Bss,LeftSide,FunCall,[]) of
		{fail, Cause} ->
		    NextPid ! {fail, self(), Cause};
		{ok, NewBss} ->
		    NextPid ! {bss, NewBss, Max, EndCntrl, Stack}
	    end,
	    funcall(NextPid, LeftSide, FunCall);
	
	{fail, SomePid, Cause} ->
	    NextPid ! {fail, self(), Cause},
	    funcall(NextPid, LeftSide, FunCall);

	{tid, Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid},
	    funcall(NextPid, LeftSide, FunCall)
    end.


	
eval_funcall([Bs|Bss], LeftSide, FunCall, Acc) ->
    {'#funcall',M,F,Args} = mnemosyne_unify:instantiate(FunCall,Bs),
    Ref = make_ref(),
    case catch {Ref, apply(M,F,Args)} of
	{Ref, Res} ->
	    case mnemosyne_unify:unify(LeftSide, apply(M,F,Args), Bs) of
		fail ->
		    eval_funcall(Bss, LeftSide, FunCall, Acc);
		NewBs ->
		    eval_funcall(Bss, LeftSide, FunCall, [NewBs|Acc])
	    end;
	{'EXIT', Cause} ->
	    {fail, Cause}; 
	Other ->
	    {fail, {throw, Other}}
    end;
    
eval_funcall([], LeftSide, FunCall, Acc) ->
    {ok, Acc}.
    

%%%----------------------------------------------------------------
%%%---- 

erl_expr(NextPid, Expr) ->
    erl_expr(NextPid, Expr, []).

erl_expr(NextPid, Expr, Sols) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} ->
	    {First, Last} = 
		split_list(Max, get_values(Bss,Expr,Sols)),
	    NextPid ! {bss, First, Max, EndCntrl, Stack},
	    erl_expr(NextPid, Expr, Last);
	
	{fail, SomePid, Cause} ->
	    NextPid ! {fail, self(), Cause},
	    erl_expr(NextPid, Expr);

	{tid, Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid},
	    erl_expr(NextPid, Expr)
    end.

get_values(Bss, Expr, Acc0) ->
    lists:foldl(
      fun(Bs,Acc) ->
	      ?debugmsg(2,"~w: Value = ~w\n",
			[self(),
			 mnemosyne_unify:instantiate(Expr#erl_expr.expr,Bs)]),
%	      case mnemosyne_unify:unify(
%		     mnemosyne_unify:instantiate(Expr#erl_expr.expr,Bs),
%		     Expr#erl_expr.alias_var,
%		     Bs) of
%		  fail -> Acc;
%		  NewBs -> [NewBs|Acc]
%	      end
	      unify_list(mnemosyne_unify:instantiate(Expr#erl_expr.expr,Bs),
			 Expr#erl_expr.alias_var, Bs, Acc)
      end, Acc0, Bss).

%%%----------------------------------------------------------------
%%%---- Multiplexes a message to many receivers

split(NextPids) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} ->
	    NewEnd = new_end_token(length(NextPids)),
	    hd(NextPids) ! {bss, Bss, Max, [NewEnd|EndCntrl], Stack},
	    send(tl(NextPids), {bss, Bss, Max, [NewEnd], Stack});
	
	{fail, SomePid, Cause} ->
	    send(NextPids, {fail, self(), Cause});
	
	{tid, Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    send(NextPids, {tid,Tid})
    end,
    split(NextPids).


%%%----------------------------------------------------------------
%%%---- Table lookup

table(NextPid, P) when record(P,pred_sym) ->
    table_loop(NextPid, P, #acc{}).


%%%----------------------------------------------------------------
%%%---- Negation

negation(NextPid, HelpPid) ->
    negation(NextPid, HelpPid, []).


negation_help() ->
    receive
	{pids, Buddie, NextPid} ->
	    ?trace_printout(negation_help, [NextPid,Buddie], 2),
	    negation_help(NextPid,Buddie,[])
    end.

%%%----------------------------------------------------------------
%%---- Recursion
%% A first solution. For each binding, get ALL solutions from SLG and
%% THEN pass them on to the DestPid.

call_recursive_op(NextPid, Goal) ->
    call_recursive_op(NextPid, Goal, [], []).

call_recursive_op(NextPid, Goal, BssAcc, Acc) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} ->
	    {BssAcc1, SolutionList} =
		get_from_slg(Max, Acc, lists:append(BssAcc,Bss), Goal),
	    {First,Acc1} =
		split_list(Max, SolutionList),
	    NextPid ! {bss, First, Max, EndCntrl, Stack},
	    call_recursive_op(NextPid, Goal, BssAcc1, Acc1);

	{fail, SomePid, Cause} ->
	    NextPid ! {fail, self(), Cause},
	    call_recursive_op(NextPid, Goal);

	{tid, Tid} ->
	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid},
	    call_recursive_op(NextPid, Goal)
    end.

    

get_from_slg(Max, Acc, [Bs|Bss], Goal) when length(Acc)<Max ->
    GoalInst = mnemosyne_unify:instantiate(Goal,Bs),
    mnemosyne_slg:slg_init(),
    Acc1 = make_bindings(mnemosyne_slg:slg(GoalInst,none), Goal, Bs, Acc),
    get_from_slg(Max, Acc1, Bss, Goal);
get_from_slg(Max, Acc, Bss, Goal) ->
    {Bss, Acc}.


make_bindings([Answ|Answs], Goal, Bs, Acc) ->
    case mnemosyne_unify:unify(Answ,Goal,Bs) of
	fail -> make_bindings(Answs, Goal, Bs, Acc);
	NewBs -> make_bindings(Answs, Goal, Bs, [NewBs|Acc])
    end;
make_bindings([], Goal, Bs, Acc) ->
    Acc.


%%%================================================================
%%% 		Private

table_loop(NextPid, F, Acc) ->
    receive
	{bss, Bss, Max, EndCntrl, Stack} ->
?debugmsg(3,"~w got ~w\n", [self(), {bss, Bss, Max, EndCntrl, Stack}]),
	    List = lists:append(Acc#acc.bss,Bss),
	    case catch fact_produce(Max, Acc#acc{bss = List}, F) of
		{'EXIT', Cause} ->
		    NextPid ! {fail, self(), Cause},
		    table(NextPid, F);
		Acc1 ->
		    {First, Sols} = split_list(Max, Acc1#acc.sols),
		    NextPid ! {bss, First, Max, EndCntrl, Stack},
		    ?debugmsg(3,"~w: ~w ! ~w\n", 
			      [self(),NextPid, {bss, First, Max, EndCntrl, Stack}]),
		    table_loop(NextPid, F, Acc1#acc{sols = Sols})
	    end;
	
	{fail, SomePid, Cause} ->
	    NextPid ! {fail, self(), Cause},
	    table(NextPid, F);

	{tid, Tid} ->
	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid},
	    table(NextPid, F)	    
    end.


%%%----------------
fact_produce(Max, Acc, F) ->
    {Sols,Bss} = fact_produce(Max, Acc#acc.sols, Acc#acc.bss, F),
    Acc#acc{bss=Bss, sols=Sols}.
    

fact_produce(Max, Sols, [Bs|Bss], F) when length(Sols)<Max ->
    fact_produce(Max, get_from_table(Bs,F,Sols), Bss, F);
fact_produce(Max, Sols, Bss, F) ->
    {Sols,Bss}.


get_from_table(Bs, F, Sols) ->
    InstPattern = mnemosyne_unify:instantiate(F#pred_sym.pattern, Bs),
    Matches = 
	case F#pred_sym.idx_method of
	    {dyn_idx,Pos} ->
		?not_yet('get_from_table: dyn_idx'),
		mnesia:match_object(InstPattern);
	    
	    {stat_idx,Pos} ->
		mnesia:index_match_object(InstPattern, Pos);

	    key ->
		mnesia:match_object(InstPattern);

	    no_idx ->				% Occurs if no statistc optim
		mnesia:match_object(InstPattern)
	end,
    NewBss = unify_list(Matches, F#pred_sym.back_pattern, Bs, Sols),
    ?debugmsg(4, "NewBss = ~s\n", 
	      [mnemosyne_pp:e(lists:map(fun(B)->{'#bindings',B} end,NewBss))]),
    NewBss.


%%%----------------------------------------------------------------
negation(NextPid, HelpPid, Acc) ->
    receive
	%%---- From the ordinary flow
	{bss, Bss, Max, EndCntrl, Stack} ->
	    HelpPid ! {neg_bss_req, Bss, Max, EndCntrl, Stack},
	    negation(NextPid, HelpPid, Acc);
		
	{fail, SomePid, Cause} ->
	    HelpPid ! {fail, self(), Cause},
	    NextPid ! {fail, self(), Cause},
	    negation(NextPid, HelpPid);
	
	{tid, Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    HelpPid ! {tid_req,Tid},
	    NextPid ! {tid,Tid},
	    negation(NextPid, HelpPid);

	%%---- From the help process
	{neg_bss_repl, Bss, Max, EndCntrl, Stack} ->
	    {First,Last} = split_list(Max, lists:append(Acc,Bss)),
	    NextPid ! {bss, First, Max, EndCntrl, Stack},
	    negation(NextPid, HelpPid, Last)
	
    end.

    
negation_help(NextPid, Buddie, Acc0) ->
    receive
	%% Requests from the negation process ("Buddie")
	{neg_bss_req, Bss, Max, EndCntrl, Stack} ->
	    eval_neg_each_binding(Bss, NextPid, [{params,Max,EndCntrl}|Stack]),
	    negation_help(NextPid, Buddie, Acc0);
	
	{tid_req,Tid} ->
%%	    mnesia:put_activity_id(Tid),
	    NextPid ! {tid,Tid}, 
	    negation_help(NextPid, Buddie, []);

	%% From the ordinary flow, that is, from the negated part
	{bss, Bss, all, [], [SendData,{params,Max,EndCntrl}|Stack]} ->
	    case SendData of
		last ->
		    Buddie ! {neg_bss_repl, Acc0, Max, EndCntrl, Stack},
		    negation_help(NextPid, Buddie, []);
		{bs,BsOrig} ->
		    case Bss of
			[] ->
			    negation_help(NextPid,Buddie,[BsOrig|Acc0]);
			[_|_] ->
			    negation_help(NextPid,Buddie,Acc0)
		    end
	    end;
	
	{fail, SomePid, Cause} ->
	    negation_help(NextPid, Buddie, Acc0);
	
	{tid, Tid} ->
	    negation_help(NextPid, Buddie, Acc0)
    end.
    


eval_neg_each_binding([Bs|Bss], NextPid, Stack) ->
    NextPid ! {bss, [Bs], all, [], [{bs,Bs}|Stack]},
    eval_neg_each_binding(Bss, NextPid, Stack);
eval_neg_each_binding([], NextPid, Stack) ->
    NextPid ! {bss, [], all, [], [last|Stack]}.

%%%----------------------------------------------------------------

bss_union(Bss, Bs0) ->
    bss_union(Bss, Bs0, []).

bss_union([Bs1|Bss], Bs0, Acc) ->
    case catch mnemosyne_unify:bs_union(Bs0,Bs1) of
	fail -> bss_union(Bss, Bs0, Acc);
	Bs -> bss_union(Bss, Bs0, [Bs|Acc])
    end;
bss_union([], _, Acc) ->
    Acc.

%%%----------------------------------------------------------------
send([Pid|Pids], Msg) when pid(Pid) ->	Pid ! Msg, send(Pids,Msg);
send([], _) -> ok;
send(Pid, Msg) when pid(Pid) -> Pid ! Msg.

%%%----
split_list(all, L) -> {L,[]};
split_list(N, L) when length(L)>N -> split_list(N, L, []);
split_list(_, L) -> {L,[]}.

split_list(N, [H|T], Acc) when N>0 -> split_list(N-1, T, [H|Acc]);
split_list(N, L, Acc) -> {Acc,L}.


%%%----
unify_list(Ms, Pattern, Bs, Sols) ->
    lists:foldl(
      fun (M,Acc) ->
	      case mnemosyne_unify:unify(Pattern, M, Bs) of
		  fail -> Acc; 
		  NewBs -> [NewBs|Acc]
	      end
      end, Sols, Ms).

