%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% MODULE
%%
%%   mnesia_tpcb - TPC-B benchmarking of Mnesia
%%
%% DESCRIPTION
%%
%%   The metrics used in the TPC-B benchmark are throughput as measured
%%   in transactions per second (TPS). The benchmark uses a single,
%%   simple update-intensive transaction to load the database system.
%%   The single transaction type provides a simple, repeatable
%%   unit of work, and is designed to exercise the basic components of
%%   a database system.
%%
%%   The definition of the TPC-B states lots of detailed rules and
%%   conditions that must be fullfilled, e.g. how the ACID (atomicity,
%%   consistency, isolation and durability) properties are verified,
%%   how the random numbers must be distributed, minimum sizes of
%%   the different types of records, minimum duration of the benchmark,
%%   formulas to calculate prices (dollars per tps), disclosure issues
%%   etc. Please, see http://www.tpc.org/ about the nitty gritty details.
%%
%%   The TPC-B benchmark is stated in terms of a hypothetical bank. The
%%   bank has one or more branches. Each branch has multiple tellers. The
%%   bank has many customers, each with an account. The database represents
%%   the cash position of each entity (branch, teller and account) and a
%%   history of recent transactions run by the bank. The transaction
%%   represents the work done when a customer makes a deposit or a
%%   withdrawal against his account. The transaction is performed by a
%%   teller at some branch.
%%
%%   Each process that performs TPC-B transactions is called a driver.
%%   Drivers generates teller_id, account_id and delta amount of
%%   money randomly. An account, a teller and a branch are read, their
%%   balances are adjusted and a history record is created. The driver
%%   measures the time for 3 reads, 3 writes and 1 create.
%%
%% GETTING STARTED
%%
%%   Generate tables and run with default configuration:
%%
%%     mnesia_tpcb:start().
%%
%%  A little bit more advanced;
%%
%%     spawn(mnesia_tpcb, start, [[[{n_drivers_per_node, 8}, {stop_after, infinity}]]),
%%     mnesia_tpcb:stop().
%%
%%  Really advanced;
%%
%%    mnesia_tpcb:init(([{n_branches, 8}, {replica_type, disc_only_copies}]),
%%    mnesia_tpcb:run(([{n_drivers_per_node, 8}]),
%%    mnesia_tpcb:run(([{n_drivers_per_node, 64}]).
%%
%% AUTHOR
%%
%%   Hkan Mattsson, hakan@erix.ericsson.se
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(mnesia_tpcb).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('hakan@erix.ericsson.se').

-export([
	 config/2,
	 conflict_test/1,
	 count_balance/0,
	 dist_test/1,
	 driver_init/1,
	 init/1,
	 replica_test/1,
	 reporter_init/2,
	 run/1,
	 start/0,
	 start/1,
	 start/2,
	 sticky_replica_test/1,
	 stop/0,
	 verify_tabs/0
	]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Account record, total size must be at least 100 bytes

-define(ACCOUNT_FILLER,
	{123456789012345678901234567890123456789012345678901234567890,
	 123456789012345678901234567890123456789012345678901234567890,
	 123456789012345678901234567890123456789012345678901234}).

-record(account,
       {
	id           = 0, % Unique account id
	branch_id    = 0, % Branch where the account is held
	balance      = 0, % Account balance
	filler       = ?ACCOUNT_FILLER  % Gap filler to ensure size >= 100 bytes
       }).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Branch record, total size must be at least 100 bytes

-define(BRANCH_FILLER,
	{123456789012345678901234567890123456789012345678901234567890,
	 123456789012345678901234567890123456789012345678901234567890,
	 123456789012345678901234567890123456789012345678901234567890}).

-record(branch,
       {
	id           = 0, % Unique branch id
	balance      = 0, % Total balance of whole branch
	filler       = ?BRANCH_FILLER  % Gap filler to ensure size >= 100 bytes
       }).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Teller record, total size must be at least 100 bytes

-define(TELLER_FILLER,
	{123456789012345678901234567890123456789012345678901234567890,
	 123456789012345678901234567890123456789012345678901234567890,
	 1234567890123456789012345678901234567890123456789012345678}).

-record(teller,
       {
	id           = 0, % Unique teller id
	branch_id    = 0, % Branch where the teller is located
	balance      = 0, % Teller balance
	filler       = ?TELLER_FILLER % Gap filler to ensure size >= 100 bytes
       }).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% History record, total size must be at least 50 bytes

-define(HISTORY_FILLER, 1234567890).

-record(history,
	{
	 history_id  = {0, 0}, % {DriverId, DriverLocalHistoryid}
	 time_stamp  = now(), % Time point during active transaction
	 branch_id   = 0, % Branch associated with teller
	 teller_id   = 0, % Teller invlolved in transaction
	 account_id  = 0, % Account updated by transaction
	 amount      = 0, % Amount (delta) specified by transaction
	 filler      = ?HISTORY_FILLER % Gap filler to ensure size >= 50 bytes
       }).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-record(tab_config,
	{
	 db_nodes = [node()],
	 replica_nodes = [node()],
	 replica_type = ram_copies,
	 use_running_mnesia = false,
	 n_branches = 1,
	 n_tellers_per_branch = 10, % Must be 10
	 n_accounts_per_branch = 100000, % Must be 100000
	 branch_filler = ?BRANCH_FILLER,
	 account_filler = ?ACCOUNT_FILLER,
	 teller_filler = ?TELLER_FILLER
	}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-record(run_config,
	{
	 driver_nodes = [node()],
	 n_drivers_per_node = 1,
	 use_running_mnesia = false,
	 stop_after = timer:minutes(15), % Minimum 15 min
	 report_interval = timer:minutes(1),
	 use_sticky_locks  = false,
	 reuse_history_id = false
	}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-record(time,
	{
	 n_trans = 0,
	 min_n = 0,
	 max_n = 0,
	 acc_time = 0,
	 max_time = 0
	}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-record(driver_state,
	{
	 driver_id,
	 seed,
	 tab_config,
	 run_config,
	 history_id,
	 time = #time{},
	 acc_time = #time{},
	 reuse_history_id
	}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-record(reporter_state,
	{
	 driver_pids,
	 starter_pid,
	 n_iters = 0,
	 prev_tps = 0,
	 curr = #time{},
	 acc = #time{},
	 run_config
	}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% One driver on this node, table replicated to all nodes.
config(replica_test, ReplicaType) ->
    Nodes = [node() | nodes()],
    [{db_nodes, Nodes},
     {driver_nodes, [node()]},
     {replica_nodes, Nodes},
     {n_accounts_per_branch, 100},
     {replica_type, ReplicaType},
     {stop_after, timer:minutes(1)},
     {report_interval, timer:seconds(10)},
     {reuse_history_id, true}];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% One driver on this node, table replicated to all nodes.
config(sticky_replica_test, ReplicaType) ->
    Nodes = [node()|nodes()],
    [{db_nodes, Nodes},
     {driver_nodes, [node()]},
     {replica_nodes, Nodes},
     {n_accounts_per_branch, 100},
     {replica_type, ReplicaType},
     {use_sticky_locks, true},
     {stop_after, timer:minutes(1)},
     {report_interval, timer:seconds(10)},
     {reuse_history_id, true}];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Ten drivers per node, tables replicated to all nodes, lots of branches
config(dist_test, ReplicaType) ->
    Nodes = [node()|nodes()],
    [{db_nodes, Nodes},
     {driver_nodes, Nodes},
     {replica_nodes, Nodes},
     {n_drivers_per_node, 10},
     {n_branches, 10 * length(Nodes) * 100},
     {n_accounts_per_branch, 10},
     {replica_type, ReplicaType},
     {stop_after, timer:minutes(1)},
     {report_interval, timer:seconds(10)},
     {reuse_history_id, true}];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Ten drivers per node, tables replicated to all nodes, single branch
config(conflict_test, ReplicaType) ->
    Nodes = [node()|nodes()],
    [{db_nodes, Nodes},
     {driver_nodes, Nodes},
     {replica_nodes, Nodes},
     {n_drivers_per_node, 10},
     {n_branches, 1},
     {n_accounts_per_branch, 10},
     {replica_type, ReplicaType},
     {stop_after, timer:minutes(1)},
     {report_interval, timer:seconds(10)},
     {reuse_history_id, true}].

start(What, ReplicaType) ->
    spawn_link(?MODULE, start, [config(What, ReplicaType)]).

replica_test(ReplicaType) ->
    start(replica_test, ReplicaType).

sticky_replica_test(ReplicaType) ->
    start(sticky_replica_test, ReplicaType).

dist_test(ReplicaType) ->
    start(dist_test, ReplicaType).

conflict_test(ReplicaType) ->
    start(conflict_test, ReplicaType).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Args is a list of {Key, Val} tuples where Key is a field name
%% in either the record tab_config or run_config. Unknown keys are ignored.
start() ->
    start([]).
start(Args) ->
    init(Args),
    run(Args).

list2rec(List, Fields, DefaultTuple) ->
    [Name|Defaults] = tuple_to_list(DefaultTuple),
    List2 = list2rec(List, Fields, Defaults, []),
    list_to_tuple([Name] ++ List2).

list2rec(List, [], [], Acc) ->
    Acc;
list2rec(List, [F|Fields], [D|Defaults], Acc) ->
    {Val, List2} =
	case lists:keysearch(F, 1, List) of
	    false ->
		{D, List};
	    {value, {F, NewVal}} ->
		{NewVal, lists:keydelete(F, 1, List)}
	end,
    list2rec(List2, Fields, Defaults, Acc ++ [Val]).

stop() ->
    case whereis(mnesia_tpcb) of
	undefined ->
	    {error, not_running};
	Pid ->
	    sync_stop(Pid)
    end.

sync_stop(Pid) ->
    Pid ! {self(), stop},
    receive
	{Pid, {stopped, Res}} ->
	    Res
    after timer:minutes(1) ->
	    exit(Pid, kill),
	    {error, brutal_kill}
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Initialization

%% Args is a list of {Key, Val} tuples where Key is a field name
%% in the record tab_config, unknown keys are ignored.
init(Args) ->
    TabConfig = list2rec(Args, record_info(fields, tab_config), #tab_config{}),
    Tags = record_info(fields, tab_config),
    Fun = fun(F, Pos) -> {{F, element(Pos, TabConfig)}, Pos + 1} end,
    {List, _} = lists:mapfoldl(Fun, 2, Tags),
    io:format("TPC-B: Table config: ~p ~n", [List]),

    DbNodes = TabConfig#tab_config.db_nodes,
    stop(),
    if
	TabConfig#tab_config.use_running_mnesia == true ->
	    ignore;
	true ->
	    rpc:multicall(DbNodes, mnesia, lkill, []),
	    case mnesia:delete_schema(DbNodes) of
		ok ->
		    case mnesia:create_schema(DbNodes) of
			ok ->
			    {Replies, BadNodes} =
				rpc:multicall(DbNodes, mnesia, start, []),
			    case [Res || Res <- Replies, Res /= ok] of
				[] when BadNodes == [] ->
				    ok;
				BadRes ->
				    io:format("TPC-B: <ERROR> "
					      "Failed to start ~p: ~p~n",
					      [BadNodes, BadRes]),
				    exit({start_failed, BadRes, BadNodes})
			    end;
			{error, Reason} ->
			    io:format("TPC-B: <ERROR> "
				      "Failed to create schema on disc: ~p~n",
				      [Reason]),
			    exit({create_schema_failed, Reason})
		    end;
		{error, Reason} ->
		    io:format("TPC-B: <ERROR> "
			      "Failed to delete schema on disc: ~p~n",
			      [Reason]),
		    exit({delete_schema_failed, Reason})
	    end
    end,
    gen_tabs(TabConfig).

gen_tabs(TC) ->
    Tabs = [branch, teller, account, history],
    Nodes = TC#tab_config.replica_nodes,
    
    Type = TC#tab_config.replica_type,
    Tail = [{Type, Nodes}, {type, set}],
    Create = fun(Name, Args) ->
		     AllArgs = Args ++ Tail,
		     mnesia:delete_table(Name),
		     case mnesia:create_table(Name, AllArgs) of
			 {atomic, ok} ->
			     ok;
			 {aborted, Reason} ->
			     io:format("TPC-B: <ERROR> "
				       "failed to create table ~w ~w: ~p~n",
				       [Name, AllArgs, Reason])
		     end
	     end,
    Create(branch, [{attributes, record_info(fields, branch)}]),
    Create(account, [{attributes, record_info(fields, account)}]),
    Create(teller, [{attributes, record_info(fields, teller)}]),
    Create(history, [{attributes, record_info(fields, history)}]),
    verify_tabs(),

    NB = TC#tab_config.n_branches,
    NT = TC#tab_config.n_tellers_per_branch,
    NA = TC#tab_config.n_accounts_per_branch,
    B = #branch{filler = TC#tab_config.branch_filler},
    A = #account{filler = TC#tab_config.account_filler},
    T = #teller{filler = TC#tab_config.teller_filler},
    H = #history{},
    io:format("TPC-B: Generating ~p branches a ~p bytes~n",
	      [NB, size(term_to_binary(B))]),
    io:format("TPC-B: Generating ~p * ~p tellers a ~p bytes~n",
	      [NB, NT, size(term_to_binary(T))]),
    io:format("TPC-B: Generating ~p * ~p accounts a ~p bytes~n",
	      [NB, NA, size(term_to_binary(A))]),
    io:format("TPC-B: Generating 0 history records a ~p bytes~n",
	      [size(term_to_binary(H))]),
    FunB =
	fun(BranchId) ->
		ok = mnesia:dirty_write(B#branch{id=BranchId}),
		FunT = fun(Id) ->
			       T2 = T#teller{id=Id, branch_id=BranchId},
			       ok = mnesia:dirty_write(T2)
		       end,
		FunA = fun(Id) ->
			       A2 = A#account{id=Id, branch_id=BranchId},
			       ok = mnesia:dirty_write(A2)
		       end,
		gen_recs((BranchId - 1) * NT + 1, BranchId * NT, FunT),
		gen_recs((BranchId - 1) * NA + 1, BranchId * NA, FunA)
	end,
    gen_recs(1, NB, FunB),
    case verify_tabs() of
	ok ->
	    ignore;
	{error, Reason} ->
	    io:format("TPC-B: <ERROR> Inconsistent tables: ~w~n",
		      [Reason]),
	    exit({inconsistent_tables, Reason})
    end.

gen_recs(First, N, Fun) when N >= First->
    Fun(N),
    gen_recs(First, N-1, Fun);
gen_recs(_, _, _) ->
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Run the benchmark

%% Args is a list of {Key, Val} tuples where Key is a field name
%% in the record run_config, unknown keys are ignored.
run(Args) ->
    RunConfig = list2rec(Args, record_info(fields, run_config), #run_config{}),
    Tags = record_info(fields, run_config),
    Fun = fun(F, Pos) -> {{F, element(Pos, RunConfig)}, Pos + 1} end,
    {List, _} = lists:mapfoldl(Fun, 2, Tags),
    io:format("TPC-B: Run config: ~p ~n", [List]),

    Pid = spawn_link(?MODULE, reporter_init, [self(), RunConfig]),
    receive
	{Pid, {stopped, Res}} ->
	    Res % Stopped by other process
    after RunConfig#run_config.stop_after ->
	    sync_stop(Pid)
    end.

reporter_init(Starter, RC) ->
    register(mnesia_tpcb, self()),
    process_flag(trap_exit, true),
    DbNodes = mnesia:system_info(db_nodes),
    if
	RC#run_config.use_running_mnesia == true ->
	    ignore;
	true ->
	    {Replies, BadNodes} =
		rpc:multicall(DbNodes, mnesia, start, []),
	    case [Res || Res <- Replies, Res /= ok] of
		[] when BadNodes == [] ->
		    ok;
		BadRes ->
		    io:format("TPC-B: <ERROR> "
			      "Failed to start ~w: ~p~n",
			      [BadNodes, BadRes]),
		    exit({start_failed, BadRes, BadNodes})
	    end,
	    verify_tabs()
    end,
    N = mnesia:table_info(branch, size),
    Type = mnesia:table_info(branch, storage_type),
    RepNodes = mnesia:table_info(branch, Type),
    NT = mnesia:table_info(teller, size) div N,
    NA = mnesia:table_info(account, size) div N,
    TC = #tab_config{n_branches=N,
		     n_tellers_per_branch=NT,
		     n_accounts_per_branch=NA,
		     db_nodes = DbNodes,
		     replica_nodes = RepNodes,
		     replica_type=Type
		    },
    erase(random_seed),
    random:seed(),
    Drivers = start_drivers(RC, TC),
    State = #reporter_state{driver_pids = Drivers,
			    run_config = RC,
			    starter_pid = Starter},
    case catch reporter_loop(State) of
	{'EXIT', Reason} ->
	    io:format("TPC-B: Abnormal termination: ~p~n", [Reason]),
	    if
		RC#run_config.use_running_mnesia == true ->
		    ignore;
		true ->
		    rpc:multicall(DbNodes, mnesia, lkill, [])
	    end,
	    unlink(Starter),
	    Starter ! {self(), {stopped, {error, Reason}}}, % To be sure
	    exit(shutdown);
	{ok, Stopper, State2} ->
	    Time = State2#reporter_state.acc,
	    Res =
		case verify_tabs() of
		    ok ->
			{ok, Time};
		    {error, Reason} ->
			io:format("TPC-B: <ERROR> Inconsistent tables, ~p~n",
				  [{error, Reason}]),
			{error, Reason}
		end,
	    if
		RC#run_config.use_running_mnesia == true -> ignore;
		true -> rpc:multicall(DbNodes, mnesia, stop, [])
	    end,
	    unlink(Starter),
	    Starter ! {self(), {stopped, Res}},
	    if
		Stopper /= Starter ->
		    Stopper ! {self(), {stopped, Res}};
		true ->
		    ignore
	    end,
	    exit(shutdown)
    end.

reporter_loop(State) ->
    RC = State#reporter_state.run_config,
    receive
	{From, stop} ->
	    {ok, From, call_drivers(State, stop)};
	{'EXIT', Pid, Reason} when Pid == State#reporter_state.starter_pid ->
	    %% call_drivers(State, stop),
	    exit({starter_died, Pid, Reason})
    after RC#run_config.report_interval ->
	    Iter = State#reporter_state.n_iters,
	    State2 = State#reporter_state{n_iters = Iter + 1},
	    case call_drivers(State2, report) of
		State3 when State3#reporter_state.driver_pids /= [] ->
		    State4 = State3#reporter_state{curr=#time{}},
		    reporter_loop(State4);
		_ ->
		    exit(drivers_died)
	    end
    end.

call_drivers(State, Msg) ->
    Drivers = State#reporter_state.driver_pids,
    lists:foreach(fun(Pid) -> Pid ! {self(), Msg} end, Drivers),
    State2 = calc_reports(Drivers, State),
    if
	Msg /= stop ->
	    ignore;
	true ->
	    Acc = State2#reporter_state.acc,
	    show_report(State2#reporter_state{n_iters = 0, curr = Acc})
    end,
    State2.


calc_reports([], State) ->
    show_report(State);
calc_reports([Pid|Drivers], State) ->
    receive
	{'EXIT', P, Reason} when P == State#reporter_state.starter_pid ->
	    exit({starter_died, P, Reason});
	{'EXIT', Pid, Reason} ->
	    exit({driver_died, Pid, Reason});
	{Pid, Time} when record(Time, time) ->
	    %% io:format("~w: ~w~n", [Pid, Time]),
	    A = add_time(State#reporter_state.acc, Time),
	    C = add_time(State#reporter_state.curr, Time),
	    State2 = State#reporter_state{acc = A, curr = C},
	    calc_reports(Drivers, State2)
    end.

add_time(Acc, New) ->
    Acc#time{n_trans = New#time.n_trans + Acc#time.n_trans,
	     min_n = lists:min([New#time.n_trans, Acc#time.min_n] -- [0]),
	     max_n = lists:max([New#time.n_trans, Acc#time.max_n]),
	     acc_time = New#time.acc_time + Acc#time.acc_time,
	     max_time = lists:max([New#time.max_time, Acc#time.max_time])}.

show_report(State) ->
    Iters = State#reporter_state.n_iters,
    Time = State#reporter_state.curr,
    Max = Time#time.max_time,
    Avg = Time#time.acc_time div Time#time.n_trans,
    N = length(State#reporter_state.driver_pids),
    Tps = (1000000 * N) div Avg,
    DiffTps =
	if
	    Iters > 1 -> Tps - State#reporter_state.prev_tps;
	    true  -> 0
	end,
    DiffSign = sign(DiffTps),
    Unfairness = Time#time.max_n / Time#time.min_n,
    if
	Iters /= 0 ->
	    io:format("TPC-B: ~p iter ~s~p diff ~p tps ~p avg micros ~p max micros ~p unfairness~n",
		      [Iters, DiffSign, DiffTps, Tps, Avg, Max, Unfairness]);
	true ->
	    io:format("TPC-B: ~p transactions per second, longest transaction was ~p seconds~n",
		      [Tps, Max div 1000000])
    end,
    State#reporter_state{prev_tps = Tps}.

sign(N) when N > 0 -> "+";
sign(N) -> "".

start_drivers(RC, TC) ->
    Nodes = RC#run_config.driver_nodes,
    N = RC#run_config.n_drivers_per_node,
    lists:append([start_drivers(Node, N, [], RC, TC) || Node <- Nodes]).
start_drivers(Node, 0, Pids, RC, TC) ->
    Pids;
start_drivers(Node, DriverId, Pids, RC, TC) ->
    State =#driver_state{driver_id=DriverId,
			 seed=get(random_seed),
			 tab_config=TC,
			 run_config=RC,
			 reuse_history_id = RC#run_config.reuse_history_id},
    Pid = spawn_link(Node, ?MODULE, driver_init, [State]),
    random:uniform(),
    start_drivers(Node, DriverId-1, [Pid|Pids], RC, TC).

driver_init(DS) ->
    io:format("TPC-B: Driver ~p started as ~p on node ~p~n",
	      [DS#driver_state.driver_id, self(), node(self())]),
    driver_loop(DS#driver_state{history_id=mnesia:table_info(history, size)}).

driver_loop(DS) ->
    receive
	{From, report} ->
	    From ! {self(), DS#driver_state.time},
	    Acc = add_time(DS#driver_state.time, DS#driver_state.acc_time),
	    DS2 = DS#driver_state{time=#time{}, acc_time = Acc}, % Reset timer
	    driver_loop(calc_trans(DS2));
	{From, stop} ->
	    Acc = add_time(DS#driver_state.time, DS#driver_state.acc_time),
	    io:format("TPC-B: Driver ~p (~p) on node ~p stopped: ~w~n",
		      [DS#driver_state.driver_id, self(), node(self()), Acc]),	
	    From ! {self(), DS#driver_state.time},
	    unlink(From),
	    exit(stopped)
    after 0 ->
	    driver_loop(calc_trans(DS))
    end.

calc_trans(DS) ->
    {Micros, DS2} = time_trans(DS),
    Time = DS2#driver_state.time,
    Time2 = Time#time{n_trans = Time#time.n_trans + 1,
		      acc_time = Time#time.acc_time + Micros,
		      max_time = lists:max([Micros, Time#time.max_time])
		     },
    HistoryId =
	if
	    DS#driver_state.reuse_history_id == false ->
		DS#driver_state.history_id + 1;
	    DS#driver_state.reuse_history_id == true ->
		DS#driver_state.history_id
	end,
    DS2#driver_state{time=Time2, history_id = HistoryId}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Generate teller_id, account_id and delta
%% Time the TPC-B transaction

time_trans(DS) ->
    OldSeed = get(random_seed), % Avoid interference with Mnesia
    put(random_seed, DS#driver_state.seed),
    Random    = random:uniform(),
    NewSeed   = get(random_seed),
    case OldSeed of
	undefined -> erase(random_seed);
	_         -> put(random_seed, OldSeed)
    end,

    DriverId = DS#driver_state.driver_id,
    TC = DS#driver_state.tab_config,
    RC = DS#driver_state.run_config,
    NB = TC#tab_config.n_branches,
    NA = TC#tab_config.n_accounts_per_branch,
    NT = TC#tab_config.n_tellers_per_branch,

    HistoryId = DS#driver_state.history_id,
    Delta = trunc(Random*1999999)-1000000, % -999999 < Delta < +999999
    TellerId = trunc(Random * NB * NT) + 1,
    BranchId = TellerId rem NB + 1,
    AccountId =
	if
	    Random > 0.85, NB > 1 ->
		%% Pick from remote account
                AccountId2 =  trunc(Random * (NB - 1) * NA) + 1,
		BranchId2 = AccountId2 rem NB + 1,
		if
		    BranchId2 >= BranchId -> AccountId2 + NA;
		    true -> AccountId2
		end;
	    true ->
		%% Pick from any account
		trunc(Random * NB * NA) + 1
	end,
    Fun =
	case RC#run_config.use_sticky_locks of
	    false ->
		fun() -> the_trans(DriverId, HistoryId, BranchId, TellerId, AccountId, Delta) end;
	    true ->
		fun() -> sticky_trans(DriverId, HistoryId, BranchId, TellerId, AccountId, Delta) end
	end,
    {Time, Res} = timer:tc(mnesia, transaction, [Fun]),
    case Res of
	{atomic, AccountBal} -> {Time, DS#driver_state{seed=NewSeed}};
	{aborted, Reason} -> exit({aborted, Reason})
    end.

%%
%% Runs the TPC-B defined transaction and returns NewAccountBalance
%%

the_trans(DriverId, HistoryId, BranchId, TellerId, AccountId, Delta) ->

    %% Add delta to branch balance
    [B]  = mnesia:wread({branch, BranchId}),
    NewB = B#branch{balance = B#branch.balance+Delta},
    ok = mnesia:write(NewB),

    %% Add delta to teller balance
    [T]  = mnesia:wread({teller, TellerId}),
    NewT = T#teller{balance = T#teller.balance+Delta},
    ok = mnesia:write(NewT),

    %% Add delta to account balance
    [A]  = mnesia:wread({account, AccountId}),
    NewA = A#account{balance = A#account.balance+Delta},
    ok = mnesia:write(NewA),

    %% Append to history log
    History = #history{history_id   = {DriverId, HistoryId},
		       account_id   = AccountId,
		       teller_id    = TellerId,
		       branch_id    = BranchId,
		       amount       = Delta
		      },
    ok = mnesia:write(History),

    %% Return account balance
    NewA#account.balance.

sticky_trans(DriverId, HistoryId, BranchId, TellerId, AccountId, Delta) ->
    %% We should really use s_wread/1

    %% Add delta to branch balance
    [B]  = mnesia:read({branch, BranchId}),
    NewB = B#branch{balance = B#branch.balance+Delta},
    ok = mnesia:s_write(NewB),

    %% Add delta to teller balance
    [T]  = mnesia:read({teller, TellerId}),
    NewT = T#teller{balance = T#teller.balance+Delta},
    ok = mnesia:s_write(NewT),

    %% Add delta to account balance
    [A]  = mnesia:read({account, AccountId}),
    NewA = A#account{balance = A#account.balance+Delta},
    ok = mnesia:s_write(NewA),

    %% Append to history log
    History = #history{history_id   = {DriverId, HistoryId},
		       account_id   = AccountId,
		       teller_id    = TellerId,
		       branch_id    = BranchId,
		       amount       = Delta
		      },
    ok = mnesia:s_write(History),

    %% Return account balance
    NewA#account.balance.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Verify table consistency

verify_tabs() ->
    Nodes = mnesia:system_info(running_db_nodes),
    case lists:member(node(), Nodes) of
	true ->
	    Tabs = [branch, teller, account, history],
	    io:format("TPC-B: Wait indefinitely for tables: ~w~n", [Tabs]),
	    rpc:multicall(Nodes, mnesia, wait_for_tables, [Tabs, infinity]),
	    
	    Fun = fun() ->
			  mnesia:write_lock_table(branch),
			  mnesia:write_lock_table(teller),
			  mnesia:write_lock_table(account),
			  mnesia:write_lock_table(history),
			  {Res, BadNodes} =
			      rpc:multicall(Nodes, ?MODULE, count_balance, []),
			  check_balance(Res, BadNodes)
		  end,
	    case mnesia:transaction(Fun) of
		{atomic, Res} -> Res;
		{aborted, Reason} -> {error, Reason}
	    end;
	false ->
	    {error, "Must be initiated from a running db_node"}
    end.

%% Returns a list of {Table, Node, Balance} tuples
%% Assumes that no updates are performed
count_balance() ->
    {ok, [
	 count_balance(branch, #branch.balance),
	 count_balance(teller, #teller.balance),
	 count_balance(account, #account.balance)
	]}.

count_balance(Tab, BalPos) ->
    count_balance(Tab, mnesia:dirty_first(Tab), 0, BalPos).

count_balance(Tab, '$end_of_table', Acc, BalPos) ->
    {Tab, node(), Acc};
count_balance(Tab, Key, Acc, BalPos) ->
    [Record] = mnesia:dirty_read({Tab, Key}),
    Bal = element(BalPos, Record),
    count_balance(Tab, mnesia:dirty_next(Tab, Key), Bal + Acc, BalPos).

check_balance(Res, []) ->
    catch check_balance2(Res, []);
check_balance(_, BadNodes) ->
    {error, {"Bad nodes", BadNodes}}.

check_balance2([{ok, List}|Tail], Acc) ->
    check_balance2(Tail, List ++ Acc);
check_balance2([], [{Tab, Node, Bal}|Acc]) ->
    case [{T, N, B} || {T, N, B} <- Acc, B /= Bal] of
	[] -> ok;
	BadRes -> throw({error, {"Bad balance", {Tab, Node, Bal}, BadRes}})
    end;
check_balance2(BadRes, Acc) ->
    throw({error, {"Bad balance", BadRes}}).

