%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% The mnesia_init process loads tables from local disc or from
%% another nodes. It also coordinates updates of the info about
%% where we can read and write tables.
%%
%% Tables may need to be loaded initially at startup of the local
%% node or when other nodes announces that they already have loaded
%% tables that we also want.
%%
%% Initially we set the load request queue to those tables that we
%% safely can load locally, i.e. tables where we have the last
%% consistent replica and we have received mnesia_down from all
%% other nodes holding the table. Then we let the mnesia_init
%% process enter its normal working state.
%%
%% When we need to load a table we append a request to the load
%% request queue. All other requests are regarded as high priority
%% and are processed immediately (e.g. update table whereabouts).
%% We processes the load request queue as a "background" job..

-module(mnesia_controller).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('klacke@erix.ericsson.se').
-author('hakan@erix.ericsson.se').

-behaviour(gen_server).

%% Mnesia internal stuff
-export([start/0,
	 i_have_tab/1,
	 info/0,
	 get_info/1,
	 force_load_table/1,
	 schema_transaction/1,
	 commit_schema_trans/3,
	 create_table/1,
	 get_disc_copy/1,
	 get_cstructs/0,
	 sync_add_table_copy_whereabouts/3,
	 sync_del_table_copy_whereabouts/2,
	 unannounce_add_table_copy/2,
	 master_nodes_updated/2,
	 mnesia_down/1,
	 add_active_replica/2,
	 add_active_replica/3,
	 change_table_access_mode/1,
	 del_active_replica/2,
	 wait_for_tables/2,
	 get_network_copy/2,
	 merge_schema/0,
	 schedule_late_disc_load/2
	]).

%% gen_server callbacks
-export([init/1,
	 handle_call/3,
	 handle_cast/2,
	 handle_info/2,
	 terminate/2,
	 code_change/3]).

%% Module internal stuff
-export([call/1,
	 cast/1,
	 run_worker/5,
	 %% do_sync_add_table_copy_whereabouts/4,
	 wait_for_tables_init/2
	]).

-import(mnesia_lib, [set/2, add/2]).
-import(mnesia_lib, [fatal/2, error/2, verbose/2, dbg_out/2]).

-include("mnesia.hrl").

-define(SERVER_NAME, ?MODULE).  

-record(state, {supervisor,
		active_workers = [],
		load_queue = [],
		late_load_queue = [],
		schema_is_merged = false,
		is_blocked = false,
		early_msgs = []}).

-record(worker, {what, 
		 pid,
		 reply_to}).

-record(worker_reply, {what,
		       pid,
		       result}).

-record(net_load, {table,
		   reason,
		   reply_to,
		   cstruct = unknown
		  }).
-record(disc_load, {table, reason, reply_to}).
-record(late_load, {table, reason, reply_to, loaders}).


val(Var) ->
    case ?catch_val(Var) of
	{'EXIT', Reason} -> mnesia_lib:other_val(Var, Reason); 
	Value -> Value
    end.

start() ->
    gen_server:start_link({local, ?SERVER_NAME}, ?MODULE, [self()],
			  [{timeout, infinity}
			   %% ,{debug, [trace]}
			  ]).

%% Wait for tables to be active
%% If needed, we will wait for Mnesia to start
%% If Mnesia stops, we will wait for Mnesia to restart
%% We will wait even if the list of tables is empty
%%
wait_for_tables(Tabs, Timeout) when list(Tabs), Timeout == infinity ->
    do_wait_for_tables(Tabs, Timeout);
wait_for_tables(Tabs, Timeout) when list(Tabs),
                                    integer(Timeout), Timeout >= 0 ->
    do_wait_for_tables(Tabs, Timeout);
wait_for_tables(Tabs, Timeout) ->
    {error, {badarg, Tabs, Timeout}}.

do_wait_for_tables(Tabs, Timeout) ->
    Pid = spawn_link(?MODULE, wait_for_tables_init, [self(), Tabs]),
    receive
	{?SERVER_NAME, Pid, Res} ->
	    Res;

	{'EXIT', Pid, _} ->
	    reply_wait(Tabs)

    after Timeout ->
	    unlink(Pid),
	    exit(Pid, timeout),
	    reply_wait(Tabs)
    end.
	
reply_wait(Tabs) ->
    case catch mnesia_lib:active_tables() of
	{'EXIT', _} ->
	    {error, {node_not_running, node()}};
	Active when list(Active) ->
	    case Tabs -- Active of
		[] ->
		    ok;
		BadTabs ->
		    {timeout, BadTabs}
	    end
    end.

wait_for_tables_init(From, Tabs) ->
    process_flag(trap_exit, true),
    Res = wait_for_init(From, Tabs, whereis(?SERVER_NAME)),
    From ! {?SERVER_NAME, self(), Res},
    unlink(From),
    exit(normal).

wait_for_init(From, Tabs, Init) ->
    case catch link(Init) of
	{'EXIT', _} ->
	    %% Mnesia is not started
	    {error, {node_not_running, node()}};
	true when pid(Init) ->
	    cast({self(), {sync_tabs, Tabs}}),
	    rec_tabs(Tabs, Tabs, From, Init)
    end.

sync_reply(Waiter, Tab) ->
    Waiter ! {?SERVER_NAME, {tab_synced, Tab}}.

rec_tabs([Tab | Tabs], AllTabs, From, Init) ->
    receive
	{?SERVER_NAME, {tab_synced, Tab}} ->
	    rec_tabs(Tabs, AllTabs, From, Init);

	{'EXIT', From, _} ->
	    %% This will trigger an exit signal
	    %% to mnesia_init
	    exit(wait_for_tables_timeout);
	
	{'EXIT', Init, _} ->
	    %% Oops, mnesia_init stopped,
	    exit(mnesia_stopped)
    end;
rec_tabs([], _, _, Init) ->
    unlink(Init),
    ok.

get_cstructs() ->
    call(get_cstructs).

mnesia_down(Node) ->
    case cast({mnesia_down, Node}) of
	{error, _} -> mnesia_monitor:mnesia_down(?SERVER_NAME, Node);
	Pid ->  ok
    end.

schema_transaction(Fun) ->
    unsafe_call({schema_transaction, Fun}).

commit_schema_trans(Tid, [], _) ->
    ok;
commit_schema_trans(Tid, Commit, Use_dumper) ->
    call({commit_schema_trans, Tid, Commit, Use_dumper}).
%commit_schema_trans(Tid, Commit, ignore_dumper) ->
%    mnesia_dumper:update(Tid, Commit, ignore_dumper).

%% Special for preparation of add table copy
get_network_copy(Tab, Cs) ->
    call({net_load, Tab, Cs}).

    %% This functions is invoked from the dumper
    %% 
    %% There are two cases here:
    %% startup ->
    %%   no need for sync, since mnesia_init is not started yet
    %% schema_trans ->
    %%   already synced with mnesia_init since the dumper
    %%   is syncronously started from mnesia_init
create_table(Tab) ->
    ok = mnesia_loader:disc_load_table(Tab, dumper_create_table).
    
get_disc_copy(Tab) ->
    disc_load_table(Tab, dumper_change_table_copy_type, undefined).

%% Returns ok instead of yes
force_load_table(Tab) when atom(Tab), Tab /= schema ->
    case ?catch_val({Tab, storage_type}) of
	ram_copies ->
	    do_force_load_table(Tab);
	disc_copies ->
	    do_force_load_table(Tab);
	disc_only_copies ->
	    do_force_load_table(Tab);
	unknown ->
	    set({Tab, load_by_force}, true),
	    cast({force_load_updated, Tab}),
	    wait_for_tables([Tab], infinity);
	{'EXIT', _} ->
	    {error, {no_exists, Tab}}
    end;
force_load_table(Tab) ->
    {error, {bad_type, Tab}}.
    
do_force_load_table(Tab) ->
    set({Tab, load_by_force}, true),
    mnesia_late_loader:async_late_disc_load(node(), [Tab], forced_by_user),
    wait_for_tables([Tab], infinity).

master_nodes_updated(schema, Masters) ->
    ignore;
master_nodes_updated(Tab, Masters) ->
    cast({master_nodes_updated, Tab, Masters}).

schedule_late_disc_load(Tabs, Reason) ->
    MsgTag = late_disc_load,
    try_schedule_late_disc_load(Tabs, Reason, MsgTag).

try_schedule_late_disc_load(Tabs, Reason, MsgTag)
  when Tabs == [], MsgTag /= schema_is_merged ->
    ignore;
try_schedule_late_disc_load(Tabs, Reason, MsgTag) ->
    GetIntents =
	fun() ->
		Item = mnesia_late_disc_load,
		Nodes = val({current, db_nodes}),
		mnesia:lock({global, Item, Nodes}, write),
		case multicall(Nodes -- [node()], disc_load_intents) of
		    {Replies, []} ->
			call({MsgTag, Tabs, Reason, Replies}),
			done;
		    {_, BadNodes} ->
			%% Some nodes did not respond, lets try again
			{retry, BadNodes}
		end
	end,
    case mnesia:transaction(GetIntents) of
	{atomic, done} ->
	    done;
	{atomic, {retry, BadNodes}} ->
	    verbose("Retry late_load_tables because bad nodes: ~p~n",
		    [BadNodes]),
	    try_schedule_late_disc_load(Tabs, Reason, MsgTag);
	{aborted, AbortReason} ->
	    fatal("Cannot late_load_tables~p: ~p~n",
		  [[Tabs, Reason, MsgTag], AbortReason])
    end.

%% Merge the local schema with the schema on other nodes.
%% But first we must let all processes that want to force
%% load tables wait until the schema merge is done.

merge_schema() ->
    AllNodes = mnesia_lib:all_nodes(),
    try_merge_schema(AllNodes),
    schema_is_merged().

try_merge_schema(Nodes) ->
    case mnesia_schema:merge_schema() of
	{atomic, not_merged} ->
	    %% No more nodes that we need to merge the schema with
	    ok;
	{atomic, {merged, OldFriends, NewFriends}} ->
	    %% Check if new nodes has been added to the schema
	    Diff = mnesia_lib:all_nodes() -- [node() | Nodes],
	    mnesia_recover:connect_nodes(Diff),

	    %% Tell everybody to adopt orphan tables
	    im_running(OldFriends, NewFriends),
	    im_running(NewFriends, OldFriends),

	    try_merge_schema(Nodes);
	{atomic, {"Cannot get cstructs", Node, Reason}} ->
	    dbg_out("Cannot get cstructs, Node ~p ~p~n", [Node, Reason]),
	    timer:sleep(1000), % Avoid a endless loop look alike	    
	    try_merge_schema(Nodes);
	Other ->
	    fatal("Failed to merge schema ~p~n", [Other])
    end.

im_running(OldFriends, NewFriends) ->
    abcast(OldFriends, {im_running, node(), NewFriends}).

schema_is_merged() ->
    MsgTag = schema_is_merged,
    SafeLoads = initial_safe_loads(),
    
    %% At this point we do not know anything about
    %% which tables that the other nodes already
    %% has loaded and therefore we let the normal
    %% processing of the load_queue take care
    %% of it, since we at that time point will
    %% know the whereabouts. We rely on the fact
    %% that all nodes tells each other directly
    %% when they have loaded a table and are
    %% willing to share it.
    
    try_schedule_late_disc_load(SafeLoads, initial, MsgTag).


cast(Msg) ->
    case whereis(?SERVER_NAME) of
	undefined ->{error, {node_not_running, node()}};
	Pid ->  gen_server:cast(Pid, Msg)
    end.

abcast(Nodes, Msg) ->
    gen_server:abcast(Nodes, ?SERVER_NAME, Msg).

unsafe_call(Msg) ->
    case whereis(?SERVER_NAME) of
	undefined -> {error, {node_not_running, node()}};
	Pid -> gen_server:call(Pid, Msg, infinity)
    end.

call(Msg) ->
    case whereis(?SERVER_NAME) of
	undefined ->
	    {error, {node_not_running, node()}};
	Pid ->
	    link(Pid),
	    Res = gen_server:call(Pid, Msg, infinity),
	    unlink(Pid),

	    %% We get an exit signal if server dies
            receive
                {'EXIT', Pid, Reason} ->
                    {error, {node_not_running, node()}}
            after 0 ->
                    ignore
            end,
	    Res
    end.

multicall(Nodes, Msg) ->
    rpc:multicall(Nodes, ?MODULE, call, [Msg]).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init([Parent]) ->
    process_flag(trap_exit, true),
    mnesia_lib:verbose("~p starting: ~p~n", [?SERVER_NAME, self()]),

    %% Handshake and initialize transaction recovery
    %% for new nodes detected in the schema
    All = mnesia_lib:all_nodes(),
    Diff = All -- [node() | val(original_nodes)],
    mnesia_lib:unset(original_nodes),
    mnesia_recover:connect_nodes(Diff),

    {ok, #state{supervisor = Parent}}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, Reply, State}     (terminate/2 is called)
%%----------------------------------------------------------------------

handle_call({schema_transaction, Fun}, From, State) ->
    What = schema_transaction,
    start_worker(What, mnesia, transaction, [Fun]),
    noreply(State);

handle_call({commit_schema_trans, Tid, Commit, UseDump}, From, State) ->
    Res = mnesia_dumper:update(Tid, Commit, UseDump),
    reply(From, Res),
    noreply(State);

handle_call(get_cstructs, From, State) ->
    Tabs = val({schema, tables}),
    Cstructs = [val({T, cstruct}) || T <- Tabs],
    Running = val({current, db_nodes}),
    reply(From, {cstructs, Cstructs, Running}),
    noreply(State);

handle_call({schema_is_merged, TabsR, Reason, RemoteLoaders}, From, State) ->
    State2 = late_disc_load(TabsR, Reason, RemoteLoaders, From, State),

    %% Handle early messages
    Msgs = State2#state.early_msgs,
    State3 = State2#state{early_msgs = [], schema_is_merged = true},
    Ns = val({current, db_nodes}),
    dbg_out("Schema is merged ~w, State ~w~n", [Ns, State3]),
    handle_early_msgs(Msgs, State3);

handle_call(disc_load_intents, From, State) ->
    Tabs = disc_load_intents(State#state.active_workers) ++
	   disc_load_intents(State#state.load_queue) ++
           disc_load_intents(State#state.late_load_queue),
    ActiveTabs = mnesia_lib:local_active_tables(),
    reply(From, {ok, node(), mnesia_lib:union(Tabs, ActiveTabs)}),
    noreply(State);

handle_call(Msg, From, State) when State#state.schema_is_merged == false ->
    %% Buffer early messages
    Msgs = State#state.early_msgs,
    noreply(State#state{early_msgs = [{call, Msg, From} | Msgs]});

handle_call({net_load, Tab, Cs}, From, State) ->
    What = #net_load{table = Tab,
		     reply_to = From,
		     reason = add_table_copy,
		     cstruct = Cs
		    },
    Queue = State#state.load_queue ++ [What],
    noreply(State#state{load_queue = Queue});

handle_call({late_disc_load, Tabs, Reason, RemoteLoaders}, From, State) ->
    State2 = late_disc_load(Tabs, Reason, RemoteLoaders, From, State),
    noreply(State2);

handle_call(Msg, From, State) ->
    error("~p got unexpected call: ~p~n", [?SERVER_NAME, Msg]),
    noreply(State).

disc_load_intents([H | T]) when record(H, disc_load) ->
    [H#disc_load.table | disc_load_intents(T)];
disc_load_intents([H | T]) when record(H, late_load) ->
    [H#late_load.table | disc_load_intents(T)];
disc_load_intents( [H | T]) when record(H, net_load) ->
    disc_load_intents(T);
disc_load_intents([]) ->
    [].

late_disc_load(TabsR, Reason, RemoteLoaders, From, State) ->
    verbose("Intend to load tables: ~p~n", [TabsR]),
    ?eval_debug_fun({?MODULE, late_disc_load},
		    [{tabs, TabsR}, 
		     {reason, Reason},
		     {loaders, RemoteLoaders}]),

    reply(From, queued),
    %% RemoteLoaders is a list of {ok, Node, Tabs} tuples

    Nodes = val({current, db_nodes}),
    LateLoaders = late_loaders(TabsR, Reason, RemoteLoaders, Nodes),
    LateQueue = State#state.late_load_queue ++ LateLoaders,
    State#state{late_load_queue = LateQueue}.

late_loaders([{Tab, Reason} | Tabs], DefaultReason, RemoteLoaders, Nodes) ->
    LoadNodes = late_load_filter(RemoteLoaders, Tab, Nodes, []),
    case LoadNodes of
	[] ->
	    cast({disc_load, Tab, Reason}); % Ugly cast
	_ ->
	    ignore
    end,
    LateLoad = #late_load{table = Tab, loaders = LoadNodes, reason = Reason},
    [LateLoad | late_loaders(Tabs, DefaultReason, RemoteLoaders, Nodes)];

late_loaders([Tab | Tabs], Reason, RemoteLoaders, Nodes) ->
    Loaders = late_load_filter(RemoteLoaders, Tab, Nodes, []),
    case Loaders of
	[] ->
	    cast({disc_load, Tab, Reason}); % Ugly cast
	_ ->
	    ignore
    end,
    LateLoad = #late_load{table = Tab, loaders = Loaders, reason = Reason},
    [LateLoad | late_loaders(Tabs, Reason, RemoteLoaders, Nodes)];
late_loaders([], _Reason, _RemoteLoaders, _Nodes) ->
    [].

late_load_filter([{error, _} | RemoteLoaders], Tab, Nodes, Acc) ->
    late_load_filter(RemoteLoaders, Tab, Nodes, Acc);
late_load_filter([{badrpc, _} | RemoteLoaders], Tab, Nodes, Acc) ->
    late_load_filter(RemoteLoaders, Tab, Nodes, Acc);
late_load_filter([RL | RemoteLoaders], Tab, Nodes, Acc) ->
    {ok, Node, Intents} = RL,
    Access = val({Tab, access_mode}),
    LocalC = val({Tab, local_content}),
    StillActive = lists:member(Node, Nodes),
    RemoteIntent = lists:member(Tab, Intents),
    if
	Access == read_write,
	LocalC == false,
	StillActive == true,
	RemoteIntent == true ->
	    Masters = mnesia_recover:get_master_nodes(Tab),
	    case lists:member(Node, Masters) of
		true ->
		    %% The other node is master node for
		    %% the table, accept his load intent
		    late_load_filter(RemoteLoaders, Tab, Nodes, [Node | Acc]);
		false when Masters == [] ->
		    %% The table has no master nodes
		    %% accept his load intent
		    late_load_filter(RemoteLoaders, Tab, Nodes, [Node | Acc]);
		false ->
		    %% Some one else is master node for
		    %% the table, ignore his load intent
		    late_load_filter(RemoteLoaders, Tab, Nodes, Acc)
	    end;
	true ->
	    late_load_filter(RemoteLoaders, Tab, Nodes, Acc)
    end;
late_load_filter([], Tab, Nodes, Acc) ->
    Acc.
    
%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

handle_cast({mnesia_down, Node}, State) ->
    maybe_log_mnesia_down(Node),
    mnesia_lib:del({current, db_nodes}, Node),
    mnesia_checkpoint:tm_mnesia_down(Node),
    Alltabs = val({schema, tables}),
    State2 = reconfigure_tables(Node, State, Alltabs),
    mnesia_monitor:mnesia_down(?SERVER_NAME, Node),
    noreply(State2);

handle_cast({im_running, Node, NewFriends}, State) ->
    Tabs = mnesia_lib:local_active_tables() -- [schema],
    Ns = mnesia_lib:intersect(NewFriends, val({current, db_nodes})),
    abcast(Ns, {adopt_orphans, node(), Tabs}),
    noreply(State);

handle_cast(Msg, State) when State#state.schema_is_merged == false ->
    %% Buffer early messages
    Msgs = State#state.early_msgs,
    noreply(State#state{early_msgs = [{cast, Msg} | Msgs]});

handle_cast({disc_load, Tab, Reason}, State) ->
    DiscLoad = #disc_load{table = Tab, reason = Reason},
    Queue = State#state.load_queue ++ [DiscLoad],
    noreply(State#state{load_queue = Queue});

handle_cast({From, {sync_tabs, Tabs}}, State) ->
    %% user initiated wait_for_tables
    handle_sync_tabs(Tabs, From),
    noreply(State);

handle_cast({i_have_tab, Tab, Node}, State) ->
    State2 = node_has_tabs([Tab], Node, State),
    Queue = State2#state.load_queue,
    noreply(State#state{load_queue = Queue});

handle_cast({force_load_updated, Tab}, State) ->
    case val({Tab, active_replicas}) of
	[] ->
	    %% No valid replicas
	    noreply(State);
	[SomeNode | _] ->
	    State2 = node_has_tabs([Tab], SomeNode, State),
	    noreply(State2)
    end;
    
handle_cast({master_nodes_updated, Tab, Masters}, State) ->
    Active = val({Tab, active_replicas}),
    Valid = 
	case val({Tab, load_by_force}) of
	    true ->
		Active;
	    false ->
		if
		    Masters == [] ->
			Active;
		    true ->
			mnesia_lib:intersect(Masters, Active)
		end
	end,
    case Valid of
	[] ->
	    %% No valid replicas
	    noreply(State);
	[SomeNode | _] ->
	    noreply(node_has_tabs([Tab], SomeNode, State))
    end;
    
handle_cast({adopt_orphans, Node, Tabs}, State) ->
    State2 = node_has_tabs(Tabs, Node, State),
    
    %% Register the other node as up and running
    mnesia_recover:log_mnesia_up(Node),
    verbose("Logging mnesia_up ~w~n", [Node]),
    mnesia_lib:report_system_event({mnesia_up, Node}),
    
    %% Load orphan tables
    LocalTabs = val({schema, local_tables}) -- [schema],
    Nodes = val({current, db_nodes}),
    {LocalOrphans, RemoteMasters} =
	orphan_tables(LocalTabs, Node, Nodes, [], []),
    Reason = {adopt_orphan, node()},
    mnesia_late_loader:async_late_disc_load(node(), LocalOrphans, Reason),
    
    Fun =
	fun(N) ->
		RemoteOrphans =
		    [Tab || {Tab, Ns} <- RemoteMasters,
			    lists:member(N, Ns)],
		mnesia_late_loader:maybe_async_late_disc_load(N, RemoteOrphans, Reason)
	  end,
    lists:foreach(Fun, Nodes),
    
    Queue = State2#state.load_queue,
    noreply(State#state{load_queue = Queue});

handle_cast(Msg, State) ->
    error("~p got unexpected cast: ~p~n", [?SERVER_NAME, Msg]),
    noreply(State).

handle_sync_tabs([Tab | Tabs], From) ->
    case val({Tab, where_to_read}) of
	nowhere ->
	    case get({sync_tab, Tab}) of
		undefined ->
		    put({sync_tab, Tab}, [From]);
		Pids ->
		    put({sync_tab, Tab}, [From | Pids])
	    end;
	_ ->
	    sync_reply(From, Tab)
    end,
    handle_sync_tabs(Tabs, From);
handle_sync_tabs([], From) ->
    ok.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

handle_info(WR, State) when record(WR, worker_reply) ->
    State2 = handle_worker_reply(WR, State),
    noreply(State2);

handle_info({'EXIT', Pid, R}, State) when Pid == State#state.supervisor  ->
    dbg_out("~p was ~p~n", [?SERVER_NAME, R]),
    {stop, shutdown, State};

handle_info({From, get_state}, State) ->
    From ! {?SERVER_NAME, State},
    noreply(State);

%% No real need for buffering
handle_info(Msg, State) when State#state.schema_is_merged == false ->
    %% Buffer early messages
    Msgs = State#state.early_msgs,
    noreply(State#state{early_msgs = [{info, Msg} | Msgs]});

handle_info(timeout, State) ->
%    dbg_out("~p timeout ~p~n", [?SERVER_NAME, State]),
    {Load, Rest} = pick_next(State#state.load_queue),
    State2 = State#state{load_queue = Rest},
%			 active_workers = [Load | State#state.active_workers]},
%                        Removed temporary..

    case load_table(Load) of
	{loaded, Tab} ->
	    LateQueue = State2#state.late_load_queue,
	    LateQueue2 = reply_late_load(Tab, LateQueue),
	    noreply(State2#state{late_load_queue = LateQueue2});
	{not_loaded, Tab} ->
	    noreply(State2)
    end;

handle_info({'EXIT', Pid, wait_for_tables_timeout}, State) ->
    sync_tab_timeout(Pid, get()),
    noreply(State);

handle_info(Msg, State) ->
    error("~p got unexpected info: ~p~n", [?SERVER_NAME, Msg]),
    noreply(State).

reply_late_load(Tab, [H | T]) when H#late_load.table == Tab ->
    reply(H#late_load.reply_to, ok),
    reply_late_load(Tab, T);
reply_late_load(Tab, [H | T])  ->
    [H | reply_late_load(Tab, T)];
reply_late_load(Tab, []) ->
    [].

sync_tab_timeout(Pid, [{{sync_tab, Tab}, Pids} | Tail]) ->
    case lists:delete(Pid, Pids) of
	[] ->
	    erase({sync_tab, Tab});
	Pids2 ->
	    put({sync_tab, Tab}, Pids2)
    end,
    sync_tab_timeout(Pid, Tail);
sync_tab_timeout(Pid, [_ | Tail]) ->
    sync_tab_timeout(Pid, Tail);
sync_tab_timeout(Pid, []) ->
    ok.

%% Pick the load record that has the highest load order
%% Returns {BestLoad, RemainingQueue} or {none, []} if queue is empty
pick_next(Queue) ->
    pick_next(Queue, none, none, []).

pick_next([Head | Tail], Load, Order, Rest) when record(Head, net_load) ->
    Tab = Head#net_load.table,
    select_best(Head, Tail, val({Tab, load_order}), Load, Order, Rest);
pick_next([Head | Tail], Load, Order, Rest) when record(Head, disc_load) ->
    Tab = Head#disc_load.table,
    select_best(Head, Tail, val({Tab, load_order}), Load, Order, Rest);
pick_next([], Load, Order, Rest) ->
    {Load, Rest}.

select_best(Load, Tail, Order, none, none, Rest) ->
    pick_next(Tail, Load, Order, Rest);
select_best(Load, Tail, Order, OldLoad, OldOrder, Rest) when Order > OldOrder ->
    pick_next(Tail, Load, Order, [OldLoad | Rest]);
select_best(Load, Tail, Order, OldLoad, OldOrder, Rest) ->
    pick_next(Tail, OldLoad, OldOrder, [Load | Rest]).

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    mnesia_monitor:terminate_proc(?SERVER_NAME, Reason, State).

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Upgrade process when its code is to be changed
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    exit(not_supported).

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

maybe_log_mnesia_down(N) ->
    HalfLoadedTabs =
	[T || T <- val({schema, local_tables}) -- [schema],
	      inactive_copy_holders(T, N)],
    if
	HalfLoadedTabs == [] ->
	    verbose("Logging mnesia_down ~w~n", [N]),
	    mnesia_recover:log_mnesia_down(N);
	true ->
	    %% Unfortunately we have not loaded some common
	    %% tables yet, so we cannot rely on the nodedown
	    ignore
    end.

inactive_copy_holders(Tab, Node) ->
    Cs = val({Tab, cstruct}),
    case mnesia_lib:cs_to_storage_type(Node, Cs) of
	unknown ->
	    false;
	_Storage ->
	    mnesia_lib:not_active_here(Tab)
    end.

orphan_tables([Tab | Tabs], Node, Ns, Local, Remote) ->
    Cs = val({Tab, cstruct}),
    CopyHolders = mnesia_lib:copy_holders(Cs),
    RamCopyHolders = Cs#cstruct.ram_copies,
    DiscCopyHolders = CopyHolders -- RamCopyHolders,
    DiscNodes = val({schema, disc_copies}),
    RamCopyHoldersOnDiscNodes = mnesia_lib:intersect(RamCopyHolders, DiscNodes),
    Active = val({Tab, active_replicas}),
    case lists:member(Node, DiscCopyHolders) of
	true when Active == [] ->
	    case DiscCopyHolders -- Ns of
		[] ->
		    %% We're last up and the other nodes have not
		    %% loaded the table. Lets load it if we are
		    %% the smallest node.
		    case lists:min(DiscCopyHolders) of
			Min when Min == node() ->
			    case mnesia_recover:get_master_nodes(Tab) of
				[] ->
				    L = [Tab | Local],
				    orphan_tables(Tabs, Node, Ns, L, Remote);
				Masters ->
				    R = [{Tab, Masters} | Remote],
				    orphan_tables(Tabs, Node, Ns, Local, R)
			    end;
			_ ->
			    orphan_tables(Tabs, Node, Ns, Local, Remote)
		    end;
		_ ->
		    orphan_tables(Tabs, Node, Ns, Local, Remote)
	    end;
	false when Active == [], DiscCopyHolders == [], RamCopyHoldersOnDiscNodes == [] ->
	    %% Special case when all replicas resides on disc less nodes
	    orphan_tables(Tabs, Node, Ns, [Tab | Local], Remote);
	_ ->
	    orphan_tables(Tabs, Node, Ns, Local, Remote)
    end;
orphan_tables([], _, _, LocalOrphans, RemoteMasters) ->
    {LocalOrphans, RemoteMasters}.

node_has_tabs([Tab | Tabs], Node, State) when Node /= node() ->
    State2 = update_whereabouts(Tab, Node, State),
    node_has_tabs(Tabs, Node, State2);
node_has_tabs([Tab | Tabs], Node, State) ->
    user_sync_tab(Tab),
    node_has_tabs(Tabs, Node, State);
node_has_tabs([], _Node, State) ->
    State.

update_whereabouts(Tab, Node, State) ->
    Storage = val({Tab, storage_type}),
    Read = val({Tab, where_to_read}),
    LocalC = val({Tab, local_content}),
    BeingCreated = ?catch_val({Tab, create_table}),
    Masters = mnesia_recover:get_master_nodes(Tab),
    ByForce = val({Tab, load_by_force}),
    GoGetIt =
	if
	    ByForce == true ->
		true;
	    Masters == [] ->
		true;
	    true ->
		lists:member(Node, Masters)
	end,
    
    dbg_out("Table ~w is loaded on ~w. s=~w, r=~w, lc=~w, f=~w, m=~w~n",
	    [Tab, Node, Storage, Read, LocalC, ByForce, GoGetIt]),
    if
	LocalC == true ->
	    %% Local contents, don't care about other node
	    State;
	Storage == unknown, Read == nowhere ->
	    %% No own copy, time to read remotely
	    %% if the other node is a good node
	    add_active_replica(Tab, Node),
	    case GoGetIt of
		true ->
		    set({Tab, where_to_read}, Node),
		    user_sync_tab(Tab),
		    State;
		false ->
		    State
	    end;
	Storage == unknown ->
	    %% No own copy, continue to read remotely
	    add_active_replica(Tab, Node),
	    user_sync_tab(Tab),
	    State;
	BeingCreated == true ->
	    %% The table is currently being created
	    %% and we shall have an own copy of it.
	    %% We will load the (empty) table locally.
	    add_active_replica(Tab, Node),
	    State;
	Read == nowhere ->
	    %% Own copy, go and get a copy of the table
	    %% if the other node is master or if there
	    %% are no master at all
	    add_active_replica(Tab, Node),
	    case GoGetIt of
		true ->
		    NetLoad = #net_load{table = Tab,
					reason = {active_remote, Node}},
		    Queue = State#state.load_queue ++ [NetLoad],
		    State#state{load_queue = Queue};
		false ->
		    State
	    end;
	true ->
	    %% We already have an own copy
	    add_active_replica(Tab, Node),
	    user_sync_tab(Tab),
	    State
    end.

initial_safe_loads() ->
    case val({schema, storage_type}) of
	ram_copies ->
	    [];
	disc_copies ->
	    Downs = mnesia_recover:get_mnesia_downs(),
	    dbg_out("mnesia_downs = ~p~n", [Downs]),
		
	    Tabs = val({schema, local_tables}) -- [schema],
	    LastC = fun(T) -> last_consistent_replica(T, Downs) end,
	    lists:zf(LastC, Tabs)
    end.
    
last_consistent_replica(Tab, Downs) ->
    Cs = val({Tab, cstruct}),
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    Ram = Cs#cstruct.ram_copies,
    Disc = Cs#cstruct.disc_copies,
    DiscOnly = Cs#cstruct.disc_only_copies,
    BetterCopies0 = mnesia_lib:remote_copy_holders(Cs) -- Downs,
    BetterCopies = BetterCopies0 -- Ram,
    AccessMode = Cs#cstruct.access_mode,
    Copies = mnesia_lib:copy_holders(Cs),
    Masters = mnesia_recover:get_master_nodes(Tab),
    LocalMaster0 = lists:member(node(), Masters),
    RemoteMaster =
	if
	    Masters == [] -> false;
	    true -> not LocalMaster0
	end,
    LocalMaster =
	if
	    Masters == [] -> false;
	    true -> LocalMaster0
	end,
    if
	Copies == [node()]  ->
	    %% Only one copy holder and it is local.
	    %% It may also be a local contents table
	    {true, {Tab, local_only}};
	LocalMaster == true ->
	    %% We have a local master
	    {true, {Tab, local_master}};
	RemoteMaster == true ->
	    %% Wait for remote master copy
	    false;
	Storage == ram_copies ->
	    if
		Disc == [], DiscOnly == [] ->
		    %% Nobody has copy on disc
		    {true, {Tab, ram_only}};
		true ->
		    %% Some other node has copy on disc
		    false
	    end;
	AccessMode == read_only ->
	    %% No one has been able to update the table,
	    %% i.e. all disc resident copies are equal
	    {true, {Tab, read_only}};
	BetterCopies /= [], Masters /= [node()] ->
	    %% There are better copies on other nodes
	    %% and we do not have the only master copy
	    false;
	true ->
	    {true, {Tab, initial}}
    end.

reconfigure_tables(N, State, [Tab |Tail]) ->
    del_active_replica(Tab, N),
    case val({Tab, where_to_read}) of
	N ->  mnesia_lib:set_remote_where_to_read(Tab);
	_ ->  ignore
    end,
    LateQ = drop_loaders(Tab, N, State#state.late_load_queue),
    reconfigure_tables(N, State#state{late_load_queue = LateQ}, Tail);

reconfigure_tables(_, State, []) ->
    State.

%% Drop loader from late load queue and possibly trigger a disc_load
drop_loaders(Tab, Node, [H | T]) when H#late_load.table == Tab ->
    %% Check if it is time to issue a disc_load request
    case H#late_load.loaders of
	[Node] ->
	    Reason = {H#late_load.reason, last_loader_down, Node},
	    cast({disc_load, Tab, Reason}); % Ugly cast
	_ ->
	    ignore
    end,
    %% Drop the node from the list of loaders
    H2 = H#late_load{loaders = H#late_load.loaders -- [Node]},
    [H2 | drop_loaders(Tab, Node, T)];
drop_loaders(Tab, Node, [H | T]) ->
    [H | drop_loaders(Tab, Node, T)];
drop_loaders(_, _, []) ->
    [].

add_active_replica(Tab, Node) ->
    add_active_replica(Tab, Node, val({Tab, cstruct})).

add_active_replica(Tab, Node, Cs) ->
    Var = {Tab, where_to_commit},
    Old = val(Var),
    Del = lists:keydelete(Node, 1, Old),
    case Cs#cstruct.access_mode of
	read_write ->
	    Storage = mnesia_lib:schema_cs_to_storage_type(Node, Cs),
	    New = lists:sort([{Node, Storage} | Del]),
	    set(Var, New),
	    add({Tab, where_to_write}, Node);
	read_only ->
	    set(Var, Del),
	    mnesia_lib:del({Tab, where_to_write}, Node)
    end,
    add({Tab, active_replicas}, Node).

del_active_replica(Tab, Node) ->
    Var = {Tab, where_to_commit},
    Old = val(Var),
    Del = lists:keydelete(Node, 1, Old),
    New = lists:sort(Del),
    set(Var, New),
    mnesia_lib:del({Tab, active_replicas}, Node),
    mnesia_lib:del({Tab, where_to_write}, Node).

change_table_access_mode(Cs) ->
    Tab = Cs#cstruct.name,
    lists:foreach(fun(N) -> add_active_replica(Tab, N, Cs) end,
		  val({Tab, active_replicas})).

%% node To now has tab loaded, but this must be undone
%% This code is rpc:call'ed from the tab_copier process
%% when it has *not* released it's table lock
unannounce_add_table_copy(Tab, To) ->
    del_active_replica(Tab, To),
    case val({Tab , where_to_read}) of
	To -> 
	    mnesia_lib:set_remote_where_to_read(Tab);
	_ -> ignore
    end.

user_sync_tab(Tab) ->
    case val(debug) of
	trace ->
	    mnesia_subscr:subscribe(whereis(mnesia_event), {table, Tab});
	_ ->
	    ignore
    end,
	
    case erase({sync_tab, Tab}) of
	undefined ->
	    ok;
	Pids ->
	    lists:foreach(fun(Pid) -> sync_reply(Pid, Tab) end, Pids)
    end.

%% Now it is time to load the table
%% but first we must check if it still is  neccessary

load_table(Load) when record(Load, net_load) ->
    Tab = Load#net_load.table,
    ReplyTo = Load#net_load.reply_to,
    Reason =  Load#net_load.reason,
    LocalC = val({Tab, local_content}),
    AccessMode = val({Tab, access_mode}),
    ReadNode = val({Tab, where_to_read}),
    Active = filter_active(Tab),
    if
	ReadNode == node() ->
	    %% Already loaded locally
	    reply(ReplyTo, ok),
            {loaded, Tab};
	LocalC == true ->
	    mnesia_lib:set_local_content_whereabouts(Tab),
	    reply(ReplyTo, ok),
            {loaded, Tab};
	AccessMode == read_only ->
	    disc_load_table(Tab, Reason, ReplyTo);
	true ->
	    %% Either we cannot read the table yet
	    %% or someone is moving a replica between
	    %% two nodes
	    Cs =  Load#net_load.cstruct,
	    Res = mnesia_loader:net_load_table(Tab, Reason, Active, Cs),
	    case Res of
		yes ->
		    user_sync_tab(Tab),
		    reply(ReplyTo, Res),
		    {loaded, Tab};
		no_need ->
		    reply(ReplyTo, Res),
		    {loaded, Tab};
		no when Reason == add_table_copy ->
		    reply(ReplyTo, Res),
		    {not_loaded, Tab};
		no ->
		    {not_loaded, Tab}
	    end
    end;

load_table(Load) when record(Load, disc_load) ->
    Tab = Load#disc_load.table,
    Reason =  Load#disc_load.reason,
    ReplyTo = Load#disc_load.reply_to,
    ReadNode = val({Tab, where_to_read}),
    Active = filter_active(Tab),
    if
	Active == [], ReadNode == nowhere ->
	    %% Not loaded anywhere, lets load it from disc
	    disc_load_table(Tab, Reason, ReplyTo);
	ReadNode == nowhere ->
	    %% Already loaded on other node, lets get it
	    Cs = val({Tab, cstruct}),
	    case mnesia_loader:net_load_table(Tab, Reason, Active, Cs) of
		yes ->
		    user_sync_tab(Tab),
		    {loaded, Tab};
		no_need ->
		    {loaded, Tab};
		no ->
		    disc_load_table(Tab, Reason, ReplyTo)
	    end;
	true ->
	    %% Already readable, do not worry be happy
	    {loaded, Tab}
    end.

filter_active(Tab) ->
    ByForce = val({Tab, load_by_force}),
    Active = val({Tab, active_replicas}),
    Masters = mnesia_recover:get_master_nodes(Tab),
    do_filter_active(Tab, ByForce, Active, Masters).

do_filter_active(Tab, true, Active, Masters) ->
    Active;
do_filter_active(Tab, false, Active, []) ->
    Active;
do_filter_active(Tab, false, Active, Masters) ->
    mnesia_lib:intersect(Active, Masters).
    
disc_load_table(Tab, Reason, ReplyTo) ->
    Res = mnesia_loader:disc_load_table(Tab, Reason),
    if
	Res == ok ->
	    announce_tab_is_here(Tab),
	    reply(ReplyTo, Res),
	    {loaded, Tab};
	ReplyTo /= undefined ->
	    reply(ReplyTo, Res),
	    {loaded, Tab};
	true ->
	    fatal("Cannot load table ~p from disc: ~p~n", [Tab, Res])
    end.

%% We have just been able to load Tab, from disc
%% Set all current variable correctly, also
%% make sure that the table is loaded on the nodes that need it
%% by multicasting the i_have_tab message
%% Returns list of nodes

announce_tab_is_here(Tab) ->
    i_have_tab(Tab),
    if
	Tab == schema ->
	    ignore;
	true ->
	    %% Local node needs to perform user_sync_tab/1
	    Ns = val({current, db_nodes}),
	    abcast(Ns, {i_have_tab, Tab, node()})
    end.

i_have_tab(Tab) ->
    case val({Tab, local_content}) of
	true ->
	    mnesia_lib:set_local_content_whereabouts(Tab);
	false ->
	    set({Tab, where_to_read}, node())
    end,
    add_active_replica(Tab, node()).

sync_add_table_copy_whereabouts(Tab, ToNode, RemoteCs) when Tab /= schema ->
    Current = val({current, db_nodes}),
    Ns =
	case lists:member(ToNode, Current) of
	    true -> Current;
	    false -> [ToNode | Current]
	end,
    Args = [Tab, ToNode, RemoteCs],
    rpc:multicall(Ns, ?MODULE, add_active_replica, Args),
    ok.

sync_del_table_copy_whereabouts(Tab, ToNode) when Tab /= schema ->
    Current = val({current, db_nodes}),
    Ns =
	case lists:member(ToNode, Current) of
	    true -> Current;
	    false -> [ToNode | Current]
	end,
    Args = [Tab, ToNode],
    rpc:multicall(Ns, ?MODULE, unannounce_add_table_copy, Args),
    ok.

%% node To now has tab loaded.
%% This code is rpc:call'ed from the tab_copier process
%% while it still holds the table lock
%%
%% do_sync_add_table_copy_whereabouts(Tab, FromNode, ToNode, Cs) ->
%%     if
%% 	   ToNode == node() ->
%% 	       i_have_tab(Tab);
%% 	   true ->
%% 	       add_active_replica(Tab, ToNode, Cs)
%%     end.
 
get_info(Timeout) ->
    case whereis(?SERVER_NAME) of
	undefined ->
	    {timeout, Timeout};
	Pid ->
	    Pid ! {self(), get_state},
	    receive
		{?SERVER_NAME, State} when record(State, state) ->
		    {info,State}
	    after Timeout ->
		    {timeout, Timeout}
	    end
    end.

info() ->
    Tabs = mnesia_lib:local_active_tables(),
    io:format( "---> Active tables <--- ~n", []),
    info(Tabs).

info([Tab | Tail]) ->
    case val({Tab, storage_type}) of
	disc_only_copies ->
	    info_format(Tab, 
			dets:info(Tab, size), 
			dets:info(Tab, file_size),
			"bytes on disc");
	_ ->
	    info_format(Tab, 
			erlang:db_info(Tab, size),
			erlang:db_info(Tab, memory),
			"words of mem")
    end,
    info(Tail);
info([]) -> ok;
info(Tab) -> info([Tab]).

info_format(Tab, Size, Mem, Media) ->
    StrT = mnesia_lib:pad_name(atom_to_list(Tab), 15, []),
    StrS = mnesia_lib:pad_name(integer_to_list(Size), 8, []),
    StrM = mnesia_lib:pad_name(integer_to_list(Mem), 8, []),
    io:format("~s: with ~s records occupying ~s ~s~n",
	      [StrT, StrS, StrM, Media]).

%% Handle early arrived messages
handle_early_msgs([Msg | Msgs], State) ->
    %% The messages are in reverse order
    case handle_early_msgs(Msgs, State) of
        {stop, Reason, Reply, State2} ->
	    {stop, Reason, Reply, State2};
        {stop, Reason, State2} ->
	    {stop, Reason, State2};
	{noreply, State2} ->
	    handle_early_msg(Msg, State2);
	{noreply, State2, Timeout} ->
	    handle_early_msg(Msg, State2)
    end;
handle_early_msgs([], State) ->
    noreply(State).

handle_early_msg({call, Msg, From}, State) ->
    handle_call(Msg, From, State);
handle_early_msg({cast, Msg}, State) ->
    handle_cast(Msg, State);
handle_early_msg({info, Msg}, State) ->
    handle_info(Msg, State).
    
noreply(State) when State#state.load_queue /= [],
		    State#state.schema_is_merged == true,
		    State#state.is_blocked == false ->
    {noreply, State, 0};
noreply(State) ->
    {noreply, State}.

reply(undefined, Reply) ->
    Reply;
reply(ReplyTo, Reply) ->
    gen_server:reply(ReplyTo, Reply),
    Reply.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Worker management

start_worker(What, Mod, Fun, Args) ->
    spawn_link(?MODULE, run_worker, [self(), What, Mod, Fun, Args]).

run_worker(LinkTo, What, Mod, Fun, Args) ->
    %% Don't trap exit
    case catch apply(Mod, Fun, Args) of
	{'EXIT', Reason} ->
	    fatal("~p -> ~p:~p(~p) crashed: ~p~n",
		  [What, Mod, Fun, Args, Reason]);
	Res ->
	    Reply = #worker_reply{what = What, pid = self(), result = Res},
	    LinkTo ! Reply,
	    unlink(LinkTo),
	    exit(normal)
    end.

handle_worker_reply(WR, State) ->
    case WR#worker_reply.what of
	schema_transaction ->
	    ReplyTo = WR#worker_reply.pid,
	    reply(ReplyTo, WR#worker_reply.result),
	    State;
	{disc_load_table, mnesia_loader} ->
	    State;
	{net_load_table, mnesia_loader} ->
	    State;
	{dump_log, mnesia_dumper} ->
	    State
    end.

pick_worker(Pid, [W | Tail], Acc) when W#worker.pid == Pid ->
    {W, Acc ++ Tail};
pick_worker(Pid, [W | Tail], Acc) ->
    pick_worker(Pid, Tail, [W | Acc]);
pick_worker(Pid, [], Acc) ->
    {no_worker, Acc}.


