%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(mnesia_dumper).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('klacke@erix.ericsson.se').
-author('hakan@erix.ericsson.se').

-behaviour(gen_fsm).

%% Public interface
-export([
	 dump_log/1,
	 get_log_writes/0,
	 incr_log_writes/0,
	 init/0,
	 raw_dump_table/2,
	 raw_named_dump_table/2,
	 start/0,
	 update/3
	]).

 %% Internal stuff
-export([perform_dump/2, regulator_init/1]).
	
%% gen_fsm callbacks
-export([init/1,
	 waiting/2,
	 waiting/3,
	 dumping/2,
	 dumping/3,
	 handle_event/3,
	 handle_sync_event/4,
	 handle_info/3,
	 terminate/3,
	 code_change/3
	]).

-include("mnesia.hrl").
-import(mnesia_lib, [fatal/2, dbg_out/2, error/2]).

-record(state, {initiated_by = nobody,
		dumper = nopid,
		regulator_pid,
		supervisor_pid,
		queue = [],
		timeout}).

start() ->
    gen_fsm:start_link({local, ?MODULE}, ?MODULE, [self()],
		       [{timeout, infinity}]).

cast(Msg) ->
    case whereis(?MODULE) of
	undefined -> ignore;
	Pid ->  gen_fsm:send_event(Pid, Msg)
    end.

call(Msg) ->
    case whereis(?MODULE) of
	undefined -> {error, {node_not_running, node()}};
	Pid -> gen_fsm:sync_send_event(Pid, Msg, infinity)
    end.

init() ->
    case mnesia_monitor:use_dir() of
	true ->
	    mnesia_log:init(),
	    P = dump_log(startup), % Dump previous log
	    L = dump_log(startup), % Dump latest log
	    Msg = "Initial dump of log during startup: ~p~n",
	    mnesia_lib:verbose(Msg, [[P, L]]);
	false ->
	    ignore
    end.

%% The InitBy arg may be one of the following:
%% scan_decisions  Initial scan for decisions
%% startup	   Initial dump during startup
%% schema_begin	   Dump initiated at beginning of schema transaction
%% schema_end	   Dump initiated at end of schema transaction
%% schema_simple   Do NOT dump, skip simple schema ops.
%% user		   Dump initiated by user
%% write_threshold Automatic dump caused by too many log writes
%% time_threshold  Automatic dump caused by timeout

dump_log(InitBy) ->
    call({sync_dump_log, InitBy}).

get_log_writes() ->
    Max = mnesia_monitor:get_env(dump_log_write_threshold),
    Prev = mnesia_lib:read_counter(trans_log_writes),
    Left = mnesia_lib:read_counter(trans_log_writes_left),
    Diff = Max - Left,
    Prev + Diff.

incr_log_writes() ->
    Left = mnesia_lib:incr_counter(trans_log_writes_left, -1),
    if
	Left > 0 ->
	    ignore;
	true ->
	    adjust_log_writes(true)
    end.

adjust_log_writes(Cast) ->
    Token = {mnesia_adjust_log_writes, self()},
    case global:set_lock(Token, [node()], 1) of 
	false -> ignore; %% Somebody else is sending a dump request
	true -> 
	    case Cast of 
		true -> 
		    cast({async_dump_log, write_threshold});
		_ -> ignore
	    end,
	    Max = mnesia_monitor:get_env(dump_log_write_threshold),
	    Left = mnesia_lib:read_counter(trans_log_writes_left),
	    %% Don't care if we lost a few writes
	    mnesia_lib:set_counter(trans_log_writes_left, Max),
	    Diff = Max - Left,
	    mnesia_lib:incr_counter(trans_log_writes, Diff),
	    global:del_lock(Token, [node()])
    end.


%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%%   init(Args)
%%     ==> {ok, StateName, StateData}
%%         {ok, StateName, StateData, Timeout}
%%         ignore
%%         {stop, Reason}
%%----------------------------------------------------------------------

init([Parent]) ->
    process_flag(trap_exit, true),
    mnesia_lib:verbose("~p starting: ~p~n", [?MODULE, self()]),
    Timeout = mnesia_monitor:get_env(dump_log_time_threshold),
    LoadRegulation = mnesia_monitor:get_env(dump_log_load_regulation),
    Regulator = init_regulator(LoadRegulation),
    State = #state{supervisor_pid = Parent,
		   regulator_pid = Regulator,
		   timeout = Timeout},
    {ok, waiting, State, Timeout}.

%%----------------------------------------------------------------------
%%   StateName(Msg, StateData)
%%
%%    ==> {next_state, NewStateName, NewStateData}
%%        {next_state, NewStateName, NewStateData, Timeout}
%%        {stop, Reason, NewStateData}
%%              Reason = normal | shutdown | Term terminate(State) is called
%%----------------------------------------------------------------------

waiting({async_dump_log, InitBy}, S) ->
    adjust_log_writes(false),    
    S2 = spawn_link_dumper(nopid, S, InitBy),
    {next_state, dumping, S2, S#state.timeout};
waiting(timeout, S) ->
    S2 = spawn_link_dumper(nopid, S, time_threshold),
    {next_state, dumping, S2, S#state.timeout}.

dumping({async_dump_log, InitBy}, S)
        when S#state.initiated_by == write_threshold ->
    DetectedBy = {dump_log, write_threshold},
    mnesia_lib:report_system_event({mnesia_overload, DetectedBy}),
    adjust_log_writes(false),
    S2  = append_q({async_dump_log, InitBy}, nopid, S),
    {next_state, dumping, S2, S#state.timeout};
dumping({async_dump_log, InitBy}, S) ->
    adjust_log_writes(false),
    S2  = append_q({async_dump_log, InitBy}, nopid, S),
    {next_state, dumping, S2, S#state.timeout};

dumping(timeout, S) when S#state.initiated_by == time_threshold ->
    InitBy = time_threshold,
    DetectedBy = {dump_log, time_threshold},
    mnesia_lib:report_system_event({mnesia_overload, DetectedBy}),
    adjust_log_writes(false),
    S2  = append_q({async_dump_log, InitBy}, nopid, S),
    {next_state, dumping, S2, S#state.timeout};
dumping(timeout, S) ->
    adjust_log_writes(false),
    S2  = append_q({async_dump_log, time_threshold}, nopid, S),
    {next_state, dumping, S2, S#state.timeout}.

%%----------------------------------------------------------------------
%%   StateName(Msg, From, StateData)
%%
%%    ==> {next_state, NewStateName, NewStateData}
%%        {next_state, NewStateName, NewStateData, Timeout}
%%        {reply, Reply, NewStateName, NewStateData}
%%        {reply, Reply, NewStateName, NewStateData, Timeout}
%%        {stop, Reason, NewStateData}
%%              Reason = normal | shutdown | Term terminate(State) is called
%%----------------------------------------------------------------------

waiting({sync_dump_log, InitBy}, From, S) ->
    adjust_log_writes(false),    
    S2 = spawn_link_dumper(From, S, InitBy),
    {next_state, dumping, S2, S#state.timeout}.

dumping({sync_dump_log, InitBy}, From, S) ->
    adjust_log_writes(false),    
    S2  = append_q({sync_dump_log, InitBy}, From, S),
    {next_state, dumping, S2, S#state.timeout}.

%%----------------------------------------------------------------------
%%   handle_event(Msg, StateName, StateData)
%%
%%    ==> {next_state, NewStateName, NewStateData}
%%        {next_state, NewStateName, NewStateData, Timeout}
%%        {stop, Reason, Reply, NewStateData}
%%        {stop, Reason, NewStateData}
%%              Reason = normal | shutdown | Term terminate(State) is called
%%----------------------------------------------------------------------

handle_event(Msg, StateName, S) ->
    error("~p got unexpected event: ~p~n", [?MODULE, Msg]),
    {next_state, StateName, S, S#state.timeout}.

%%----------------------------------------------------------------------
%%   handle_sync_event(Msg, From, StateName, StateData)
%%
%%    ==> {next_state, NewStateName, NewStateData}
%%        {next_state, NewStateName, NewStateData, Timeout}
%%        {reply, Reply, NewStateName, NewStateData}
%%        {reply, Reply, NewStateName, NewStateData, Timeout}
%%        {stop, Reason, Reply, NewStateData}
%%        {stop, Reason, NewStateData}
%%              Reason = normal | shutdown | Term terminate(State) is called
%%----------------------------------------------------------------------

handle_sync_event(Msg, From, StateName, StateData) ->
    error("~p got unexpected sync_event: ~p~n", [?MODULE, Msg]),
    {next_state, StateName, StateData, StateData#state.timeout}.

%%----------------------------------------------------------------------
%%   handle_info(Info, StateName, StateData)
%%      (e.g. {'EXIT', P, R}, {nodedown, N}, ...
%%
%%    ==> {next_state, NewStateName, NewStateData}
%%        {next_state, NewStateName, NewStateData, Timeout}
%%        {stop, Reason, NewStateData}
%%              Reason = normal | shutdown | Term terminate(State) is called
%%----------------------------------------------------------------------

handle_info({'EXIT', P, _}, waiting, S) when P == S#state.supervisor_pid ->
    {stop, shutdown, S};

handle_info({'EXIT', P, _}, N, S) when P == S#state.supervisor_pid ->
    S2  = append_q(stop, nopid, S),
    {next_state, N, S2, S#state.timeout};

handle_info({'EXIT', Pid, shutdown}, N, S)
        when N == dumping, Pid == S#state.dumper ->
    handle_q(N, S);

handle_info({'EXIT', Pid, Reason}, N, S)
        when N == dumping, Pid == S#state.dumper ->
    {stop, {dumper_crashed, Reason}, S};

handle_info({'EXIT', Pid, Reason}, StateName, S) ->
    Q = S#state.queue,
    case lists:keymember(Pid, 2, Q) of
	true ->
	    S2 = S#state{queue = lists:keydelete(Pid, 2, Q)},
	    dbg_out("Dumper client ~p died: ~p~n", [Pid, Reason]),
	    {next_state, StateName, S2, S#state.timeout};
	false ->
	    error("Dumper got {'EXIT', ~p, ~p} ~p~n", [Pid, Reason, S]),
	    {next_state, StateName, S, S#state.timeout}
    end;

handle_info(Msg, StateName, S) ->
    error("~p got unexpected info: ~p~n", [?MODULE, Msg]),
    {next_state, StateName, S, S#state.timeout}.

%%----------------------------------------------------------------------
%%   terminate(Reason, StateName, StateData) Let the user module clean up
%%        always called when server terminates
%%
%%    ==> the return value is ignored
%%----------------------------------------------------------------------

terminate(Reason, StateName , StateData) ->
    mnesia_monitor:terminate_proc(?MODULE, Reason, {StateName, StateData}).

%%----------------------------------------------------------------------
%% Func: code_change/3
%% Purpose: Upgrade process when its code is to be changed
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(OldVsn, State, Extra) ->
    exit(not_supported).

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

append_q(Event, From, StateData) ->
    catch link(element(1, From)), % Don't bother about nopid's
    ToDo = [{Event, From}],
    Q = StateData#state.queue,
    StateData#state{queue = Q ++ ToDo}.

%% Handle queue after dump
handle_q(dumping, S) ->
    handle_q(dumping, S, S#state.queue).

handle_q(dumping, S, [{{sync_dump_log, InitBy}, ReplyTo} | Q]) ->
    reply(ReplyTo, dumped),
    handle_q(waiting, S, Q);
handle_q(waiting, S, [{{sync_dump_log, InitBy}, ReplyTo} | Q]) ->
    S2 = spawn_link_dumper(ReplyTo, S#state{queue = Q}, InitBy),
    {next_state, dumping, S2, S#state.timeout};

handle_q(dumping, S, [{{async_dump_log, InitBy}, nopid} | Q]) ->
    handle_q(waiting, S, Q);
handle_q(waiting, S, [{{async_dump_log, InitBy}, nopid} | Q]) ->
    S2 = spawn_link_dumper(nopid, S#state{queue = Q}, InitBy),
    {next_state, dumping, S2, S#state.timeout};

handle_q(waiting, S, [{stop, ReplyTo} | Q]) ->
    reply(ReplyTo, stopped),
    {stop, shutdown, S#state{queue = Q}};
handle_q(waiting, S, []) ->
    S2 = S#state{initiated_by = nobody, dumper = nopid, queue = []},
    {next_state, waiting, S2, S#state.timeout}.

reply(nopid, Reply) ->
    ignore;
reply(ReplyTo, Reply) ->
    {Pid, _} = ReplyTo,
    unlink(Pid),
    gen_fsm:reply(ReplyTo, Reply).

spawn_link_dumper(From, S, InitBy) ->
    Args = [InitBy, S#state.regulator_pid],
    Pid = proc_lib:spawn_link(?MODULE, perform_dump, Args),
    S2 = S#state{initiated_by = InitBy, dumper = Pid},
    append_q({sync_dump_log, InitBy}, From, S2).

%% Scan for decisions
perform_dump(InitBy, Regulator) when InitBy == scan_decisions ->
    ?eval_debug_fun({?MODULE, perform_dump}, [InitBy]),
    
    dbg_out("Transaction log dump initiated by ~w~n", [InitBy]),
    scan_decisions(mnesia_log:previous_log_file(), InitBy, Regulator),
    scan_decisions(mnesia_log:latest_log_file(), InitBy, Regulator),
    exit(shutdown);

%% Propagate the log into the DAT-files
perform_dump(InitBy, Regulator) ->
    ?eval_debug_fun({?MODULE, perform_dump}, [InitBy]),
    LogState = mnesia_log:prepare_log_dump(InitBy),
    dbg_out("Transaction log dump initiated by ~w: ~w~n",
	    [InitBy, LogState]),
    mnesia_recover:allow_garb(),
    Res =
	case LogState of
	    already_dumped ->
		shutdown;
	    {needs_dump, Diff} ->
		U = mnesia_monitor:get_env(dump_log_update_in_place),
		Cont = mnesia_log:init_log_dump(),
		case do_perform_dump(Cont, U, InitBy, Regulator) of
		    ok ->
			?eval_debug_fun({?MODULE, post_dump}, [InitBy]),
			case mnesia_monitor:use_dir() of
			    true ->
				mnesia_recover:dump_decision_tab();
			    false ->
				mnesia_log:purge_some_logs()
			end,
			%% And now to the crucial point...
			mnesia_log:confirm_log_dump(Diff);
		    {error, Reason} ->
			{error, Reason}
		end;
	    {error, Reason} ->
		{error, {"Cannot prepare log dump", Reason}}
	end,
    exit(Res).

scan_decisions(Fname, InitBy, Regulator) ->
    Exists = mnesia_lib:exists(Fname),
    case Exists of
	false ->
	    ok;
	true ->
	    Header = mnesia_log:trans_log_header(),
	    Name = previous_log,
	    mnesia_log:open_log(Name, Header, Fname, Exists),
	    Cont = start,
	    Res = do_perform_dump(Cont, false, InitBy, Regulator),
	    mnesia_log:close_log(Name),
	    case Res of
		ok ->
		    Res;
		_ ->
		    exit(Res)
	    end
    end.

do_perform_dump(Cont, InPlace, InitBy, Regulator) ->
    case mnesia_log:chunk_log(Cont) of
	{C2, Recs} ->
	    case catch insert_recs(Recs, InPlace, InitBy, Regulator) of
		{'EXIT', Reason} ->
		    Error = {error, {"File access error", Reason}},
		    close_files(InPlace, Error, InitBy),
		    Error;
		_ ->
		    do_perform_dump(C2, InPlace, InitBy, Regulator)
	    end;
	eof ->
	    close_files(InPlace, ok, InitBy),
	    ok
    end.

insert_recs([Rec | Recs], InPlace, InitBy, Regulator) ->
    regulate(Regulator),
    insert_rec(Rec, InPlace, InitBy),
    insert_recs(Recs, InPlace, InitBy, Regulator);
insert_recs([], _InPlace, _InitBy, _Regulator) ->
    ok.

insert_rec(Rec, InPlace, scan_decisions) ->
    if 
	record(Rec, commit) ->
	    ignore;
	record(Rec, log_header) ->
	    ignore;
	true ->
	    mnesia_recover:note_log_decision(Rec, scan_decisions)
    end;
insert_rec(Rec, InPlace, InitBy) when record(Rec, commit) ->
    %% Determine the Outcome of the transaction and recover it
    D = Rec#commit.decision,
    case mnesia_recover:wait_for_decision(D, InitBy) of
	{Tid, committed} ->
	    do_insert_rec(Tid, Rec, InPlace, InitBy);
	{Tid, aborted} ->
	    mnesia_schema:undo_prepare_commit(Tid, Rec)
    end;
insert_rec(H, InPlace, InitBy) when record(H, log_header) ->
    V = mnesia_log:version(),
    if
        H#log_header.log_kind /= trans_log ->
	    fatal("Bad kind of transaction log: ~p~n", [H]);
	H#log_header.log_version /= V,
	H#log_header.log_version /= "4.0" -> % Backwards compatibility
	    fatal("Bad version of transaction log: ~p~n", [H]);
	true ->
	    ok
    end;

insert_rec(Rec, InPlace, InitBy) ->
    ok.

do_insert_rec(Tid, Rec, InPlace, InitBy) ->
    case Rec#commit.schema_ops of
	[] ->
	    ignore;
	SchemaOps ->
	    case val({schema, storage_type}) of 
		ram_copies -> 
		    insert_ops(Tid, schema_ops, SchemaOps, InPlace, InitBy);
	        Storage ->
		    true = open_files(schema, Storage, InPlace, InitBy),
		    insert_ops(Tid, schema_ops, SchemaOps, InPlace, InitBy)
	    end
    end,
    D = Rec#commit.disc_copies,
    insert_ops(Tid, disc_copies, D, InPlace, InitBy),
	    case InitBy of
	startup ->
	    DO = Rec#commit.disc_only_copies,
	    insert_ops(Tid, disc_only_copies, DO, InPlace, InitBy);
	_ ->
	    ignore
    end.
    

update(Tid, SchemaOps, UseDumper) ->
    UseDir = mnesia_monitor:use_dir(),
    InPlace = mnesia_monitor:get_env(dump_log_update_in_place),
    if 
	UseDir == false ->
	    %% Perform the updates here and now, since we have no log
%	    InitBy = schema_simple, 
	    InitBy = schema_end,
	    case catch insert_ops(Tid, schema_ops, SchemaOps, InPlace, InitBy) of
		{'EXIT', Reason} ->
		    Error = {error, {"Schema update error", Reason}},
		    close_files(InPlace, Error, InitBy),
		    Error;
		_ ->
		    ok
	    end;
%%  Removed temporary until debugged proparly

%       UseDumper == ignore_dumper ->
%	    %% Perform the updates here and now, since we don't care about the log
%	    InitBy = schema_simple,
%	    case catch insert_ops(Tid, schema_ops, SchemaOps, InPlace, InitBy) of
%		{'EXIT', Reason} ->
%		    Error = {error, {"Schema update error", Reason}},
%		    close_files(InPlace, Error, InitBy),
%		    Error;
%		_ ->		    
%		    close_files(InPlace, ok, InitBy),
%		    ok
%	    end;
%       UseDumper == use_dumper ->
       true ->
	    InitBy = schema_end,
	    ?eval_debug_fun({?MODULE, dump_schema_op}, [InitBy]),
	    %% Force a dump of the transaction log in order to let the
	    %% dumper perform needed updates
	    dump_log(InitBy)
    end.

%% Insert the items in reverse order
insert_ops(Tid, Storage, [Op | Ops], InPlace, InitBy) ->
    insert_ops(Tid, Storage, Ops, InPlace, InitBy),
    insert_op(Tid, Storage, Op, InPlace, InitBy);
insert_ops(Tid, Storage, [], InPlace, InitBy) ->
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Normal ops

disc_insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy) ->
    case open_files(Tab, Storage, InPlace, InitBy) of
	true ->
	    case Op of
		write ->
		    ok = dets:insert(Tab, Val);
		delete ->
		    ok = dets:delete(Tab, Key);
		update_counter ->
		    {RecName, Incr} = 
			if
			    integer(Val) ->
				%% For backward compatibility
				{mnesia_lib:val({Tab, record_name}), Val};
			    true ->
				Val
			end,
		    case catch dets:update_counter(Tab, Key, Incr) of
			CounterVal when integer(CounterVal) ->
			    ok;
			_ ->
			    Zero = {RecName, Key, 0},
			    ok = dets:insert(Tab, Zero)
		    end;
		delete_object ->
		    ok = dets:delete_object(Tab, Val);
		clear_table ->
		    ok = dets:match_delete(Tab, '_')
	    end;
	false ->
	    ignore
    end.

insert(Tid, Storage, Tab, Key, [Val | Tail], Op, InPlace, InitBy) ->
    insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy),
    insert(Tid, Storage, Tab, Key, Tail, Op, InPlace, InitBy);

insert(Tid, Storage, Tab, Key, [], Op, InPlace, InitBy) ->
    ok;
    
insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy) ->
    Item = {{Tab, Key}, Val, Op},
    case InitBy of
	startup ->
	    disc_insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy);

	_ when Storage == ram_copies ->
	    mnesia_tm:do_update_op(Tid, Storage, Item),
	    Snmp = mnesia_tm:prepare_snmp(Tab, Key, [Item]),
	    mnesia_tm:do_snmp(Tid, Snmp);

	_ when Storage == disc_copies ->
	    disc_insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy),
	    mnesia_tm:do_update_op(Tid, Storage, Item),
	    Snmp = mnesia_tm:prepare_snmp(Tab, Key, [Item]),
	    mnesia_tm:do_snmp(Tid, Snmp);

	_ when Storage == disc_only_copies ->
	    mnesia_tm:do_update_op(Tid, Storage, Item),
	    Snmp = mnesia_tm:prepare_snmp(Tab, Key, [Item]),
	    mnesia_tm:do_snmp(Tid, Snmp);

	_ when Storage == unknown ->
	    ignore
    end.

disc_delete_table(Tab, Storage) ->
    case mnesia_monitor:use_dir() of
	true ->
	    mnesia_monitor:unsafe_close_dets(Tab),
	    Dat = mnesia_lib:tab2dat(Tab),
	    file:delete(Dat),
	    erase({?MODULE, Tab});
	false ->
	    ignore
    end.

disc_delete_indecies(Tab, Cs, Storage) when Storage /= disc_only_copies ->
    ignore;
disc_delete_indecies(Tab, Cs, disc_only_copies) ->
    Indecies = Cs#cstruct.index,
    lists:foreach(fun(Pos) ->
			  disc_delete_index(Tab, disc_only_copies, Pos)
		  end,
		  Indecies).

disc_delete_index(Tab, Storage, Pos) when Storage /= disc_only_copies -> 
    ignore;
disc_delete_index(Tab, Storage, Pos) ->
    IxTag = {Tab, index, Pos},
    mnesia_lib:del({Tab, index_tables}, {Pos, {dets, IxTag}}),
    mnesia_lib:unset({Tab, {index, Pos}}),
    mnesia_monitor:unsafe_close_dets(IxTag),
    IxName = mnesia_index:tab2filename(Tab, Pos),
    file:delete(IxName).

insert_op(Tid, Storage, {{Tab, Key}, Val, Op}, InPlace, InitBy) ->
    %% Propagate to disc only
    disc_insert(Tid, Storage, Tab, Key, Val, Op, InPlace, InitBy);

insert_op(Tid, _, {op, rec, Storage, Item}, InPlace, InitBy) ->
    {{Tab, Key}, ValList, Op} = Item,
    insert(Tid, Storage, Tab, Key, ValList, Op, InPlace, InitBy);

insert_op(Tid, _, {op, change_table_copy_type, N, FromS, ToS, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Val = mnesia_schema:insert_cstruct(Tid, Cs, true), % Update ram only
    {schema, Tab, _} = Val,
    if
	InitBy /= startup ->
	    mnesia_controller:add_active_replica(Tab, N, Cs);
	true ->
	    ignore
    end,
    if
	N == node() ->
	    Dmp  = mnesia_lib:tab2dmp(Tab),
	    Dat  = mnesia_lib:tab2dat(Tab),
	    case {FromS, ToS} of
		{ram_copies, disc_copies} ->
		    ok = ensure_rename(Dmp, Dat);
		{disc_copies, ram_copies} when Tab == schema ->
		    mnesia_monitor:unsafe_close_dets(Tab),
		    file:delete(Dat),
		    mnesia_lib:set(use_dir, false);
		{disc_copies, ram_copies} ->
		    file:delete(Dat);
		{ram_copies, disc_only_copies} ->
		    ok = ensure_rename(Dmp, Dat),
		    PosList = Cs#cstruct.index,
		    true = open_files(Tab, disc_only_copies, InPlace, InitBy),
		    %% ram_delete_table must be done before init_indecies, 
		    %% it uses info which is reset in init_indecies,
		    %% it doesn't matter, because init_indecies don't use 
		    %% the ram replica of the table when creating the disc 
		    %% index; Could be improved :)
		    mnesia_schema:ram_delete_table(Tab, FromS, PosList),
		    mnesia_index:init_indecies(Tab, disc_only_copies, PosList);
		{disc_only_copies, ram_copies} ->
		    mnesia_monitor:unsafe_close_dets(Tab),
		    disc_delete_indecies(Tab, Cs, disc_only_copies),
		    case InitBy of 
			startup -> 
			    ignore;
			_ -> 
			    mnesia_controller:get_disc_copy(Tab)
		    end,
		    disc_delete_table(Tab, disc_only_copies);
		{disc_copies, disc_only_copies} ->
		    PosList = Cs#cstruct.index,
		    true = open_files(Tab, disc_only_copies, InPlace, InitBy),
		    mnesia_schema:ram_delete_table(Tab, FromS, PosList),
		    mnesia_index:init_indecies(Tab, disc_only_copies, PosList);
		{disc_only_copies, disc_copies} ->
		    mnesia_monitor:unsafe_close_dets(Tab),
		    disc_delete_indecies(Tab, Cs, disc_only_copies),
		    mnesia_controller:get_disc_copy(Tab)
	    end;
	true ->
	    ignore
    end,
    S = val({schema, storage_type}),
    disc_insert(Tid, S, schema, Tab, Val, write, InPlace, InitBy);

insert_op(Tid, _, {op, dump_table, Size, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    case lists:member(node(), Cs#cstruct.ram_copies) of
	true ->
	    %% Assume that table files already are closed
	    Tab = Cs#cstruct.name,
	    Dmp = mnesia_lib:tab2dmp(Tab),
	    Dat = mnesia_lib:tab2dat(Tab),
	    file:delete(Dat),
	    if
		Size == 0 ->       file:delete(Dmp);
		Size /= unknown -> ok = ensure_rename(Dmp, Dat)
	    end;
	false ->
	    ignore
    end;

insert_op(Tid, S, {op, transform, Fun, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

						%insert_op(Tid, schema_ops, _OP, InPlace, Initby) when 
						%  Initby /= startup, Initby /= schema_simple -> 
						%    ok;

%%%  Operations below this are handled without using the logg.

insert_op(Tid, _, {op, create_table, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, false, InPlace, InitBy),
    Tab = Cs#cstruct.name,

    Copies = mnesia_lib:copy_holders(Cs),
    Active = mnesia_lib:intersect(Copies, val({current, db_nodes})),
    [mnesia_controller:add_active_replica(Tab, N, Cs) || N <- Active],

    case mnesia_lib:cs_to_storage_type(node(), Cs) of
	unknown ->
	    case Cs#cstruct.local_content of
                true ->
		    ignore;
		false ->
		    case Active of
			[SomeNode | _] ->
			    mnesia_lib:set({Tab, where_to_read}, SomeNode);
			[] ->
			    ignore
		    end
	    end;
	Storage ->
	    case Cs#cstruct.local_content of
		true ->
                    mnesia_lib:set_local_content_whereabouts(Tab);
		false ->
		    mnesia_lib:set({Tab, where_to_read}, node())
	    end,
	    disc_delete_indecies(Tab, Cs, Storage),
	    disc_delete_table(Tab, Storage),

	    %% Update whereabouts and create table
	    mnesia_controller:create_table(Tab),
	    case InitBy of
		startup ->		    
		    mnesia_schema:ram_delete_table(Tab, Storage),
		    case Storage of 
			disc_only_copies ->
			    mnesia_monitor:unsafe_close_dets(Tab);
			_ -> 
			    ignore
		    end,
		    mnesia_controller:unannounce_add_table_copy(Tab, node());		
		_ ->
		    ignore
	    end
    end;

insert_op(Tid, _, {op, delete_table, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    case mnesia_lib:cs_to_storage_type(node(), Cs) of
	unknown ->
	    ignore;
	Storage ->
	    disc_delete_table(Tab, Storage),
	    disc_delete_indecies(Tab, Cs, Storage),
	    case InitBy of
		startup ->
		    ignore;
		_ ->
		    mnesia_schema:ram_delete_table(Tab, Storage),
		    mnesia_checkpoint:tm_del_copy(Tab, node())
	    end
    end,
    delete_cstruct(Tid, Cs, InPlace, InitBy);

insert_op(Tid, _, {op, clear_table, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    case mnesia_lib:cs_to_storage_type(node(), Cs) of
	unknown ->
	    ignore;
	Storage ->
	    Oid = '_', %%val({Tab, wild_pattern}),
	    insert(Tid, Storage, Tab, '_', Oid, clear_table, InPlace, InitBy)
    end;

insert_op(Tid, _, {op, merge_schema, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, false, InPlace, InitBy);

insert_op(Tid, _, {op, del_table_copy, Storage, Node, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    if
	Tab == schema, Storage == ram_copies ->
	    insert_cstruct(Tid, Cs, true, InPlace, InitBy);
        Tab /= schema ->
	    mnesia_controller:del_active_replica(Tab, Node),
	    mnesia_lib:del({Tab, Storage}, Node),
	    if
		Node == node() ->
		    case Cs#cstruct.local_content of
			true -> mnesia_lib:set({Tab, where_to_read}, nowhere);
			false -> mnesia_lib:set_remote_where_to_read(Tab)
		    end,
		    mnesia_lib:del({schema, local_tables}, Tab),
		    mnesia_lib:set({Tab, storage_type}, unknown),
		    insert_cstruct(Tid, Cs, true, InPlace, InitBy),
		    disc_delete_table(Tab, Storage),
		    disc_delete_indecies(Tab, Cs, Storage),
		    mnesia_schema:ram_delete_table(Tab, Storage),
		    mnesia_checkpoint:tm_del_copy(Tab, Node);
		true ->
		    case val({Tab, where_to_read}) of
			Node ->
			    mnesia_lib:set_remote_where_to_read(Tab);
			_  ->
			    ignore
		    end,
		    insert_cstruct(Tid, Cs, true, InPlace, InitBy)
	    end
    end;

insert_op(Tid, _, {op, add_table_copy, Storage, Node, TabDef}, InPlace, InitBy) ->
    %% During prepare commit, the files was created
    %% and the replica was announced
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, add_snmp, Us, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, del_snmp, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    if
	InitBy /= startup,
	Storage /= unknown ->
	    Stab = val({Tab, {index, snmp}}),
	    mnesia_snmp_hook:delete_table(Tab, Stab),
	    mnesia_lib:unset({Tab, {index, snmp}});
	true ->
	    ignore
    end,
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, add_index, Pos, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = insert_cstruct(Tid, Cs, true, InPlace, InitBy),
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    case InitBy of
	startup when Storage == disc_only_copies ->
	    mnesia_index:init_indecies(Tab, Storage, [Pos]);
	startup ->
	    ignore;  %% Dan ???
	_  ->
	    mnesia_index:init_indecies(Tab, Storage, [Pos])
    end;

insert_op(Tid, _, {op, del_index, Pos, TabDef}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    case InitBy of
	startup when Storage == disc_only_copies ->
	    mnesia_index:del_index(Tab, Storage, Pos);
	startup -> 
	    ignore;
	_ ->
	    mnesia_index:del_index(Tab, Storage, Pos)
    end,
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, change_table_access_mode,TabDef, OldAccess, Access}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    case InitBy of
	startup -> ignore;
	_ -> mnesia_controller:change_table_access_mode(Cs)
    end,
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, change_table_load_order, TabDef, OldLevel, Level}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, delete_property, TabDef, PropKey}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    Tab = Cs#cstruct.name,
    mnesia_lib:unset({Tab, user_property, PropKey}),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy);

insert_op(Tid, _, {op, write_property, TabDef, Prop}, InPlace, InitBy) ->
    Cs = mnesia_schema:list2cs(TabDef),
    insert_cstruct(Tid, Cs, true, InPlace, InitBy).

open_files(Tab, Storage, UpdateInPlace, InitBy)
  when Storage /= unknown, Storage /= ram_copies ->
    case get({?MODULE, Tab}) of
	undefined ->
	    case ?catch_val({Tab, setorbag}) of
		{'EXIT', _} ->
		    false;
		Type ->
		    Fname = prepare_open(Tab, UpdateInPlace, InitBy),
		    Args = [{file, Fname},
			    {keypos, 2},
			    {repair, mnesia_monitor:get_env(auto_repair)},
			    {type, Type}],
		    {ok, _} = mnesia_monitor:open_dets(Tab, Args),
		    put({?MODULE, Tab}, opened_dumper),		    
		    true
	    end;
	opened_dumper ->
	    true
    end;
open_files(Tab, Storage, UpdateInPlace, InitBy) ->
    false.

prepare_open(Tab, UpdateInPlace, InitBy) ->
    Dat =  mnesia_lib:tab2dat(Tab),
    case UpdateInPlace of
	true ->
	    Dat;
	false ->
	    Tmp = mnesia_lib:tab2tmp(Tab),
	    case catch mnesia_lib:copy_file(Dat, Tmp) of
		ok ->
		    Tmp;
		Error ->
		    fatal("Cannot copy dets file ~p to ~p: ~p~n",
			  [Dat, Tmp, Error])
	    end
	end.

close_files(UpdateInPlace, Outcome, InitBy) -> % Update in place
    close_files(UpdateInPlace, Outcome, InitBy, get()).

close_files(InPlace, Outcome, InitBy, [{{?MODULE, Tab}, _} | Tail]) ->
    erase({?MODULE, Tab}),
    case val({Tab, storage_type}) of
	disc_only_copies when InitBy /= startup ->
	    ignore;
	_ ->
	    do_close(InPlace, Outcome, Tab)
    end,
    close_files(InPlace, Outcome, InitBy, Tail);

close_files(InPlace, Outcome, InitBy, [_ | Tail]) ->
    close_files(InPlace, Outcome, InitBy, Tail);
close_files(_, _, InitBy, []) ->
    ok.

do_close(InPlace, Outcome, Tab) ->
    mnesia_monitor:close_dets(Tab),
    if
	InPlace == true ->
	    %% Update in place
	    ok;
	Outcome == ok ->
	    %% Success: swap tmp files with dat files
	    TabDat = mnesia_lib:tab2dat(Tab),
	    ok = file:rename(mnesia_lib:tab2tmp(Tab), TabDat);
	true ->
	    file:delete(mnesia_lib:tab2tmp(Tab))
    end.

ensure_rename(From, To) ->
    case mnesia_lib:exists(From) of
	true ->
	    file:rename(From, To);
	false ->
	    case mnesia_lib:exists(To) of
		true ->
		    ok;
		false ->
		    {error, {rename_failed, From, To}}
	    end
    end.
    
insert_cstruct(Tid, Cs, KeepWhereabouts, InPlace, InitBy) ->
    Val = mnesia_schema:insert_cstruct(Tid, Cs, KeepWhereabouts),
    {schema, Tab, _} = Val,
    S = val({schema, storage_type}),
    disc_insert(Tid, S, schema, Tab, Val, write, InPlace, InitBy),
    Tab.

delete_cstruct(Tid, Cs, InPlace, InitBy) ->
    Val = mnesia_schema:delete_cstruct(Tid, Cs),
    {schema, Tab, _} = Val,
    S = val({schema, storage_type}),
    disc_insert(Tid, S, schema, Tab, Val, delete, InPlace, InitBy),
    Tab.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Raw dump of table. Dumper must have unique access to the ets table.

raw_named_dump_table(Tab, Ftype) ->
    mnesia_lib:lock_table(Tab),
    case mnesia_monitor:use_dir() of
	true ->
	    TmpFname = mnesia_lib:tab2tmp(Tab),
	    Fname =
		case Ftype of
		    dat -> mnesia_lib:tab2dat(Tab);
		    dmp -> mnesia_lib:tab2dmp(Tab)
		end,
	    file:delete(TmpFname),
	    file:delete(Fname),
	    TabSize = erlang:db_info(Tab, size),
	    TabRef = Tab,
	    Args = [{file, TmpFname},
		    {keypos, 2},
%%		    {ram_file, true},
		    {estimated_no_objects, TabSize + 256},
		    {repair, mnesia_monitor:get_env(auto_repair)},
		    {type, val({Tab, setorbag})}],
	    case mnesia_lib:dets_sync_open(TabRef, Args) of
		{ok, TabRef} ->
		    Storage = ram_copies,
		    mnesia_lib:db_fixtable(Storage, Tab, true),
		
		    case catch raw_dump_table(TabRef, Tab) of
			{'EXIT', Reason} ->
			    mnesia_lib:db_fixtable(Storage, Tab, false),
			    mnesia_lib:dets_sync_close(Tab),
			    file:delete(TmpFname),
			    exit({"Dump of table to disc failed", Reason});
			ok ->
			    mnesia_lib:db_fixtable(Storage, Tab, false),
			    mnesia_lib:dets_sync_close(Tab),
			    ok = file:rename(TmpFname, Fname)
		    end;
		{error, Reason} ->
		    mnesia_lib:unlock_table(Tab),
		    exit({"Open of file before dump to disc failed", Reason})
	    end;
	false ->
	    exit({has_no_disc, node()})
    end.

raw_dump_table(DetsRef, EtsRef) ->
    raw_dump_table(DetsRef, EtsRef, 0).

raw_dump_table(DetsRef, EtsRef, Slot) ->
    case erlang:db_slot(EtsRef, Slot) of
	'$end_of_table' ->
	    ok;
	Recs ->
	    dets_insert(Recs, DetsRef),
	    raw_dump_table(DetsRef, EtsRef, Slot + 1)
    end.

dets_insert([Rec | Recs], DetsRef) ->
    ok = dets:insert(DetsRef, Rec),
    dets_insert(Recs, DetsRef);
dets_insert([], DetsRef) ->
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load regulator
%% 
%% This is a poor mans substitute for a fair scheduler algorithm 
%% in the Erlang emulator. The mnesia_dumper process performs many 
%% costly BIF invokations and must pay for this. But since the 
%% Emulator does not handle this properly we must compensate for 
%% this with some form of load regulation of ourselves in order to
%% not steal all computation power in the Erlang Emulator ans make
%% other processes starve. Hopefully this is a temporary solution.

init_regulator(false) ->
    nopid;
init_regulator(true) ->
    N = mnesia_dumper_load_regulator,
    case mnesia_monitor:start_proc(N, ?MODULE, regulator_init, [self()]) of
	{ok, Pid} ->
	    Pid;
	{error, Reason} ->
	    fatal("Failed to start ~n: ~p~n", [N, Reason])
    end.

regulator_init(Parent) ->
    %% No need for trapping exits.
    %% Using low priority causes the regulation
    process_flag(priority, low), 
    proc_lib:init_ack(Parent, {ok, self()}),
    regulator_loop().

regulator_loop() ->
    receive
	{regulate, From} ->
	    From ! {regulated, self()},
	    regulator_loop();
	{stop, From} ->
	    From ! {stopped, self()},
	    exit(normal)
    end.

regulate(nopid) ->
    ok;
regulate(RegulatorPid) ->
    RegulatorPid ! {regulate, self()},
    receive
	{regulated, RegulatorPid} -> ok
    end.

val(Var) ->
    case ?catch_val(Var) of
	{'EXIT', Reason} -> mnesia_lib:other_val(Var, Reason); 
	Value -> Value 
    end.
