%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%----------------------------------------------------------------------
%%% File    : mnesia_frag.erl
%%% Author  : H}kan Mattsson <hakan@erix.ericsson.se>
%%% Purpose : Support tables so large that they need
%%%           to be divided into several fragments.
%%%
%%% Created : 2 Apr 1998 by Hkan Mattsson <hakan@erix.ericsson.se>
%%%----------------------------------------------------------------------

%header_doc_include

-module(mnesia_frag).
-author('hakan@erix.ericsson.se').
-behaviour(mnesia_access).

%header_doc_include

-export([
	 fragment/1,
	 defragment/1,
	 add_fragment/1,
	 del_fragment/1,
	 add_nodes/2,
	 del_nodes/2
	]).

%% Callback functions when accessed within an activity
-export([
	 lock/4,
	 write/5, delete/5, delete_object/5,
	 read/5, match_object/5, all_keys/4,
	 index_match_object/6, index_read/6,
	 table_info/4
       ]).

-import(mnesia_lib, [val/1]).

-import(mnesia_schema,
	[schema_transaction/1,
	 insert_schema_ops/2,
	 get_tid_ts_and_lock/2,
	 make_write_table_properties/2,
	 make_write_table_properties/3,
	 make_delete_table_properties/2,
	 verify/3
	]).

-include("mnesia.hrl").

-record(mnesia_fragment_hash,
	{n_fragments = 1,
	 next_n_to_split = 1,
	 n_doubles = 0
	}).

-record(mnesia_fragment,
	{name,
	 storage_type,
	 n_replicas,
	 node_pool
	}).

-record(mnesia_fragment_master,
	{name}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Access functions

%impl_doc_include

%% Callback functions which provides transparent
%% access of fragmented tables from any activity
%% access context.

lock(ActivityId, Opaque, LockItem, LockKind) ->
    mnesia:lock(ActivityId, Opaque, LockItem, LockKind).

write(ActivityId, Opaque, Tab, Rec, LockKind) ->
    Key = element(2, Rec),
    Frag = key_to_frag_name(Tab, Key),
    mnesia:write(ActivityId, Opaque, Frag, Rec, LockKind).

delete(ActivityId, Opaque, Tab, Key, LockKind) ->
    Frag = key_to_frag_name(Tab, Key),
    mnesia:delete(ActivityId, Opaque, Frag, Key, LockKind).

delete_object(ActivityId, Opaque, Tab, Rec, LockKind) ->
    Key = element(2, Rec),
    Frag = key_to_frag_name(Tab, Key),
    mnesia:delete_object(ActivityId, Opaque, Frag, Rec, LockKind).

read(ActivityId, Opaque, Tab, Key, LockKind) ->
    Frag = key_to_frag_name(Tab, Key),
    mnesia:read(ActivityId, Opaque, Frag, Key, LockKind).

match_object(ActivityId, Opaque, Tab, Pat, LockKind) ->
    Match = [mnesia:match_object(ActivityId, Opaque, Frag, Pat, LockKind)
	     || Frag <- frag_names(Tab)],
    lists:flatten(Match).

all_keys(ActivityId, Opaque, Tab, LockKind) ->
    Match = [mnesia:all_keys(ActivityId, Opaque, Frag, LockKind)
	     || Frag <- frag_names(Tab)],
    lists:flatten(Match).

index_match_object(ActivityId, Opaque, Tab, Pat, Attr, LockKind) ->
    Match =
	[mnesia:index_match_object(ActivityId, Opaque, Frag, Pat, Attr, LockKind)
	 || Frag <- frag_names(Tab)],
    lists:flatten(Match).

index_read(ActivityId, Opaque, Tab, Key, Attr, LockKind) ->
    Match =
	[mnesia:index_read(ActivityId, Opaque, Frag, Key, Attr, LockKind)
	     || Frag <- frag_names(Tab)],
    lists:flatten(Match).

table_info(ActivityId, Opaque, {Tab, Key}, Item) ->
    Frag = key_to_frag_name(Tab, Key),
    table_info2(ActivityId, Opaque, Tab, Frag, Item);
table_info(ActivityId, Opaque, Tab, Item) ->
    table_info2(ActivityId, Opaque, Tab, Tab, Item).

table_info2(ActivityId, Opaque, Tab, Frag, Item) ->
    case Item of
	n_fragments ->
	    FH = lookup_fragment_hash(Tab),
	    FH#mnesia_fragment_hash.n_fragments;
        n_replicas ->
	    F = lookup_fragment(Tab),
	    F#mnesia_fragment.n_replicas;
	node_pool ->
	    F = lookup_fragment(Tab),
	    F#mnesia_fragment.node_pool;
	size ->
	    Sizes = [mnesia:table_info(ActivityId, Opaque, F, Item)
		     || F <- frag_names(Tab)],
	    lists:sum(Sizes);
	memory ->
	    Memories = [mnesia:table_info(ActivityId, Opaque, F, Item)
			|| F <- frag_names(Tab)],
	    lists:sum(Memories);
	fragments ->
	    frag_names(Tab);
	distribution ->
	    frag_distribution(Tab);
	_ ->
	    mnesia:table_info(ActivityId, Opaque, Frag, Item)
    end.

%impl_doc_include

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Turn a normal table into a fragmented table
%% 
%% The storage type must be the same on all nodes

fragment(Tab) ->
    schema_transaction(fun() -> do_fragment(Tab) end).

do_fragment(Tab) when atom(Tab), Tab /= schema ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, none),
    insert_schema_ops(TidTs, make_fragment(Tab));
do_fragment(Tab) ->
    mnesia:abort({bad_type, Tab}).

make_fragment(Tab) ->
    Cs = val({Tab, cstruct}),
    case catch lookup_fragment(Tab) of
	{'EXIT', _} ->
	    LC = Cs#cstruct.local_content,
	    verify(false, LC, {bad_type, Tab, local_content, LC}),
	    
	    Ns = mnesia_lib:cs_to_nodes(Cs),
	    Ss = [mnesia_lib:schema_cs_to_storage_type(N, Cs) || N <- Ns],
	    case mnesia_lib:uniq(Ss) of
		[S] ->
		    F = #mnesia_fragment{name = Tab,
					 node_pool = Ns,
					 n_replicas = erlang:length(Ns),
					 storage_type = S},
		    FH = #mnesia_fragment_hash{},
		    FM = #mnesia_fragment_master{name = Tab},
		    make_write_table_properties(Tab, [FM, F, FH], Cs);
		BadStorage ->
		    mnesia:abort({combine_error, BadStorage})
	    end;
	F ->
	    mnesia:abort({already_exists, Tab, mnesia_fragment})
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Turn a table into a normal defragmented table

defragment(Tab) ->
    schema_transaction(fun() -> do_defragment(Tab) end).

do_defragment(Tab) when atom(Tab), Tab /= schema ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, none),
    insert_schema_ops(TidTs, make_defragment(Tab));
do_defragment(Tab) ->
    mnesia:abort({bad_type, Tab}).

make_defragment(Tab) ->
    F = lookup_fragment(Tab),
    case catch lookup_fragment_hash(Tab) of
	{'EXIT', _} ->
	    make_delete_table_properties(Tab, [mnesia_fragment]);
	FH ->
	    if
		FH#mnesia_fragment_hash.n_fragments > 1 ->
		    mnesia:abort({combine_error, Tab,
				  "Too many fragments"});
		true ->
		    Keys = [mnesia_fragment, mnesia_fragment_hash],
		    make_delete_table_properties(Tab, Keys)
	    end
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Add a fragment to a fragmented table  and fill it with half of
%% the records from one of the old fragments
    
add_fragment(Tab) ->
    schema_transaction(fun() -> do_add_fragment(Tab) end).

do_add_fragment(Tab) when atom(Tab), Tab /= schema ->
    TidTs = get_tid_ts_and_lock(schema, write),
    F = lookup_fragment(Tab),
    FH = lookup_fragment_hash(Tab),
    do_add_fragment2(TidTs, Tab, F, FH);
do_add_fragment(Tab) ->
    mnesia:abort({bad_type, Tab}).

do_add_fragment2(TidTs, Tab, F, FH) ->
    SplitN = FH#mnesia_fragment_hash.next_n_to_split,
    SplitFrag = n_to_frag_name(Tab, SplitN),
    get_tid_ts_and_lock(SplitFrag, write),

    FH2 = adjust_before_split(FH),
    NewN = FH2#mnesia_fragment_hash.n_fragments,
    NewFrag = n_to_frag_name(Tab, NewN),
    FM = #mnesia_fragment_master{name = Tab},
    Storage = F#mnesia_fragment.storage_type,
    N = F#mnesia_fragment.n_replicas,
    FragDist = frag_distribution(Tab),
    Ram = frag_replicas(Storage, ram_copies, N, FragDist, Tab),
    Disc = frag_replicas(Storage, disc_copies, N, FragDist, Tab),
    DiscOnly = frag_replicas(Storage, disc_only_copies, N, FragDist, Tab),
    Cs = val({Tab, cstruct}),
    FragCs = Cs#cstruct{name = NewFrag,
			user_properties = [FM],
			ram_copies = Ram,
			disc_copies = Disc,
			disc_only_copies = DiscOnly},
    mnesia_schema:do_create_table(FragCs),
    insert_schema_ops(TidTs, make_write_table_properties(Tab, [FH2], Cs)),
    insert_schema_ops(TidTs, make_add_fragment(TidTs, Tab, FH2, SplitFrag, NewFrag, SplitN)).

frag_replicas(Storage, Storage, N, FragDist, Tab) ->
    pick_frag_replicas(N, FragDist, [], Tab);
frag_replicas(_, _, _, _, _) ->
    [].

%% Pick first N nodes in FragDist
pick_frag_replicas(0, _FragDist, Acc, _Tab) ->
    Acc;
pick_frag_replicas(N, [{Node, _Count} | Tail], Acc, Tab) ->
    pick_frag_replicas(N - 1, Tail, [Node | Acc], Tab);
pick_frag_replicas(N, [], Acc, Tab) ->
    mnesia:abort({too_few_pool_nodes, Tab}).
    
%% Adjust table info before split
adjust_before_split(FH) ->
    P = FH#mnesia_fragment_hash.next_n_to_split + 1,
    L = FH#mnesia_fragment_hash.n_doubles,
    N = FH#mnesia_fragment_hash.n_fragments + 1,
    P2= trunc(math:pow(2, L)) + 1,
    if
	P == P2 ->
	    FH#mnesia_fragment_hash{n_doubles = L + 1,
				    next_n_to_split = 1,
				    n_fragments = N};
	true ->
	    FH#mnesia_fragment_hash{next_n_to_split = P,
				    n_fragments = N}
    end.

make_add_fragment({_, Tid, Ts}, Tab, FH, SplitFrag, NewFrag, SplitN) ->
    Pat = mnesia:table_info(SplitFrag, wild_pattern),
    Recs = mnesia:match_object(Tid, Ts, SplitFrag, Pat, read),
    split(FH, SplitFrag,NewFrag, SplitN, Recs).

%% Perform the split of the table
split(FH, SplitFrag, NewFrag, SplitN, [Rec | Recs]) ->
    Key = element(2, Rec),
    case key_to_n(FH, Key) of
	N when N == SplitN ->
	    split(FH, SplitFrag, NewFrag, SplitN, Recs);
	_ ->
	    SplitOid = {SplitFrag, Key},
	    [{op, rec, unknown, {{NewFrag, Key}, [Rec], write}}, 
	     {op, rec, unknown, {SplitOid, [SplitOid], delete}} |
	     split(FH, SplitFrag, NewFrag, SplitN, Recs)]
    end;
split(FH, SplitFrag, NewFrag, SplitN, []) ->
    [].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Delete a fragment from a fragmented table
%% and merge its records with an other fragment
    
del_fragment(Tab) ->
    schema_transaction(fun() -> do_del_fragment(Tab) end).

do_del_fragment(Tab) when atom(Tab), Tab /= schema ->
    TidTs = get_tid_ts_and_lock(schema, write),
    F = lookup_fragment(Tab),
    case lookup_fragment_hash(Tab) of
	FH when FH#mnesia_fragment_hash.n_fragments > 1 ->
	    LastN = FH#mnesia_fragment_hash.n_fragments,
	    LastFrag = n_to_frag_name(Tab, LastN),
	    FH2 = adjust_before_merge(FH),

	    MergeN = FH2#mnesia_fragment_hash.next_n_to_split,
	    MergeFrag = n_to_frag_name(Tab, MergeN),
	    get_tid_ts_and_lock(MergeFrag, write),
	    mnesia_schema:do_delete_table(LastFrag),
		    
	    insert_schema_ops(TidTs, make_write_table_properties(Tab, [FH2])),
	    insert_schema_ops(TidTs, make_del_fragment(TidTs, Tab, FH2, MergeFrag, LastFrag));
	FH ->
	    mnesia:abort({combine_error, Tab, "Too few fragments"})
    end;
do_del_fragment(Tab) ->
    mnesia:abort({bad_type, Tab}).

%% Adjust tab info before merge
adjust_before_merge(FH) ->
    P = FH#mnesia_fragment_hash.next_n_to_split - 1,
    L = FH#mnesia_fragment_hash.n_doubles,
    N = FH#mnesia_fragment_hash.n_fragments - 1,
    if
	P < 1 ->
	    L2 = L - 1,
	    FH#mnesia_fragment_hash{n_doubles = L2,
				    next_n_to_split = trunc(math:pow(2, L2)),
				    n_fragments = N};
	true ->
	    FH#mnesia_fragment_hash{next_n_to_split = P,
				    n_fragments = N}
    end.

make_del_fragment({_, Tid, Ts}, Tab, FH, MergeFrag, LastFrag) ->
    Pat = mnesia:table_info(LastFrag, wild_pattern),
    Recs = mnesia:match_object(Tid, Ts, LastFrag, Pat, read),
    merge(MergeFrag, LastFrag, Recs).

%% Perform the merge of the table
merge(MergeFrag, LastFrag, [Rec | Recs]) ->
    Key = element(2, Rec),
    LastOid = {LastFrag, Key},
    [{op, rec, unknown, {{MergeFrag, Key}, [Rec], write}} |
     merge(MergeFrag, LastFrag, Recs)];
merge(_MergeFrag, _LastFrag, []) ->
    [].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Add nodes to the node pool of a fragmented table
    
add_nodes(Tab, Nodes) ->
    schema_transaction(fun() -> do_add_nodes(Tab, Nodes) end).

do_add_nodes(Tab, Nodes) when atom(Tab), Tab /= schema, list(Nodes)  ->
    TidTs = get_tid_ts_and_lock(schema, write),
    F = lookup_fragment(Tab),
    Pool = F#mnesia_fragment.node_pool,
    case mnesia_lib:intersect(Nodes, Pool) of
	[] when Nodes /= [] ->
	    Cs = val({Tab, cstruct}),
	    %% mnesia_schema:ensure_running(Cs, Node),
	    F2 =  F#mnesia_fragment{node_pool = Pool ++ Nodes},
	    insert_schema_ops(TidTs, make_write_table_properties(Tab, [F2]));
	BadNodes ->
	    mnesia:abort({already_exists, Tab, BadNodes})
    end;
do_add_nodes(Tab, Nodes) ->
    mnesia:abort({bad_type, Tab, Nodes}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Delete nodes to the node pool of a fragmented table

del_nodes(Tab, Nodes) ->
    schema_transaction(fun() -> do_del_nodes(Tab, Nodes) end).

do_del_nodes(Tab, Nodes) when atom(Tab), Tab /= schema, list(Nodes) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    F = lookup_fragment(Tab),
    Pool = F#mnesia_fragment.node_pool,
    case Nodes -- Pool of
	[] when Nodes /= [] ->
	    F2 =  F#mnesia_fragment{node_pool = Pool -- Nodes},
	    insert_schema_ops(TidTs, make_write_table_properties(Tab, [F2]));
	BadNodes ->
	    mnesia:abort({no_exists, Tab, BadNodes})
    end;
do_del_nodes(Tab, Nodes) ->
    mnesia:abort({bad_type, Tab, Nodes}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Helpers

lookup_fragment(Tab) ->
    case catch mnesia:read_table_property(Tab, mnesia_fragment) of
	F when record(F, mnesia_fragment) ->
	    F;
	{'EXIT', _} ->
	    mnesia:abort({no_exists, Tab, mnesia_fragment})
    end.

lookup_fragment_hash(Tab) ->
    case catch mnesia:read_table_property(Tab, mnesia_fragment_hash) of
	FH when record(FH, mnesia_fragment_hash) ->
	    FH;
	{'EXIT', _} ->
	    mnesia:abort({no_exists, Tab, mnesia_fragment_hash})
    end.

%% Returns name of fragment table
key_to_frag_name(Tab, Key) ->
    case catch mnesia:read_table_property(Tab, mnesia_fragment_hash) of
	{'EXIT', _} ->
	    Tab;
	FH ->
	    N = key_to_n(FH, Key),
	    n_to_frag_name(Tab, N)
    end.

%% Returns name of fragment table
n_to_frag_name(Tab, 1) ->
    Tab;
n_to_frag_name(Tab, N) ->
    list_to_atom(atom_to_list(Tab) ++ "_frag" ++ integer_to_list(N)).

%% Returns fragment number
key_to_n(FH, Key) ->
    L = FH#mnesia_fragment_hash.n_doubles,
    A = erlang:hash(Key, trunc(math:pow(2, L))),
    P = FH#mnesia_fragment_hash.next_n_to_split,
    if
	A < P ->
	    erlang:hash(Key, trunc(math:pow(2, L + 1)));
	true ->
	    A
    end.

%% Returns a list of frament table names
frag_names(Tab) ->
    case catch mnesia:read_table_property(Tab, mnesia_fragment_hash) of
	{'EXIT', _} ->
	    [Tab];
	FH ->
	    N = FH#mnesia_fragment_hash.n_fragments,
	    frag_names(Tab, N, [])
    end.

frag_names(Tab, 1, Acc) ->
    [Tab | Acc];
frag_names(Tab, N, Acc) ->
    Frag = n_to_frag_name(Tab, N),
    frag_names(Tab, N - 1, [Frag | Acc]).

%% Returns a sorted list of {Node, FragCount} tuples
frag_distribution(Tab) ->
    Ets = erlang:db_create(mnesia_frag, [private]),
    Res = (catch frag_distribution(Tab, Ets)),
    erlang:db_erase(Ets),
    case Res of
	{'EXIT', Reason} -> exit(Reason);
	Other -> Other
    end.

frag_distribution(Tab, Ets) ->
    F = lookup_fragment(Tab),
    Pool = F#mnesia_fragment.node_pool,
    init_frag_count(Pool, Ets),
    count_frag(frag_names(Tab), Ets),
    lists:keysort(2, mnesia_lib:erlang_db_match_object(Ets, '_')).

init_frag_count([Node | Tail], Ets) ->
    erlang:db_put(Ets, {Node, 0}),
    init_frag_count(Tail, Ets);
init_frag_count([], Ets) ->
    ok.

count_frag([Frag | Tail], Ets) ->
    incr_frag_count(val({Frag, ram_copies}), Ets),
    count_frag(Tail, Ets);
count_frag([], Ets) ->
    ok.

incr_frag_count([Node | Tail], Ets) ->
    case catch erlang:db_update_counter(Ets, Node, 1) of
	{'EXIT', _} ->
	    erlang:db_put(Ets, {Node, 1});
	_ ->
	    ignore
    end,
    incr_frag_count(Tail, Ets);
incr_frag_count([], Ets) ->
    ok.

