%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1996, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose:   
-module(mnesia_index).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-export([read/5,
	 add_index/4,
	 delete_index/2,
	 del_object_index/4,
	 clear_index/3,
	 match_object/5,
	 dirty_match_object/3,
	 dirty_read/3,
	 dirty_read2/3,
	 realkeys/3,

	 db_put/2,
	 db_get/2,
	 db_match_erase/2,
	 get_index_table/2,
	 get_index_table/3,
	 
	 tab2filename/2,
	 tab2tmp_filename/2,
	 init_index/2,
	 init_indecies/3,
	 create_index/3,
	 create_empty_disc_indecies/1,
	 create_disc_indecies/2,
	 del_ix_files/2,
	 add_ix_files/2,
	 del_index/3]).
	 
	 
-import(mnesia_lib, [verbose/2]).

-include("mnesia.hrl").


val(Var) ->
    case ?catch_val(Var) of
	{'EXIT', _ReASoN_} -> mnesia_lib:other_val(Var, _ReASoN_); 
	_VaLuE_ -> _VaLuE_ 
    end.

%% read an object list throuh its index table
%% we assume that table Tab has index on attribute number Pos

read(Tid, Store, Tab, IxKey, Pos) ->
    ResList = mnesia_locker:ixrlock(Tid, Store, Tab, IxKey, Pos),
    %% Remove all tuples which don't include Ixkey, happens when Tab is a bag
    case val({Tab, setorbag}) of
	set -> 
	    ResList;
	bag -> 
	    mnesia_lib:key_search_all(IxKey, Pos, ResList)
    end.

add_index(Tab, Key, Obj, Old) ->
    add_index(Tab, Key, Obj, val({Tab, index_tables}), Old).

add_index(Tab, K, Obj, [], _) -> ok;
add_index(Tab, K, Obj, [{Pos, Ixt} |Tail], OldRecs) when integer(Pos) ->
    case val({Tab, setorbag}) of 
	set -> 
	    %% Remove old tuples in index if Tab is updated
	    case OldRecs of 
		undefined -> 
		    Old = mnesia_lib:db_get(Tab, K),
		    del_ixes(Ixt, Old, Pos, K);
		Old -> 
		    del_ixes(Ixt, Old, Pos, K)
	    end;
	bag ->  %% Old tuples should be left in Index if type is bag
	    ok
    end,
    db_put(Ixt, {element(Pos, Obj), K}),
    add_index(Tab, K, Obj, Tail, OldRecs).

delete_index(Tab, K) ->
    delete_index(Tab, K,  val({Tab, index_tables})).

delete_index(Tab, K, []) -> ok;
delete_index(Tab, K, [{Pos, Ixt} | Tail]) when integer(Pos) ->
    DelObjs = mnesia_lib:db_get(Tab, K), 
    del_ixes(Ixt, DelObjs, Pos, K),
    delete_index(Tab, K, Tail).

del_ixes(Ixt, [], Pos, L) -> ok;
del_ixes(Ixt, [Obj | Tail], Pos, Key) ->
    db_match_erase(Ixt, {element(Pos, Obj), Key}),
    del_ixes(Ixt, Tail, Pos, Key).

del_object_index(Tab, K, Obj, Old) ->
    del_object_index(Tab, K, Obj,  val({Tab, index_tables}), Old).

del_object_index(Tab, K, Obj, [], Old) -> ok;
del_object_index(Tab, K, Obj, [{Pos, Ixt} | Tail], Old) when integer(Pos) ->
    case val({Tab, setorbag}) of
	set -> %% If set remove the tuple in index table
	    del_ixes(Ixt, [Obj], Pos, K);
	bag -> 
	    del_object_bag(Tab, K, Obj, Pos, Ixt, Old)
    end,
    del_object_index(Tab, K, Obj, Tail, Old).

del_object_bag(Tab, Key, Obj, Pos, Ixt, undefined) -> 
    Old = mnesia_lib:db_get(Tab, Key),
    del_object_bag(Tab, Key, Obj, Pos, Ixt, Old);
%% If Tab type is bag we need remove index identifier if Tab 
%% contains less than 2 elements. 
del_object_bag(Tab, Key, Obj, Pos, Ixt, Old) when length(Old) < 2 ->
    del_ixes(Ixt, [Obj], Pos, Key);
del_object_bag(Tab, Key, Obj, Pos, Ixt, Old) -> ok.

clear_index(Tab, K, Obj) ->
    clear_index(Tab, K, Obj,  val({Tab, index_tables})).

clear_index(Tab, K, Obj, []) -> ok;
clear_index(Tab, K, Obj, [{Pos, Ixt} | Tail]) when integer(Pos) ->
    db_match_erase(Ixt, Obj),
    clear_index(Tab, K, Obj, Tail).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

realkeys(Tab, Pos, IxKey) ->
    Index = get_index_table(Tab, Pos),
    db_get(Index, IxKey).
    %% a list on the form [{IxKey, RealKey1} , ....
    
match_object(Tid, Store, Tab, Pat, Pos) ->
    %% Assume that we are on the node where the replica is
    mnesia_locker:rlock_table(Tid, Store, Tab), %% ignore return value
    dirty_match_object(Tab, Pat, Pos).

dirty_match_object(Tab, Pat, Pos) ->
    %% Assume that we are on the node where the replica is
    IxKey = element(Pos, Pat),
    RealKeys = mnesia_index:realkeys(Tab, Pos, IxKey),
    merge(RealKeys, Tab, Pat, []).

merge([{IxKey, RealKey}|Tail], Tab, Pat0, Ack) ->
    %% Assume that we are on the node where the replica is
    Pat = setelement(2, Pat0, RealKey),
    L0 = mnesia_lib:db_match_object(Tab, Pat),
    merge(Tail, Tab, Pat0, L0 ++ Ack);
merge([], _, _, Ack) -> Ack.

dirty_read(Tab, IxKey, Pos) ->
    Node = val({Tab, where_to_read}),
    ResList = mnesia_lib:tr_rpc(Node, mnesia_index, dirty_read2, [Tab, IxKey, Pos]),
    %% Remove all tuples which don't include Ixkey, happens when Tab is a bag
    case val({Tab, setorbag}) of
	set -> 
	    ResList;
	bag -> 
	    mnesia_lib:key_search_all(IxKey, Pos, ResList)
    end.

dirty_read2(Tab, IxKey, Pos) ->
    Ix = get_index_table(Tab, Pos),
    Keys = db_match(Ix, {IxKey, '$1'}),
    r_keys(Keys, Tab, []).

r_keys([[H]|T],Tab,Ack) -> 
    V = mnesia_lib:db_get(Tab, H),
    r_keys(T, Tab, V ++ Ack);
r_keys([], _, Ack) ->
    Ack.
	    

%%%%%%% Creation, Init and deletion routines for index tables
%% We can have several indexes on the same table
%% this can be a fairly costly operationif table is *very* large


tab2filename(Tab, Pos) ->
    mnesia_lib:dir(Tab) ++ "_" ++ integer_to_list(Pos) ++ ".DAT".

tab2tmp_filename(Tab, Pos) ->
    mnesia_lib:dir(Tab) ++ "_" ++ integer_to_list(Pos) ++ ".TMP".

%% Initialize an old index disc or ram
init_indecies(Tab, Storage, PosList) ->
    case Storage of
	unknown ->
	    ignore;
	disc_only_copies ->
	    init_disc_index(Tab, PosList);
	ram_copies ->
	    make_ram_index(Tab, PosList);
	disc_copies ->
	    make_ram_index(Tab, PosList)
    end.
    
    
init_index(Tab, Storage) ->
    PosList = val({Tab, index}),
    init_indecies(Tab, Storage, PosList).

%% Create a new index
create_index(Tab, Pos, Storage) ->
    case Storage of
	disc_only_copies ->
	    create_disc_indecies(Tab, [Pos]);
	ram_copies ->
	    ignore;
	disc_copies ->
	    ignore
    end.

%% works for both ram and disc indexes
del_index(Tab, Storage, Pos) ->
    PosList = val({Tab, index}),
    case lists:member(Pos, PosList) of
	true ->
	    Ixt0 = get_index_table(Tab, Storage, Pos),
	    case Ixt0 of
		{ram, Ixt} ->
		    erlang:db_erase(Ixt),
		    mnesia_lib:del({Tab, index_tables}, {Pos, Ixt0});
		{dets, Tag} ->
		    mnesia_monitor:close_dets(Tag),
		    file:delete(tab2filename(Tab, Pos)),
		    mnesia_lib:del({Tab, index_tables}, {Pos, Ixt0});
		Else -> %% Index doesn't exist
		    ignore
	    end,
	    mnesia_lib:unset({Tab, {index, Pos}}),
	    mnesia_lib:del({Tab, index}, Pos);
	false ->
	    ok
    end.

%%%%% misc functions for the index create/init/delete functions above
%%%%


%% assuming that the file exists.
init_disc_index(Tab, []) ->
    done;
init_disc_index(Tab, [Pos | Tail]) when integer(Pos) ->
    Fn = tab2filename(Tab, Pos),
    IxTag = {Tab, index, Pos},
    file:delete(Fn),
    Args = [{file, Fn}, {keypos, 1}, {type, bag}],
    mnesia_monitor:open_dets(IxTag, Args),
    insert_disc_index(Tab, Pos, IxTag, dets:first(Tab)),
    mnesia_lib:set({Tab, {index, Pos}}, IxTag),
    mnesia_lib:add({Tab, index_tables}, {Pos, {dets, IxTag}}).
%    mnesia_monitor:close_dets(IxTag).

%% Returns a list of index tags
create_empty_disc_indecies(Tab) ->
    create_empty_disc_indecies(Tab, val({Tab, index})).

create_empty_disc_indecies(Tab, [Pos | Tail]) ->
    Fname = tab2filename(Tab, Pos),
    IxTag = {Tab, index, Pos},
    file:delete(Fname),
    Args = [{file, Fname}, {keypos, 1},{type, bag}],
    mnesia_monitor:open_dets(IxTag, Args),
    [IxTag | create_empty_disc_indecies(Tab, Tail)];
create_empty_disc_indecies(Tab, []) ->
    [].
    
create_disc_indecies(Tab, PosList) ->
    IxTags = create_empty_disc_indecies(Tab, PosList),
    do_create_disc_indecies(Tab, IxTags).
    
do_create_disc_indecies(Tab, [IxTag | Tail]) ->
    {_, index, Pos} = IxTag,
    insert_disc_index(Tab, Pos, IxTag, mnesia_lib:db_first(Tab)),
    mnesia_lib:set({Tab, {index, Pos}}, IxTag),
    mnesia_lib:add({Tab, index_tables}, {Pos, {dets, IxTag}}),
    mnesia_monitor:close_dets(IxTag),
    do_create_disc_indecies(Tab, Tail);
do_create_disc_indecies(Tab, []) ->
    ok.

insert_disc_index(Tab, Pos, Ix, '$end_of_table') ->
    done;
insert_disc_index(Tab, Pos, Ix, K) ->
    Vals = dets:lookup(Tab, K),
    insert_disc_index_objects(Ix, Vals, Pos),
    insert_disc_index(Tab, Pos, Ix, mnesia_lib:db_next_key(Tab, K)).

insert_disc_index_objects(Ix, [], Pos) -> 
    done;
insert_disc_index_objects(Ix, [Obj|Tail], Pos) -> 
    O2 = {element(Pos, Obj), element(2, Obj)},
    dets:insert(Ix, O2),
    insert_disc_index_objects(Ix, Tail, Pos).

make_ram_index(_, []) -> 
    done;
make_ram_index(Tab, [Pos | Tail]) ->
    add_ram_index(Tab, Pos),
    make_ram_index(Tab, Tail).

add_ram_index(Tab, Pos) when integer(Pos) ->
    verbose("Creating index for ~w ~n", [Tab]),
    Pat0 = val({Tab, wild_pattern}),
    Pat1 = setelement(2, Pat0, '$1'),
    Pat2 = setelement(Pos, Pat1, '$2'),
    Keys = erlang:db_match(Tab, Pat2),
    Index = mnesia_monitor:mktab(mnesia_index, [bag, public]),
    mnesia_lib:set({Tab, {index, Pos}}, Index),
    mnesia_lib:add({Tab, index_tables}, {Pos, {ram, Index}}),
    insert_ram_index(Index, Keys);
add_ram_index(Tab, snmp) ->
    ok.

insert_ram_index(Index, [[RealKey, IxKey] | Tail]) ->
    erlang:db_put(Index, {IxKey, RealKey}),
    insert_ram_index(Index, Tail);
insert_ram_index(Index, []) -> done.
    

db_put({ram, Ixt}, V) ->
    erlang:db_put(Ixt, V);
db_put({dets, Ixt}, V) ->
    dets:insert(Ixt, V).

db_get({ram, Ixt}, K) ->
    erlang:db_get(Ixt, K);
db_get({dets, Ixt}, K) ->
    dets:lookup(Ixt, K).

db_match_erase({ram, Ixt}, Pat) ->
    erlang:db_match_erase(Ixt, Pat);
db_match_erase({dets, Ixt}, Pat) ->
    dets:match_delete(Ixt, Pat).
    
db_match({ram, Ixt}, Pat) ->
    erlang:db_match(Ixt, Pat);
db_match({dets, Ixt}, Pat) ->
    dets:match(Ixt, Pat).
    
get_index_table(Tab, Pos) ->
    get_index_table(Tab,  val({Tab, storage_type}), Pos).

get_index_table(Tab, ram_copies, Pos) ->
    {ram,  val({Tab, {index, Pos}})};
get_index_table(Tab, disc_copies, Pos) ->
    {ram,  val({Tab, {index, Pos}})};
get_index_table(Tab, disc_only_copies, Pos) ->
    {dets, val({Tab, {index, Pos}})};
get_index_table(_Tab, unknown, _Pos) ->
    unknown.

del_ix_files(Tab, Storage) ->
    PosList = val({Tab, index}),
    del_ix_files(Tab, PosList, Storage).

del_ix_files(_, [], _) ->
    done;
del_ix_files(Tab, [Pos | Tail], Storage) ->
    del_index(Tab, Storage, Pos),
    del_ix_files(Tab, Tail, Storage).

add_ix_files(Tab, Storage) ->
    PosList = val({Tab, index}),
    add_ix_files(Tab, PosList, Storage).

add_ix_files(_, [], Storage) ->
    done;
add_ix_files(Tab, [Pos|Tail], Storage) ->
    create_index(Tab, Pos, Storage),
    add_ix_files(Tab, Tail, Storage).

