%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%% File    : mnesia_loader.erl
%%% Author  : H}kan Mattsson <hakan@sam>
%%% Purpose : Loads tables from local disc or from remote node
%%% Created : 28 Aug 1998 by H}kan Mattsson <hakan@eric.ericsson.se>

-module(mnesia_loader).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('hakan@erix.ericsson.se').

%% Mnesia internal stuff
-export([disc_load_table/2,
	 net_load_table/4]).

%% Module internal stuff
-export([tab_copier/4]).

-import(mnesia_lib, [set/2, fatal/2, verbose/2, dbg_out/2]).

-include("mnesia.hrl").

val(Var) ->
    case ?catch_val(Var) of
	{'EXIT', Reason} -> mnesia_lib:other_val(Var, Reason); 
	Value -> Value
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load a table from local disc

disc_load_table(Tab, Reason) ->
    Storage =  val({Tab, storage_type}),
    Type = val({Tab, setorbag}),
    dbg_out("Getting table ~p (~p~) from disc: ~p~n",
	    [Tab, Storage, Reason]),
    ?eval_debug_fun({?MODULE, do_get_disc_copy},
		    [{tab, Tab},
		     {reason, Reason},
		     {storage, Storage}, 
		     {type, Type}]),
    do_get_disc_copy2(Tab, Reason, Storage, Type).

do_get_disc_copy2(Tab, Reason, Storage, Type) when Storage == unknown ->
    verbose("Local table copy of ~p has recently been deleted, ignored.~n",
	    [Tab]),
    ok;
do_get_disc_copy2(Tab, Reason, Storage, Type) when Storage == disc_copies ->
    %% NOW we create the actual table
    Repair = mnesia_monitor:get_env(auto_repair),
    Args = [{keypos, 2}, public, named_table, Type],
    mnesia_monitor:mktab(Tab, Args),
    Fname = mnesia_lib:tab2dat(Tab),
    case mnesia_lib:dat_to_ets(Tab, Tab, Fname, Type, Repair, yes) of
	loaded ->
	    mnesia_index:init_index(Tab, Storage),
	    snmpify(Tab, Storage),
	    set({Tab, load_node}, node()),
	    set({Tab, load_reason}, Reason),
	    ok;
	{error, Error} ->
	    {error, {"Failed to load table from disc", [Tab, Error]}}
    end;
do_get_disc_copy2(Tab, Reason, Storage, Type) when Storage == ram_copies ->
    Args = [{keypos, 2}, public, named_table, Type],
    mnesia_monitor:mktab(Tab, Args),
    Fname = mnesia_lib:tab2dat(Tab),
    DiscLoad =
	case mnesia_monitor:use_dir() of
	    true ->
		case mnesia_lib:exists(Fname) of
		    true -> true;
		    false -> false
		end;
	    false ->
		false
	end,
    case DiscLoad of
	true ->
	    Repair = mnesia_monitor:get_env(auto_repair),
	    case mnesia_lib:dat_to_ets(Tab, Tab, Fname, Type, Repair, yes) of
		loaded ->
		    mnesia_index:init_index(Tab, Storage),
		    snmpify(Tab, Storage),
		    set({Tab, load_node}, node()),
		    set({Tab, load_reason}, Reason),
		    ok;
		{error, Error} ->
		    {error, {"Failed to load table from disc", [Tab, Error]}}
	    end;
	false ->
	    mnesia_index:init_index(Tab, Storage),
	    snmpify(Tab, Storage),
	    set({Tab, load_node}, node()),
	    set({Tab, load_reason}, Reason),
	    ok
    end;

do_get_disc_copy2(Tab, Reason, Storage, Type) when Storage == disc_only_copies ->
    Args = [{file, mnesia_lib:tab2dat(Tab)},
	    {type, Type},
	    {keypos, 2},
	    {repair, mnesia_monitor:get_env(auto_repair)}],
    case mnesia_monitor:open_dets(Tab, Args) of
	{ok, _} ->
	    mnesia_index:init_index(Tab, Storage),
	    snmpify(Tab, Storage),
	    set({Tab, load_node}, node()),
	    set({Tab, load_reason}, Reason),
	    ok;
	{error, Error} ->
	    {error, {"Failed to create dets table", Error}}
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load a table from a remote node
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Receiver                             Sender
%% --------                             ------
%% Grab schema lock on table
%%                                      Determine table size
%% Create empty pre-grown table
%%                                      Grab read lock on table
%%                                      Let receiver subscribe on updates done on sender node
%%                                      Disable rehashing of table
%%                                      Release read lock on table
%%                                      Send table to receiver in chunks
%% 
%%                                      Grab read lock on table
%% Block dirty updates
%%                                      Update wherabouts
%% 
%%                                      Cancel the update subscription
%% Process the subscription events
%% Optionally dump to disc
%% Unblock dirty updates
%%                                      Release read lock on table
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-define(MAX_TRANSFER_SIZE, 4000).
-define(MAX_RAM_FILE_SIZE, 400000).
-define(MAX_RAM_TRANSFERS, (?MAX_RAM_FILE_SIZE div ?MAX_TRANSFER_SIZE) + 1).

net_load_table(Tab, Reason, Ns, Cs)
        when Reason == add_table_copy ->
    try_net_load_table(Tab, Reason, Ns, Cs);
net_load_table(Tab, Reason, Ns, _Cs) ->
    try_net_load_table(Tab, Reason, Ns, val({Tab, cstruct})).

try_net_load_table(Tab, Reason, [], Cs) ->
    verbose("Copy failed. No active replicas of ~p are available.~n", [Tab]),
    no;
try_net_load_table(Tab, Reason, Ns, Cs) ->
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    do_get_network_copy(Tab, Reason, Ns, Storage, Cs).

do_get_network_copy(Tab, Reason, Ns, unknown, Cs) ->
    verbose("Local table copy of ~p has recently been deleted, ignored.~n", [Tab]),
    no_need;
do_get_network_copy(Tab, Reason, Ns, Storage, Cs) ->
    [Node | Tail] = Ns,
    dbg_out("Getting table ~p (~p) from node ~p: ~p~n",
	    [Tab, Storage, Node, Reason]),
    ?eval_debug_fun({?MODULE, do_get_network_copy},
		    [{tab, Tab}, {reason, Reason}, 
		     {nodes, Ns}, {storage, Storage}]),
    Args = [self(), Tab, Storage, Cs],
    P = spawn_link(Node, ?MODULE, tab_copier, Args),
    case tab_receiver(P, Tab, undefined, Storage, Cs, ?MAX_RAM_TRANSFERS, first) of
	ok ->
	    set({Tab, load_node}, Node),
	    set({Tab, load_reason}, Reason),
	    mnesia_controller:i_have_tab(Tab),
	    dbg_out("Table ~p copied from ~p to ~p~n", [Tab, Node, node()]),
	    yes;
	down ->
	    try_net_load_table(Tab, Reason, Tail, Cs) 
    end.

snmpify(Tab, Storage) ->
    do_snmpify(Tab, val({Tab, snmp}), Storage).

do_snmpify(Tab, [], _Storage) ->
    ignore;
do_snmpify(Tab, Us, Storage) ->
    Snmp = mnesia_snmp_hook:create_table(Us, Tab, Storage),
    set({Tab, {index, snmp}}, Snmp).
    
tab_receiver(Pid, Tab, TabRef, Storage, Cs, RamLeft, State) ->
    receive
	{Pid, {first, TabSize}} when State == first ->
	    Pid ! {self(), more},
	    TabRef2 = init_table(Tab, TabSize, Storage, Cs),
	    mnesia_tm:block_tab(Tab),
	    tab_receiver(Pid, Tab, TabRef2, Storage, Cs, RamLeft, more);

	{Pid, {more, Recs}} when State == more ->
	    Pid ! {self(), more},
	    insert_records(TabRef, Recs),
	    opt_ram_to_disc(Tab, Storage, Cs, RamLeft),
	    tab_receiver(Pid, Tab, TabRef, Storage, Cs, RamLeft - 1, State);

	{Pid, {no_more, DatBin}} when State == more ->
	    subscr_receiver(Tab, TabRef, Storage, Cs#cstruct.record_name),
	    case handle_last(Tab, Storage, Cs#cstruct.type, DatBin) of
		ok ->
%%	            mnesia_index:add_ix_files(Tab, Storage),
		    mnesia_index:init_index(Tab, Storage),
		    snmpify(Tab, Storage),
		    Pid ! {self(), no_more},
		    mnesia_tm:unblock_tab(Tab),
		    ok;
		{error, Reason} ->
		    Msg = "Failed to handle last",
		    dbg_out("~s: ~p: ~p~n", [Msg, Tab, Reason]),
		    down(Tab, Storage)
	    end;
	
	{'EXIT', Pid, Reason} ->
	    dbg_out("Failed to copy ~p from ~p failed (~p): ~p~n",
		    [Tab, node(Pid), State, Reason]),
	    down(Tab, Storage)
    end.

init_table(Tab, TabSize, Storage, Cs) ->
    case Storage of
	unknown ->
	    no_need;
	disc_only_copies ->
	    mnesia_lib:lock_table(Tab) ,
	    Tmp = mnesia_lib:tab2tmp(Tab),
	    Args = [{file, Tmp},
		    {keypos, 2},
		    {ram_file, true},
		    {estimated_no_objects, TabSize + 256},
		    {repair, mnesia_monitor:get_env(auto_repair)},
		    {type, Cs#cstruct.type}],
	    file:delete(Tmp),
	    case mnesia_lib:dets_sync_open(Tab, Args) of
		{ok, _} ->
		    {dets, Tab};
		Else ->
		    fatal("do_get_network_copy open ~w Tab ~w cstruct ~w ~s ~n",
			  [Else, Tab, Cs, Tmp])
	    end;
	_ ->
	    Args = [{keypos, 2}, public, named_table, Cs#cstruct.type],
	    mnesia_monitor:mktab(Tab, Args), % exits upon failure
	    {ets, Tab}
    end.

opt_ram_to_disc(Tab, disc_only_copies, Cs, 0) ->
    dets:close(Tab), % Flush dets table in ram to disc
    Tmp = mnesia_lib:tab2tmp(Tab),
    Args = [{file, Tmp},
	    {keypos, 2},
	    {repair, mnesia_monitor:get_env(auto_repair)},
	    {type, Cs#cstruct.type}],
    case dets:open_file(Tab, Args) of
	{ok, _} ->
	    ok;
	Else ->
	    fatal("do_get_network_copy reopen ~w Tab ~w cstruct ~w ~s ~n",
		  [Else, Tab, Cs, Tmp])
    end;
opt_ram_to_disc(Tab, Storage, Cs, RamLeft) ->
    ignore.

subscr_receiver(Tab, TabRef, Storage, RecName) ->
    receive
	{mnesia_table_event, {Op, Val, _Tid}} ->
	    if
		Tab == RecName ->
		    handle_event(TabRef, Op, Val);
		true ->
		    handle_event(TabRef, Op, setelement(1, Val, RecName))
	    end,
	    subscr_receiver(Tab, TabRef, Storage, RecName)
    after 0 ->
	    ok
    end.

handle_event(TabRef, write, Rec) ->	    
    db_put(TabRef, Rec);
handle_event(TabRef, delete, {_Tab, Key}) ->
    db_erase(TabRef, Key);
handle_event(TabRef, delete_object, OldRec) ->
    db_match_erase(TabRef, OldRec);
handle_event(TabRef, clear_table, {_Tab, _Key}) ->
    db_match_erase(TabRef, '_').

handle_last(Tab, disc_copies, _Type, nobin) ->
    mnesia_dumper:raw_named_dump_table(Tab, dat);
handle_last(Tab, disc_only_copies, Type, nobin) ->
    case mnesia_lib:swap_tmp_files([Tab]) of
	[] ->
	    Args = [{file, mnesia_lib:tab2dat(Tab)},
		    {type, Type},
		    {keypos, 2},
		    {repair, mnesia_monitor:get_env(auto_repair)}],
	    mnesia_monitor:open_dets(Tab, Args),
	    ok;
	L when list(L) ->
	    {error, {"Cannot swap tmp files", Tab, L}}
    end;

handle_last(Tab, ram_copies, _Type, nobin) ->
    ok;
handle_last(Tab, ram_copies, _Type, DatBin) ->
    case mnesia_monitor:use_dir() of
	true ->
	    mnesia_lib:lock_table(Tab),
	    Tmp = mnesia_lib:tab2tmp(Tab),
	    ok = file:write_file(Tmp, DatBin),
	    ok = file:rename(Tmp, mnesia_lib:tab2dat(Tab)),
	    mnesia_lib:unlock_table(Tab),
	    ok;
	false ->
	    ok
    end.

down(Tab, Storage) ->
    %% dbg_out("DOWN ~p (~p)~n", [Tab, Storage]),
    case Storage of
	ram_copies ->
	    catch erlang:db_erase(Tab);
	disc_copies ->
	    catch erlang:db_erase(Tab);
	disc_only_copies ->
	    mnesia_lib:cleanup_tmp_files([Tab])
    end,
    mnesia_checkpoint:tm_del_copy(Tab, node()),
    mnesia_controller:sync_del_table_copy_whereabouts(Tab, node()),
    mnesia_tm:unblock_tab(Tab),
    down.

insert_records(TabRef, [RecsWithSameKey | Tail]) ->
    do_insert_records(TabRef, RecsWithSameKey),
    insert_records(TabRef, Tail);
insert_records(TabRef, []) ->
    ok.

do_insert_records(TabRef, [Rec | Recs]) ->
    db_put(TabRef, Rec),
    do_insert_records(TabRef, Recs);
do_insert_records(TabRef, []) ->
    ok.
    
db_erase({dets, Tab}, Key) ->
    ok = dets:delete(Tab, Key);
db_erase({ets, Tab} , Key) ->
    true = erlang:db_erase(Tab, Key).

db_match_erase({dets, Tab}, Pat) ->
    ok = dets:match_delete(Tab, Pat);
db_match_erase({ets, Tab} , Pat) ->
    true = erlang:db_match_erase(Tab, Pat).

db_put({dets, Tab}, Val) ->
    ok = dets:insert(Tab, Val);
db_put({ets, Tab}, Val) ->
    true = erlang:db_put(Tab, Val).

%% This code executes at the remote site where the data is
%% executes in a special copier process.

tab_copier(Pid, Tab, RemoteS, RemoteCs) ->
    process_flag(trap_exit, true), % Avoid dangling subscription and fixtable

    %% Send first
    Storage = val({Tab, storage_type}),
    TabSize = mnesia:table_info(Tab, size),
    Pid ! {self(), {first, TabSize}},

    %% Calculate #keys per transfer
    Key = mnesia_lib:db_first(Storage, Tab),
    Recs = mnesia_lib:db_get(Storage, Tab, Key),
    BinSize = size(term_to_binary(Recs)),
    KeysPerTransfer = (?MAX_TRANSFER_SIZE div BinSize) + 1,

    case catch tab_sender(Pid, Tab, Key, Storage, RemoteS, RemoteCs, KeysPerTransfer) of
	{atomic, no_more} ->
	    unlink(Pid), % Will be obsolete when loader and controller is separate processes
	    exit(normal);
	Reason ->
	    cleanup_tab_copier(Pid, Storage, Tab),
	    exit({error, Reason})
    end.

tab_sender(Pid, Tab, Key, Storage, RemoteS, RemoteCs, KeysPerTransfer) ->
    prepare_copy(Pid, Tab, Storage),
    send_more(Pid, Tab, Storage, Key, KeysPerTransfer),
    finish_copy(Pid, Tab, Storage, RemoteS, RemoteCs).

prepare_copy(Pid, Tab, Storage) ->
    Trans =
	fun() ->
		mnesia:read_lock_table(Tab),
		mnesia_subscr:subscribe(Pid, {table, Tab}),
		mnesia_lib:db_fixtable(Storage, Tab, true),
		ok
	end,
    case mnesia:transaction(Trans) of
	{atomic, ok} ->
	    ok;
	{aborted, Reason} ->
	    exit({tab_copier_prepare, Tab, Reason})
    end.

send_more(Pid, Tab, Storage, '$end_of_table', _KeysPerTransfer) ->
    ok; 
send_more(Pid, Tab, Storage, Key, KeysPerTransfer) ->
    {NextKey, Recs} = get_records(Tab, Storage, Key, KeysPerTransfer, []),
    receive
	{Pid, more} ->
	    Pid ! {self(), {more, Recs}},
	    send_more(Pid, Tab, Storage, NextKey, KeysPerTransfer);
	{'EXIT', Pid, Reason} ->
	    verbose("Tab receiver ~p crashed (more): ~p~n", [Tab, Reason]),
	    exit({tab_receiver_more, Tab, Reason})
    end.

get_records(Tab, Storage, Key, KeysLeft, AccRecs)
  when KeysLeft > 0,  Key /= '$end_of_table'  ->
    Recs = mnesia_lib:db_get(Storage, Tab, Key),
    NextKey = mnesia_lib:db_next_key(Storage, Tab, Key),
    get_records(Tab, Storage, NextKey, KeysLeft - 1, [Recs | AccRecs]);
get_records(Tab, Storage, Key, KeysLeft, AccRecs) ->
    {Key, AccRecs}.

finish_copy(Pid, Tab, Storage, RemoteS, RemoteCs) ->
    RecNode = node(Pid),
    DatBin = dat2bin(Tab, Storage, RemoteS),
    Trans =
	fun() ->
		mnesia:read_lock_table(Tab),
		mnesia_controller:sync_add_table_copy_whereabouts(Tab, RecNode, RemoteCs),
		cleanup_tab_copier(Pid, Storage, Tab),
		ok = mnesia_checkpoint:tm_add_copy(Tab, RecNode),
		Pid ! {self(), {no_more, DatBin}},
		receive
		    {Pid, no_more} -> % Dont bother about the spurious 'more' message
			no_more;
		    {'EXIT', Pid, Reason} ->
			{tab_receiver_no_more, Reason}
		end
	end,
    mnesia:transaction(Trans).

cleanup_tab_copier(Pid, Storage, Tab) ->
    mnesia_lib:db_fixtable(Storage, Tab, false),
    mnesia_subscr:unsubscribe(Pid, {table, Tab}).

dat2bin(Tab, ram_copies, ram_copies) ->
    mnesia_lib:lock_table(Tab),
    Res = file:read_file(mnesia_lib:tab2dat(Tab)),
    mnesia_lib:unlock_table(Tab),
    case Res of
	{ok, DatBin} -> DatBin;
	_ -> nobin
    end;
dat2bin(_Tab, _LocalS, RemoteS) ->
    nobin.

