%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% This module administers three kinds of log files:
%%
%% 1 The transaction log
%%   mnesia_tm appends to the log (via mnesia_log) at the
%%   end of each transaction (or dirty write) and
%%   mnesia_dumper reads the log and performs the ops in
%%   the dat files. The dump_log is done  at startup and
%%   at intervals controlled by the user.
%%
%% 2 The mnesia_down log
%%   mnesia_tm appends to the log (via mnesia_log) when it
%%   realizes that mnesia goes up or down on another node.
%%   mnesia_init reads the log (via mnesia_log) at startup.
%%
%% 3 The backup log
%%   mnesia_schema produces one tiny log when the schema is
%%   initially created. mnesia_schema also reads the log
%%   when the user wants tables (possibly incl the schema)
%%   to be restored. mnesia_log appends to the log when the
%%   user wants to produce a real backup.
%%
%%   The actual access to the backup media is performed via the
%%   mnesia_backup module for both read and write. mnesia_backup
%%   uses the disk_log (*), BUT the user may write an own module
%%   with the same interface as mnesia_backup and configure
%%   Mnesia so the alternate module performs the actual accesses
%%   to the backup media. This means that the user may put the
%%   backup on medias that Mnesia does not know about possibly on
%%   hosts where Erlang is not running.
%%
%% All these logs have to some extent a common structure.
%% They are all using the disk_log module (*) for the basic
%% file structure. The disk_log has a repair feature that
%% can be used to skip erroneous log records if one comes to
%% the conclusion that it is more important to reuse some
%% of the log records than the risque of obtaining inconsistent
%% data. If the data becomes inconsistent it is solely up to the
%% application to make it consistent again. The automatic
%% reparation of the disk_log is very powerful, but use it
%% with extreme care.
%%
%% First in all Mnesia's log file is a mnesia log header.
%% It contains a list with a log_header record as single
%% element. The structure of the log_header may never be
%% changed since it may be written to very old backup files.
%% By holding this record definition stable we can be
%% able to comprahend backups from timepoint 0. It also
%% allows us to use the backup format as an interchange
%% format between Mnesia releases.
%%
%% An op-list is a list of tuples with arity 3. Each tuple
%% has this structure: {Oid, Recs, Op} where Oid is the tuple
%% {Tab, Key}, Recs is a (possibly empty) list of records and
%% Op is an atom.
%%
%% The log file structure for the transaction log is as follows.
%%
%%    After the mnesia log section follows an extended record section
%%    containing op-lists. There are several values that Op may
%%    have, such as write, delete, update_counter, delete_object,
%%    and replace. There is no special end of section marker.
%%
%%    +-----------------+
%%    | mnesia log head |
%%    +-----------------+
%%    | extended record |
%%    | section         |
%%    +-----------------+
%%
%% The log file structure for the mnesia_down log is as follows.
%%
%%    After the mnesia log section follows a mnesia_down section
%%    containg lists with yoyo records as single element.
%%
%%    +-----------------+
%%    | mnesia log head |
%%    +-----------------+
%%    | mnesia_down     |
%%    | section         |
%%    +-----------------+
%%
%% The log file structure for the backup log is as follows.
%%
%%    After the mnesia log section follows a schema section
%%    containing record lists. A record list is a list of tuples
%%    where {schema, Tab} is interpreted as a delete_table(Tab) and
%%    {schema, Tab, CreateList} are interpreted as create_table.
%%
%%    The record section also contains record lists. In this section
%%    {Tab, Key} is interpreted as delete({Tab, Key}) and other tuples
%%    as write(Tuple). There is no special end of section marker.
%%
%%    +-----------------+
%%    | mnesia log head |
%%    +-----------------+
%%    | schema section  |
%%    +-----------------+
%%    | record section  |
%%    +-----------------+
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

-module(mnesia_log).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('klacke@erix.ericsson.se').
-author('hakan@erix.ericsson.se').

-export([
	 append/2,
	 backup/1,
	 backup/2,
	 backup_checkpoint/2,
	 backup_checkpoint/3,
	 backup_log_header/0,
	 backup_master/4,
	 chunk_decision_log/1,
	 chunk_decision_tab/1,
	 chunk_log/1,
	 close_decision_log/0,
	 close_decision_tab/0,
	 close_log/1,
	 confirm_log_dump/1,
	 confirm_decision_log_dump/0,
	 previous_log_file/0,
	 previous_decision_log_file/0,
	 latest_log_file/0,
	 decision_log_version/0,
	 decision_log_file/0,
	 decision_tab_file/0,
	 decision_tab_version/0,
	 init/0,
	 init_log_dump/0,
	 log/1,
	 log_decision/1,
	 log_files/0,
	 open_decision_log/0,
	 trans_log_header/0,
	 open_decision_tab/0,
	 open_log/4,
	 prepare_decision_log_dump/0,
	 prepare_log_dump/1,
	 save_decision_tab/1,
	 purge_all_logs/0,
	 purge_some_logs/0,
	 stop/0,
	 tab_copier/3,
	 version/0,
	 view/0,
	 view/1,
	 write_trans_log_header/0
	]).
	
	
-include("mnesia.hrl").
-import(mnesia_lib, [val/1, dir/1]).
-import(mnesia_lib, [exists/1, fatal/2, error/2, verbose/2, dbg_out/2]).

trans_log_header() -> log_header(trans_log, version()).
backup_log_header() -> log_header(backup_log, "1.2").
decision_log_header() -> log_header(decision_log, decision_log_version()).
decision_tab_header() -> log_header(decision_tab, decision_tab_version()).

log_header(Kind, Version) ->
    #log_header{log_version=Version,
		log_kind=Kind,
		mnesia_version=mnesia:system_info(version),
		node=node(),
		now=now()}.

version() -> "4.1".

decision_log_version() -> "3.0".

decision_tab_version() -> "1.0".

append(Log, Bin) when binary(Bin) ->
    %% dbg_out("balog ~p ~p~n", [Log, binary_to_term(Bin)]),
    disk_log:balog(Log, Bin);
append(Log, Term) ->
    %% dbg_out("alog  ~p ~p~n", [Log, Term]),
    disk_log:alog(Log, Term).

%% Write commit records to the latest_log
log(C) when  C#commit.disc_copies == [],
             C#commit.disc_only_copies  == [],
             C#commit.schema_ops == [] ->
    ignore;
log(C) ->
    case mnesia_monitor:use_dir() of
        true ->
	    mnesia_dumper:incr_log_writes(),
	    if
		record(C, commit) ->
		    C2 =  C#commit{ram_copies = [], snmp = []},
		    append(latest_log, C2);
		true ->
		    %% Either a commit record as binary
		    %% or some decision related info
		    append(latest_log, C)
		end;
	false ->
	    ignore
    end.

%% Stuff related to the file LOG

%% Returns a list of logfiles. The oldest is first.
log_files() -> [previous_log_file(),
		latest_log_file(),
		decision_tab_file()
	       ].

latest_log_file() -> dir(latest_log_name()).

previous_log_file() -> dir("PREVIOUS.LOG").

decision_log_file() -> dir(decision_log_name()).

decision_tab_file() -> dir(decision_tab_name()).
    
previous_decision_log_file() -> dir("PDECISION.LOG").

latest_log_name() -> "LATEST.LOG".

decision_log_name() -> "DECISION.LOG".

decision_tab_name() -> "DECISION_TAB.LOG".

init() ->
    open_log(latest_log, trans_log_header(), latest_log_file()).

open_log(Name, Header, Fname) ->
    open_log(Name, Header, Fname, exists(Fname)).

open_log(Name, Header, Fname, Exists) ->
    Repair = mnesia_monitor:get_env(auto_repair),
    open_log(Name, Header, Fname, Exists, Repair).

open_log(Name, Header, Fname, Exists, Repair) ->
    Args = [{file, Fname}, {name, Name}, {repair, Repair}],
    case mnesia_monitor:open_log(Args) of
	{ok, Log} when Exists == true ->
	    Log;
	{ok, Log} ->
	    write_header(Log, Header),
	    Log;
	{repaired, Log, _, {badbytes, 0}} when Exists == true ->
	    Log;
	{repaired, Log, _, {badbytes, 0}} ->
	    write_header(Log, Header),
	    Log;
	{repaired, Log, Recover, BadBytes} ->
	    verbose("Log ~p repaired: ~p ~p~n", [Fname, Recover, BadBytes]),
	    Log;
	{error, Reason} when Repair == true ->
	    Res = file:delete(Fname),
	    verbose("Logfile ~p deleted: ~p, ~p~n", [Args, Res, Reason]),
	    open_log(Name, Header, Fname, Exists, false);
	{error, Reason} ->
	    fatal("Cannot open log file ~p: ~p~n", [Fname, Reason])
    end.

write_header(Log, Header) -> 
    append(Log, Header).

write_trans_log_header() ->
    write_header(latest_log, trans_log_header()).

stop() ->
    case mnesia_monitor:use_dir() of
        true ->
	    close_log(latest_log);
	false ->
	    ok
    end.

close_log(Log) ->
    disk_log:sync(Log),
    mnesia_monitor:close_log(Log).

purge_some_logs() ->
    mnesia_monitor:unsafe_close_log(latest_log), 
    file:delete(latest_log_file()),
    file:delete(decision_tab_file()).

purge_all_logs() ->
    file:delete(previous_log_file()),
    purge_some_logs().


%% Prepare dump by renaming the open logfile if possible
%% Returns a tuple on the following format: {Res, OpenLog}
%% where OpenLog is the file descriptor to log file, ready for append
%% and Res is one of the following: already_dumped, needs_dump or {error, Reason}
prepare_log_dump(InitBy) ->
    Diff = mnesia_dumper:get_log_writes() -
	mnesia_lib:read_counter(trans_log_writes_prev),
    if
	Diff == 0, InitBy /= startup ->
	    already_dumped;
	true ->
	    case mnesia_monitor:use_dir() of
		true ->
		    Prev = previous_log_file(),
		    prepare_prev(Diff, InitBy, Prev, exists(Prev));
		false ->
		    already_dumped
	    end
    end.

prepare_prev(Diff, _, _, true) ->
    {needs_dump, Diff};
prepare_prev(Diff, InitBy, Prev, false) ->
    Head = trans_log_header(),
    case mnesia_monitor:reopen_log(latest_log, Prev, Head) of
	ok ->
	    {needs_dump, Diff};
	{error, Reason} ->
	    Latest = latest_log_file(),
	    {error, {"Cannot rename log file",
		     [Latest, Prev, Reason]}}
    end.

%% Init dump and return PrevLogFileDesc or exit.
init_log_dump() ->
    Fname = previous_log_file(),
    open_log(previous_log, trans_log_header(), Fname),
    start.

chunk_log(eof) ->
    eof;
chunk_log(Cont) ->
    case catch disk_log:chunk(previous_log, Cont) of
	{error, Reason} ->
	    fatal("Possibly truncated previous_log file: ~p~n",
		  [Reason]);
	Other ->
	    Other
    end.

%% Confirms the dump by closing prev log and delete the file
confirm_log_dump(Updates) ->
    case mnesia_monitor:close_log(previous_log) of
	ok ->
	    file:delete(previous_log_file()),
	    mnesia_lib:incr_counter(trans_log_writes_prev, Updates),
	    shutdown;
	{error, Reason} ->
	    {error, Reason}
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Decision log

open_decision_log() ->
    Latest = decision_log_file(),
    open_log(decision_log, decision_log_header(), Latest),
    start.
    
prepare_decision_log_dump() ->
    Prev = previous_decision_log_file(),
    prepare_decision_log_dump(exists(Prev), Prev).

prepare_decision_log_dump(false, Prev) ->
    Head = decision_log_header(),
    case mnesia_monitor:reopen_log(decision_log, Prev, Head) of
	ok ->
	    prepare_decision_log_dump(true, Prev);
	{error, Reason} ->
	    fatal("Cannot rename decision log file ~p -> ~p: ~p~n",
		     [decision_log_file(), Prev, Reason])
    end;
prepare_decision_log_dump(true, Prev) ->
    open_log(previous_decision_log, decision_log_header(), Prev),
    start.

chunk_decision_log(Cont) ->
    %% dbg_out("chunk log ~p~n", [Cont]),
    case disk_log:chunk(previous_decision_log, Cont) of
	{error, Reason} ->
	    fatal("Possibly truncated mnesia_decision log file: ~p~n",
		  [Reason]);
	Other ->
	    Other
    end.

%% Confirms dump of the decision log
confirm_decision_log_dump() ->
    case mnesia_monitor:close_log(previous_decision_log) of
	ok ->
	    file:delete(previous_decision_log_file());
	{error, Reason} ->
	    fatal("Cannot confirm decision log dump: ~p~n",
		  [Reason])
    end.

save_decision_tab(Decisions) ->
    Log = decision_tab,
    Tmp = mnesia_lib:dir("DECISION_TAB.TMP"),
    file:delete(Tmp),
    open_log(Log, decision_tab_header(), Tmp),
    append(Log, Decisions),
    close_log(Log),
    TabFile = decision_tab_file(),
    ok = file:rename(Tmp, TabFile).

open_decision_tab() ->
    TabFile = decision_tab_file(),
    open_log(decision_tab, decision_tab_header(), TabFile),
    start.

close_decision_tab() ->
    close_log(decision_tab).

chunk_decision_tab(Cont) ->
    %% dbg_out("chunk tab ~p~n", [Cont]),
    case disk_log:chunk(decision_tab, Cont) of
	{error, Reason} ->
	    fatal("Possibly truncated mnesia_decision tab file: ~p~n",
		  [Reason]);
	Other ->
	    Other
    end.

close_decision_log() ->
    close_log(decision_log).

log_decision(Decision) ->
    append(decision_log, Decision).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Debug functions

view() ->
    lists:foreach(fun(F) -> view(F) end, log_files()).

view(File) ->
    mnesia_lib:show("*****  ~p ***** ~n", [File]),
    case exists(File) of
	false ->
	    nolog;
	true ->
	    Repair = mnesia_monitor:get_env(auto_repair),
	    N = view_only,
	    Args = [{file, File}, {name, N}, {repair, Repair}],
	    case disk_log:open(Args) of
		{ok, N} ->
		    view_file(start, N);
		{repaired, _, _, _} ->
		    view_file(start, N);
		{error, Reason} ->
		    error("Cannot open log ~p: ~p~n", [File, Reason])
	    end
    end.

view_file(C, Log) ->
    case disk_log:chunk(Log, C) of
	{error, Reason} ->
	    error("** Possibly truncated FILE ~n", []),
	    error;
	eof ->
	    disk_log:close(Log),
	    eof;
	{C2, Terms} ->
	    lists:foreach(fun(X) -> mnesia_lib:show("~p~n", [X]) end, Terms),
	    view_file(C2, Log)
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Backup

-record(backup, {cp_name, bup_module, bup_data}).

backup(Opaque) ->
    backup(Opaque, mnesia_monitor:get_env(backup_module)).

backup(Opaque, Mod) -> % Backup all tables, max redundancy
    Args = [{ram_overrides_dump, false}, {max, val({schema, tables})}],
    case mnesia_checkpoint:activate(Args) of
	{ok, Name, Nodes} ->
	    Res = backup_checkpoint(Name, Opaque, Mod),
	    mnesia_checkpoint:deactivate(Name),
	    Res;
	{error, Reason} ->
	    {error, Reason}
    end.

backup_checkpoint(Name, Opaque) ->
    backup_checkpoint(Name, Opaque, mnesia_monitor:get_env(backup_module)).

backup_checkpoint(Name, Opaque, Mod) ->
    Self = self(),
    Pid = spawn_link(?MODULE, backup_master, [Self, Name, Opaque, Mod]),
    receive
	{Pid, Self, Res} -> Res
    end.

backup_master(ClientPid, Name, Opaque, Mod) ->
    process_flag(trap_exit, true),
    B = #backup{cp_name=Name, bup_module=Mod, bup_data=Opaque},
    case catch backup_master(B) of
	{'EXIT', Reason} ->
	    ClientPid ! {self(), ClientPid, {error, {'EXIT', Reason}}};
	Res ->
	    ClientPid ! {self(), ClientPid, Res}
    end,
    unlink(ClientPid),
    exit(normal).

backup_master(B) ->
    Name = B#backup.cp_name,
    B2 = safe_apply(B, open_write, [B#backup.bup_data]),
    B3 = safe_apply(B2, write, [B2#backup.bup_data, [backup_log_header()]]),
    case mnesia_checkpoint:tables(Name) of
	{ok, Tabs} ->
	    %% Always put schema first in backup file
	    B4 =
		case lists:member(schema, Tabs) of
		    true ->
			backup_tab(schema, B3);
		    false ->
			Defs = [{schema, T, mnesia_schema:get_create_list(T)} || T <- Tabs],
			safe_apply(B3, write, [B3#backup.bup_data, Defs])
		end,
	    B5 = lists:foldl(fun backup_tab/2, B4, Tabs -- [schema]),
	    safe_apply(B5, commit_write, [B5#backup.bup_data]),
	    ok;
	{error, Reason} ->
	    abort_write(B3, {?MODULE, backup_master}, [B], {error, Reason})
    end.

safe_apply(B, write, [_, Items]) when Items == [] ->
    B;
safe_apply(B, What, Args) ->
    Abort = fun(R) -> abort_write(B, What, Args, R) end,
    receive
	{'EXIT', Pid, R} -> Abort({'EXIT', Pid, R})
    after 0 ->
	    Mod = B#backup.bup_module,
	    case catch apply(Mod, What, Args) of
		{ok, Opaque} -> B#backup{bup_data=Opaque};
		{error, R} -> Abort(R);
		R -> Abort(R)
	    end
    end.

abort_write(B, What, Args, Reason) ->
    Mod = B#backup.bup_module,
    Opaque = B#backup.bup_data,
    dbg_out("Failed to perform backup. M=~p:F=~p:A=~p -> ~p~n",
	    [Mod, What, Args, Reason]),
    case catch apply(Mod, abort_write, [Opaque]) of
	{ok, Res} ->
	    throw({error, Reason});
	Other ->
	    error("Failed to abort backup. ~p:~p~p -> ~p~n",
		  [Mod, abort_write, [Opaque], Other]),
	    throw({error, Reason})
    end.
	
backup_tab(Tab, B) ->
    Name = B#backup.cp_name,
    case mnesia_checkpoint:most_local_node(Name, Tab) of
	{ok, Node} ->
	    Pid = spawn_link(Node, ?MODULE, tab_copier, [self(), B, Tab]),
	    RecName = val({Tab, record_name}),
	    tab_receiver(Pid, B, Tab, RecName, 0);
	{error, Reason} ->
	    abort_write(B, {?MODULE, backup_tab}, [Tab, B], {error, Reason})
    end.

tab_receiver(Pid, B, Tab, RecName, Slot) ->
    Pid ! {self(), more, Slot},
    receive
	{Pid, {more, Next, Recs}} ->
	    Recs2 = rec_filter(Tab, RecName, Recs),
	    B2 = safe_apply(B, write, [B#backup.bup_data, Recs2]),
	    tab_receiver(Pid, B2, Tab, RecName, Next);
	
	{Pid, {last, {ok, _}}} ->
	    B;

	{Pid, {last, {error, R}}} ->
	    Reason = {error, {"Tab copier crashed", R}},
	    abort_write(B, {?MODULE, tab_copier}, [self(), B, Tab], Reason);
	
	{'EXIT', Pid, {error, R}} ->
	    Reason = {error, {"Tab copier crashed", R}},
	    abort_write(B, {?MODULE, tab_copier}, [self(), B, Tab], Reason);
	{'EXIT', Pid, R} ->
	    Reason = {error, {"Tab copier crashed", {'EXIT', R}}},
	    abort_write(B, {?MODULE, tab_copier}, [self(), B, Tab], Reason);
	Msg ->
	    fatal("Unexpected msg tab_receiver: ~p~n", [Msg])
    end.

rec_filter(schema, _RecName, Recs) ->
    case catch mnesia_bup:refresh_cookie(Recs) of
	Recs2 when list(Recs2) ->
	    Recs2;
	{error, _Reason} ->
	    %% No schema table cookie
	    Recs
    end;
rec_filter(Tab, Tab, Recs) ->
    Recs;
rec_filter(Tab, RecName, Recs) ->
    [setelement(1, Rec, Tab) || Rec <- Recs].

tab_copier(Pid, B, Tab) ->
    %% Intentional crash at exit
    Name = B#backup.cp_name,
    ?eval_debug_fun({?MODULE, tab_copier, pre}, [{name, Name}, {tab, Tab}]),
    Res2 =
	case mnesia_checkpoint:really_retain(B#backup.cp_name, Tab) of
	    true ->
		Res = iterate(Name, Tab, Pid, table, first, {ok, 0}),
		iterate(Name, Tab, Pid, retainer, last, Res);
	    false ->
		%% Put the dumped file in the backup
		%% instead of the ram table. Does
		%% only apply to ram_copies.
		iterate(Name, Tab, Pid, retainer, last, {ok, 0})
	end,
    ?eval_debug_fun({?MODULE, tab_copier, post}, [{name, Name}, {tab, Tab}]),
    unlink(Pid),
    Pid ! {self(), {last, Res2}},
    exit(normal).

iterate(_Name, _Tab, _Pid, _What, _Pass, {error, Reason}) ->
    {error, Reason};
iterate(Name, Tab, Pid, What, Pass, {ok, Acc}) ->
    Fun = fun(Recs, A) -> send_records(Pid, Tab, Recs, Pass, A) end,
    mnesia_checkpoint:iterate(Name, Tab, Fun, Acc, What, checkpoint).

send_records(Pid, Tab, [], Pass, Acc) ->
    receive
	{Pid, more, Acc} when Pass == last ->
	    Acc;
	{Pid, more, Acc} when Pass == first ->
	    Next = Acc + 1,
	    Pid ! {self(), {more, Next, []}},
	    Next;
	Msg ->
	    exit({send_records_unexpected_msg, Tab, Msg})
    end;
send_records(Pid, Tab, Recs, _Last, Acc) ->
    receive
	{Pid, more, Acc} ->
	    Next = Acc + 1,
	    Pid ! {self(), {more, Next, Recs}},
	    Next;
	Msg ->
	    exit({send_records_unexpected_msg2, Tab, Msg})
    end.
