%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% In this module we provide a number of explicit functions
%% to maninpulate the schema. All these functions are called
%% within a special schema transaction.
%%
%% We also have an init/1 function defined here, this func is
%% used by mnesia:start() to initialize the entire schema.

-module(mnesia_schema).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('klacke@erix.ericsson.se').
-author('hakan@erix.ericsson.se').

-export([
         add_snmp/2,
         add_table_copy/3,
         add_table_index/2,
	 arrange_restore/3,
         attr_tab_to_pos/2,
         change_table_copy_type/3,
         change_table_access_mode/2,
         change_table_load_order/2,
	 clear_table/1,
         create_table/1,
	 cs2list/1,
         del_snmp/1,
         del_table_copy/2,
         del_table_index/2,
         delete_cstruct/2,
         delete_schema/1,
         delete_schema2/0,
         delete_table/1,
         delete_table_property/2,
         dump_tables/1,
         ensure_no_schema/1,
	 ensure_running/2,
	 get_create_list/1,
         get_initial_schema/2,
	 get_table_properties/1,
         info/0,
         info/1,
         init/1,
         insert_cstruct/3,
	 is_remote_member/1,
         list2cs/1,
         lock_schema/0,
         merge_schema/0,
         move_table/3,
         prepare_commit/2,
         purge_tmp_files/0,
         ram_delete_table/2,
         ram_delete_table/3,
	 read_cstructs_from_disc/0,
         read_nodes/0,
         remote_read_schema/0,
	 restore/1,
         restore/2,
         restore/3,
         opt_create_dir/2,
         purge_dir/2,
         transform_table/3,
         undo_prepare_commit/2,
         unlock_schema/0,
	 verify/3,
         version/0,
         write_table_property/2
        ]).

%% Exports for mnesia_frag
-export([schema_transaction/1, insert_schema_ops/2,
	 get_tid_ts_and_lock/2,
	 make_write_table_properties/2, make_delete_table_properties/2,
	 make_write_table_properties/3, make_delete_table_properties/3,
	 do_create_table/1, do_delete_table/1,
	 do_add_table_copy/3, do_del_table_copy/2,
	 do_clear_table/1, do_change_table_copy_type/3,
	 do_transform_table/3, do_dump_tables/1,
	 do_change_table_access_mode/2, do_change_table_load_order/2,
	 do_write_table_property/2, do_delete_table_property/2
	]).

-include("mnesia.hrl").

-import(mnesia_lib, [set/2, del/2, verbose/2, dbg_out/2, fatal/2]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Here comes the init function which also resides in
%% this module, it is called upon by the trans server
%% at startup of the system
%% 
%% We have a meta table which looks like
%% {table, schema, 
%%    {type, set},
%%    {disc_copies, all},
%%    {arity, 2}
%%    {attributes, [key, val]}
%%
%% This means that we have a series of {schema, Name, Cs} tuples
%% in a table called schema !!

init(IgnoreFallback) ->
    Res = read_schema(true, false, IgnoreFallback),
    {ok, Source, _CreateList} = exit_on_error(Res),
    verbose("Schema initiated from: ~p~n", [Source]),
    set({schema, tables}, []),
    set({schema, local_tables}, []),
    Tabs = set_schema(erlang:db_first(schema)),
    lists:foreach(fun(Tab) -> clear_whereabouts(Tab) end, Tabs),
    set({schema, where_to_read}, node()),
    set({schema, load_node}, node()),
    set({schema, load_reason}, initial),
    mnesia_controller:add_active_replica(schema, node()).

exit_on_error({error, Reason}) ->
    exit(Reason);
exit_on_error(GoodRes) ->
    GoodRes.

val(Var) ->
    case ?catch_val(Var) of
	{'EXIT', Reason} -> mnesia_lib:other_val(Var, Reason); 
	Value -> Value 
    end.

%% This function traverses all cstructs in the schema and
%% sets all values in mnesia_gvar accordingly for each table/cstruct

set_schema('$end_of_table') -> 
    [];
set_schema(Tab) ->
    do_set_schema(Tab),
    [Tab | set_schema(erlang:db_next_key(schema, Tab))].

get_create_list(Tab) ->
    erlang:db_get_element(schema, Tab, 3).

do_set_schema(Tab) ->
    List = get_create_list(Tab),
    Cs = list2cs(List),

    Type = Cs#cstruct.type,
    set({Tab, setorbag}, Type),
    set({Tab, local_content}, Cs#cstruct.local_content),
    set({Tab, ram_copies}, Cs#cstruct.ram_copies),
    set({Tab, disc_copies}, Cs#cstruct.disc_copies),
    set({Tab, disc_only_copies}, Cs#cstruct.disc_only_copies),
    set({Tab, load_order}, Cs#cstruct.load_order),
    set({Tab, access_mode}, Cs#cstruct.access_mode),
    set({Tab, snmp}, Cs#cstruct.snmp),
    set({Tab, user_properties}, Cs#cstruct.user_properties),
    [set({Tab, user_property, element(1, P)}, P) || P <- Cs#cstruct.user_properties],
    set({Tab, attributes}, Cs#cstruct.attributes),
    Arity = length(Cs#cstruct.attributes) + 1,
    set({Tab, arity}, Arity),
    RecName =  Cs#cstruct.record_name,
    set({Tab, record_name}, RecName),
    set({Tab, record_validation}, {RecName, Arity, Type}),
    set({Tab, wild_pattern}, wild(RecName, Arity)),
    set({Tab, index}, Cs#cstruct.index),
    %% create actual index tabs later
    set({Tab, cookie}, Cs#cstruct.cookie),
    set({Tab, version}, Cs#cstruct.version),
    set({Tab, cstruct}, Cs),
    Storage = mnesia_lib:schema_cs_to_storage_type(node(), Cs),
    set({Tab, storage_type}, Storage),
    mnesia_lib:add({schema, tables}, Tab),
    Ns = mnesia_lib:cs_to_nodes(Cs),
    case lists:member(node(), Ns) of
        true ->
            mnesia_lib:add({schema, local_tables}, Tab);
        false when Tab == schema ->
            mnesia_lib:add({schema, local_tables}, Tab);
        false ->
            ignore
    end.

wild(RecName, Arity) ->
    Wp0 = list_to_tuple(lists:duplicate(Arity, '_')),
    Wp = setelement(1, Wp0, RecName).

%% Temporarily read the local schema and return a list
%% of all nodes mentioned in the schema.DAT file
read_nodes() ->
    %% Ensure that we access the intended Mnesia
    %% directory. This function may not be called
    %% during startup since it will cause the
    %% application_controller to get into deadlock
    case mnesia_lib:ensure_loaded(?APPLICATION) of
	ok ->
	    case read_schema(false, false) of
		{ok, Source, CreateList} ->
		    Cs = list2cs(CreateList),
		    {ok, Cs#cstruct.disc_copies ++ Cs#cstruct.ram_copies};
		{error, Reason} ->
		    {error, Reason}
	    end;
	{error, Reason} ->
	    {error, Reason}
    end.

%% Returns Version from the tuple {Version,MasterNodes}
version() ->
    case read_schema(false, false) of
        {ok, Source, CreateList} when Source /= default ->
	    Cs = list2cs(CreateList),
            {Version, Details} = Cs#cstruct.version,
            Version;
        _ ->
            case dir_exists(mnesia_lib:dir()) of
                true -> {1,0};
                false -> {0,0}
            end
    end.

%% Calculate next table version from old cstruct
incr_version(Cs) ->
    {{Major, Minor}, _} = Cs#cstruct.version,
    Nodes = mnesia_lib:intersect(val({schema, disc_copies}),
                                 mnesia_lib:cs_to_nodes(Cs)), 
    V = 
        case Nodes -- val({Cs#cstruct.name, active_replicas}) of
            [] -> {Major + 1, 0};    % All replicas are active
            _ -> {Major, Minor + 1}  % Some replicas are inactive
        end,
    Cs#cstruct{version = {V, {node(), now()}}}.

%% Returns table name
insert_cstruct(Tid, Cs, KeepWhereabouts) ->
    Tab = Cs#cstruct.name,
    TabDef = cs2list(Cs),
    Val = {schema, Tab, TabDef},
    Item = {{schema, Tab}, Val, write},
    mnesia_checkpoint:tm_retain(Tid, Item),
    Active = val({Tab, active_replicas}),

    case KeepWhereabouts of
        true ->
            ignore;
        false when Active == [] ->
            clear_whereabouts(Tab);
        false ->
            %% Someone else has initiated table
            ignore
    end,
    set({Tab, cstruct}, Cs),
    erlang:db_put(schema, Val),
    do_set_schema(Tab),
    Val.

clear_whereabouts(Tab) ->
    set({Tab, checkpoints}, []),
    set({Tab, subscribers}, []),
    set({Tab, index_tables}, []),
    set({Tab, where_to_read}, nowhere),
    set({Tab, active_replicas}, []),
    set({Tab, commit_work}, []),
    set({Tab, where_to_write}, []),
    set({Tab, where_to_commit}, []),
    set({Tab, load_by_force}, false),
    set({Tab, load_node}, unknown),
    set({Tab, load_reason}, unknown).

%% Returns table name
delete_cstruct(Tid, Cs) ->
    Tab = Cs#cstruct.name,
    TabDef = cs2list(Cs),
    Val = {schema, Tab, TabDef},
    mnesia_checkpoint:tm_retain(Tid,{{schema,Tab}, Val, delete}),
    erlang:db_match_erase(mnesia_gvar, {{Tab, '_'}, '_'}),
    del({schema, local_tables}, Tab),
    del({schema, tables}, Tab),
    erlang:db_erase(schema, Tab),
    Val.

%% Delete the Mnesia directory on all given nodes
%% Requires that Mnesia is not running anywhere
%% Returns ok | {error,Reason}
delete_schema(Ns) when list(Ns), Ns /= [] ->
    RunningNs = mnesia_lib:running_nodes(Ns),
    Reason = "Cannot delete schema on all nodes",
    if
        RunningNs == [] ->
	    case rpc:multicall(Ns, ?MODULE, delete_schema2, []) of
		{Replies, []} ->
		    case [R || R <- Replies, R /= ok]  of
			[] ->
			    ok;
			BadReplies ->
			    verbose("~s: ~p~n", [Reason, BadReplies]),
			    {error, {"All nodes not running", BadReplies}}
		    end;
		{Replies, BadNs} ->
                    verbose("~s: ~p~n", [Reason, BadNs]),
                    {error, {"All nodes not running", BadNs}}
            end;
        true ->
            verbose("~s: ~p~n", [Reason, RunningNs]),
            {error, {"Mnesia is not stopped everywhere", RunningNs}}
    end;
delete_schema(Ns) ->
    {error, {badarg, Ns}}.

delete_schema2() ->
    %% Ensure that we access the intended Mnesia
    %% directory. This function may not be called
    %% during startup since it will cause the
    %% application_controller to get into deadlock
    case mnesia_lib:ensure_loaded(?APPLICATION) of
	ok ->
	    case mnesia_lib:is_running() of
		no ->
		    Dir = mnesia_lib:dir(),
		    purge_dir(Dir, []),
		    ok;
		_ ->
		    {error, {"Mnesia still running", node()}}
	    end;
	{error, Reason} ->
	    {error, Reason}
    end.
    
ensure_no_schema([H|T]) when atom(H) ->
    case rpc:call(H, ?MODULE, remote_read_schema, []) of
        {badrpc, Reason} ->
            {H, {"All nodes not running", H, Reason}};
        {ok,Source, _} when Source /= default ->
            {H, {already_exists, H}};
        _ ->
            ensure_no_schema(T)
    end;
ensure_no_schema([H|_]) ->
    {error,{badarg, H}};
ensure_no_schema([]) ->
    ok.

remote_read_schema() ->
    %% Ensure that we access the intended Mnesia
    %% directory. This function may not be called
    %% during startup since it will cause the
    %% application_controller to get into deadlock
    case mnesia_lib:ensure_loaded(?APPLICATION) of
	ok ->
	    case mnesia_monitor:get_env(schema_location) of
		opt_disc ->
		    read_schema(false, true);
		_ ->
		    read_schema(false, false)
	    end;
	{error, Reason} ->
	    {error, Reason}
    end.

dir_exists(Dir) ->
    dir_exists(Dir, mnesia_monitor:use_dir()).
dir_exists(Dir, true) ->
    case file:file_info(Dir) of
        {ok, _} -> true;
        _ -> false
    end;
dir_exists(Dir, false) ->
    false.

opt_create_dir(UseDir, Dir) when UseDir == true->
    case dir_exists(Dir, UseDir) of
        true ->
            check_can_write(Dir);
        false ->
            case file:make_dir(Dir) of
                ok -> 
                    verbose("Create Directory ~p~n", [Dir]),
                    ok;
                {error, Reason} ->
                    verbose("Cannot create mnesia dir ~p~n", [Reason]),
                    {error, {"Cannot create Mnesia dir", Dir, Reason}}
            end
    end;
opt_create_dir(false, _) ->
    {error, {has_no_disc, node()}}.
    
check_can_write(Dir) ->
    case file:file_info(Dir) of
        {ok, {_, directory, read_write, _,_,_,_}} ->
            ok;
        {ok, _} ->
            {error, "Not allowed to write in Mnesia dir", Dir};
        _ ->
            {error, "Non existent Mnesia dir", Dir}
    end.

lock_schema() ->
    mnesia_lib:lock_table(schema).

unlock_schema() ->
    mnesia_lib:unlock_table(schema).

read_schema(Keep, UseDirAnyway) ->
    read_schema(Keep, false, false).

%% The schema may be read for several reasons.
%% If Mnesia is not already started the read intention
%% we normally do not want the ets table named schema
%% be left around.
%% If Keep == true, the ets table schema is kept
%% If Keep == false, the ets table schema is removed
%%
%% Returns {ok, Source, SchemaCstruct} or {error, Reason}
%% Source may be: default | ram | disc | fallback

read_schema(Keep, UseDirAnyway, IgnoreFallback) ->
    lock_schema(),
    Res = 
        case mnesia:system_info(is_running) of
            yes ->
                {ok, ram, get_create_list(schema)};
            IsRunning ->
                    case mnesia_monitor:use_dir() of
                        true ->
                            read_disc_schema(Keep, IgnoreFallback);
                        false when UseDirAnyway == true ->
                            read_disc_schema(Keep, IgnoreFallback);
                        false when Keep == true ->
                            Args = [{keypos, 2}, public, named_table, set],
                            mnesia_monitor:mktab(schema, Args),
                            CreateList = get_initial_schema(ram_copies, []),
                            erlang:db_put(schema,{schema, schema, CreateList}),
                            {ok, default, CreateList};
                        false when Keep == false ->
			    CreateList = get_initial_schema(ram_copies, []),
                            {ok, default, CreateList}
                    end
        end,
    unlock_schema(),
    Res.

read_disc_schema(Keep, IgnoreFallback) ->
    Running = mnesia:system_info(is_running),
    case mnesia_bup:fallback_exists() of
        true when IgnoreFallback == false, Running /= yes ->
             mnesia_bup:fallback_to_schema();
        _ ->  
            %% If we're running, we read the schema file even
            %% if fallback exists
            Dat = mnesia_lib:tab2dat(schema),
            case mnesia_lib:exists(Dat) of
                true ->
                    do_read_disc_schema(Dat, Keep);
                false ->
		    Dmp = mnesia_lib:tab2dmp(schema),
		    case mnesia_lib:exists(Dmp) of
			true ->
			    %% May only happen when toggling of
			    %% schema storage type has been
			    %% interrupted
			    do_read_disc_schema(Dmp, Keep);
			false ->
			    {error, "No schema file exists"}
		    end
            end
    end.

do_read_disc_schema(Fname, Keep) ->
    T = 
        case Keep of
            false ->
                Args = [{keypos, 2}, public, set],
                erlang:db_create(schema, Args);
            true ->
                Args = [{keypos, 2}, public, named_table, set],
                mnesia_monitor:mktab(schema, Args)
        end,
    Repair = mnesia_monitor:get_env(auto_repair),
    Res =
        case mnesia_lib:dat_to_ets(schema, T, Fname, set, Repair, no) of
            loaded -> {ok, disc, erlang:db_get_element(T, schema, 3)};
            Other -> {error, {"Cannot read schema", Fname, Other}}
        end,
    case Keep of
        true -> ignore;
        false -> erlang:db_erase(T)
    end,
    Res.

get_initial_schema(SchemaStorage, Nodes) ->
    Cs = #cstruct{name = schema,
		  record_name = schema,
		  attributes = [table, cstruct]},    
    Cs2 =
	case SchemaStorage of
        ram_copies -> Cs#cstruct{ram_copies = Nodes};
        disc_copies -> Cs#cstruct{disc_copies = Nodes}
    end,
    cs2list(Cs2).

read_cstructs_from_disc() ->
    %% Assumptions: 
    %% - local schema lock in global
    %% - use_dir is true
    %% - Mnesia is not running
    %% - Ignore fallback

    Fname = mnesia_lib:tab2dat(schema),
    case mnesia_lib:exists(Fname) of
	true ->
	    Args = [{file, Fname},
		    {keypos, 2},
		    {repair, mnesia_monitor:get_env(auto_repair)},
		    {type, set}],
	    case dets:open_file(make_ref(), Args) of
		{ok, Tab} ->
		    Fun = fun({_, _, List}) ->
				  {continue, list2cs(List)}
			  end,
		    Cstructs = dets:traverse(Tab, Fun),
		    dets:close(Tab),
		    {ok, Cstructs};
		{error, Reason} ->
		    {error, Reason}
	    end;
	false ->
	    {error, "No schema file exists"}
    end.
    
%% We run a very special type of transactions when we
%% we want to manipulate the schema.

get_tid_ts_and_lock(Tab, Intent) ->
    TidTs = get(mnesia_activity_state),
    case TidTs of
	{Mod, Tid, Ts} when record(Ts, tidstore)->
	    Store = Ts#tidstore.store,
	    case Intent of
		read -> mnesia_locker:rlock_table(Tid, Store, Tab);
		write -> mnesia_locker:wlock_table(Tid, Store, Tab);
		none -> ignore
	    end,
	    TidTs;
	_ ->
	    mnesia:abort(no_transaction)
    end.

schema_transaction(Fun) ->
    case get(mnesia_activity_state) of
	undefined ->
            mnesia:transaction(Fun);
	_ ->
            {aborted, nested_transaction}
    end.

%% The make* rotines return a list of ops, this function
%% inserts em all in the Store and maintains the local order
%% of ops.

insert_schema_ops({_Mod, _Tid, Ts}, SchemaIOps) ->
    do_insert_schema_ops(Ts#tidstore.store, SchemaIOps).
    
do_insert_schema_ops(Store, [Head | Tail]) ->
    case element(1, Head) of
	op -> 
	    erlang:db_put(Store, Head),
	    do_insert_schema_ops(Store, Tail);
	restore_op ->
	    erlang:db_put(Store, Head),
	    do_insert_schema_ops(Store, Tail)
    end;
do_insert_schema_ops(_Store, []) ->
    ok.

cs2list(Cs) when record(Cs, cstruct) ->
    Tags = record_info(fields, cstruct),
    rec2list(Tags, 2, Cs);
cs2list(CreateList) when list(CreateList) ->
    CreateList.

rec2list([Tag | Tags], Pos, Rec) ->
    Val = element(Pos, Rec),
    [{Tag, Val} | rec2list(Tags, Pos + 1, Rec)];
rec2list([], _Pos, _Rec) ->
    [].

list2cs(List) ->
    Name = pick(name, List, must),
    Type = pick(type, List, set),
    Rc0 = pick(ram_copies, List, []),
    Dc = pick(disc_copies, List, []),
    Doc = pick(disc_only_copies, List, []),
    L = pick(local_content, List, false),
    RecName = pick(record_name, List, Name),
    Attrs = pick(attributes, List, [key, val]),
    Rc = case {Rc0,Dc,Doc} of
             {[],[],[]} -> [node()];
             _ -> Rc0
         end,
    Snmp = pick(snmp, List, []),
    LoadOrder = pick(load_order, List, 0),
    AccessMode = pick(access_mode, List, read_write),
    UserProps = pick(user_properties, List, []),
    verify({alt, [nil, list]}, mnesia_lib:etype(UserProps), {bad_type, Name, UserProps}),
    Cookie = pick(cookie, List, {erlang:now(), node()}),
    Version = pick(version, List, {{2, 0}, []}),
    Ix = pick(index, List, []),
    verify({alt, [nil, list]}, mnesia_lib:etype(Ix), {bad_type, Name, Ix}),
    Ix2 = lists:map(fun(I) -> attr_to_pos(I, Attrs) end, Ix),
    Keys = mnesia_lib:elems(1, List),
    check_keys(Keys, record_info(fields, cstruct)),
    check_duplicates(Keys),
    #cstruct{name = Name,
             ram_copies = Rc,
             disc_copies = Dc,
             disc_only_copies = Doc,
             type = Type,
             index = Ix2,
             snmp = Snmp,
             load_order = LoadOrder,
             access_mode = AccessMode,
             local_content = L,
	     record_name = RecName,
             attributes = Attrs,
             user_properties = lists:sort(UserProps),
             cookie = Cookie,
             version = Version}.

pick(Key, List, Default) ->
    case lists:keysearch(Key, 1, List) of
        false  when Default == must ->
            throw(error);
        false ->
            Default;
        {value, {Key, Value}} ->
            Value
    end.

%% Convert attribute name to integer if neccessary
attr_tab_to_pos(_Tab, Pos) when integer(Pos) ->
    Pos;
attr_tab_to_pos(Tab, Attr) ->
    attr_to_pos(Attr, val({Tab, attributes})).
    
%% Convert attribute name to integer if neccessary
attr_to_pos(Pos, _Attrs) when integer(Pos) ->
    Pos;
attr_to_pos(Attr, Attrs) when atom(Attr) ->
    attr_to_pos(Attr, Attrs, 2);
attr_to_pos(Attr, _) ->
    mnesia:abort({bad_type, Attr}).

attr_to_pos(Attr, [Attr | Attrs], Pos) ->
    Pos;
attr_to_pos(Attr, [_ | Attrs], Pos) ->
    attr_to_pos(Attr, Attrs, Pos + 1);
attr_to_pos(Attr, _, _) ->
    mnesia:abort({bad_type, Attr}).
    
check_keys([], _) -> ok;
check_keys([H|T], Items) ->
    case lists:member(H, Items) of
        true ->  check_keys(T, Items);
        false -> mnesia:abort({badarg, H})
    end.

check_duplicates(Keys) ->
    case has_duplicates(Keys) of
        false -> ok;
        true -> mnesia:abort({badarg, "Duplicate keys", Keys})
    end.

has_duplicates([H | T]) ->
    case lists:member(H, T) of
        true -> true;
        false -> has_duplicates(T)
    end;
has_duplicates([]) ->
    false.

%% This is the only place where we check the validity of data
check_cstruct(Cs) ->
    verify_nodes(Cs),

    Tab = Cs#cstruct.name,
    verify(atom, mnesia_lib:etype(Tab), {bad_type, Tab}),
    Type = Cs#cstruct.type,
    verify(true, lists:member(Type, [set, bag]), {bad_type, Tab, Type}),
    RecName = Cs#cstruct.record_name,
    verify(atom, mnesia_lib:etype(RecName), {bad_type, Tab, RecName}),

    Attrs = Cs#cstruct.attributes,
    verify(list, mnesia_lib:etype(Attrs), {bad_type, Tab, Attrs}),

    Arity = length(Attrs) + 1,
    verify(true, Arity > 2, {bad_type, Tab, Attrs}),
    
    lists:foldl(fun(Attr,Other) when Attr == snmp ->
                        mnesia:abort({bad_type, Tab, Attr});
                   (Attr,Other) -> 
                        verify(atom, mnesia_lib:etype(Attr),
                               {bad_type, Tab, Attr}),
                        verify(false, lists:member(Attr,Other),
                               {combine_error, Tab, Attr}),
                        [Attr|Other]
                end,
                [],
                Attrs),

    Index = Cs#cstruct.index,
    verify({alt, [nil, list]}, mnesia_lib:etype(Index), {bad_type, Tab, Index}),
    
    IxFun =
        fun(Pos) ->
                verify(true, fun() ->
                                     if
                                         Pos > 2, integer(Pos),
                                         Pos =< Arity ->
                                             true;
                                         true -> false
                                     end
                             end,
                       {bad_index, Tab, Pos})
        end,
    lists:foreach(IxFun, Index),
    
    LC = Cs#cstruct.local_content,
    verify({alt, [true, false]}, LC, {bad_type, Tab, LC}),
    Access = Cs#cstruct.access_mode,
    verify({alt, [read_write, read_only]}, Access, {bad_type, Tab, Access}),

    Snmp = Cs#cstruct.snmp,
    verify(true, mnesia_snmp_hook:check_ustruct(Snmp), {badarg, Tab, Snmp}),

    CheckProp = fun(Prop) when tuple(Prop), size(Prop) >= 1 -> ok;
		   (Prop) -> mnesia:abort({bad_type, Tab, Prop})
		end,
    lists:foreach(CheckProp, Cs#cstruct.user_properties),

    case Cs#cstruct.cookie of
    {{MegaSecs, Secs, MicroSecs}, Node}
                when integer(MegaSecs), integer(Secs),
                     integer(MicroSecs), atom(node) ->
            ok;
        Cookie ->
            mnesia:abort({bad_type, Tab, Cookie})
    end,
    case Cs#cstruct.version of
        {{Major, Minor}, Detail}
                when integer(Major), integer(Minor) ->
            ok;
        Version ->
            mnesia:abort({bad_type, Tab, Version})
    end,
    Cs.

verify_nodes(Cs) ->
    Tab = Cs#cstruct.name,
    Ram = Cs#cstruct.ram_copies,
    Disc = Cs#cstruct.disc_copies,
    DiscOnly = Cs#cstruct.disc_only_copies,
    LoadOrder = Cs#cstruct.load_order,

    verify({alt, [nil, list]}, mnesia_lib:etype(Ram), {bad_type, Tab, Ram}),
    verify({alt, [nil, list]}, mnesia_lib:etype(Disc), {bad_type, Tab, Disc}),
    verify({alt, [nil, list]}, mnesia_lib:etype(DiscOnly), {bad_type, Tab, DiscOnly}),
    verify(integer, mnesia_lib:etype(LoadOrder), {bad_type, Tab, LoadOrder}),
    
    Nodes = Ram ++ Disc ++ DiscOnly,
    verify(false, has_duplicates(Nodes), {combine_error, Tab, Nodes}),
    AtomCheck = fun(N) -> verify(atom, mnesia_lib:etype(N), {bad_type, Tab, N}) end, 
    lists:foreach(AtomCheck, Nodes).
    
verify(Expected, Fun, Error) when function(Fun) ->
    do_verify(Expected, catch Fun(), Error);
verify(Expected, Actual, Error) ->
    do_verify(Expected, Actual, Error).

do_verify({alt, Values}, Value, Error) ->
    case lists:member(Value, Values) of
        true -> ok;
        false -> mnesia:abort(Error)
    end;
do_verify(Value, Value, _) ->
    ok;
do_verify(Value, _, Error) ->
     mnesia:abort(Error).

ensure_writable(Tab) ->
    case val({Tab, where_to_write}) of
        [] -> mnesia:abort({read_only, Tab});
        _ -> ok
    end.

%% Ensure that all replicas on disk full nodes are active
ensure_active(Cs) ->
    ensure_active(Cs, active_replicas).

ensure_active(Cs, What) ->
    Tab = Cs#cstruct.name,
    case val({Tab, What}) of
        [] -> mnesia:abort({no_exists, Tab});
        _ -> ok
    end,
    Nodes = mnesia_lib:intersect(val({schema, disc_copies}),
                                 mnesia_lib:cs_to_nodes(Cs)), 
    W = {Tab, What},
    case Nodes -- val(W) of
        [] ->
            ok;
        Ns ->
            Expl = "All replicas on diskfull nodes are not active yet",
            case val({Tab, local_content}) of
                true ->
		    case rpc:multicall(Ns, ?MODULE, is_remote_member, [W]) of
			{Replies, []} ->
			    check_active(Replies, Expl, Tab);
			{_Replies, BadNs} ->
			    mnesia:abort({not_active, Expl, Tab, BadNs})
                    end;
                false ->
                    mnesia:abort({not_active, Expl, Tab, Ns})
            end
    end.

is_remote_member(Key) ->
    IsActive = lists:member(node(), val(Key)),
    {IsActive, node()}.

check_active([{true, Node} | Replies], Expl, Tab) ->
    check_active(Replies, Expl, Tab);
check_active([{false, Node} | _Replies], Expl, Tab) ->
    mnesia:abort({not_active, Expl, Tab, [Node]});
check_active([{badrpc, Reason} | _Replies], Expl, Tab) ->
    mnesia:abort({not_active, Expl, Tab, Reason});
check_active([], _Expl, _Tab) ->
    ok.

%% Ensure that node is running or that the new replica
%% is ram_copies which not has been dumped to disc
ensure_running(Cs, Node) ->
    IsRunning = lists:member(Node, val({current, db_nodes})),
    Storage =  mnesia_lib:schema_cs_to_storage_type(Node, Cs),
    Tab = Cs#cstruct.name,
    Reason = {not_active, Tab, Node},
    if
        IsRunning == true ->
            ok;
        Storage == ram_copies ->
            case Cs#cstruct.disc_copies ++ Cs#cstruct.disc_only_copies of
                [] ->
		    case mnesia_lib:schema_cs_to_storage_type(node(), Cs) of
                        unknown ->
                            %% Cannot check if table is dumped to disc
                            mnesia:abort(Reason);
                        ram_copies  ->
                            Dat = mnesia_lib:tab2dat(Tab),
                            case mnesia_lib:exists(Dat) of
                                true ->
                                    mnesia:abort(Reason);
                                false ->
                                    ok
                            end
                    end;
                _ ->
                    ok
            end;
        true ->
            mnesia:abort(Reason)
    end.

%% Here's the real interface function to create a table

create_table(X) ->
    schema_transaction(fun() -> do_create_table(X) end).

%% And the corresponding do routines ....

do_create_table(Cs) ->
    do_create_table(Cs, val({schema, tables})).

do_create_table(Cs, _) when Cs#cstruct.name == schema ->
    mnesia:abort({bad_type, schema});
do_create_table(Cs, Tabs) when record(Cs, cstruct) ->
    ensure_writable(schema),
    {_Mod, Tid, Ts} = get_tid_ts_and_lock(schema, write),
    check_cstruct(Cs),
    Tab = Cs#cstruct.name,
    verify(false, lists:member(Tab, Tabs), {already_exists, Tab}),
        
    %% Check that we have all disc replica nodes running
    DiscNodes = Cs#cstruct.disc_copies ++ Cs#cstruct.disc_only_copies,
    RunningNodes = val({current, db_nodes}),
    CheckDisc = fun(N) ->
			verify(true, lists:member(N, RunningNodes), {not_active, N})
		end,
    lists:foreach(CheckDisc, DiscNodes),

    Nodes = mnesia_lib:intersect(mnesia_lib:cs_to_nodes(Cs), RunningNodes),
    Store = Ts#tidstore.store,
    mnesia_locker:wlock_no_exist(Tid, Store, Tab, Nodes),
    Ops = [{op, create_table, cs2list(Cs)}],
    do_insert_schema_ops(Store, Ops);

do_create_table(List, Tabs) when list(List) ->  %% Key value list
    case catch list2cs(List) of
        {'EXIT', {aborted, Reason}} ->
            mnesia:abort(Reason);
        {'EXIT', Reason} ->
            mnesia:abort({badarg, List, Reason});
        Cs ->
            do_create_table(Cs, Tabs)
    end;
do_create_table(Other, _) ->
    mnesia:abort({badarg, Other}).

%% Delete a table entirely on all nodes.

delete_table(Tab) ->
    schema_transaction(fun() -> do_delete_table(Tab) end).

do_delete_table(schema) ->
    mnesia:abort({bad_type, schema});
do_delete_table(Tab) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, write),
    insert_schema_ops(TidTs, make_delete_table(Tab)).

make_delete_table(Tab) ->
    ensure_writable(schema),
    Cs = val({Tab, cstruct}),
    ensure_active(Cs),
    ensure_writable(Tab),
    [{op, delete_table, cs2list(Cs)}].

%% Clear a table
clear_table(Tab) ->
    schema_transaction(fun() -> do_clear_table(Tab) end).

do_clear_table(schema) ->
    mnesia:abort({bad_type, schema});
do_clear_table(Tab) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, write),
    insert_schema_ops(TidTs, make_clear_table(Tab)).

make_clear_table(Tab) ->
    ensure_writable(schema),
    Cs = val({Tab, cstruct}),
    ensure_active(Cs),
    ensure_writable(Tab),
    [{op, clear_table, cs2list(Cs)}].

%% endof create/delete functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% 

add_table_copy(Tab, Node, Storage) ->
    schema_transaction(fun() -> do_add_table_copy(Tab, Node, Storage) end).

do_add_table_copy(Tab, Node, Storage) when atom(Tab), atom(Node) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    insert_schema_ops(TidTs, make_add_table_copy(Tab, Node, Storage));
do_add_table_copy(Tab,Node,_) ->
    mnesia:abort({badarg, Tab, Node}).

make_add_table_copy(Tab, Node, Storage) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    Ns = mnesia_lib:cs_to_nodes(Cs),
    verify(false, lists:member(Node, Ns), {already_exists, Tab, Node}),
    Cs2 = new_cs(Cs, Node, Storage, add),
    ensure_running(Cs2, Node),
    check_cstruct(Cs2),
    [{op, add_table_copy, Storage, Node, cs2list(Cs2)}].

del_table_copy(Tab, Node) ->
    schema_transaction(fun() -> do_del_table_copy(Tab, Node) end).

do_del_table_copy(Tab, Node) when atom(Node)  ->
    TidTs = get_tid_ts_and_lock(schema, write),
%%    get_tid_ts_and_lock(Tab, read), %%dan
    insert_schema_ops(TidTs, make_del_table_copy(Tab, Node));
do_del_table_copy(Tab, Node) ->
    mnesia:abort({badarg, Tab, Node}).
   
make_del_table_copy(Tab, Node) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    Storage = mnesia_lib:schema_cs_to_storage_type(Node, Cs),
    Cs2 = new_cs(Cs, Node, Storage, del),
    case mnesia_lib:cs_to_nodes(Cs2) of
        [] when Tab == schema ->
            mnesia:abort({combine_error, Tab, "Last replica"});
        [] ->
            dbg_out("Last replica deleted in table ~p~n",  [Tab]),
            make_delete_table(Tab);
        _ when Tab == schema, Storage == disc_copies ->
            make_change_table_copy_type(Tab, Node, ram_copies);
        _ ->
            check_cstruct(Cs2),
            [{op, del_table_copy, Storage, Node, cs2list(Cs2)}]
    end.

new_cs(Cs, Node, ram_copies, add) ->
    Cs#cstruct{ram_copies = opt_add(Node, Cs#cstruct.ram_copies)};
new_cs(Cs, Node, disc_copies, add) ->
    Cs#cstruct{disc_copies = opt_add(Node, Cs#cstruct.disc_copies)};
new_cs(Cs, Node, disc_only_copies, add) ->
    Cs#cstruct{disc_only_copies = opt_add(Node, Cs#cstruct.disc_only_copies)};

new_cs(Cs, Node, ram_copies, del) ->
    Cs#cstruct{ram_copies = lists:delete(Node , Cs#cstruct.ram_copies)};
new_cs(Cs, Node, disc_copies, del) ->
    Cs#cstruct{disc_copies = lists:delete(Node , Cs#cstruct.disc_copies)};
new_cs(Cs, Node, disc_only_copies, del) ->
    Cs#cstruct{disc_only_copies = 
               lists:delete(Node , Cs#cstruct.disc_only_copies)};
new_cs(Cs, Node, Storage, Op) ->
    mnesia:abort({badarg, Cs#cstruct.name, Storage}).


opt_add(N, L) -> [N | lists:delete(N, L)].

move_table(Tab, FromNode, ToNode) ->
    schema_transaction(fun() -> do_move_table(Tab, FromNode, ToNode) end).

do_move_table(schema, FromNode, ToNode) ->
    mnesia:abort({bad_type, schema});
do_move_table(Tab, FromNode, ToNode) when atom(FromNode), atom(ToNode) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    insert_schema_ops(TidTs, make_move_table(Tab, FromNode, ToNode));
do_move_table(Tab, FromNode, ToNode) ->
    mnesia:abort({badarg, Tab, FromNode, ToNode}).

make_move_table(Tab, FromNode, ToNode) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    Ns = mnesia_lib:cs_to_nodes(Cs),
    verify(false, lists:member(ToNode, Ns), {already_exists, Tab, ToNode}),
    verify(true, lists:member(FromNode, val({Tab, where_to_write})),
           {not_active, Tab, FromNode}),
    verify(false, val({Tab,local_content}),
           {"Cannot move table with local content", Tab}),
    ensure_active(Cs),
    Running = val({current, db_nodes}),
    verify(true, lists:member(ToNode, Running), {"Node not running", ToNode}),
    Storage = mnesia_lib:schema_cs_to_storage_type(FromNode, Cs),

    Cs2 = new_cs(Cs, ToNode, Storage, add),
    Cs3 = new_cs(Cs2, FromNode, Storage, del),
    check_cstruct(Cs3),
    [{op, add_table_copy, Storage, ToNode, cs2list(Cs2)},
     {op, del_table_copy, Storage, FromNode, cs2list(Cs3)}].

%% end of functions to add and delete nodes to tables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%

change_table_copy_type(Tab, Node, ToS) when Node == node() ->
    schema_transaction(fun() -> do_change_table_copy_type(Tab, Node, ToS) end);
change_table_copy_type(Tab, Node, ToS) ->
    Args = [Tab, Node, ToS],
    case rpc:call(Node, ?MODULE, change_table_copy_type, Args) of
        {badrpc, _Reason} ->
            {aborted, {node_not_running, Node}};
        Res ->
            Res
    end.

do_change_table_copy_type(Tab, Node, ToS) when atom(Node) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, write), % ensure global sync
    %% get_tid_ts_and_lock(Tab, read),
    insert_schema_ops(TidTs, make_change_table_copy_type(Tab, Node, ToS));
do_change_table_copy_type(Tab, Node, ToS) ->
    mnesia:abort({badarg, Tab, Node}).

make_change_table_copy_type(Tab, Node, ToS) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    FromS = mnesia_lib:schema_cs_to_storage_type(Node, Cs),
    IsActive = mnesia_lib:active_remote_copy_holders(Cs, [Node]),
    case compare_storage_type(false, FromS, ToS) of
        _ when IsActive == false ->
            mnesia:abort({not_active, Tab, Node}); % Is not active remote
        {same, _} ->
            mnesia:abort({already_exists, Tab, Node, ToS});
        {diff, _} ->
            ignore;
        incompatible ->
            ensure_active(Cs)
    end,

    Cs2 = new_cs(Cs, Node, FromS, del),
    Cs3 = new_cs(Cs2, Node, ToS, add),
    check_cstruct(Cs3),
    case Tab of
        schema when ToS == disc_only_copies ->
            mnesia:abort({combine_error, Tab, ToS});
        schema when ToS == disc_copies, FromS /= ram_copies ->
            mnesia:abort({combine_error, Tab, ToS});
        schema when ToS == ram_copies, FromS /= disc_copies ->
            mnesia:abort({combine_error, Tab, ToS});
        _ ->
            ignore
    end,
    [{op, change_table_copy_type, Node, FromS, ToS, cs2list(Cs3)}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% change index functions ....
%% Pos is allready added by 1 in both of these functions 

add_table_index(Tab, Pos) ->
    schema_transaction(fun() -> do_add_table_index(Tab, Pos) end).

do_add_table_index(schema, Attr) ->
    mnesia:abort({bad_type, schema});
do_add_table_index(Tab, Attr) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, read),
    Pos = attr_tab_to_pos(Tab, Attr),
    insert_schema_ops(TidTs, make_add_table_index(Tab, Pos)).

make_add_table_index(Tab, Pos) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    Ix = Cs#cstruct.index,
    verify(false, lists:member(Pos, Ix), {already_exists, Tab, Pos}),
    Ix2 = lists:sort([Pos | Ix]),
    Cs2 = Cs#cstruct{index = Ix2},
    check_cstruct(Cs2),
    [{op, add_index, Pos, cs2list(Cs2)}].

del_table_index(Tab, Pos) ->
    schema_transaction(fun() -> do_del_table_index(Tab, Pos) end).

do_del_table_index(schema, Attr) ->
    mnesia:abort({bad_type, schema});
do_del_table_index(Tab, Attr) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, read),
    Pos = attr_tab_to_pos(Tab, Attr),
    insert_schema_ops(TidTs, make_del_table_index(Tab, Pos)).

make_del_table_index(Tab, Pos) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    Ix = Cs#cstruct.index,
    verify(true, lists:member(Pos, Ix), {no_exists, Tab, Pos}),
    Cs2 = Cs#cstruct{index = lists:delete(Pos, Ix)},
    check_cstruct(Cs2),
    [{op, del_index, Pos, cs2list(Cs2)}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

add_snmp(Tab, Ustruct) ->
    schema_transaction(fun() -> do_add_snmp(Tab, Ustruct) end).

do_add_snmp(schema, Ustruct) ->
    mnesia:abort({bad_type, schema});
do_add_snmp(Tab, Ustruct) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, read),
    insert_schema_ops(TidTs, make_add_snmp(Tab, Ustruct)).

make_add_snmp(Tab, Ustruct) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    verify([], Cs#cstruct.snmp, {already_exists, Tab, snmp}),
    Error = {badarg, Tab, snmp, Ustruct},
    verify(true, mnesia_snmp_hook:check_ustruct(Ustruct), Error),
    Cs2 = Cs#cstruct{snmp = Ustruct},
    check_cstruct(Cs2),
    [{op, add_snmp, Ustruct, cs2list(Cs2)}].

del_snmp(Tab) ->
    schema_transaction(fun() -> do_del_snmp(Tab) end).

do_del_snmp(schema) ->
    mnesia:abort({bad_type, schema});
do_del_snmp(Tab) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, read),
    insert_schema_ops(TidTs, make_del_snmp(Tab)).

make_del_snmp(Tab) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    Cs2 = Cs#cstruct{snmp = []},
    check_cstruct(Cs2),
    [{op, del_snmp, cs2list(Cs2)}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  

transform_table(Tab, Fun, NewAttrs) when function(Fun), list(NewAttrs) -> 
    schema_transaction(fun() -> do_transform_table(Tab, Fun, NewAttrs) end);

transform_table(Tab, Fun, NewAttrs) ->
    {aborted,{bad_type, Tab, Fun, NewAttrs}}.

do_transform_table(schema, Fun, NewAttrs) ->
    mnesia:abort({bad_type, schema});
do_transform_table(Tab, Fun, NewAttrs) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, write),
    insert_schema_ops(TidTs, make_transform(Tab, Fun, NewAttrs)).

make_transform(Tab, Fun, NewAttrs) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    ensure_writable(Tab),
    verify([], mnesia_lib:val({Tab, index}), {index_exists, Tab}),
    Cs2 = Cs#cstruct{attributes = NewAttrs},
    check_cstruct(Cs2),
    [{op, transform, Fun, cs2list(Cs2)}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  

change_table_access_mode(Tab, Mode) ->
    schema_transaction(fun() -> do_change_table_access_mode(Tab, Mode) end).

do_change_table_access_mode(Tab, Mode) ->
    {_Mod, Tid, Ts} = get_tid_ts_and_lock(schema, write),
    Store = Ts#tidstore.store,
    mnesia_locker:wlock_no_exist(Tid, Store, schema, val({schema, active_replicas})),
    mnesia_locker:wlock_no_exist(Tid, Store, Tab, val({Tab, active_replicas})),
    do_insert_schema_ops(Store, make_change_table_access_mode(Tab, Mode)).

make_change_table_access_mode(Tab, Mode) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    OldMode = Cs#cstruct.access_mode,
    verify(false, OldMode ==  Mode, {already_exists, Tab, Mode}),
    Cs2 = Cs#cstruct{access_mode = Mode},
    check_cstruct(Cs2),
    [{op, change_table_access_mode, cs2list(Cs2), OldMode, Mode}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

change_table_load_order(Tab, LoadOrder) ->
    schema_transaction(fun() -> do_change_table_load_order(Tab, LoadOrder) end).

do_change_table_load_order(schema, LoadOrder) ->
    mnesia:abort({bad_type, schema});
do_change_table_load_order(Tab, LoadOrder) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, none),
    insert_schema_ops(TidTs, make_change_table_load_order(Tab, LoadOrder)).

make_change_table_load_order(Tab, LoadOrder) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    OldLoadOrder = Cs#cstruct.load_order,
    Cs2 = Cs#cstruct{load_order = LoadOrder},
    check_cstruct(Cs2),
    [{op, change_table_load_order, cs2list(Cs2), OldLoadOrder, LoadOrder}].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_table_property(Tab, Prop) when tuple(Prop), size(Prop) >= 1 ->
    schema_transaction(fun() -> do_write_table_property(Tab, Prop) end);
write_table_property(Tab, Prop) ->
    {aborted, {bad_type, Tab, Prop}}.

do_write_table_property(Tab, Prop) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, none),
    insert_schema_ops(TidTs, make_write_table_properties(Tab, [Prop])).

make_write_table_properties(Tab, Props) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    make_write_table_properties(Tab, Props, Cs).

make_write_table_properties(Tab, [Prop | Props], Cs) ->
    OldProps = Cs#cstruct.user_properties,
    PropKey = element(1, Prop),
    DelProps = lists:keydelete(PropKey, 1, OldProps),
    MergedProps = lists:merge(DelProps, [Prop]),
    Cs2 = Cs#cstruct{user_properties = MergedProps},
    check_cstruct(Cs2),
    [{op, write_property, cs2list(Cs2), Prop} |
     make_write_table_properties(Tab, Props, Cs2)];
make_write_table_properties(Tab, [], Cs) ->
    [].

delete_table_property(Tab, PropKey) ->
    schema_transaction(fun() -> do_delete_table_property(Tab, PropKey) end).

do_delete_table_property(Tab, PropKey) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    get_tid_ts_and_lock(Tab, none),
    insert_schema_ops(TidTs, make_delete_table_properties(Tab, [PropKey])).

make_delete_table_properties(Tab, PropKeys) ->
    ensure_writable(schema),
    Cs = incr_version(val({Tab, cstruct})),
    ensure_active(Cs),
    make_delete_table_properties(Tab, PropKeys, Cs).

make_delete_table_properties(Tab, [PropKey | PropKeys], Cs) ->
    OldProps = Cs#cstruct.user_properties,
    Props = lists:keydelete(PropKey, 1, OldProps),
    Cs2 = Cs#cstruct{user_properties = Props},
    check_cstruct(Cs2),
    [{op, delete_property, cs2list(Cs2), PropKey} |
     make_delete_table_properties(Tab, PropKeys, Cs2)];
make_delete_table_properties(Tab, [], Cs) ->
    [].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Ensure that the transaction can be committed even 
%% if the node crashes and Mnesia is restarted
prepare_commit(Tid, Commit) when Commit#commit.schema_ops /= [] ->
    {Modified, Ops, UseDumper} = prepare_ops(Tid, Commit#commit.schema_ops),
    Begin = schema_begin,
    if
	Ops == [] ->
	    dbg_out("Transaction log dump skipped: ~w~n", [Begin]),
	    {Modified, Commit#commit{schema_ops = Ops}, UseDumper};
	UseDumper == use_dumper ->
	    case mnesia_dumper:dump_log(Begin) of
		dumped -> 
		    {Modified, Commit#commit{schema_ops = Ops}, UseDumper};
		{error, Reason} ->
		    mnesia:abort(Reason)
	    end;
	true ->
	    {Modified, Commit#commit{schema_ops = Ops}, UseDumper}
    end;
prepare_commit(_Tid, Commit) ->
    {false, Commit, ignore_dumper}.

prepare_ops(Tid, [Op | Ops]) ->
    {Modified2, Acc2, Dump} = prepare_ops(Tid, Ops),
    case prepare_op(Tid, Op) of
        {true, use_dumper} -> 
	    {Modified2, [Op | Acc2], use_dumper};
        {true, ignore_dumper} -> 
	    {Modified2, [Op | Acc2], Dump};
        {true, Ops2, use_dumper} -> 
	    {true, Ops2 ++ Acc2, use_dumper};
        {true, Ops2, ignore_dumper} -> 
	    {true, Ops2 ++ Acc2, Dump};
	{false, use_dumper} -> 
	    {true, Acc2, use_dumper};
	{false, ignore_dumper} -> 
	    {true, Acc2, Dump}
    end;
prepare_ops(_Tid, []) ->
    {false, [], ignore_dumper}.
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Prepare for commit
%% returns true if Op should be included, i.e. unmodified
%%         {true, Operation} if NewRecs should be included, i.e. modified
%%         false if Op should NOT be included, i.e. modified
%%
prepare_op(_Tid, {op, rec, unknown, Rec}) ->
    {{Tab, Key}, Items, Op} = Rec,
    case val({Tab, storage_type}) of
        unknown ->
            {false, ignore_dumper};
        Storage ->
            mnesia_tm:prepare_snmp(Tab, Key, Items), % May exit
            {true, [{op, rec, Storage, Rec}], ignore_dumper}
    end;

prepare_op(_Tid, {op, announce_im_running, Node, SchemaDef, Running, RemoteRunning}) ->
    SchemaCs = list2cs(SchemaDef),
    case lists:member(node(), Running) of
        true ->
            announce_im_running(RemoteRunning -- Running, SchemaCs);
        false ->
            announce_im_running(Running -- RemoteRunning, SchemaCs)
    end,
    {false, ignore_dumper};

prepare_op(Tid, {op, create_table, TabDef}) ->
    Cs = list2cs(TabDef),
    Storage = mnesia_lib:cs_to_storage_type(node(), Cs),
    UseDir = mnesia_monitor:use_dir(),
    Reason = {bad_type, Cs#cstruct.name, Storage, node()},
    case Storage of
        disc_copies when UseDir == false ->
            mnesia:abort(Reason);
        disc_only_copies when UseDir == false ->
            mnesia:abort(Reason);
        _ ->
	    insert_cstruct(Tid, Cs, false),
            {true, ignore_dumper}
    end;
                                                
prepare_op(Tid, {op, add_table_copy, Storage, Node, TabDef}) ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    if
	Tab == schema ->
	    {true, ignore_dumper}; % Nothing to prepare
	Node == node() ->
	    insert_cstruct(Tid, Cs, true),
	    case mnesia_lib:val({schema, storage_type}) of 
		ram_copies when Storage /= ram_copies -> 
		    Error = {"has no disc", Node},
		    mnesia:abort(Error);
		_  -> ok
	    end,   
	    case mnesia_controller:get_network_copy(Tab, Cs) of
		yes ->
		    {true, ignore_dumper};
		no ->
		    Reason = {"Cannot copy table", Tab, Node},
		    mnesia:abort(Reason)
	    end;
	Node /= node() ->
	    insert_cstruct(Tid, Cs, true),
	    {prepare_verify_dump(Storage, Cs), ignore_dumper}
    end;
            
prepare_op(_Tid, {op, change_table_copy_type,  N, From, To, TabDef}) when N == node() ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    
    case {From, To} of
        {ram_copies, disc_copies} when Tab == schema ->
            verify(opt_disc, fun() -> mnesia:system_info(schema_location) end,
                   {"schema_location must be opt_disc", N}),
	    Dir = mnesia_lib:dir(),    
            case opt_create_dir(true, Dir) of
                ok ->
		    purge_dir(Dir, []),
		    mnesia_log:purge_all_logs(),
                    mnesia_log:init(),
                    set(use_dir, true),
                    Ns = mnesia_lib:running_nodes(),
                    F = fun(U) -> mnesia_recover:log_mnesia_up(U) end,
                    lists:foreach(F, Ns),
                    
                    prepare_change_table_copy_type(Tab, From, To);
                {error, Reason} ->
                    mnesia:abort(Reason)
            end;
        {disc_copies, ram_copies} when Tab == schema ->
            Ltabs = val({schema, local_tables}) -- [schema],
            Dtabs = [L || L <- Ltabs,
                          val({L, storage_type}) /= ram_copies],
            verify([], Dtabs, {"Disc resident tables", Dtabs, N});
        {ram_copies, disc_copies} when Tab /= schema -> 
            prepare_change_table_copy_type(Tab, From, To);
        {ram_copies, disc_only_copies} ->       
            prepare_change_table_copy_type(Tab, From, To);
        _ ->
            ignore
    end,
    {true, use_dumper};

prepare_op(_Tid, {op, change_table_copy_type,  N, From, To, TabDef}) when N /= node() ->
    Cs = list2cs(TabDef),
    {prepare_verify_dump(To, Cs), use_dumper};

prepare_op(_Tid, {op, dump_table, unknown, TabDef}) ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    case lists:member(node(), Cs#cstruct.ram_copies) of
        true ->
            mnesia_dumper:raw_named_dump_table(Tab, dmp),
	    Size = mnesia:table_info(Tab, size),
	    {true, [{op, dump_table, Size, TabDef}], ignore_dumper};
        false ->
            {true, ignore_dumper}
    end;

prepare_op(_Tid, {op, add_snmp, Ustruct, TabDef}) ->
    Cs = list2cs(TabDef),
    case mnesia_lib:cs_to_storage_type(node(), Cs) of
        unknown ->
            {true, ignore_dumper};
        Storage ->
            Tab = Cs#cstruct.name,
            Stab = mnesia_snmp_hook:create_table(Ustruct, Tab, Storage),
            mnesia_lib:set({Tab, {index, snmp}}, Stab),
            {true, ignore_dumper}
    end;

prepare_op(_Tid, {op, transform, Fun, TabDef}) ->
    Cs = list2cs(TabDef),
    case mnesia_lib:cs_to_storage_type(node(), Cs) of
        unknown ->
            {true, use_dumper};
        Storage ->
            Tab = Cs#cstruct.name,
            RecName = Cs#cstruct.record_name,
            NewArity = length(Cs#cstruct.attributes) + 1,
            Key = mnesia_lib:db_first(Tab),
	    Tail = [{op, transform, Fun, TabDef}],
            case catch transform_objs(Fun, Tab, RecName, Key, NewArity, Storage, Tail) of
                {'EXIT', Reason} ->
                    exit({"Bad transform function", Tab, Fun, Reason});
                Objs ->
                    {true, Objs, use_dumper}
            end
    end;

prepare_op(_Tid, Op) ->
    {true, ignore_dumper}.

prepare_verify_dump(Storage, Cs) ->
    if
        Storage /= disc_copies, Storage /= disc_only_copies ->
            true;
        true ->
            UseDir = mnesia_monitor:use_dir(),
            case mnesia_lib:cs_to_storage_type(node(), Cs) of
                ram_copies when UseDir == true ->
                    Tab = Cs#cstruct.name,
                    case mnesia_lib:exists(mnesia_lib:tab2dat(Tab)) of
                        false ->
                            true;
                        true ->
                            Reason = {combine_error, Tab, node(),
                                      "Table dump exists"},
                            mnesia:abort(Reason)
                    end;
                _ ->
                    true
            end
    end.

prepare_change_table_copy_type(Tab, FromS, ToS) ->
    mnesia_dumper:raw_named_dump_table(Tab, dmp),
    mnesia_checkpoint:tm_change_table_copy_type(Tab, FromS, ToS).

%% Build the list in reverse order
transform_objs(Fun, _Tab, _RT, '$end_of_table', _NewArity, _Storage, Tail) ->
    Tail;
transform_objs(Fun, Tab, RecName, Key, A, Storage, Tail) ->
    Objs = mnesia_lib:db_get(Tab, Key),
    NewObjs = [transform_obj(Tab, RecName, Key, Fun, Obj, A) || Obj <- Objs],
    NextKey = mnesia_lib:db_next_key(Tab, Key),
    Oid = {Tab, Key},
    [{op, rec, Storage, {Oid, NewObjs, write}},
     {op, rec, Storage, {Oid, [Oid], delete}} |
     transform_objs(Fun, Tab, RecName, NextKey, A, Storage, Tail)].

transform_obj(Tab, RecName, Key, Fun, Obj, NewArity) ->
    NewObj = Fun(Obj),
    if
        size(NewObj) /= NewArity ->
            exit({"Bad arity", Obj, NewObj});
        RecName == element(1, NewObj), Key == element(2, NewObj) ->
            NewObj;
        true ->
            exit({"Bad key", Obj, NewObj})
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Undo prepare of commit
undo_prepare_commit(Tid, Commit) when Commit#commit.schema_ops /= [] ->
    undo_prepare_ops(Tid, Commit#commit.schema_ops),
    Commit;
undo_prepare_commit(_Tid, Commit) ->
    Commit.

%% Undo in reverse order
undo_prepare_ops(Tid, [Op | Ops]) ->
    case element(1, Op) of	
	TheOp when TheOp /= op, TheOp /= restore_op ->
	    undo_prepare_ops(Tid, Ops);
	_ ->
	    undo_prepare_ops(Tid, Ops),
	    undo_prepare_op(Tid, Op)
    end;
undo_prepare_ops(_Tid, []) ->
    [].

undo_prepare_op(_Tid, {op, announce_im_running, Node, _, Running, RemoteRunning}) ->
    case lists:member(node(), Running) of
        true ->
            unannounce_im_running(RemoteRunning -- Running);
        false ->
            unannounce_im_running(Running -- RemoteRunning)
    end;

undo_prepare_op(Tid, {op, create_table, TabDef}) ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    mnesia_lib:unset({Tab, create_table}),
    delete_cstruct(Tid, Cs);

undo_prepare_op(Tid, {op, add_table_copy, Storage, Node, TabDef}) ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    if
	Tab == schema ->
	    true; % Nothing to prepare
	Node == node() ->
	    mnesia_checkpoint:tm_del_copy(Tab, Node),
	    mnesia_controller:unannounce_add_table_copy(Tab, Node),
	    mnesia_monitor:close_dets(Tab),
	    mnesia_index:del_ix_files(Tab, Storage),
	    file:delete(mnesia_lib:tab2dat(Tab)),
	    ram_delete_table(Tab, Storage),
	    Cs2 = new_cs(Cs, Node, Storage, del),
	    insert_cstruct(Tid, Cs2, true); % Don't care about the version
	Node /= node() ->
	    mnesia_controller:unannounce_add_table_copy(Tab, Node),
	    Cs2 = new_cs(Cs, Node, Storage, del),
	    insert_cstruct(Tid, Cs2, true) % Don't care about the version
    end;
    
undo_prepare_op(_Tid, {op, change_table_copy_type, N, FromS, ToS, TabDef}) 
        when N == node() ->
    Cs = list2cs(TabDef),
    Tab = Cs#cstruct.name,
    mnesia_checkpoint:tm_change_table_copy_type(Tab, ToS, FromS),
    Dmp = mnesia_lib:tab2dmp(Tab),
    
    case {FromS, ToS} of
        {ram_copies, disc_copies} when Tab == schema ->
            file:delete(Dmp),
            mnesia_log:purge_some_logs(),
	    set(use_dir, false);
	{ram_copies, disc_copies} ->
	    file:delete(Dmp);
	{ram_copies, disc_only_copies} ->
	    file:delete(Dmp);
	_ ->
	    ignore
    end;

undo_prepare_op(_Tid, {op, dump_table, Size, TabDef}) ->
    Cs = list2cs(TabDef),
    case lists:member(node(), Cs#cstruct.ram_copies) of
	true ->
	    Tab = Cs#cstruct.name,
	    Dmp = mnesia_lib:tab2dmp(Tab),
	    file:delete(Dmp);
	false ->
	    ignore
    end;

undo_prepare_op(_Tid, {op, add_snmp, Ustruct, TabDef}) ->
    Cs = list2cs(TabDef),
    case mnesia_lib:cs_to_storage_type(node(), Cs) of
	unknown ->
	    true;
	Storage ->
	    Tab = Cs#cstruct.name,
	    case ?catch_val({Tab, {index, snmp}}) of
		{'EXIT',_} ->
		    ignore;
		Stab ->
		    mnesia_snmp_hook:delete_table(Tab, Stab),
		    mnesia_lib:unset({Tab, {index, snmp}})
	    end
    end;

undo_prepare_op(_Tid, _Op) ->
    ignore.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% delete possible index files .....
ram_delete_table(Tab, Storage) ->
    case Storage of
	unknown ->
	    ignore;
	ram_copies ->
	    PosList = val({Tab, index}),
	    ram_delete_table(Tab, Storage, PosList);
	disc_copies ->
	    PosList = val({Tab, index}),
	    ram_delete_table(Tab, Storage, PosList);
	disc_only_copies ->
	    ignore
    end.

ram_delete_table(Tab, Storage, []) ->
    case ?catch_val({Tab, {index, snmp}}) of
	{'EXIT', _} ->
	    ignore;
	Etab ->
	    catch mnesia_snmp_hook:delete_table(Tab, Etab)
    end,
    catch erlang:db_erase(Tab);
ram_delete_table(Tab, Storage, [Pos | Tail]) when integer(Pos) ->
    Etab = val({Tab, {index, Pos}}),
    IxTag = mnesia_index:get_index_table(Tab, Storage, Pos),
    catch erlang:db_erase(Etab),
    mnesia_lib:unset({Tab, {index, Pos}}),    
    mnesia_lib:del({Tab, index_tables}, {Pos, IxTag}),
    ram_delete_table(Tab, Storage, Tail).
    
purge_dir(Dir, KeepFiles) ->
    Suffixes = known_suffixes(),
    purge_dir(Dir, KeepFiles, Suffixes).

purge_dir(Dir, KeepFiles, Suffixes) ->
    case dir_exists(Dir) of
	true ->
	    {ok, AllFiles} = file:list_dir(Dir),
	    purge_known_files(AllFiles, KeepFiles, Dir, Suffixes);
	false ->
	    ok
    end.

purge_tmp_files() ->
    case mnesia_monitor:use_dir() of
	true ->
	    Dir = mnesia_lib:dir(),
	    KeepFiles = [],
	    Exists = mnesia_lib:exists(mnesia_lib:tab2dat(schema)),
	    case Exists of
		true ->
		    Suffixes = tmp_suffixes(),
		    purge_dir(Dir, KeepFiles, Suffixes);
		false ->
		    %% Interrupted change of storage type
		    %% for schema table
		    Suffixes = known_suffixes(),
		    purge_dir(Dir, KeepFiles, Suffixes),
		    mnesia_lib:set(use_dir, false)
		end;

	false ->
	    ok
    end.

purge_known_files([File | Tail], KeepFiles, Dir, Suffixes) ->
    case lists:member(File, KeepFiles) of
	true ->
	    ignore;
	false ->
	    case has_known_suffix(File, Suffixes, false) of
		false ->
		    ignore;
		true ->
		    AbsFile = filename:join([Dir, File]),
		    file:delete(AbsFile)
	    end
    end,
    purge_known_files(Tail, KeepFiles, Dir, Suffixes);
purge_known_files([], KeepFiles, Dir, Suffixes) ->
    ok.

has_known_suffix(File, Suffixes, true) ->
    true;
has_known_suffix(File, [Suffix | Tail], false) ->
    has_known_suffix(File, Tail, lists:suffix(Suffix, File));
has_known_suffix(File, [], Bool) ->
    Bool.
	    
known_suffixes() -> real_suffixes() ++ tmp_suffixes().

real_suffixes() ->  [".DAT", ".LOG", ".BUP"].

tmp_suffixes() -> [".TMP", ".BUPTMP", ".RET", ".DMP"].

info() ->
    io:format("---- Table properties ------ ~n",[]),
    lists:foreach(fun(T) -> info(T) end, lists:sort(val({schema, tables}))), 
    ok.

info(Tab) ->
    io:format("-- table ~-15w properties --- ~n",[Tab]),
    Props = get_table_properties(Tab),
    info2(Tab, Props).
info2(Tab, [{P, V} | Tail]) ->
    io:format("~-20w -> ~p~n",[P,V]),
    info2(Tab, Tail);
info2(_, []) -> 
    io:format("~n", []).

get_table_properties(Tab) ->
    RawGvar =
	mnesia_lib:erlang_db_match_object(mnesia_gvar, {{Tab, '_'}, '_'}),
    case [{Item, Val} || {{_Tab, Item}, Val} <- RawGvar] of
	[] ->
	    [];
	Gvar ->
	    Size = {size, mnesia:table_info(Tab, size)},
	    Memory = {memory, mnesia:table_info(Tab, memory)},
	    Master = {master_nodes, mnesia:table_info(Tab, master_nodes)},
	    lists:sort([Size, Memory, Master | Gvar])
    end.

%%%%%%%%%%% RESTORE %%%%%%%%%%%

-record(r, {iter = schema,
	    module,
	    table_options = [],
	    default_op = clear_tables,
	    tables = [],
	    opaque,
	    insert_op = error_fun,
	    recs = error_recs
	   }).

restore(Opaque) -> 
    restore(Opaque, [], mnesia_monitor:get_env(backup_module)).
restore(Opaque, Args) when list(Args) -> 
    restore(Opaque, Args, mnesia_monitor:get_env(backup_module));
restore(_Opaque, BadArg) ->
    {aborted, {badarg, BadArg}}.
restore(Opaque, Args, Module) when list(Args), atom(Module) ->
    InitR = #r{opaque = Opaque, module = Module},
    case catch lists:foldl(fun check_restore_arg/2, InitR, Args) of
	R when record(R, r) ->
	    case mnesia_bup:read_schema(Module, Opaque) of
		{error, Reason} ->
		    {aborted, Reason};
		BupSchema -> 
		    schema_transaction(fun() -> do_restore(R, BupSchema) end)
	    end;
	{'EXIT', Reason} ->
	    {aborted, Reason}
    end;
restore(_Opaque, Args, Module) ->
    {aborted, {badarg, Args, Module}}.

check_restore_arg({module, Mod}, R) when atom(Mod) ->
    R#r{module = Mod};

check_restore_arg({clear_tables, List}, R) when list(List) ->
    case lists:member(schema, List) of
	false ->
	    TableList = [{Tab, clear_tables} || Tab <- List],
	    R#r{table_options = R#r.table_options ++ TableList};
	true ->
	    exit({badarg, {clear_tables, schema}})
    end;
check_restore_arg({recreate_tables, List}, R) when list(List) ->
    case lists:member(schema, List) of
	false ->
	    TableList = [{Tab, recreate_tables} || Tab <- List],
	    R#r{table_options = R#r.table_options ++ TableList};
	true ->
	    exit({badarg, {recreate_tables, schema}})
    end;
check_restore_arg({keep_tables, List}, R) when list(List) ->
    TableList = [{Tab, keep_tables} || Tab <- List],
    R#r{table_options = R#r.table_options ++ TableList};
check_restore_arg({skip_tables, List}, R) when list(List) ->
    TableList = [{Tab, skip_tables} || Tab <- List],
    R#r{table_options = R#r.table_options ++ TableList};
check_restore_arg({default_op, Op}, R) -> 
    case Op of 
	clear_tables -> ok;
	recreate_tables -> ok;
	keep_tables -> ok;
	skip_tables -> ok;
	Else ->
	    exit({badarg, {bad_default_op, Else}})
    end,
    R#r{default_op = Op};

check_restore_arg(BadArg,_) ->
    exit({badarg, BadArg}).

do_restore(R, BupSchema) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    R2 = restore_schema(BupSchema, R),
    insert_schema_ops(TidTs, [{restore_op, R2}]),
    [element(1, TabStruct) || TabStruct <- R2#r.tables].

arrange_restore(R, Fun, Recs) ->
    R2 = R#r{insert_op = Fun, recs = Recs},
    case mnesia_bup:iterate(R#r.module, fun restore_items/4, R#r.opaque, R2) of
	{ok, R3} -> R3#r.recs;
	{error, Reason} -> mnesia:abort(Reason);
	Reason -> mnesia:abort(Reason)
    end.

restore_items([Rec | Recs], Header, Schema, R) ->
    Tab = element(1, Rec),
    case lists:keysearch(Tab, 1, R#r.tables) of
	{value, {Tab, Where, Snmp, RecName}} ->
	    {Rest, NRecs} =
		restore_tab_items([Rec | Recs], Tab, RecName, Where, Snmp,
				  R#r.recs, R#r.insert_op),
	    restore_items(Rest, Header, Schema, R#r{recs = NRecs});
	false ->
	    Rest = skip_tab_items(Recs, Tab),
	    restore_items(Rest, Header, Schema, R)
    end;
    
restore_items([], Header, Schema, R) ->
    R.

restore_func(Tab, R) ->
    case lists:keysearch(Tab, 1, R#r.table_options) of 
	{value, {Tab, OP}} ->
	    OP;
	false ->
	    R#r.default_op
    end.

where_to_commit(CsList) ->
    Ram =   [{N, ram_copies} || N <- pick(ram_copies, CsList, [])],
    Disc =  [{N, disc_copies} || N <- pick(disc_copies, CsList, [])],
    DiscO = [{N, disc_only_copies} || N <- pick(disc_only_copies, CsList, [])],
    Ram ++ Disc ++ DiscO.
    
%% Changes of the Meta info of schema itself is not allowed
restore_schema([{schema, schema, List} | Schema], R) ->
    restore_schema(Schema, R);
restore_schema([{schema, Tab, List} | Schema], R) ->
    case restore_func(Tab, R) of
	clear_tables -> 
	    do_clear_table(Tab),
	    Where = val({Tab, where_to_commit}),
	    Snmp = val({Tab, snmp}), 
	    RecName = val({Tab, record_name}), 
	    R2 = R#r{tables = [{Tab, Where, Snmp, RecName} | R#r.tables]},
	    restore_schema(Schema, R2);
	recreate_tables -> 
	    NC = {cookie, {erlang:now(), node()}},
	    List2 = lists:keyreplace(cookie, 1, List, NC),		
	    Where = where_to_commit(List2),
	    Snmp = pick(snmp, List2, []),
	    RecName = pick(record_name, List2, Tab),
	    case ?catch_val({Tab, cstruct}) of
		OldCs when record(OldCs, cstruct) -> 
		    do_delete_table(Tab),
		    do_create_table(List2, []);
		{'EXIT', _} -> 
		    do_create_table(List2)
	    end,
	    R2 = R#r{tables = [{Tab, Where, Snmp, RecName} | R#r.tables]},
	    restore_schema(Schema, R2);
	keep_tables -> 
	    get_tid_ts_and_lock(Tab, write),
	    Where = val({Tab, where_to_commit}),
	    Snmp = val({Tab, snmp}),
	    RecName = val({Tab, record_name}), 
	    R2 = R#r{tables = [{Tab, Where, Snmp, RecName} | R#r.tables]},
	    restore_schema(Schema, R2);
	skip_tables -> 
	    restore_schema(Schema, R)
    end;

restore_schema([{schema, Tab} | Schema], R) ->
    do_delete_table(Tab),
    Tabs = lists:delete(Tab,R#r.tables),
    restore_schema(Schema, R#r{tables = Tabs});
restore_schema([], R) ->
    R.

restore_tab_items([Rec | Rest], Tab, RecName, Where, Snmp, Recs, Op) 
  when element(1, Rec) == Tab ->
    NewRecs = Op(Rec, Recs, RecName, Where, Snmp),
    restore_tab_items(Rest, Tab, RecName, Where, Snmp, NewRecs, Op);

restore_tab_items(Rest, _Tab, _RecName, _Where, Snmp, Recs, _Op) ->
    {Rest, Recs}.

skip_tab_items([Rec| Rest], Tab) 
  when element(1, Rec) == Tab ->
    skip_tab_items(Rest, Tab);
skip_tab_items(Recs, _) ->
    Recs.

%%%%%%%%% Dump tables %%%%%%%%%%%%%
dump_tables(Tabs) when list(Tabs) ->
    schema_transaction(fun() -> do_dump_tables(Tabs) end);
dump_tables(Tabs) ->
    {aborted, {bad_type, Tabs}}.

do_dump_tables(Tabs) ->
    TidTs = get_tid_ts_and_lock(schema, write),
    insert_schema_ops(TidTs, make_dump_tables(Tabs)).

make_dump_tables([schema | Tabs]) ->
    mnesia:abort({bad_type, schema});
make_dump_tables([Tab | Tabs]) ->
    get_tid_ts_and_lock(Tab, read),
    TabDef = get_create_list(Tab),
    DiscResident =  val({Tab, disc_copies}) ++ val({Tab, disc_only_copies}),
    verify([], DiscResident,
	   {"Only allowed on ram_copies", Tab, DiscResident}),
    [{op, dump_table, unknown, TabDef} | make_dump_tables(Tabs)];
make_dump_tables([]) ->
    [].

%% Merge the local schema with the schema on other nodes
merge_schema() ->
    schema_transaction(fun() -> do_merge_schema() end).

do_merge_schema() ->
    {_Mod, Tid, Ts} = get_tid_ts_and_lock(schema, write),
    Connected = val(recover_nodes),
    Running = val({current, db_nodes}),
    Store = Ts#tidstore.store,
    case Connected -- Running of
	[Node | _] ->
	    %% Time for a schema merging party!
	    mnesia_locker:wlock_no_exist(Tid, Store, schema, [Node]),
	    
	    case rpc:call(Node, mnesia_controller, get_cstructs, []) of
		{cstructs, Cstructs, RemoteRunning} ->
		    LockedAlready = Running ++ [Node],
		    NeedsLock = RemoteRunning -- LockedAlready,
		    mnesia_locker:wlock_no_exist(Tid, Store, schema, NeedsLock),

		    {value, SchemaCs} =
			lists:keysearch(schema, #cstruct.name, Cstructs),

		    %% Announce that Node is running
		    A = [{op, announce_im_running, node(),
			  cs2list(SchemaCs), Running, RemoteRunning}],
		    do_insert_schema_ops(Store, A),

		    %% Introduce remote tables to local node
		    do_insert_schema_ops(Store, make_merge_schema(Node, Cstructs)),
		    
		    %% Introduce local tables to remote nodes
		    Tabs = val({schema, tables}),
		    Ops = [{op, merge_schema, get_create_list(T)}
			   || T <- Tabs,
			      not lists:keymember(T, #cstruct.name, Cstructs)],
		    do_insert_schema_ops(Store, Ops),
		    
		    %% Ensure that the txn will be committed on all nodes
		    announce_im_running(RemoteRunning, SchemaCs), 
		    {merged, Running, RemoteRunning};
		{error, Reason} ->
		    {"Cannot get cstructs", Node, Reason};
		{badrpc, Reason} ->
		    {"Cannot get cstructs", Node, {badrpc, Reason}}
	    end;
	[] ->
	    %% No more nodes to merge schema with
	    not_merged
    end.
		    
make_merge_schema(Node, [Cs | Cstructs]) ->
    Ops = do_make_merge_schema(Node, Cs),
    Ops ++ make_merge_schema(Node, Cstructs);
make_merge_schema(Node, []) ->
    [].

%% Merge definitions of schema table
do_make_merge_schema(Node, RemoteCs)
        when RemoteCs#cstruct.name == schema ->
    Cs = val({schema, cstruct}),
    Masters = mnesia_recover:get_master_nodes(schema),
    HasRemoteMaster = lists:member(Node, Masters),
    HasLocalMaster = lists:member(node(), Masters),
    Force = HasLocalMaster or HasRemoteMaster,
    if
	Cs#cstruct.cookie == RemoteCs#cstruct.cookie,
	Cs#cstruct.version == RemoteCs#cstruct.version ->
	    %% Great, we have the same cookie and version
	    %% and do not need to merge cstructs
	    [];
	
	Cs#cstruct.cookie /= RemoteCs#cstruct.cookie,
	Cs#cstruct.disc_copies /= [],
	RemoteCs#cstruct.disc_copies /= [] ->
	    %% Both cstructs involves disc nodes
	    %% and we cannot merge them
	    if
		HasLocalMaster == true,
		HasRemoteMaster == false ->
		    %% Choose local cstruct,
		    %% since it's the master
		    [{op, merge_schema, cs2list(Cs)}];

		HasRemoteMaster == true,
		HasLocalMaster == false ->
		    %% Choose remote cstruct,
		    %% since it's the master
		    [{op, merge_schema, cs2list(RemoteCs)}];
		
		true ->
		    fatal("Incompatible schema cookies. "
			  "Please, restart from old backup."
			  "~w = ~w, ~w = ~w~n",
			  [Node, cs2list(RemoteCs), node(), cs2list(Cs)])
		end;

	Cs#cstruct.disc_copies /= [] ->
	    %% Choose local cstruct,
	    %% since it involves disc nodes
	    MergedCs = merge_cstructs(Cs, RemoteCs, Force),
	    [{op, merge_schema, cs2list(MergedCs)}];
	
	RemoteCs#cstruct.disc_copies /= [] ->
	    %% Choose remote cstruct,
	    %% since it involves disc nodes
	    MergedCs = merge_cstructs(RemoteCs, Cs, Force),
	    [{op, merge_schema, cs2list(MergedCs)}];

	Cs > RemoteCs ->
	    %% Choose remote cstruct
	    MergedCs = merge_cstructs(RemoteCs, Cs, Force),
	    [{op, merge_schema, cs2list(MergedCs)}];
	
	true ->
	    %% Choose local cstruct
	    MergedCs = merge_cstructs(Cs, RemoteCs, Force),
	    [{op, merge_schema, cs2list(MergedCs)}]
    end;

%% Merge definitions of normal table
do_make_merge_schema(Node, RemoteCs) ->
    Tab = RemoteCs#cstruct.name,
    Masters = mnesia_recover:get_master_nodes(schema),
    HasRemoteMaster = lists:member(Node, Masters),
    HasLocalMaster = lists:member(node(), Masters),
    Force = HasLocalMaster or HasRemoteMaster,
    case ?catch_val({Tab, cstruct}) of
	{'EXIT', _} ->
	    %% A completely new table, created while Node was down
	    [{op, merge_schema, cs2list(RemoteCs)}];
	Cs when Cs#cstruct.cookie == RemoteCs#cstruct.cookie ->
	    if
		Cs#cstruct.version == RemoteCs#cstruct.version ->
		    %% We have exactly the same version of the
		    %% table def
		    [];
		
		Cs#cstruct.version > RemoteCs#cstruct.version ->
		    %% Oops, we have different versions
		    %% of the table def, lets merge them.
		    %% The only changes that may have occurred
		    %% is that new replicas may have been added.
		    MergedCs = merge_cstructs(Cs, RemoteCs, Force),
		    [{op, merge_schema, cs2list(MergedCs)}];
		
		Cs#cstruct.version < RemoteCs#cstruct.version ->
		    %% Oops, we have different versions
		    %% of the table def, lets merge them
		    MergedCs = merge_cstructs(RemoteCs, Cs, Force),
		    [{op, merge_schema, cs2list(MergedCs)}]
	    end;
	Cs ->
	    %% Different cookies, not possible to merge
	    if
		HasLocalMaster == true,
		HasRemoteMaster == false ->
		    %% Choose local cstruct,
		    %% since it's the master
		    [{op, merge_schema, cs2list(Cs)}];

		HasRemoteMaster == true,
		HasLocalMaster == false ->
		    %% Choose remote cstruct,
		    %% since it's the master
		    [{op, merge_schema, cs2list(RemoteCs)}];
		
		true ->
		    fatal("Bad cookie in table definition w: ~w = ~w, ~w = ~w~n",
			  [Tab, node(), Cs, Node, RemoteCs])
	    end
    end.

%% Change of table definitions (cstructs) requires all replicas
%% of the table to be active. New replicas, db_nodes and tables
%% may however be added even if some replica is inactive. These
%% invariants must be enforced in order to allow merge of cstructs.
%%
%% Returns a new cstruct or issues a fatal error
merge_cstructs(Cs, RemoteCs, Force) ->
    check_cstruct(Cs),
    case catch do_merge_cstructs(Cs, RemoteCs, Force) of
	{'EXIT', {aborted, Reason}} when Force == true ->
	    Cs;
	{'EXIT', Reason} ->
	    exit(Reason);
	MergedCs when record(MergedCs, cstruct) ->
	    MergedCs;
	Other ->
	    throw(Other)
    end.

do_merge_cstructs(Cs, RemoteCs, Force) ->
    check_cstruct(RemoteCs),
    Ns = mnesia_lib:uniq(mnesia_lib:cs_to_nodes(Cs) ++
			 mnesia_lib:cs_to_nodes(RemoteCs)),
    {AnythingNew, MergedCs} =
	merge_storage_type(Ns, false, Cs, RemoteCs, Force),
    MergedCs2 = merge_versions(AnythingNew, MergedCs, RemoteCs, Force),
    check_cstruct(MergedCs2),
    MergedCs2.

merge_storage_type([N | Ns], AnythingNew, Cs, RemoteCs, Force) ->
    Local = mnesia_lib:cs_to_storage_type(N, Cs),
    Remote = mnesia_lib:cs_to_storage_type(N, RemoteCs),
    case compare_storage_type(true, Local, Remote) of
	{same, Storage} ->
	    merge_storage_type(Ns, AnythingNew, Cs, RemoteCs, Force);
	{diff, Storage} ->
	    Cs2 = change_storage_type(N, Storage, Cs),
	    merge_storage_type(Ns, true, Cs2, RemoteCs, Force);
	incompatible when Force == true ->
	    merge_storage_type(Ns, AnythingNew, Cs, RemoteCs, Force);
	Other ->
	    fatal("Cannot merge storage type for node ~w "
		  "in cstruct ~w with remote cstruct ~w (~w)~n",
		  [N, Cs, RemoteCs, Other])
    end;
merge_storage_type([], AnythingNew, MergedCs, _RemoteCs, _Force) ->
    {AnythingNew, MergedCs}.

compare_storage_type(_Retry, Any, Any) ->
    {same, Any};
compare_storage_type(_Retry, unknown, Any) ->
    {diff, Any};
compare_storage_type(_Retry, ram_copies, disc_copies) ->
    {diff, disc_copies};
compare_storage_type(_Retry, disc_copies, disc_only_copies) ->
    {diff, disc_only_copies};
compare_storage_type(true, One, Another) ->
    compare_storage_type(false, Another, One);
compare_storage_type(false, _One, _Another) ->
    incompatible.
  
change_storage_type(N, ram_copies, Cs) ->
    Nodes = [N | Cs#cstruct.ram_copies],
    Cs#cstruct{ram_copies = mnesia_lib:uniq(Nodes)};
change_storage_type(N, disc_copies, Cs) ->
    Nodes = [N | Cs#cstruct.disc_copies],
    Cs#cstruct{disc_copies = mnesia_lib:uniq(Nodes)};
change_storage_type(N, disc_only_copies, Cs) ->
    Nodes = [N | Cs#cstruct.disc_only_copies],
    Cs#cstruct{disc_only_copies = mnesia_lib:uniq(Nodes)}.

merge_versions(AnythingNew, Cs, RemoteCs, Force) ->
    if
	Cs#cstruct.name == schema ->
	    ok;
	Cs#cstruct.name /= schema,
	Cs#cstruct.cookie == RemoteCs#cstruct.cookie ->
	    ok;
	Force == true ->
	    ok;
	true ->
	    fatal("Bad cookies. Cannot merge definitions of "
		  "table ~w. Local = ~w, Remote = ~w~n",
		  [Cs#cstruct.name, Cs, RemoteCs])
    end,
    if
	Cs#cstruct.name == RemoteCs#cstruct.name,
	Cs#cstruct.type == RemoteCs#cstruct.type,
	Cs#cstruct.local_content == RemoteCs#cstruct.local_content,
	Cs#cstruct.attributes == RemoteCs#cstruct.attributes,
	Cs#cstruct.index == RemoteCs#cstruct.index,
	Cs#cstruct.snmp == RemoteCs#cstruct.snmp,
	Cs#cstruct.access_mode == RemoteCs#cstruct.access_mode,
	Cs#cstruct.load_order == RemoteCs#cstruct.load_order,
	Cs#cstruct.user_properties == RemoteCs#cstruct.user_properties ->
	    do_merge_versions(AnythingNew, Cs, RemoteCs);
	Force == true ->
	    do_merge_versions(AnythingNew, Cs, RemoteCs);
	true ->
	    fatal("Cannot merge definitions of "
		  "table ~w. Local = ~w, Remote = ~w~n",
		  [Cs#cstruct.name, Cs, RemoteCs])
    end.

do_merge_versions(AnythingNew, MergedCs, RemoteCs) ->
    {{Major1, Minor1}, Detail1} = MergedCs#cstruct.version,
    {{Major2, Minor2}, Detail2} = RemoteCs#cstruct.version,
    if
	MergedCs#cstruct.version == RemoteCs#cstruct.version ->
	    MergedCs;
	AnythingNew == false ->
	    MergedCs;
	Major1 == Major2 ->
	    Minor = lists:max([Minor1, Minor2]),
	    V = {{Major1, Minor}, dummy},
	    incr_version(MergedCs#cstruct{version = V});
	Major1 /= Major2 ->
	    Major = lists:max([Major1, Major2]),
	    V = {{Major, 0}, dummy},
	    incr_version(MergedCs#cstruct{version = V})
    end.

announce_im_running([N | Ns], SchemaCs) ->
    mnesia_lib:add({current, db_nodes}, N),
    mnesia_controller:add_active_replica(schema, N, SchemaCs),
    announce_im_running(Ns, SchemaCs);
announce_im_running([], _) ->
    [].

unannounce_im_running([N | Ns]) ->
    mnesia_lib:del({current, db_nodes}, N),
    mnesia_controller:del_active_replica(schema, N),
    unannounce_im_running(Ns);
unannounce_im_running([]) ->
    [].

