%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: cdr_encode.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains all encoding functions for the CDR
%%    format.
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
-module(cdr_encode).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/include/corba.hrl").
-include_lib("orber/src/orber_iiop.hrl").

%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([enc_giop_msg_type/1, enc_request/7, enc_reply/5,
	 enc_type/2, enc_type/4, enc_cancel_request/1, enc_close_connection/0,
	 enc_message_error/0]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([]).

%%-----------------------------------------------------------------
%% External functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: enc_giop_message_header/3
%%-----------------------------------------------------------------
enc_giop_message_header(MessType, MessSize, Message) ->
    %% The header size is known so we know that the size will be aligned.
    %% MessSize already includes the header length.
    Message1 = cdrlib:enc_unsigned_long(MessSize, Message), 
    {Message2, _}  = enc_type('tk_octet', enc_giop_msg_type(MessType),
			      Message1, 0),
    {Message3, _} = enc_byte_order(Message2),
    Message4 = enc_giop_version(Message3),
    enc_magic(Message4).

enc_magic(Message) ->
    [$G, $I, $O, $P | Message].

enc_giop_version(Message) ->
    [?GIOP_MAJOR, ?GIOP_MINOR | Message].

enc_byte_order(Message) ->
    enc_type('tk_boolean', 'false', Message, 0).


%%-----------------------------------------------------------------
%% Func: enc_parameters/2
%%-----------------------------------------------------------------
enc_parameters([], [], Message, Len) ->
    {Message, Len};
enc_parameters([], _, _, _) -> exit({'enc_parameters', 'wrong_number_of_parameters'});
enc_parameters([PT1 |TypeList], [ P1 | Parameters], Message, Len) ->
    {Message1, Len1} = enc_type(PT1, P1, Message, Len),
    enc_parameters(TypeList, Parameters, Message1, Len1).

%%-----------------------------------------------------------------
%% Func: enc_request/7
%%-----------------------------------------------------------------
enc_request(ObjectKey, RequestId, ResponseExpected, Op, Parameters,
	    Context, TypeCodes) ->
    Operation = atom_to_list(Op),
    {Message0, Len0} = enc_service_context(Context, [], 0),
    {Message, Len} = enc_request_id(RequestId, Message0, Len0),
    {Message1, Len1} = enc_response(ResponseExpected, Message, Len),
    {Message2, Len2} = enc_object_key(ObjectKey, Message1, Len1),
    {Message3, Len3} = enc_operation(Operation, Message2, Len2),
    {Message4, Len4} = enc_principal(Message3, Len3),
    {Message5, Len5} = enc_request_body(TypeCodes, Parameters,
					Message4, Len4),
    enc_giop_message_header('request', Len5, lists:reverse(Message5)).

enc_principal(Mess, Len) ->
    enc_type({'tk_string', 0}, atom_to_list(node()), Mess, Len).
    
enc_operation(Op, Mess, Len) ->
    enc_type({'tk_string', 0}, Op, Mess, Len).
    
enc_object_key(ObjectKey, Mess, Len) ->
    enc_type({'tk_sequence', 'tk_octet', 0}, ObjectKey, Mess, Len).
    
enc_response(ResponseExpected, Mess, Len) ->
    enc_type('tk_boolean', ResponseExpected, Mess, Len).
    
enc_request_id(RequestId, Mess, Len) ->  
    enc_type('tk_ulong', RequestId, Mess, Len).

enc_service_context(Context, Message, Len) ->
    enc_type({'tk_sequence',
	      {'tk_struct', 0, 'ServiceContext',
	       [{"context_id", 'tk_long'},
		{"context_data",
		 {'tk_sequence', 'tk_octet', 0}}]}, 0},
	     Context, Message, Len).
    
%%-----------------------------------------------------------------
%% Func: enc_request_body/5
%%-----------------------------------------------------------------
enc_request_body({RetType, InParameters, OutParameters}, Parameters,
		 Message, Len) ->
    enc_parameters(InParameters, Parameters, Message, Len).

%%-----------------------------------------------------------------
%% Func: enc_reply/5
%%-----------------------------------------------------------------
enc_reply(ReqId, RepStatus, TypeCodes, Result, OutParameters) ->
    {Message, Len} = enc_service_context([], [],0),
    {Message1, Len1} = enc_request_id(ReqId, Message, Len), 
    {Message2, Len2} = enc_reply_status(RepStatus, Message1, Len1),
    {Message3, Len3} = enc_reply_body(TypeCodes, Result, OutParameters,
				      Message2, Len2),
    enc_giop_message_header('reply', Len3, lists:reverse(Message3)).

enc_reply_status(Status, Mess, Len) ->
    L = enc_giop_reply_status_type(Status),
    enc_type('tk_ulong', L, Mess, Len).

%%-----------------------------------------------------------------
%% Func: enc_reply_body/5
%%-----------------------------------------------------------------
enc_reply_body({RetType, InParameters, OutParameters}, Result, Parameters,
	       Message, Len) ->
    {Message1, Len1}  = enc_type(RetType, Result, Message, Len),
    enc_parameters(OutParameters, Parameters, Message1, Len1).

%%-----------------------------------------------------------------
%% Func: enc_cancel_request/1
%%-----------------------------------------------------------------
enc_cancel_request(RequestId) ->
    {Message, Len} = enc_request_id(RequestId, [], 0),
    enc_giop_message_header('cancel_request', Len, lists:reverse(Message)).

%%-----------------------------------------------------------------
%% Func: enc_locate_request/2
%%-----------------------------------------------------------------
enc_locate_request(ObjectKey, RequestId) ->
    {Message, Len} = enc_request_id(RequestId, [], 0), 
    {Message1, Len1} = enc_object_key(ObjectKey, Message, Len),
    enc_giop_message_header('locate_request', Len1, lists:reverse(Message1)).

%%-----------------------------------------------------------------
%% Func: enc_locate_reply/2
%%-----------------------------------------------------------------
enc_locate_reply(RequestId, LocStatus) ->
    {Message, Len} = enc_request_id(RequestId, [], 0), 
    {Message1, Len1} = enc_locate_status(LocStatus, Message, Len),
    enc_giop_message_header('locate_reply', Len1, lists:reverse(Message1)).

enc_locate_status(Status, Mess, Len) ->
    L = enc_giop_locate_status_type(Status),
    enc_type('tk_ulong', L, Mess, Len).
%%-----------------------------------------------------------------
%% Func: enc_close_connection/0
%%-----------------------------------------------------------------
enc_close_connection() ->
    enc_giop_message_header('close_connection', 0, []).

%%-----------------------------------------------------------------
%% Func: enc_message_error/0
%%-----------------------------------------------------------------
enc_message_error() ->
    enc_giop_message_header('message_error', 0, []).

%%-----------------------------------------------------------------
%% Func: enc_giop_msg_type
%% Args: An integer message type code
%% Returns: An atom which is the message type code name
%%-----------------------------------------------------------------
enc_giop_msg_type('request') ->
    0;
enc_giop_msg_type('reply') ->
    1;
enc_giop_msg_type('cancel_request') ->
    2;
enc_giop_msg_type('locate_request') ->
    3;
enc_giop_msg_type('locate_reply') ->
    4;
enc_giop_msg_type('close_connection') ->
    5;
enc_giop_msg_type('message_error') ->
    6.


%%-----------------------------------------------------------------
%% Func: enc_giop_reply_status_type
%% Args: An atom which is the reply status
%% Returns: An integer status code
%%-----------------------------------------------------------------
enc_giop_reply_status_type(?NO_EXCEPTION) ->
    0;
enc_giop_reply_status_type(?USER_EXCEPTION) ->
    1;
enc_giop_reply_status_type(?SYSTEM_EXCEPTION) ->
    2;
enc_giop_reply_status_type('location_forward') ->
    3.

%%-----------------------------------------------------------------
%% Func: enc_giop_locate_status_type
%% Args: An integer status code
%% Returns: An atom which is the reply status
%%-----------------------------------------------------------------
enc_giop_locate_status_type('unknown_object') ->
    0;
enc_giop_locate_status_type('object_here') ->
    1;
enc_giop_locate_status_type('object_forward') ->
    2.

%%-----------------------------------------------------------------
%% Func: enc_type/2
%% Args: TypeCode - A type specication.
%%       Value - value to encode.
%%-----------------------------------------------------------------
enc_type(TypeCode, Value) ->
    {Bytes, Len} = enc_type(TypeCode, Value, [], 0),
    Bytes1 = lists:reverse(Bytes),
    lists:flatten(Bytes1).
%%-----------------------------------------------------------------
%% Func: enc_type/4
%% Args: A type specication
%%       Bytes - a bytelist containing the message
%%       ByteOrder - big or little endian
%%-----------------------------------------------------------------
enc_type('tk_null', null, Bytes, Len) ->
    {Bytes, Len}; 
enc_type('tk_void', ok, Bytes, Len) ->
    {Bytes, Len}; 
enc_type('tk_short', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 2),
    {cdrlib:enc_r_short(Value, Rest), Len1 + 2};
enc_type('tk_long', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 4),
    {cdrlib:enc_r_long(Value, Rest ), Len1 + 4};
enc_type('tk_ushort', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 2),
    {cdrlib:enc_r_unsigned_short(Value, Rest), Len1 + 2};
enc_type('tk_ulong', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 4),
    {cdrlib:enc_r_unsigned_long(Value, Rest), Len1 + 4};
enc_type('tk_float', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 4),
    {cdrlib:enc_r_float(Value, Rest), Len1 + 4};
enc_type('tk_double', Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 8),
    {cdrlib:enc_r_double(Value, Rest), Len1 + 8};
enc_type('tk_boolean', Value, Bytes, Len) ->
    {cdrlib:enc_bool(Value, Bytes), Len + 1};
enc_type('tk_char', Value, Bytes, Len) ->
    {cdrlib:enc_char(Value, Bytes), Len + 1};
enc_type('tk_octet', Value, Bytes, Len) ->
    {cdrlib:enc_octet(Value, Bytes), Len + 1};
enc_type('tk_any', Any, Bytes, Len) when record(Any, any) ->
    {Rest, Len1} = enc_type('tk_TypeCode', Any#any.typecode, Bytes, Len),
    enc_type(Any#any.typecode, Any#any.value, Rest, Len1);
enc_type('tk_TypeCode', Value, Bytes, Len) ->
    enc_type_code(Value, Bytes, Len);
enc_type('tk_Principal', Value, Bytes, Len) ->
    enc_sequence(Value, 'tk_octet', Bytes, Len);
enc_type({'tk_objref', IFRId, Name}, Value, Bytes, Len) ->
    enc_objref(Name,Value, Bytes, Len);
enc_type({'tk_struct', IFRId, Name, ElementList}, Value, Bytes, Len) -> 
    enc_struct(Value, ElementList, Bytes, Len);
enc_type({'tk_union', IFRId, Name, DiscrTC, Default, ElementList},
	Value, Bytes, Len) ->
    enc_union(Value, DiscrTC, ElementList, Bytes, Len);
enc_type({'tk_enum', IFRId, Name, ElementList}, Value, Bytes, Len) ->
    {Rest, Len1} = enc_align(Bytes, Len, 4),
    {cdrlib:enc_r_enum(atom_to_list(Value), ElementList, Rest), Len1 + 4};
enc_type({'tk_string', MaxLength}, Value, Bytes, Len) -> % MaxLength not used
    enc_string(Value, Bytes, Len);
enc_type({'tk_sequence', ElemTC, MaxLength}, Value, Bytes, Len) -> % MaxLength not used
    enc_sequence(Value, ElemTC, Bytes, Len);
enc_type({'tk_array', ElemTC, Size}, Value, Bytes, Len) -> 
    enc_array(Value, Size, ElemTC, Bytes, Len);
enc_type({'tk_alias', IFRId, Name, TC}, Value, Bytes, Len) ->
    enc_type(TC, Value, Bytes, Len);
enc_type({'tk_except', IFRId, Name, ElementList}, Value, Bytes, Len) ->
    enc_exception(Name, IFRId, Value, ElementList, Bytes, Len).

%%-----------------------------------------------------------------
%% Func: enc_sequence/4
%%-----------------------------------------------------------------
enc_sequence(Sequence, TypeCode, Bytes, Len) ->
    {ByteSequence, Len1} = enc_align(Bytes, Len, 4),
    ByteSequence1 = cdrlib:enc_r_unsigned_long(length(Sequence), ByteSequence),
    enc_sequence1(Sequence, TypeCode, ByteSequence1, Len1 + 4).

%%-----------------------------------------------------------------
%% Func: enc_sequence1/4
%%-----------------------------------------------------------------
enc_sequence1([], TypeCode, Bytes, Len) ->
    {Bytes, Len};
enc_sequence1(CharSeq, 'tk_char', Bytes, Len) -> 
    {[CharSeq |Bytes], Len + length(CharSeq)};
enc_sequence1(OctetSeq, 'tk_octet', Bytes, Len) -> 
    {[OctetSeq |Bytes], Len + length(OctetSeq)};
enc_sequence1([Object| Rest], TypeCode, Bytes, Len) -> 
    {ByteSequence, Len1} = enc_type(TypeCode, Object, Bytes, Len),
    enc_sequence1(Rest, TypeCode, ByteSequence, Len1).


%%-----------------------------------------------------------------
%% Func: enc_array/4
%%-----------------------------------------------------------------
enc_array(Array, Size, TypeCode, Bytes, Len) when size(Array) == Size ->
    Sequence = tuple_to_list(Array),
    enc_sequence1(Sequence, TypeCode, Bytes, Len);
enc_array(_, _, _, _, _) ->
    corba:raise(#'MARSHAL'{completion_status=?COMPLETED_NO}).

%%-----------------------------------------------------------------
%% Func: enc_string/3
%%-----------------------------------------------------------------
enc_string(String, Bytes, Len) ->
    {ByteSequence, Len1} = enc_align(Bytes, Len, 4),
    ByteSequence1 = cdrlib:enc_r_unsigned_long(length(String) + 1,
					       ByteSequence),
    {cdrlib:enc_octet(0, [String | ByteSequence1]), Len1 + length(String) + 5}.

%%-----------------------------------------------------------------
%% Func: enc_union/5
%%-----------------------------------------------------------------
enc_union({_, Label, Value}, DiscrTC, TypeCodeList, Bytes, Len) ->
    {ByteSequence, Len1} = enc_type(DiscrTC, Label, Bytes, Len),
    Label2 = stringify_enum(DiscrTC,Label),
    enc_union({Label2, Value},TypeCodeList, ByteSequence, Len1).

enc_union(_, [], _, _) -> corba:raise(#'MARSHAL'{completion_status=?COMPLETED_NO});
enc_union({Label,Value} ,[{Label, Name, Type} |List], Bytes, Len) ->
    enc_type(Type, Value, Bytes, Len);
enc_union(Union,[_ | List], Bytes, Len) ->
    enc_union(Union, List, Bytes, Len).

stringify_enum({tk_enum, _,_,_}, Label) ->
    atom_to_list(Label);
stringify_enum(_, Label) ->
    Label.
%%-----------------------------------------------------------------
%% Func: enc_struct/4
%%-----------------------------------------------------------------
enc_struct(Struct, TypeCodeList, Bytes, Len) ->
    [Name | StructList] = tuple_to_list(Struct),
    enc_struct1(StructList, TypeCodeList, Bytes, Len).

enc_struct1([], [], Bytes, Len) ->
    {Bytes, Len};
enc_struct1([Object | Rest], [{ElemName, ElemType} | TypeCodeList], Bytes,
	    Len) ->
    {ByteSequence, Len1} = enc_type(ElemType, Object, Bytes, Len),
    enc_struct1(Rest, TypeCodeList, ByteSequence, Len1).

%%-----------------------------------------------------------------
%% Func: enc_objref/4
%%-----------------------------------------------------------------
enc_objref(Name, Value, Bytes, Len) ->
     iop_ior:code(Value, Bytes, Len).
      
%%-----------------------------------------------------------------
%% Func: enc_exception/5
%%-----------------------------------------------------------------
enc_exception(Name, IFRId, Value, ElementList, Bytes, Len) ->
    [Name1, TypeId | Args] = tuple_to_list(Value),
    {Bytes1, Len1} = enc_type({'tk_string', 0}, IFRId , Bytes, Len),
    enc_exception_1(Args, ElementList, Bytes1, Len1).

enc_exception_1([], [], Bytes, Len) ->
    {Bytes, Len};
enc_exception_1([Arg |Args], [{ElemName, ElemType} |ElementList],
		Bytes, Len) ->
    {Bytes1, Len1} = enc_type(ElemType, Arg, Bytes, Len),
    enc_exception_1(Args, ElementList, Bytes1, Len1).
    

%%-----------------------------------------------------------------
%% Func: enc_type_code/3
%%-----------------------------------------------------------------
enc_type_code('tk_null', Message, Len) ->
    enc_type('tk_ulong', 0, Message, Len);
enc_type_code('tk_void', Message, Len) ->
    enc_type('tk_ulong', 1, Message, Len);
enc_type_code('tk_short', Message, Len) ->
    enc_type('tk_ulong', 2, Message, Len);
enc_type_code('tk_long', Message, Len) ->
    enc_type('tk_ulong', 3, Message, Len);
enc_type_code('tk_ushort', Message, Len) ->
    enc_type('tk_ulong', 4, Message, Len);
enc_type_code('tk_ulong', Message, Len) ->
    enc_type('tk_ulong', 5, Message, Len);
enc_type_code('tk_float', Message, Len) ->
    enc_type('tk_ulong', 6, Message, Len);
enc_type_code('tk_double', Message, Len) ->
    enc_type('tk_ulong', 7, Message, Len);
enc_type_code('tk_boolean', Message, Len) ->
    enc_type('tk_ulong', 8, Message, Len);
enc_type_code('tk_char', Message, Len) ->
    enc_type('tk_ulong', 9, Message, Len);
enc_type_code('tk_octet', Message, Len) ->
    enc_type('tk_ulong', 10, Message, Len);
enc_type_code('tk_any', Message, Len) ->
    enc_type('tk_ulong', 11, Message, Len);
enc_type_code('tk_TypeCode', Message, Len) ->
    enc_type('tk_ulong', 12, Message, Len);
enc_type_code('tk_Principal', Message, Len) ->
    enc_type('tk_ulong', 13, Message, Len);
enc_type_code({'tk_objref', RepId, Name}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 14, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				    {"name", {'tk_string', 0}}]},
				     {"", RepId, Name},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_struct', RepId, SimpleName, ElementList}, Message, Len) ->
    Name = ifrid_to_name(RepId),
    {Message1, Len1} = enc_type('tk_ulong', 15, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				    {"name", {'tk_string', 0}},
				    {"element list",
				     {'tk_sequence', {'tk_struct', "","",
						      [{"member name", {'tk_string', 0}},
						       {"member type", 'tk_TypeCode'}]},
				      0}}]},
				     {"", RepId, Name,
				      lists:map(fun({N,T}) -> {"",N,T} end, ElementList)},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_union', RepId, Name, DiscrTC, Default, ElementList},
	      Message, Len) ->
    NewElementList =
	case check_enum(DiscrTC) of
	    true ->
		lists:map(fun({L,N,T}) -> {"",list_to_atom(L),N,T} end, ElementList);
	    false ->
		lists:map(fun({L,N,T}) -> {"",L,N,T} end, ElementList)
	end,
    {Message1, Len1} = enc_type('tk_ulong', 16, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				    {"name", {'tk_string', 0}},
				    {"discriminant type", 'tk_TypeCode'},
				    {"default used", 'tk_long'},
				    {"element list",
				     {'tk_sequence', {'tk_struct', "","",
						      [{"label value", DiscrTC},
						       {"memeber name", {'tk_string', 0}},
						       {"memver type", 'tk_TypeCode'}]},
				      0}}]},
				     {"", RepId, Name, DiscrTC, Default, NewElementList},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_enum', RepId, Name, ElementList}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 17, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				    {"name", {'tk_string', 0}},
				    {"element list",
				     {'tk_sequence', {'tk_string', 0}, 0}}]},
				     {"", RepId, Name, ElementList},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_string', MaxLength}, Message, Len) ->
     enc_type({'tk_struct', "", "", [{"TCKind", 'tk_ulong'},
				    {"max length", 'tk_ulong'}]},
	      {"", 18, MaxLength},
	      Message, Len);
enc_type_code({'tk_sequence', ElemTC, MaxLength}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 19, Message, Len),
    {Message2, _} = enc_byte_order([]),
     {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"element type", 'tk_TypeCode'},
				    {"max length", 'tk_ulong'}]},
				      {"", ElemTC, MaxLength},
				      Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_array', ElemTC, Length}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 20, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"element type", 'tk_TypeCode'},
				    {"length", 'tk_ulong'}]},
				     {"", ElemTC, Length},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_alias', RepId, Name, TC}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 21, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				     {"name", {'tk_string', 0}},
				     {"TypeCode", 'tk_TypeCode'}]},
				     {"", RepId, Name, TC},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'tk_except', RepId, Name, ElementList}, Message, Len) ->
    {Message1, Len1} = enc_type('tk_ulong', 22, Message, Len),
    {Message2, _} = enc_byte_order([]),
    {ComplexParams, Len2} = enc_type({'tk_struct', "", "", [{"repository ID", {'tk_string', 0}},
				    {"name", {'tk_string', 0}},
				    {"element list",
				     {'tk_sequence',
				      {'tk_struct', "", "",
				       {"member name", {'tk_string', 0}},
				       {"member type", 'tk_TypeCode'}}, 0}}]},
				     {"", RepId, Name,
				      lists:map(fun({N,T}) -> {"",N,T} end, ElementList)},
				     Message2, 1),
    encode_complex_tc_paramters(lists:reverse(ComplexParams), Len2, Message1, Len1);
enc_type_code({'none', Indirection}, Message, Len) ->  %% placeholder
     enc_type({'tk_struct', "", "", [{"TCKind", 'tk_ulong'},
				    {"indirection", 'tk_ulong'}]},
	      {"", 16#ffffffff, Indirection},
	      Message, Len).


check_enum({'tk_enum', _, _, _}) ->
    true;
check_enum(_) ->
    false.

encode_complex_tc_paramters(Value, Value_length, Message, Len) ->
    {Message1, Len1} = enc_align(Message, Len, 4),
    Message2 = cdrlib:enc_r_unsigned_long(Value_length, Message1),
    {[Value |Message2], Len+Value_length+4}.


ifrid_to_name([]) ->
        corba:raise(#'MARSHAL'{completion_status=?COMPLETED_NO});
ifrid_to_name(Id) ->
    Rep = orber_ifr:find_repository(),
    Key = orber_ifr:'Repository_lookup_id'(Rep, Id),
    [$:, $: |N] = orber_ifr:'Contained__get_absolute_name'(Key),
    change_colons_to_underscore(N, []).

change_colons_to_underscore([$:, $: | T], Acc) ->
    change_colons_to_underscore(T, [$_ |Acc]);
change_colons_to_underscore([H |T], Acc) ->
    change_colons_to_underscore(T, [H |Acc]);
change_colons_to_underscore([], Acc) ->
    lists:reverse(Acc).

%%-----------------------------------------------------------------
%% Func: enc_align/1
%%-----------------------------------------------------------------
enc_align(R, Len, Alignment) ->
    Rem = Len rem Alignment,
    if Rem == 0 ->
	    {R, Len};
       true ->
	    {add_bytes(R, Alignment - Rem), Len + (Alignment - Rem)}
    end.

add_bytes(R, 0) ->
    R;
add_bytes(R,N) ->
    add_bytes([0 | R], N - 1).

