%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the Orber application interface
%%
%% Creation date: 970407
%%
%%-----------------------------------------------------------------
-module(orber).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/src/orber_debug.hrl").
%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([start/0, stop/0, install/1, orber_nodes/0, iiop_port/0, domain/0,
	 bootstrap_port/0]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([host/0, start/2, init/1, start_naming_service/1]).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
start() ->
    application:start(mnesia),
    application:start(orber).

stop() ->
    application:stop(orber).


iiop_port() ->
    case application:get_env(orber, iiop_port) of
	{ok, Port} when integer(Port) ->
	    Port;
	_ ->
	    4001
    end.

bootstrap_port() ->
    case application:get_env(orber, bootstrap_port) of
	{ok, Port} when integer(Port) ->
	    Port;
	_ ->
	    iiop_port()
    end.

orber_nodes() ->
    case application:get_env(orber, orber_nodes) of
	{ok, Nodes} when list(Nodes) ->
	    Nodes;
	_ ->
	    [node()]
    end.

domain() -> 
    case application:get_env(orber, domain) of
	{ok, Domain} when list(Domain) ->
	    Domain;
	{ok, Domain} when atom(Domain) ->
	    atom_to_list(Domain);
	_ ->
	    "ORBER"
    end.

%%-----------------------------------------------------------------
%% Installation interface functions
%%-----------------------------------------------------------------
install([]) ->
    install([node()]);
install(Nodes) when list(Nodes) ->
    install(Nodes,
	    [{ifr_storage_type, disc_copies},
	     {install_timeout, infinity}]).
   
install(Nodes, Options) when list(Nodes), list(Options)->
    case mnesia:system_info(is_running) of
	no ->
	    application:start(mnesia),
	    install_orber(Nodes, Options),
	    application:stop(mnesia);
	yes ->
	    install_orber(Nodes, Options)
    end.

install_orber(Nodes, Options) ->
    Timeout = get_option(install_timeout, Options),
    IFR_storage_type = get_option(ifr_storage_type, Options),
    orber_ifr:init(Nodes, Timeout, IFR_storage_type),
    orber_objectkeys:install(Nodes, Timeout),
    'CosNaming_NamingContext_impl':install(Nodes, Timeout),
    application:start(orber),
    oe_cos_naming:oe_register(),
    oe_erlang:oe_register(),
    create_default_contexts(),
    mnesia:dump_tables(['orber_CosNaming']),
    application:stop(orber).

%%-----------------------------------------------------------------
%% UnInstallation interface functions
%%-----------------------------------------------------------------
uninstall() ->
    ok.

get_option(Key, OptionList) ->
    case lists:keysearch(Key, 1, OptionList) of
	{value,{Key,Value}} ->
	    Value;
	_ ->
	    exit(io_lib:format("Option ~w not found in optionlist", [Key]))
    end.

%%-----------------------------------------------------------------
%% Internal interface functions
%%-----------------------------------------------------------------
host() ->
    net_adm:localhost().

%%-----------------------------------------------------------------
%% Server functions
%%-----------------------------------------------------------------
start(_, _) ->
    supervisor:start_link({local, orber_sup}, orber, orb_init).

init(orb_init) ->
    ?PRINTDEBUG("init orber supervisor"),
    case mnesia:wait_for_tables([ir_IRObject,
				 ir_Contained,
				 ir_Container,
				 ir_IDLType,
				 ir_Repository,
				 ir_ModuleDef,
				 ir_ConstantDef,
				 ir_TypedefDef,
				 ir_StructDef,
				 ir_UnionDef,
				 ir_EnumDef,
				 ir_AliasDef,
				 ir_PrimitiveDef,
				 ir_StringDef,
				 ir_SequenceDef,
				 ir_ArrayDef,
				 ir_ExceptionDef,
				 ir_AttributeDef,
				 ir_OperationDef,
				 ir_InterfaceDef,
				 ir_ORB], infinity) of
	ok ->
	    SupFlags = {one_for_one, 500, 1000},  %Max 5 restarts in 1 second
	    ChildSpec = [
                 {orber_iiop_sup, {orber_iiop, start_sup, [[]]}, permanent, 
                  10000, supervisor, [orber_iiop]},
		 {orber_init, {orber_initial_references, start, [[]]},
		  permanent, 
                  10000, worker, [orber_initial_references]},
		 {orber_reqno, {orber_request_number, start, [[]]},
		  permanent, 
                  10000, worker, [orber_request_number]},
		 {orber_objkeyserver, {orber_objectkeys, start, [[orber_nodes(), 0]]},
		  permanent, 
                  10000, worker, [orber_objectkeys]},
		 {orber_nameservice, {orber, start_naming_service, 
				      [[]]},
		  permanent, 
                  10000, worker, ['CosNaming_NamingContext']}
		],
	    {ok, {SupFlags, ChildSpec}};
	StopReason ->
	    {stop, StopReason}
    end.

start_naming_service(_) ->
    case catch 'CosNaming_NamingContext':oe_create_link([], {local, orber_nameservice}) of 
	{'EXCEPTION', E} ->
	    {error, {'EXCEPTION', E}};
	{'EXIT', R} ->
	    {error, {'EXIT', R}};
	Key ->
	    {ok, corba:get_pid(Key)}
    end.



create_default_contexts() ->
    HostComponent = lname_component:set_id(lname_component:create(),
					   "host"),
    HostsComponent = lname_component:set_id(lname_component:create(),
					    "hosts"),
    ResourcesComponent = lname_component:set_id(lname_component:create(),
						"resources"),
    DevelopmentComponent = lname_component:set_id(lname_component:create(),
					   "development"),
    FactoriesComponent = lname_component:set_id(lname_component:create(),
					   "factories"),
    WGComponent = lname_component:set_id(lname_component:create(),
					   "workgroup"),
    %% Creation of Naming Context host and it's subcontexts
    NS = corba:resolve_initial_references("NameService"),
    H = 'CosNaming_NamingContext':bind_new_context(NS,
		lname:insert_component(lname:create(), 1, HostComponent)),
    HR = 'CosNaming_NamingContext':bind_new_context(H,
		lname:insert_component(lname:create(), 1, ResourcesComponent)),
    'CosNaming_NamingContext':bind_new_context(HR,
		lname:insert_component(lname:create(), 1, FactoriesComponent)),
    HD = 'CosNaming_NamingContext':bind_new_context(H,
	lname:insert_component(lname:create(), 1, DevelopmentComponent)),
    HDR = 'CosNaming_NamingContext':bind_new_context(HD,
		lname:insert_component(lname:create(), 1, ResourcesComponent)),
    'CosNaming_NamingContext':bind_new_context(HDR,
		lname:insert_component(lname:create(), 1, FactoriesComponent)),
      %% Creation of Naming Context workgroup and it's subcontexts
    W = 'CosNaming_NamingContext':bind_new_context(NS,
		lname:insert_component(lname:create(), 1, WGComponent)),
    'CosNaming_NamingContext':bind_new_context(W,
		lname:insert_component(lname:create(), 1, HostsComponent)),
    WR = 'CosNaming_NamingContext':bind_new_context(W,
		lname:insert_component(lname:create(), 1, ResourcesComponent)),
    'CosNaming_NamingContext':bind_new_context(WR,
		lname:insert_component(lname:create(), 1, FactoriesComponent)),
    WD = 'CosNaming_NamingContext':bind_new_context(W,
	lname:insert_component(lname:create(), 1, DevelopmentComponent)),
    WDR = 'CosNaming_NamingContext':bind_new_context(WD,
		lname:insert_component(lname:create(), 1, ResourcesComponent)),
    'CosNaming_NamingContext':bind_new_context(WDR,
		lname:insert_component(lname:create(), 1, FactoriesComponent)),
    ok.



