%%--------------------------------------------------------------------
%%
%%    COPYRIGHT Ericsson Telecom AB 1998
%%
%%    126 25 STOCKHOLM
%%    SWEDEN
%%    tel int + 46 8 719 0000
%%
%%    The copyright to the computer program herein is the property of
%%    Ericsson Telecom AB, Sweden. The program may be used and/or
%%    copied only with the written permission from Ericsson Telecom AB
%%    or in accordance with the terms and conditions stipulated in the
%%    agreement/contract under which the program have been supplied.
%%
%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%%-----------------------------------------------------------------
%% File: orber_iiop.erl
%% Author: Lars Thorsen
%% 
%% Description:
%%    This file contains the IIOP interface
%%
%% Creation date: 970115
%%
%%-----------------------------------------------------------------
%%
%% Interface to the iiop operations.
%%
%% The iiop is organised as one process being the center piece from
%% which connections to other nodes can be started. 
%%-----------------------------------------------------------------
-module(orber_iiop).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-author('lars@erix.ericsson.se').

-include_lib("orber/include/corba.hrl").
-include_lib("orber/src/orber_iiop.hrl").
-include_lib("orber/src/orber_debug.hrl").

-behaviour(supervisor).
%%-----------------------------------------------------------------
%% External exports
%%-----------------------------------------------------------------
-export([init/1, start_sup/1, connect/1, request/5]).

%%-----------------------------------------------------------------
%% Internal exports
%%-----------------------------------------------------------------
-export([init/1, terminate/2, handle_call/3]).

%%-----------------------------------------------------------------
%% Server state record
%%-----------------------------------------------------------------
-record(state, {db=[]}).

%%-----------------------------------------------------------------
%% External interface functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: start_sup/1
%%-----------------------------------------------------------------
start_sup(Opts) ->
    supervisor:start_link({local, orber_iiop_sup}, ?MODULE,
			  {orber_iiop_sup, Opts}).

%%%-----------------------------------------------------------------
%%% Func: connect/1
%%%-----------------------------------------------------------------
connect(OrbName) ->
    orber_iiop_net:connect(OrbName).

%%%-----------------------------------------------------------------
%%% Func: request/5
%%%-----------------------------------------------------------------
request({Host, IIOP_port, ObjKey}, Op, Parameters, TypeCodes, ResponseExpected) ->
    RequestId = orber_request_number:get(),
    {NewConnection, Socket} = case get({Host, IIOP_port}) of
		 undefined ->
		     S = ?IIOP_SOCKET_MOD:connect(Host, IIOP_port),
		     case ObjKey of
			 "INIT" ->
			     true;
			 _ ->
			     put({Host, IIOP_port}, S)
		     end,
		     {true, S};
		 S ->
		     {false, S}
	     end,
    case catch cdr_encode:enc_request(ObjKey,  RequestId, ResponseExpected,
				      Op, Parameters,
				      [], TypeCodes) of
	{'EXCEPTION', E} ->
	    corba:raise(E);
	{'EXIT', R} ->
	    corba:raise(#'MARSHAL'{completion_status=?COMPLETED_NO});
	Request ->
	    ?IIOP_SOCKET_MOD:write(Socket, Request),
	    case ResponseExpected of
		'true' ->
		    read_reply(Socket, NewConnection, RequestId, ObjKey, Op,
			       Parameters, TypeCodes);
		'false' ->
		    'ok'
	    end
    end.

read_reply(Socket, NewConnection, RequestId, ObjKey, Op, Parameters, TypeCodes) ->
    Reply = ?IIOP_SOCKET_MOD:read(Socket),
    case catch cdr_decode:dec_message(TypeCodes, Reply) of
	{'EXCEPTION', DecodeException} ->
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_MAYBE});
	{'EXIT', message_error} ->
	    ME = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, ME),
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO});
	{'EXIT', R} ->
	    corba:raise(#'UNKNOWN'{minor=40, completion_status=?COMPLETED_MAYBE});
	'message_error' ->
	    %% Perhaps a resend should be done when a message error occurs
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO});
	{ReplyHdr, Result, Par} ->
	    case {NewConnection, ObjKey} of
		{true, "INIT"} ->
		    ?IIOP_SOCKET_MOD:close(Socket);
		_ ->
		    true
	    end,
	    %% Check request id 
	    if 
		RequestId == ReplyHdr#reply_header.request_id ->
		    case ReplyHdr#reply_header.reply_status of
			'no_exception' ->
			    case Par of
				[] ->
				    Result;
				_ ->
				    L = [Result | Par],
				    list_to_tuple(L)
			    end;
			'system_exception' ->
			    corba:raise(Result);
			'user_exception' ->
			    corba:raise(Result);
			'location_forward' ->
			    ?IIOP_SOCKET_MOD:close(Socket),
			    erase(ObjKey),
			    corba:call(Result, Op, Parameters, TypeCodes)
		    end;		  
		true ->
		    read_reply(Socket, NewConnection, RequestId, ObjKey,
			       Op, Parameters, TypeCodes)
	    end
    end.

%%-----------------------------------------------------------------
%% Func: locate/1
%%-----------------------------------------------------------------
locate({Host, IIOP_port, ObjKey}) ->
    RequestId = orber_request_number:get(),
    {NewConnection, Socket} = case get({Host, IIOP_port}) of
		 undefined ->
		     S = ?IIOP_SOCKET_MOD:connect(Host, IIOP_port),
		     case ObjKey of
			 "INIT" ->
			     true;
			 _ ->
			     put({Host, IIOP_port}, S)
		     end,
		     {true, S};
		 S ->
		     {false, S}
	     end,
    case catch cdr_encode:enc_locate_request(ObjKey,  RequestId) of
	{'EXCEPTION', E} ->
	    corba:raise(E);
	{'EXIT', R} ->
	    corba:raise(#'MARSHAL'{completion_status=?COMPLETED_NO});
	Request ->
	    ?IIOP_SOCKET_MOD:write(Socket, Request),
	    read_locate_reply(Socket, NewConnection, RequestId, ObjKey)
    end.

read_locate_reply(Socket, NewConnection, RequestId, ObjKey) ->
    Reply = ?IIOP_SOCKET_MOD:read(Socket),
    case catch cdr_decode:dec_message([], Reply) of
	{'EXCEPTION', DecodeException} ->
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_MAYBE});
	{'EXIT', message_error} ->
	    ME = cdr_encode:enc_message_error(),
	    ?IIOP_SOCKET_MOD:write(Socket, ME),
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO});
	{'EXIT', R} ->
	    corba:raise(#'UNKNOWN'{completion_status=?COMPLETED_MAYBE});
	'message_error' ->
	    %% Perhaps a resend should be done when a message error occurs
	    corba:raise(#'COMM_FAILURE'{completion_status=?COMPLETED_NO});
	{ReplyHdr, Result} ->
	    case {NewConnection, ObjKey} of
		{true, "INIT"} ->
		    ?IIOP_SOCKET_MOD:close(Socket);
		_ ->
		    true
	    end,
	    %% Check request id 
	    if 
		RequestId == ReplyHdr#locate_reply_header.request_id ->
		    case ReplyHdr#locate_reply_header.locate_status of
			'unkown_object' ->
			    'unkown_object';
			'object_here' ->
			    'object_here';
			'object_forward' ->
			    {'object_forward', Result}
		    end;		  
		true ->
		    read_locate_reply(Socket, NewConnection, RequestId, ObjKey)
	    end
    end.

%%%-----------------------------------------------------------------
%%% Func: cancel/1
%%%-----------------------------------------------------------------
%cancel(X) ->
%	ok.

%%%-----------------------------------------------------------------
%%% Func: message_error/1
%%%-----------------------------------------------------------------
%message_error(X) ->
%	ok.

%%-----------------------------------------------------------------
%% Server functions
%%-----------------------------------------------------------------
%%-----------------------------------------------------------------
%% Func: init/1
%%-----------------------------------------------------------------
init({orber_iiop_sup, Opts}) ->
    ?PRINTDEBUG("init iiop supervisor"),
    IIOP_port =  orber:iiop_port(),
    Bootstrap_port =  orber:bootstrap_port(),
    SupFlags = {one_for_one, 5, 1},	%Max 5 restarts in 1 second
    ChildSpec1 = if  
		     Bootstrap_port == IIOP_port ->
			 [{orber_iiop_net, {orber_iiop_net, start,
					   [[{port, IIOP_port}]]},
			  permanent, 10000, worker, [orber_iiop_net]}];
		     Bootstrap_port < 1024 ->
%		     Bootstrap_port < 2024 -> % Used for testing without being root
			 [{orber_iiop_net, {orber_iiop_net, start,
					    [[{port, IIOP_port}]]},
			   permanent, 10000, worker, [orber_iiop_net]},
			  {orber_bootstrap, {orber_bootstrap,
					     start, [{port, Bootstrap_port}]}, permanent, 
			   10000, worker, [orber_bootstrap]}];
		     true ->
			 [{orber_iiop_net, {orber_iiop_net, start,
					   [[{port, IIOP_port},
					     {port, Bootstrap_port}]]},
			  permanent, 10000, worker, [orber_iiop_net]}]
		 end,
    ChildSpec = [
		 {orber_iiop_netsup, {orber_iiop_netsup, start,
				      [sup, Opts]},
		  permanent, 10000, supervisor, [orber_iiop_netsup]} |
		 ChildSpec1
		 ],
    {ok, {SupFlags, ChildSpec}}.





%%-----------------------------------------------------------------
%% Func: terminate/2
%%-----------------------------------------------------------------
terminate(Reason, State) ->
    ok.

%%-----------------------------------------------------------------
%% Func: handle_call/3
%%-----------------------------------------------------------------
handle_call(Req, From, State) ->
    {reply, ok, State}.


