%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_acm).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-export([init_check_access/2, init_check_access/3,
	 get_root_mib_view/0, get_mib_view/1,
	 validate_mib_view/2, validate_all_mib_view/2,
	 is_definitely_not_in_mib_view/2]).

-include("snmp_types.hrl").
-include("STANDARD-MIB.hrl").
-include("OTP-SNMPEA-MIB.hrl").
-define(any_ip, [0,0,0,0]).

%%%-----------------------------------------------------------------
%%% This module implements the Access Control Model part of the
%%% multi-lingual SNMP agent.
%%%
%%% It supports different access control models for different
%%% SNMP versions.  E.g. v1 and v2c uses the mibviews defined in
%%% OTP-SNMPEA-MIB, and v3 may use the SNMPV3-AC-MIB.
%%%
%%% Access checking is done in several steps.  First, the version-
%%% specific MPC (see snmp_mpc) decides which ACM to use, and
%%% creates data for that model.  This means that the format of this
%%% data is known by both the MPC and the ACM.  When the master
%%% agent wants to check the access to a Pdu, it first calls
%%% init_check_access/3, which returns a MibView
%%% that can be used to check access of individual variables.
%%%-----------------------------------------------------------------

%%-----------------------------------------------------------------
%% Func: init_check_access(ACMModule, Pdu, ACMData) ->
%%       {ok, MibView} | {discarded, Variable, Reason}
%%       Variable = snmpInBadCommunityNames |
%%                  snmpInBadCommunityUses |
%%                  snmpInASNParseErrs
%% The authenticationFailure trap is sent (maybe) when the auth.
%% procedure evaluates to unauthentic,
%% Reason is one of the following:
%%    snmp_message_decoding
%%    {bad_community_name, Address, Community}}
%%    {invalid_access, Access, Op}
%% NOTE: This function is executed in the Master agents's context
%%-----------------------------------------------------------------
init_check_access(ACMModule, Pdu, ACMData) ->
    check(catch apply(ACMModule, init_check_access, [Pdu, ACMData]),
	  ACMModule).

%%-----------------------------------------------------------------
%% Func: check(Res) -> {ok, MibView} | {discarded, Variable, Reason}
%% Args: Res = {ok, AccessFunc} | {authentication_failure, Variable, Reason}
%%-----------------------------------------------------------------
check({ok, MibView}, _ACMModule) ->
    {ok, MibView};
check({authentication_failure, Variable, Reason}, _ACMModule) ->
    handle_authentication_failure(),
    {discarded, Variable, Reason};
check(X, ACMModule) ->
    {discarded, false, {bad_return_value, {ACMModule, init_check_access, [X]}}}.

%%-----------------------------------------------------------------
%% Func: init_check_access(Pdu, ACMData) ->
%%       {ok, MibView} | {authentication_failure, Variable, Reason}
%% Args: Pdu = #pdu
%%       ACMData = acm_data() = {community, Community, Address}
%%       Community = string()
%%       Address = ip() (list)
%% Purpose: Default init_check_access function.  Uses the mechanism in
%%          OTP-SNMPEA-MIB for authentication.
%%-----------------------------------------------------------------
init_check_access(Pdu, {community, Community, Ip}) ->
    CommunityKey = [length(Community) | Community],
    case get_community_info(Ip, CommunityKey) of
	[{value, ViewIndex}, {value, Access}] -> 
	    auth(Access, ViewIndex, Pdu, Community);
	_ -> 
	    % now, check for wildcard ip address
	    case get_community_info(?any_ip, CommunityKey) of
		[{value, ViewIndex}, {value, Access}] -> 
		    auth(Access, ViewIndex, Pdu, Community);
		_ ->
		    {authentication_failure, snmpInBadCommunityNames,
		     {discarded, {bad_community_name, Ip, Community}}}
	    end
    end.

get_community_info(IP, CommunityKey) ->
    Key = lists:append(IP, CommunityKey),
    Elements = [?intCommunityViewIndex, ?intCommunityAccess],
    case snmp_internal_mib:check_status(intCommunityTable,
                                        Key, ?intCommunityStatus) of
	true -> snmp_internal_mib:intCommunityTable(get, Key, Elements);
	false -> undefined
    end.
    
auth(Access, ViewIndex, Pdu, Community) ->
    #pdu{type = Op} = Pdu,
    case check_access(Access, Op) of
	true ->
	    MibView = get_mib_view(ViewIndex),
	    {ok, MibView};
	false ->
	    {authentication_failure, snmpInBadCommunityUses,
	     {discarded, {invalid_access, Access, Op}}}
    end.

check_access(?intCommunityAccess_readWrite, _Op) -> true;
check_access(?intCommunityAccess_read, 'get-request') -> true;
check_access(?intCommunityAccess_read, 'get-next-request') -> true;
check_access(?intCommunityAccess_read, 'get-bulk-request') -> true;
check_access(_Access, _Op) -> false.

%%-----------------------------------------------------------------
%% NOTE: This function is executed in the Master agents's context
%% Do a GET to retrieve the value for snmpEnableAuthenTraps.  A
%% user may have another impl. than default for this variable.
%%-----------------------------------------------------------------
handle_authentication_failure() ->
    case snmp_agent:do_get(get_root_mib_view(),
			   [#varbind{oid = ?snmpEnableAuthenTraps_instance}],
			   true) of
	{noError, _, [#varbind{value = ?snmpEnableAuthenTraps_enabled}]} ->
	    snmp:send_notification(self(), authenticationFailure,
				   no_receiver);
	_ ->
	    ok
    end.

%%%-----------------------------------------------------------------
%%% MIB View handling
%%%-----------------------------------------------------------------

%%-----------------------------------------------------------------
%% Returns a list of {ViewSubtree, ViewMask, ViewType}
%% The view table is index by ViewIndex, ViewSubtree,
%% so a next on ViewIndex returns the first
%% key in the table >= ViewIndex.
%%-----------------------------------------------------------------
get_mib_view(ViewIndex) ->
    case snmp_internal_mib:table_next(intViewTable, [ViewIndex]) of
	endOfTable ->
	    snmp_error:config_err("No such viewIndex ~w", [ViewIndex]),
	    [];
	[ViewIndex | Subtree] ->
	    loop_mib_view(ViewIndex, Subtree, []);
	Indexes ->
	    snmp_error:config_err("No such viewIndex ~w", [ViewIndex]),
	    []
    end.

get_root_mib_view() ->
    [{[1], [], ?intViewType_included}].

loop_mib_view(ViewIndex, Subtree, MibView) ->
    Indexes = [ViewIndex | Subtree],
    [{value, Mask}, {value, Type}, {value, Status}] = 
	snmp_internal_mib:intViewTable(get, Indexes, [?intViewMask, 
						      ?intViewType,
						      ?intViewStatus]),
    NextMibView = 
	case Status of
	    ?'RowStatus_active' ->
		[Length | Tree] = Subtree,
		[{Tree, Mask, Type} | MibView];
	    _ ->
		MibView
	end,
    case snmp_internal_mib:table_next(intViewTable, Indexes) of
	endOfTable -> NextMibView;
	[ViewIndex | NextSubTree] -> 
	    loop_mib_view(ViewIndex, NextSubTree, NextMibView);
	Else -> NextMibView
    end.
	
%%-----------------------------------------------------------------
%% Returns true if Oid is in the MibView, false
%% otherwise.
%% Alg: (defined in rfc1445 section 4.3)
%% For each family (= {SubTree, Mask, Type}), check if Oid
%% belongs to that family. For each family that Oid belong to,
%% get the longest. If two or more are longest, get the
%% lexicografically greatest. Check the type of this family. If
%% included, then Oid belongs to the MibView, otherwise it
%% does not.
%% Optimisation: Do only one loop, and kepp the largest sofar.
%% When we find a family that Oid belongs to, check if it is
%% larger than the largest.
%%-----------------------------------------------------------------
validate_mib_view(Oid, MibView) ->
    case get_largest_family(MibView, Oid, undefined) of
	{_, _, ?intViewType_included} -> true;
	_ -> false
    end.

get_largest_family([{SubTree, Mask, Type} | T], Oid, Res) ->
    case check_mask(Oid, SubTree, Mask) of
	true -> get_largest_family(T, Oid, add_res(length(SubTree), SubTree,
						   Type, Res));
	false -> get_largest_family(T, Oid, Res)
    end;
get_largest_family([], Oid, Res) -> Res.

%%-----------------------------------------------------------------
%% We keep only the largest (first longest SubTree, and then 
%% lexicografically greatest) SubTree.
%%-----------------------------------------------------------------
add_res(Len, SubTree, Type, undefined) ->
    {Len, SubTree, Type};
add_res(Len, SubTree, Type, {MaxLen, MaxS, MaxT}) when Len > MaxLen ->
    {Len, SubTree, Type};
add_res(Len, SubTree, Type, {MaxLen, MaxS, MaxT}) when Len == MaxLen ->
    if
	SubTree > MaxS -> {Len, SubTree, Type};
	true -> {MaxLen, MaxS, MaxT}
    end;
add_res(_Len, _SubTree, _Type, MaxRes) -> MaxRes.


%% 1 in mask is exact match, 0 is wildcard.
%% If mask is shorter than SubTree, its regarded
%% as being all ones.
check_mask(Oid, [], Mask) -> true;
check_mask([X | Xs], [X | Ys], [1 | Ms]) ->
    check_mask(Xs, Ys, Ms);
check_mask([X | Xs], [X | Ys], []) ->
    check_mask(Xs, Ys, []);
check_mask([X | Xs], [Y | Ys], [0 | Ms]) ->
    check_mask(Xs, Ys, Ms);
check_mask(_, _, _) -> false.

%%-----------------------------------------------------------------
%% Validates all oids in the Varbinds list towards the MibView.
%%-----------------------------------------------------------------
validate_all_mib_view([#varbind{oid = Oid, org_index = Index} | Varbinds],
		      MibView) ->
    case validate_mib_view(Oid, MibView) of
	true -> validate_all_mib_view(Varbinds, MibView);
	false -> {false, Index}
    end;
validate_all_mib_view([], MibView) ->
    true.

%%-----------------------------------------------------------------
%% This function is used to optimize the next operation in
%% snmp_mib_data. If we get to a node in the tree where we can
%% determine that we are guaranteed to be outside the mibview,
%% we don't have to continue the search in the that tree (Actually
%% we will, because we only check this at leafs. But we won't
%% go into tables or subagents, and that's the important
%% optimization.) For now, this function isn't that sophisticated;
%% it just checks that there is really no family in the mibview
%% that the Oid (or any other oids with Oid as prefix) may be
%% included in. Perhaps this function easily could be more
%% intelligent.
%%-----------------------------------------------------------------
is_definitely_not_in_mib_view(Oid, [{SubTree, Mask,?intViewType_included}|T]) ->
    case check_maybe_mask(Oid, SubTree, Mask) of
	true -> false;
	false -> is_definitely_not_in_mib_view(Oid, T)
    end;
is_definitely_not_in_mib_view(Oid, [{SubTree, Mask,?intViewType_excluded}|T]) ->
    is_definitely_not_in_mib_view(Oid, T);
is_definitely_not_in_mib_view(_Oid, []) ->
    true.
    
%%-----------------------------------------------------------------
%% As check_mask, BUT if Oid < SubTree and sofar good, we
%% return true. As Oid get larger we may decide.
%%-----------------------------------------------------------------
check_maybe_mask(Oid, [], Mask) -> true;
check_maybe_mask([X | Xs], [X | Ys], [1 | Ms]) ->
    check_maybe_mask(Xs, Ys, Ms);
check_maybe_mask([X | Xs], [X | Ys], []) ->
    check_maybe_mask(Xs, Ys, []);
check_maybe_mask([X | Xs], [Y | Ys], [0 | Ms]) ->
    check_maybe_mask(Xs, Ys, Ms);
check_maybe_mask([X | Xs], [Y | Ys], _) ->
    false;
check_maybe_mask(_, _, _) -> 
    true.
