%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_conf).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').
-revision('$Revision: 1.4 $').
-date('$Date: 1997/08/20 14:13:57 $').

-include("SNMPv2-TC.hrl").
-include("SNMPv2-TM.hrl").
-include("SNMPv2-MIB.hrl").
-include("OTP-SNMPEA-MIB.hrl").
-include("snmp_types.hrl").

%%%-----------------------------------------------------------------
%%% Handles the configuration files.
%%%-----------------------------------------------------------------

%% External exports
-export([read_internal_config_files/1, read_agent/1, read_standard/1,
	 read_target_config_files/1, read_notify_config_files/1]).
-export([check_mandatory/2]).

%% Internal exports
-export([check_community/1, check_trap/1, check_view/1, check_address/1]).
-export([check_target_addr/1, check_target_params/1]).
-export([check_notify/1]).
-export([check_agent/1, check_standard/1]).

%%-----------------------------------------------------------------
%% Func: read_internal_config_files/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads the configuration files.
%% Returns: A tuple with lists of rows for each
%%          table, {Communities, Traps, Views, Addresses}
%% Fails: If an error occurs, the process will die with Reason
%%        configuration_error.
%%-----------------------------------------------------------------
read_internal_config_files(Dir) ->
    Communities = snmp_misc:read(lists:append(Dir, "community.conf"), 
				 {snmp_conf, check_community}),
    Views = snmp_misc:read(lists:append(Dir, "view.conf"), 
			   {snmp_conf, check_view}),
    Addresses = snmp_misc:read(lists:append(Dir, "address.conf"), 
			       {snmp_conf, check_address}),
    {Communities, Views, Addresses}.

read_target_config_files(Dir) ->
    TAFile = lists:append(Dir, "target_addr.conf"),
    case file:read_file_info(TAFile) of
	{ok, _} -> ok;
	{error, _} -> convert_old(Dir)
    end,
    Addrs = snmp_misc:read(TAFile,
			   {snmp_conf, check_target_addr}),
    Params = snmp_misc:read(lists:append(Dir, "target_params.conf"), 
			    {snmp_conf, check_target_params}),
    {Addrs, Params}.
    
read_notify_config_files(Dir) ->
    Notifs = snmp_misc:read(lists:append(Dir, "notify.conf"), 
			    {snmp_conf, check_notify}),
    Notifs.

%%-----------------------------------------------------------------
%% Func: read_agent/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads the agent configuration files.
%% Returns: A list of {Variable, Value}.
%%-----------------------------------------------------------------
read_agent(Dir) ->
    Agent = snmp_misc:read(lists:append(Dir, "agent.conf"), 
			   {snmp_conf, check_agent}),
    sort_agent(Agent).

%%-----------------------------------------------------------------
%% Make sure that each mandatory agent attribute is present, and
%% provide default values for the other non-present attributes.
%%-----------------------------------------------------------------
sort_agent(L) ->
    L2 = check_mandatory(L, [{intAgentIpAddress, mandatory},
			     {intAgentUDPPort, mandatory},
			     {intAgentMaxPacketSize, mandatory}]),
    lists:keysort(1, L2).

%%-----------------------------------------------------------------
%% Func: read_standard/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads th standard configuration file.
%% Returns: A list of standard variables
%% Fails: If an error occurs, the process will die with Reason
%%        configuration_error.
%%-----------------------------------------------------------------
read_standard(Dir) ->
    Standard = snmp_misc:read(lists:append(Dir, "standard.conf"), 
			      {snmp_conf, check_standard}),
    sort_standard(Standard).

%%-----------------------------------------------------------------
%% Make sure that each mandatory standard attribute is present, and
%% provide default values for the other non-present attributes.
%%-----------------------------------------------------------------
sort_standard(L) ->
    L2 = check_mandatory(L, [{sysContact, {value, ""}},
			     {sysDescr, {value, ""}},
			     {sysLocation, {value, ""}},
			     {sysName, {value, ""}},
			     {sysObjectID, mandatory},
			     {sysServices, mandatory},
			     {snmpEnableAuthenTraps, mandatory}]),
    lists:keysort(1, L2).

check_mandatory(L, [{Key, Value}|T]) ->
    case lists:keymember(Key, 1, L) of
	true -> check_mandatory(L, T);
	false when Value == mandatory -> 
	    error("No value specified for mandatory attribute ~w", [Key]);
	false ->
	    {value, V} = Value,
	    check_mandatory([{Key, V} | L], T)
    end;
check_mandatory(L, []) -> L.

%%-----------------------------------------------------------------
%%  Community
%%  {ManagerIpAddress, CommunityString, ViewIndex, Access}.
%%-----------------------------------------------------------------
check_community({Ip, CommunityString, ViewIndex, Access}) ->
    check_ip(Ip),
    check_string(CommunityString),
    check_integer(ViewIndex),
    {ok, AccessVal} =
	check_atom(Access, [{read, ?intCommunityAccess_read},
			    {readWrite, ?intCommunityAccess_readWrite}]),
    {ok, {Ip, CommunityString, ViewIndex, AccessVal, ?'RowStatus_active'}};
check_community(X) ->
    {invalid_community, X}.


%%-----------------------------------------------------------------
%%  View
%%  {viewIndex, viewSubtree, viewType, viewMask}.
%%-----------------------------------------------------------------
check_view({ViewIndex, Subtree, Type, Mask}) ->
    check_integer(ViewIndex),
    check_oid(Subtree),
    {ok, TypeVal} =
	check_atom(Type, [{included, ?intViewType_included},
			  {excluded, ?intViewType_excluded}]),
    case catch check_atom(Mask, [{null, []}]) of
	{ok, X} -> MaskVal = X;
	_  -> 
	    check_oid(Mask),
	    MaskVal = Mask
    end,
    {ok, {ViewIndex, Subtree, MaskVal,
	  TypeVal, ?'StorageType_volatile', ?'RowStatus_active'}};
check_view(X) ->
    {invalid_view, X}.


%%-----------------------------------------------------------------
%%  Address
%%  {ManagerIpAddress, UDPPort, MaxPacketSize}.
%%-----------------------------------------------------------------
check_address({Ip, UDP, PacketSize}) ->
    check_ip(Ip),
    check_integer(UDP),
    check_packet_size(PacketSize),
    {ok, {Ip, UDP, PacketSize, ?'RowStatus_active'}};
check_address(X) ->
    {invalid_address, X}.

%%-----------------------------------------------------------------
%%  Agent
%%  {Name, Value}.
%%-----------------------------------------------------------------
check_agent({intAgentIpAddress, Value}) -> check_ip(Value);
check_agent({intAgentMaxPacketSize, Value}) -> check_integer(Value);
check_agent({intAgentUDPPort, Value}) -> check_integer(Value);
check_agent(X) -> {invalid_agent_attribute, X}.
    
%%-----------------------------------------------------------------
%%  TargetAddr
%%  {Name, Ip, Udp, Timeout, RetryCount, TagList, Params}
%%-----------------------------------------------------------------
check_target_addr({Name, Ip, Udp, Timeout, RetryCount, TagList, Params}) ->
    check_string(Name),
    check_ip(Ip),
    check_integer(Udp),
    check_integer(Timeout),
    if
	Timeout < 0 -> throw({invalid_timeout, Timeout});
	true -> ok
    end,
    check_integer(RetryCount),
    if
	RetryCount < 0 -> throw({invalid_retry_count, RetryCount});
	true -> ok
    end,
    check_string(TagList),
    check_string(Params),
    {ok, {Name, ?snmpUDPDomain, Ip ++ [Udp div 256, Udp rem 256], Timeout,
	  RetryCount, TagList, Params,
	  ?'StorageType_volatile', ?'RowStatus_active'}};
check_target_addr(X) ->
    {invalid_target_addr, X}.

%%-----------------------------------------------------------------
%%  TargetParams
%%  {Name, MPModel, SecurityModel, SecurityName, SecurityLevel}
%%-----------------------------------------------------------------
check_target_params({Name, MPModel, SecModel, SecName, SecLevel}) ->
    check_string(Name),
    {ok, MP} = check_atom(MPModel, [{v1, 0},
				     {v2c, 1}]),
    {ok, SecM} = check_atom(SecModel, [{v1, 1},
				       {v2c, 2},
				       {otp_sec, ?OTP_SEC}]),
    check_string(SecName),
    {ok, SecL} = check_atom(SecLevel, [{noAuthNoPriv, 1},
				       {authNoPriv, 2},
				       {authPriv, 3}]),
    {ok, {Name, MP, SecM, SecName, SecL, 
	  ?'StorageType_volatile', ?'RowStatus_active'}};
check_target_params(X) ->
    {invalid_target_params, X}.

%%-----------------------------------------------------------------
%%  Notify
%%  {Name, Tag, Type}
%%-----------------------------------------------------------------
check_notify({Name, Tag, Type}) ->
    check_string(Name),
    check_string(Tag),
    {ok, Val} = check_atom(Type, [{trap, 1},
				  {inform, 2}]),
    {ok, {Name, Tag, Type,
	  ?'StorageType_volatile', ?'RowStatus_active'}};
check_notify(X) ->
    {invalid_notify, X}.

%%-----------------------------------------------------------------
%%  Standard
%%  {Name, Value}.
%%-----------------------------------------------------------------
check_standard({sysDescr, Value}) -> check_string(Value);
check_standard({sysObjectID, Value}) -> check_oid(Value);
check_standard({sysContact, Value}) -> check_string(Value);
check_standard({sysName, Value}) -> check_string(Value);
check_standard({sysLocation, Value}) -> check_string(Value);
check_standard({sysServices, Value}) -> check_integer(Value);
check_standard({snmpEnableAuthenTraps, Value}) -> 
    {ok, Val} = check_atom(Value, [{enabled, ?snmpEnableAuthenTraps_enabled},
				   {disabled, ?snmpEnableAuthenTraps_disabled}]),
    {ok, {snmpEnableAuthenTraps, Val}};
check_standard({Attrib, Value}) -> {unknown_attribute, Attrib};
check_standard(X) -> {invalid_standard_specification, X}.
    

check_integer(X) when integer(X) -> true;
check_integer(X) -> throw({invalid_integer, X}).

check_string(X) when list(X) -> true;
check_string(X) -> throw({invalid_string, X}).

check_oid([E1,E2|R]) when E1 * 40 + E2 =< 255 ->
    X = [E1,E2|R],
    case all_integer(X) of
	true -> true;
	_ -> throw({invalid_object_identifier, X})
    end;
check_oid(X) -> throw({invalid_object_identifier, X}).

all_integer([H|T]) when integer(H) -> all_integer(T);
all_integer([H|T]) -> false;
all_integer([]) -> true.

check_ip(X) ->
    case catch all_integer(X) of
	true when length(X) == 4 -> true;
	_ -> throw({invalid_ip_address, X})
    end.

check_packet_size(X) when integer(X), X >= 484, X =< 65535 -> true;
check_packet_size(X) -> throw({invalid_packet_size, X}).

check_atom(X, Atoms) ->
    case lists:keysearch(X, 1, Atoms) of
	{value, {X, Val}} -> {ok, Val};
	_ -> throw({"Expected", Atoms, "got", X})
    end.

error(Format, Error) ->
    snmp_error:config_err(Format, Error),
    exit(configuration_error).
error(File, Line, Error) ->
    snmp_error:config_err("~p:~w: ~p", [File, Line, Error]),
    exit(configuration_error).


%%-----------------------------------------------------------------
%% Converts old trap_dest.conf to the new target*.conf
%%-----------------------------------------------------------------
convert_old(Dir) ->
    error_logger:info_msg("snmp: Converting old style trap_dest file "
			  "to new target and notify files.\n", []),
    case catch c_old(Dir) of
	ok ->
	    ok;
	Else ->
	    file:delete("target_addr.conf"),
	    file:delete("target_params.conf"),
	    file:delete("notify.conf"),
	    {Reason, Str} = case Else of
				{error, R, S} -> {R, S};
				_ -> {Else, io_lib:format("~p", [Else])}
			    end,
	    error_logger:error_msg("snmp: conversion failed. ~s\n", [Str]),
	    exit({conversion_failed, Reason})
    end.

c_old(Dir) ->
    Traps = snmp_misc:read(lists:append(Dir, "trap_dest.conf"), 
			   {snmp_conf, check_trap}),
    Addresses = snmp_misc:read(lists:append(Dir, "address.conf"), 
			       {snmp_conf, check_address}),
    
    %% Trap = {CommunityString, Ip, ?'RowStatus_active', VsnVal}}
    %% Address = {Ip, UDP, PacketSize, ?'RowStatus_active'}}

    SortedTraps = sort_traps(Traps, []),
    Communities = get_communities(Traps, []),
    
    wr_addr(Dir, SortedTraps, Addresses),
    wr_params(Dir, [1,2]),  % always write both of them
    wr_notify(Dir, Communities).

%%-----------------------------------------------------------------
%%  Trap
%%  {CommunityString, ManagerIpAddress}.
%%-----------------------------------------------------------------
check_trap({CommunityString, Ip, Vsn}) ->
    check_ip(Ip),
    check_string(CommunityString),
    {ok, VsnVal} =
	check_atom(Vsn, [{v1, 1},
			 {v2, 2}]),
    {ok, {CommunityString, Ip, ?'RowStatus_active', VsnVal}};
check_trap({CommunityString, Ip}) ->
    % Old format
    check_ip(Ip),
    check_string(CommunityString),
    {ok, {CommunityString, Ip, ?'RowStatus_active', 1}};
check_trap(X) ->
    {invalid_trap, X}.

%% Ret: [{ {Ip,Vsn}, [Community] }]
sort_traps([{Comm, Ip, _, Vsn} | T], Res) ->
    case lists:keymember({Ip, Vsn}, 1, Res) of
	false ->
	    Comms = [Comm | [C1 || {C1, I, _, V} <- T,
				   I == Ip,
				   V == Vsn]],
	    sort_traps(T, [{{Ip, Vsn},  Comms} | Res]);
	true ->
	    sort_traps(T, Res)
    end;
sort_traps([], Res) ->
    Res.
    
get_communities([{Comm, _, _, _} | T], Res) ->
    case lists:member(Comm, Res) of
	true ->
	    get_communities(T, Res);
	false ->
	    get_communities(T, [Comm | Res])
    end;
get_communities([], Res) ->
    Res.

%% {Name, Ip, Udp, Timeout, RetryCount, TagList, ParamsName}
wr_addr(Dir, SortedTraps, Addresses) ->
    {ok, Fid} = file:open(filename:join(Dir,"target_addr.conf"),write),
    io:format(Fid, "~s\n", [header()]),
    lists:foreach(fun({{Ip, Vsn}, Communities}) ->
			  UDP = get_udp(Ip, Addresses),
			  AA = mk_ip(Ip, Vsn),
			  BB = mk_tag_list(Communities, []),
			  ok = io:format(Fid, "{\"~s\", ~w, ~w, 1500, 3, "
					 "\"~s\", \"~s\"}.~n",
					 [mk_ip(Ip, Vsn), Ip, UDP,
					  mk_tag_list(Communities, []),
					  mk_param(Vsn)])
		  end,
		  SortedTraps),
    file:close(Fid).

get_udp(Ip, Addresses) ->
    case lists:keysearch(Ip, 1, Addresses) of
	{value, {_, UDP, _, _}} ->
	    UDP;
	false ->
	    Str = io_lib:format("No UDP port found for ~w in address.conf",
				[Ip]),
	    throw({error, {no_udp_port, Ip}, Str})
    end.

mk_ip([A,B,C,D], Ver) ->
    io_lib:format("~w.~w.~w.~w v~w", [A,B,C,D,Ver]).

mk_param(1) -> "otp_v1";
mk_param(2) -> "otp_v2".
     
mk_tag_list([Community | T], Res) ->
    mk_tag_list(T, [32 | mk_tag(Community)] ++  Res);
mk_tag_list([], []) ->
    [];
mk_tag_list([], Res) ->
    tl(Res).

mk_tag([32 | T]) -> [$_ | mk_tag(T)];
mk_tag([9 | T]) -> [$_ | mk_tag(T)];
mk_tag([13 | T]) -> [$_ | mk_tag(T)];
mk_tag([11 | T]) -> [$_ | mk_tag(T)];
mk_tag([Char | T]) -> [Char | mk_tag(T)];
mk_tag([]) -> [].

wr_params(Dir, Vers) ->
    {ok, Fid} = file:open(filename:join(Dir,"target_params.conf"),write),
    io:format(Fid, "~s\n", [header()]),
    lists:foreach(fun(Ver) ->
			  SM = if Ver == 1 -> v1;
				  Ver == 2 -> v2c
			       end,
			  ok = io:format(Fid, "{\"otp_v~w\", ~w, otp_sec, "
					 "\"\", noAuthNoPriv}.~n",
					 [Ver, SM])
		  end,
		  Vers),
    file:close(Fid).

wr_notify(Dir, Communities) ->
    {ok, Fid} = file:open(filename:join(Dir,"notify.conf"),write),
    io:format(Fid, "~s\n", [header()]),
    lists:foreach(fun(Community) ->
			  ok = io:format(Fid, "{\"~s\", \"~s\", trap}.~n",
					 [Community, mk_tag(Community)])
		  end,
		  Communities).


header() ->
    {Y,Mo,D} = date(),
    {H,Mi,S} = time(),
    io_lib:format("%% This file was automatically converted from old style\n"
		  "%% trap_dest.conf by snmp_config (v~s)  ~w-~w-~w ~w:~w:~w~n",
		  [?version,Y,Mo,D,H,Mi,S]).
