%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_internal_mib).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').
-revision('$Revision: 1.2 $').
-date('$Date: 1997/06/24 07:44:40 $').

-include("snmp_types.hrl").
-include("OTP-SNMPEA-MIB.hrl").
-include("STANDARD-MIB.hrl").

%%%-----------------------------------------------------------------
%%% This module implements the init- configure-functions
%%% for the OTP-SNMPEA-MIB.  
%%%-----------------------------------------------------------------
%% External exports
-export([init/0, configure/1]).
-export([get_mms/1, get_address_info/1]).
-export([intCommunityTable/1, intCommunityTable/3,
	 intAddressTable/1, intAddressTable/3,
	 intViewTable/1, intViewTable/3,
	 intAgentUDPPort/1, intAgentUDPPort/2,
	 intAgentIpAddress/1, intAgentIpAddress/2,
	 intAgentMaxPacketSize/1, intAgentMaxPacketSize/2,
	 table_next/2, check_status/3]).

%% Internal exports
-export([init_var/1]).

%%-----------------------------------------------------------------
%% Func: get_mms(ManagerIp) -> MMS
%% Types: ManagerIp = [int(), int), int(), int()]
%%        MMS = int()
%% Purpose: Gets the maximum message size for use when the agent
%%          communicates with the manager ManagerIp (i.e. the
%%          minimum of the intAgentMaxPacketSize and the max packet
%%          size for the manager).
%%-----------------------------------------------------------------
get_mms(ManagerIp) ->
    {value, AgentMax} = snmp_internal_mib:intAgentMaxPacketSize(get),
    case intAddressTable(get, ManagerIp, [?intAddressMaxPacketSize]) of
	[{value, MgrMax}] when MgrMax =< AgentMax -> MgrMax;
	_ -> AgentMax
    end.

%%-----------------------------------------------------------------
%% Func: get_address_info(ManagerIp) -> {UdpPort, MMS} | false
%% Types: ManagerIp = [int(), int), int(), int()]
%%        UdpPort = int()
%%        MMS = int()
%% Purpose: Gets the UDP port and MMS for the ManagerIp.
%%-----------------------------------------------------------------
get_address_info(ManagerIp) ->
    case check_status(intAddressTable, ManagerIp, ?intAddressStatus) of
	true ->
	    case intAddressTable(get, ManagerIp,
				 [?intAddressUDPPort,
				  ?intAddressMaxPacketSize]) of
		[{value, UdpPort}, {value, Max}] ->
		    {UdpPort, Max};
		_ ->
		    false
	    end;
	false ->
	    false
    end.

%%-----------------------------------------------------------------
%% Func: init/0
%% Purpose: Creates the tables and variables necessary for the SNMP
%%          mechanism to work properly.
%%          Note that this function won't destroy any old values.
%%          This function should be called only once.
%%-----------------------------------------------------------------
init() ->
    maybe_create_table(intCommunityTable),
    maybe_create_table(intViewTable),
    maybe_create_table(intAddressTable).

%%-----------------------------------------------------------------
%% Func: configure/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads the config-files for the internal tables, and
%%          inserts the data.  Makes sure that all old data in
%%          the tables are deleted, and the new data inserted.
%%          This function makes sure that all (and only) 
%%          config-file-data are in the tables. 
%% Returns: ok
%% Fails: exit(configuration_error)
%% PRE: init/1 has been successfully called
%%-----------------------------------------------------------------
configure(Dir) ->
    {Communities, Views, Addresses} = 
	snmp_conf:read_internal_config_files(Dir),
    Agent = snmp_conf:read_agent(Dir),
    init_vars(Agent),
    init_tabs(Communities, Views, Addresses),
    ok.

init_vars(Agent) ->
    lists:map({snmp_internal_mib, init_var}, [], Agent).

maybe_create_table(Name) ->
    case snmp_local_db:table_exists(db(Name)) of
	true -> ok;
	_ -> snmp_local_db:table_create(db(Name))
    end.

init_var({Var, Val}) ->
    snmp_generic:variable_set(Var, Val).    

init_tabs(Communities, Views, Addresses) ->
    snmp_local_db:table_delete(db(intCommunityTable)),
    snmp_local_db:table_create(db(intCommunityTable)),
    init_community_table(Communities),
    snmp_local_db:table_delete(db(intAddressTable)),
    snmp_local_db:table_create(db(intAddressTable)),
    init_address_table(Addresses),
    snmp_local_db:table_delete(db(intViewTable)),
    snmp_local_db:table_create(db(intViewTable)),
    init_view_table(Views).
    
init_community_table([Row | T]) ->
    Key1 = element(1, Row),
    Key2 = element(2, Row),
    Key = lists:append(Key1, [length(Key2) | Key2]),
    snmp_local_db:table_create_row(db(intCommunityTable), Key, Row),
    init_community_table(T);
init_community_table([]) -> true.

init_view_table([Row | T]) ->
    Key1 = element(1, Row),
    Key2 = element(2, Row),
    snmp_local_db:table_create_row(db(intViewTable), [Key1, length(Key2) |Key2],
				  Row),
    init_view_table(T);
init_view_table([]) -> true.

init_address_table([Row | T]) ->
    Key = element(1, Row),
    snmp_local_db:table_create_row(db(intAddressTable), Key, Row),
    init_address_table(T);
init_address_table([]) -> true.

%%-----------------------------------------------------------------
%% Instrumentation functions
%% Retreive functions are also used internally by the agent, so
%% don't change the interface without changing those functions.
%% Note that if these functions implementations are changed,
%% an error can make the agent crash, as no error detection is
%% performed for the internal data.
%% These functions cannot use the default functions as is, because
%% the default functions rely on that the mib is loaded, and
%% these functions must work even if the OTP-SNMPEA-MIB isn't loaded.
%% So we hardcode the information necessary for the functions
%% called by the default functions in snmp_generic. This info is
%% normally provided my the mib compiler, and inserted into 
%% snmp_symbolic_store at load mib time.
%%-----------------------------------------------------------------
%% Op == new | delete
intCommunityTable(Op) ->
    snmp_generic:table_func(Op, db(intCommunityTable)).

%% Op == get | is_set_ok | set | get_next
intCommunityTable(get, RowIndex, Cols) ->
    get(intCommunityTable, RowIndex, Cols);
intCommunityTable(get_next, RowIndex, Cols) ->
    next(intCommunityTable, RowIndex, Cols);
intCommunityTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(intCommunityTable)).

intAddressTable(Op) ->
    snmp_generic:table_func(Op, db(intAddressTable)).
intAddressTable(get_next, RowIndex, Cols) ->
    next(intAddressTable, RowIndex, Cols);
intAddressTable(get, RowIndex, Cols) ->
    get(intAddressTable, RowIndex, Cols);
intAddressTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(intAddressTable)).

intViewTable(Op) ->
    snmp_generic:table_func(Op, db(intViewTable)).
intViewTable(get_next, RowIndex, Cols) ->
    next(intViewTable, RowIndex, Cols);
intViewTable(get, RowIndex, Cols) ->
    get(intViewTable, RowIndex, Cols);
intViewTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(intViewTable)).

table_next(Name, RestOid) ->
    snmp_generic:table_next(db(Name), RestOid).


check_status(Name, Indexes, StatusNo) ->
    case snmp_local_db:table_get_element(db(Name), Indexes, StatusNo) of
	{value, ?'RowStatus_active'} -> true;
	_ -> false
    end.

db(X) -> {X, volatile}.

fa(intCommunityTable) -> 3;
fa(intAddressTable) -> 1;
fa(intViewTable) -> 3.
 
foi(intCommunityTable) -> 1;
foi(intAddressTable) -> 1;
foi(intViewTable) -> 1.
 
noc(intCommunityTable) -> 5;
noc(intAddressTable) -> 4;
noc(intViewTable) -> 6.
 
next(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_next(db(Name), RowIndex, Cols,
                                   fa(Name), foi(Name), noc(Name)).
 
get(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_get(db(Name), RowIndex, Cols, foi(Name)).

%% Op == new | delete | get
intAgentUDPPort(Op) ->
    snmp_generic:variable_func(Op, db(intAgentUDPPort)).

%% Op == is_set_ok | set
intAgentUDPPort(Op, Val) ->
    snmp_generic:variable_func(Op, Val, db(intAgentUDPPort)).

intAgentIpAddress(Op) ->
    snmp_generic:variable_func(Op, db(intAgentIpAddress)).
intAgentIpAddress(Op, Val) ->
    snmp_generic:variable_func(Op, Val, db(intAgentIpAddress)).

intAgentMaxPacketSize(Op) ->
    snmp_generic:variable_func(Op, db(intAgentMaxPacketSize)).
intAgentMaxPacketSize(Op, Val) ->
    snmp_generic:variable_func(Op, Val, db(intAgentMaxPacketSize)).
