%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_log).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-export([create_log/0, log/4]).
-export([log_to_txt/2, log_to_txt/3]).

-include("snmp_types.hrl").

-define(log_name, "snmp log").
-define(log_file, "snmp.log").

%%%-----------------------------------------------------------------
%%% This module contains Audit Trail Logging functions.
%%%-----------------------------------------------------------------

%%-----------------------------------------------------------------
%% Func: create_log() -> {ok, Log} | false | {error, Error}
%% Types: Log = log()
%% Purpose: Checks the snmp application's parameters to see
%%          if a log should be created, and creates the log.
%%-----------------------------------------------------------------
create_log() ->
    LogType = case application:get_env(snmp, audit_trail_log) of
		  {ok, write_log} -> w_log;
		  {ok, read_write_log} -> rw_log;
		  _ -> false
	      end,
    case application:get_env(snmp, audit_trail_log_dir) of
	{ok, LogDir} when list(LogDir), LogType /= false ->
	    case application:get_env(snmp, audit_trail_log_size) of
		{ok, Size} ->
		    case disk_log:open([{name, ?log_name},
					{file, filename:join(LogDir,?log_file)},
					{type, wrap},
					{format, internal},
					{size, Size}]) of
			{error, Reason} ->
			    {error, {snmp, disk_log_error, Reason}};
			_ ->
			    {ok, LogType}
		    end;
		_ ->
		    {error, {snmp, undefined_parameter, audit_trail_log_size}}
	    end;
	_ when LogType == false ->
	    false;
	{ok, Dir1} ->
	    {error, {snmp, bad_type, {audit_trail_log_dir, Dir1}}};
	_ ->
	    {error, {snmp, undefined_parameter, audit_trail_log_dir}}
    end.

log(w_log, 'set-request', Packet, Addr) ->
    disk_log:alog(?log_name, {Packet, Addr});
log(w_log, _, _, _) -> ok;
log(_, _, Packet, Addr) ->
    disk_log:alog(?log_name, {Packet, Addr}).
     

log_to_txt(LogDir, Mibs) -> log_to_txt(LogDir, Mibs, "./snmp_log.txt").
log_to_txt(LogDir, Mibs, TxtFile) ->
    case disk_log:open([{name, ?log_name},
			{file, filename:join(LogDir, ?log_file)},
			{type, wrap},
			{format, internal},
			%% We should open the log in read-only mode without
			%% size.
			{size, {100000,255}}]) of
	{ok, _} ->
	    Res = case file:open(TxtFile, write) of
		      {ok, Fd} ->
			  Mib = snmp_misc:make_mini_mib(Mibs),
			  Res2 = (catch loop(disk_log:chunk(?log_name, start),
					    Mib, Fd)),
			  file:close(Fd),
			  Res2;
		      {error, Reason} ->
			  {error, {TxtFile, Reason}}
		  end,
	    disk_log:close(?log_name),
	    Res;
	{error, Reason} ->
	    {error, {?log_name, Reason}}
    end.

loop({Cont, Terms}, Mib, Fd) ->
    lists:foreach(fun(X) -> write(X, Mib, Fd) end, Terms),
    loop(disk_log:chunk(?log_name, Cont), Mib, Fd);
loop(eof, _Mib, _Fd) ->
    ok;
loop(Error, _Mib, _Fd) ->
    Error.
			  
write({Packet, Address}, Mib, Fd) ->
    case catch snmp_pdus:dec_message(binary_to_list(Packet)) of
	Msg when record(Msg, message) ->
	    w(Msg, Address, Mib, Fd);
	{'EXIT', R} ->
	    io:format(Fd, "** error in log file ~p\n\n", [R])
    end.

w(#message{version = Vsn, community = Community, data = Pdu}, Addr, Mib, Fd) ->
    Str = snmp_misc:format_pdu(Pdu, Mib),
    if
	record(Pdu, trappdu) ->
	    w_trap(Vsn, Community, Str, Addr, Fd);
	Pdu#pdu.type == 'snmpv2-trap' ->
	    w_trap(Vsn, Community, Str, Addr, Fd);
	Pdu#pdu.type == 'get-response' ->
	    w_response(Vsn, Community, Str, Addr, Fd);
	true ->
	    w_request(Vsn, Community, Str, Addr, Fd)
    end.

w_request(Vsn, Community, Str, {Ip, Port}, Fd) ->
    io:format(Fd, "Got ~w message from ~s:~w, community: ~s\n~s",
	      [Vsn, ip(Ip), Port, Community, Str]).

w_response(Vsn, Community, Str, {Ip, Port}, Fd) ->
    io:format(Fd, "Sent ~w response to ~s:~w, community: ~s\n~s",
	      [Vsn, ip(Ip), Port, Community, Str]).
    
w_trap('version-1', Community, Str, Addrs, Fd) ->
    io:format(Fd, "Sent 'version-1' trap to ~s, community: ~s\n~s",
	      [addr(Addrs), Community, Str]);
w_trap('version-2', Community, Str, Addrs, Fd) ->
    io:format(Fd, "Sent 'version-2' trap to ~s, community: ~s\n~s",
	      [addr(Addrs), Community, Str]).

addr(Addrs) -> addr(Addrs, "").
addr([{snmpUDPDomain, {Ip, Port}}], Str) ->
    [Str, io_lib:format("~s:~w", [ip(Ip), Port])];
addr([{snmpUDPDomain, {Ip, Port}} | Addrs], Str) ->
    addr(Addrs, [io_lib:format("~s:~w, ", [ip(Ip), Port])]);
%% Handle old format as well...
addr([{snmpUDPDomain, Ip, Port}], Str) ->
    [Str, io_lib:format("~s:~w", [ip(Ip), Port])];
addr([{snmpUDPDomain, Ip, Port} | Addrs], Str) ->
    addr(Addrs, [io_lib:format("~s:~w, ", [ip(Ip), Port])]);
addr([], Str) ->
    Str.

ip({A,B,C,D}) ->
    io_lib:format("~w.~w.~w.~w", [A,B,C,D]).
