%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_mpc).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-export([init_mpc/1, reinit/0, process_packet/4, discarded_pdu/1,
	 generate_response_msg/4, generate_msg/4]).

-include("snmp_types.hrl").

-define(empty_msg_size, 24).

-record(state, {v1, v2c, v3}).

%%%-----------------------------------------------------------------
%%% This module implemets the Message Processing and Control part of
%%% the multi-lingual SNMP agent.
%%%
%%% The MPC is responsible for:
%%%   *) security issues (auth/priv).
%%%   *) decoding the message into a PDU.
%%%   *) decide a suitable Access Control Model, and provide it with
%%%      the data it needs.
%%%   *) maintaining SNMP counters.
%%%
%%% In order to take care of the different versions of counters, it
%%% implements and maintains the union of all SNMP counters (i.e. from
%%% rfc1213 and from rfc1907).  It is up to the administrator of the
%%% agent to load the correct MIB.  Note that this module implements
%%% the counters only, it does not provide instrumentation functions
%%% for the counters.
%%%-----------------------------------------------------------------
init_mpc(Opts) ->
    init_vars(),
    lists:foldl(fun(v1, S)  -> S#state{v1 = true};
		   (v2, S) -> S#state{v2c = true};
		   (v3, S)  -> S#state{v3 = true}
		end, #state{}, Opts).

%%-----------------------------------------------------------------
%% Purpose: We must calculate the length of a
%%          message with an empty Pdu, and zero-length community
%%          string.  This length is used to calculate the max
%%          pdu size allowed for each request. This size is 
%%          dependent on two dynamic fields, the community string
%%          and the pdu (varbinds actually). It is calculated
%%          as EmptySize + length(CommunityString) + 4.
%%          We assume that the length of the CommunityString is
%%          less than 128 (thus requiring just one octet for the
%%          length field (the same as the zero-length community
%%          string)). 4 comes from the fact that the maximum pdu
%%          size needs 31 bits which needs 5 * 7 bits to be
%%          expressed. One 7bit octet is already present in the
%%          empty msg, leaving 4 more 7bit octets.
%% Actually, this function is not used, we use a constant instead.
%%-----------------------------------------------------------------
%% Ret: 24
%empty_msg() ->
%    M = #message{version = 'version-1', community = "", data = 
%		 #pdu{type = 'get-response', request_id = 1,
%		      error_status = noError, error_index = 0, varbinds = []}},
%    length(snmp_pdus:enc_message(M)) + 4.

%%-----------------------------------------------------------------
%% Func: process_packet(Packet, TDomain, TAddress, State) ->
%%       {ok, SnmpVsn, Pdu, PduMS, ACMData} | {discarded, Reason}
%% Types: Packet = binary()
%%        TDomain = snmpUDPDomain | atom()
%%        TAddress = {Ip, Udp}
%%        State = #state
%% Purpose: 
%%-----------------------------------------------------------------
process_packet(Packet, TDomain, TAddress, State) ->
    inc(snmpInPkts),
    case catch snmp_pdus:dec_message_only(binary_to_list(Packet)) of
	#message{version = 'version-1', community = Community, data = Data} 
	  when State#state.v1 == true ->
	    HS = ?empty_msg_size + length(Community),
	    v1_v2c_mpc('version-1', Community, TDomain, TAddress, Data, HS);
	#message{version = 'version-2', community = Community, data = Data}
	  when State#state.v2c == true ->
	    HS = ?empty_msg_size + length(Community),
	    v1_v2c_mpc('version-2', Community, TDomain, TAddress, Data, HS);
	{'EXIT', {bad_version, _Vsn}} ->
	    inc(snmpInBadVersions),
	    {discarded, snmpInBadVersions};
	{'EXIT', Reason} ->
	    inc(snmpInASNParseErrs),
	    {discarded, Reason};
	_Message ->
	    inc(snmpInBadVersions),
	    {discarded, snmpInBadVersions}
    end.

discarded_pdu(false) -> ok;
discarded_pdu(Variable) -> inc(Variable).

%%-----------------------------------------------------------------
%% Handles a Community based message (v1 or v2c).
%%-----------------------------------------------------------------
v1_v2c_mpc(Vsn, Community, snmpUDPDomain, {Ip, Udp}, Data, HS) ->
    PduMS = snmp_internal_mib:get_mms(tuple_to_list(Ip)) - HS,
    case catch snmp_pdus:dec_pdu(Data) of
	{'EXIT', Reason} ->
	    inc(snmpInASNParseErrs),
	    {discarded, Reason};
	Pdu when record(Pdu, pdu) ->
	    inc_snmp_in_vars(Pdu),
	    #pdu{request_id = ReqId} = Pdu,
	    case snmp_note_store:get_note({Ip, ReqId}) of
		undefined -> 
		    % Set the note _after_ pdu processing. This makes
		    % duplicated requests be ignored even if pdu
		    % processing took long time.
		    snmp_note_store:set_note(100, {Ip, ReqId}, true),
		    % Uses ACMData that snmp_acm knows of.
		    {ok, Vsn, Pdu, PduMS,
		     {community, Community, tuple_to_list(Ip)}};
		true ->
		    {discarded, duplicate_pdu}
	    end;
	TrapPdu ->
	    {discarded, trap_pdu}
    end;
v1_v2c_mpc(Vsn, Community, snmpUDPDomain, TAddress, Data, HS) ->
    {discarded, {badarg, TAddress}};
v1_v2c_mpc(Vsn, Community, TDomain, TAddress, Data, HS) ->
    {discarded, {badarg, TDomain}}.

%%-----------------------------------------------------------------
%% Executed when a response message should be generated.
%%-----------------------------------------------------------------
generate_response_msg(Vsn, RePdu, Type, {community, Community, Ip}) ->
    Message = #message{version = Vsn, community = Community, data = RePdu},
    case catch list_to_binary(snmp_pdus:enc_message(Message)) of
	{'EXIT', Reason} ->
	    snmp_error:user_err("~p (pdu: ~p, community: ~p)",
				[Reason, RePdu, Community]),
	    {discarded, encoding_error};
	Packet ->
	    MMS = snmp_internal_mib:get_mms(Ip),
	    case size(Packet) of
		Len when Len =< MMS ->
		    inc_snmp_cnt_vars(Type, RePdu),
		    inc_snmp_out_vars(RePdu),
		    {ok, Packet};
		Len ->
		    too_big(Vsn, RePdu, Community)
	    end
    end.

too_big(Vsn, Pdu, Community) when Pdu#pdu.type == 'get-response' ->
    ErrPdu =
	if 
	    Vsn == 'version-1' ->
		%% In v1, the varbinds should be identical to the incoming
		%% request.  It isn't identical now!
		%% Make acceptable (?) approximation.
		V = set_vb_null(Pdu#pdu.varbinds),
		Pdu#pdu{error_status = tooBig, error_index = 0, varbinds = V};
	    true ->
		%% In v2, varbinds should be empty (reasonable!)
		Pdu#pdu{error_status = tooBig, error_index = 0, varbinds = []}
	end,
    Message = #message{version = Vsn, community = Community, data = ErrPdu},
    case catch snmp_pdus:enc_message(Message) of
	{'EXIT', Reason} ->
	    snmp_error:user_err(Reason, []),
	    {discarded, Reason};
	Packet -> 
	    inc_snmp_out_vars(ErrPdu),
	    {ok, list_to_binary(Packet)}
    end;
too_big(Vsn, Pdu, _Community) ->
    snmp_error:user_err("Pdu ~p too big, not sent.", [Pdu]),
    {discraded, too_big}.

set_vb_null([Vb | Vbs]) ->
    [Vb#varbind{variabletype = 'NULL', value = 'NULL'} | set_vb_null(Vbs)];
set_vb_null([]) ->
    [].

%%-----------------------------------------------------------------
%% Executed when a message that isn't a response is generated, i.e.
%% a trap.
%%-----------------------------------------------------------------
generate_msg(Vsn, Pdu, {community, Community}, To) ->
    Message = #message{version = Vsn, community = Community, data = Pdu},
    case catch list_to_binary(snmp_pdus:enc_message(Message)) of
	{'EXIT', Reason} ->
	    snmp_error:user_err("~p (pdu: ~p, community: ~p)",
				[Reason, Pdu, Community]);
	Packet ->
	    {value, AgentMax} = snmp_internal_mib:intAgentMaxPacketSize(get),
	    case size(Packet) of
		Len when Len =< AgentMax ->
		    send_pdu_list(To, Packet, Len, Pdu, []);
		Len ->
		    {discarded, toobig}
	    end
    end.

send_pdu_list([{snmpUDPDomain, {ManagerIp, Udp}} | T], Packet, Len, Pdu, Res) ->
    case snmp_internal_mib:get_address_info(ManagerIp) of
	{_UdpPort, Max} when Max >= Len ->
	    % Sending from default UDP port
	    inc_snmp_out_vars(Pdu),
	    send_pdu_list(T, Packet, Len, Pdu,
			  [{snmpUDPDomain, {list_to_tuple(ManagerIp), Udp}} |
			   Res]);
	{_UdpPort, Max} ->
	    send_pdu_list(T, Packet, Len, Pdu, Res);
	_ ->
	    snmp_error:user_err("No Address info for Manager ~w "
				"in intAddressTable", [ManagerIp]),
	    send_pdu_list(T, Packet, Len, Pdu, Res)
    end;
send_pdu_list([], Packet, _Len, _Pdu, Res) ->
    {ok, Packet, Res}.

%%-----------------------------------------------------------------
%% Counter functions
%%-----------------------------------------------------------------
init_vars() -> lists:map(fun maybe_create_var/1, vars()).

reinit() -> lists:map(fun init_var/1, vars()).

maybe_create_var(Var) ->
    case ets:lookup(snmp_agent_table, Var) of
	[_] -> ok;
	_ -> init_var(Var)
    end.

init_var(Var) -> ets:insert(snmp_agent_table, {Var, 0}).

vars() ->
    [snmpInPkts,
     snmpOutPkts,
     snmpInBadVersions,
     snmpInBadCommunityNames,
     snmpInBadCommunityUses,
     snmpInASNParseErrs,
     snmpInTooBigs,
     snmpInNoSuchNames,
     snmpInBadValues,
     snmpInReadOnlys,
     snmpInGenErrs,
     snmpInTotalReqVars,
     snmpInTotalSetVars,
     snmpInGetRequests,
     snmpInGetNexts,
     snmpInSetRequests,
     snmpInGetResponses,
     snmpInTraps,
     snmpOutTooBigs,
     snmpOutNoSuchNames,
     snmpOutBadValues,
     snmpOutGenErrs,
     snmpOutGetRequests,
     snmpOutGetNexts,
     snmpOutSetRequests,
     snmpOutGetResponses,
     snmpOutTraps,
     snmpSilentDrops,
     snmpProxyDrops].
    
%%-----------------------------------------------------------------
%%  inc(VariableName) increments the variable (Counter) in
%%  the local mib. (e.g. snmpInPkts)
%%-----------------------------------------------------------------
inc(Name) -> ets:update_counter(snmp_agent_table, Name, 1).
inc(Name, N) -> ets:update_counter(snmp_agent_table, Name, N).

inc_snmp_in_vars(Pdu) ->
    inc_in_type(Pdu#pdu.type).

inc_snmp_cnt_vars(_, #pdu{error_status = ErrStat}) when ErrStat /= noError ->
    ok;
inc_snmp_cnt_vars('get-request', #pdu{varbinds = Vbs}) ->
    inc(snmpInTotalReqVars, length(Vbs));
inc_snmp_cnt_vars('get-next-request', #pdu{varbinds = Vbs}) ->
    inc(snmpInTotalReqVars, length(Vbs));
inc_snmp_cnt_vars('set-request', #pdu{varbinds = Vbs}) ->
    inc(snmpInTotalSetVars, length(Vbs));
inc_snmp_cnt_vars(_, _) ->
    ok.

inc_snmp_out_vars(Pdu) when record(Pdu, pdu) ->
    inc(snmpOutPkts),
    ErrorStatus = Pdu#pdu.error_status,
    inc_out_err(ErrorStatus),
    inc_out_vars_2(Pdu#pdu.type);
inc_snmp_out_vars(TrapPdu) when record(TrapPdu, trappdu) ->
    inc(snmpOutPkts),
    inc(snmpOutTraps).

inc_out_vars_2('get-response') -> inc(snmpOutGetResponses);
inc_out_vars_2('get-request') -> inc(snmpOutGetRequests);
inc_out_vars_2('get-next-request') -> inc(snmpOutGetNexts);
inc_out_vars_2('set-request') -> inc(snmpOutSetRequests);
inc_out_vars_2(_) -> ok.

inc_out_err(genErr) -> inc(snmpOutGenErrs);
inc_out_err(tooBig) -> inc(snmpOutTooBigs);
inc_out_err(noSuchName) -> inc(snmpOutNoSuchNames);
inc_out_err(badValue) -> inc(snmpOutBadValues);
% snmpOutReadOnlys is not used any more (rfc1213)
%inc_out_err(readOnly) -> inc(snmpOutReadOnlys);
inc_out_err(_) -> ok.

inc_in_type('get-request') -> inc(snmpInGetRequests);
inc_in_type('get-next-request') -> inc(snmpInGetNexts);
inc_in_type('set-request') -> inc(snmpInSetRequests);
inc_in_type(_) -> ok.
