%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_net_if).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/2').
-author('mbj@erlang.ericsson.se').

-export([start_link/1, get_log_type/1, set_log_type/2]).
-export([system_continue/3, system_terminate/4, system_code_change/4]).
-export([init/2]).

-include("snmp_types.hrl").

-record(state, {parent, master_agent, usock, mpc_state, log,
		reqs = [], debug = false}).

%%%-----------------------------------------------------------------
%%% This module implements the default Network Interface part
%%% of the SNMP agent. It uses UDP, and OTP-SNMPEA-MIB for finding
%%% the UDP port.
%%%-----------------------------------------------------------------
start_link(MasterAgent) ->
    proc_lib:start_link(?MODULE, init, [MasterAgent, self()]).

get_log_type(Pid) ->
    Pid ! {self(), get_log_type},
    receive
	{Pid, Type} ->
	    Type
    after
	5000 ->
	    undefined
    end.

set_log_type(Pid, Type) ->
    Pid ! {set_log_type, Type}.


init(MasterAgent, Parent) ->
    process_flag(trap_exit, true),
    {value, UDPPort} = snmp_internal_mib:intAgentUDPPort(get),
    Vsns = snmp_misc:get_vsns(),
    Log = case snmp_log:create_log() of
	      {ok, L} -> L;
	      false -> false;
	      {error, Err} -> exit(Err)
	  end,
    case gen_udp_open(UDPPort, [binary, {reuseaddr, true}]) of
	{ok, Id} ->
	    MpcState = snmp_mpc:init_mpc(Vsns),
	    S = #state{parent = Parent,
		       master_agent = MasterAgent, mpc_state = MpcState,
		       usock = Id, log = Log},
	    proc_lib:init_ack({ok, self()}),
	    loop(S);
	{error, Reason} ->
	    proc_lib:init_ack({error, {udp, UDPPort, Reason}})
    end.

gen_udp_open(Port, Opts) ->
    case init:get_argument(snmp_fd) of
	{ok, [[FdStr]]} ->
	    Fd = list_to_integer(FdStr),
	    gen_udp:open(0, [{fd, Fd}|Opts]);
	error ->
	    gen_udp:open(Port, Opts)
    end.
	
loop(S) ->
    receive
	{udp, _UdpId, Ip, Port, Packet} ->
%	    io:format("net_if got packet from ~w:~w~n", [Ip, Port]),
%	    put(n1,erlang:now()),
	    case snmp_mpc:process_packet(Packet, snmpUDPDomain, {Ip, Port}, 
					 S#state.mpc_state) of
		{ok, Vsn, Pdu, PduMS, ACMData} ->
		    pdebug(S#state.debug, "got pdu from ~w:~w ~p",
			   [Ip, Port, Pdu]),
		    Type = Pdu#pdu.type,
		    log(S#state.log, Type, Packet, {Ip, Port}),
		    case Type of
			'get-response' ->
			    handle_response(Vsn, Pdu, {Ip, Port}, S);
			_ ->
			    S#state.master_agent !
				{snmp_pdu, Vsn, Pdu, PduMS,
				 ACMData, {Ip, Port}, []}
		    end;
		{discarded, Reason} ->
		    pdebug(S#state.debug,
			   "got packet from ~p, discarded for reason: ~p", 
			   [{Ip, Port}, Reason])
		end,
	    loop(S);
	{snmp_response, Vsn, RePdu, Type, ACMData, {Ip, Port}, []} ->
	    pdebug(S#state.debug, "reply pdu: ~p", [RePdu]),
	    case snmp_mpc:generate_response_msg(Vsn, RePdu, Type, ACMData) of
		{ok, Packet} ->
		    log(S#state.log, Type, Packet, {Ip, Port}),
%		    N2 = erlang:now(),
%		    NT = subtr(N2, get(n1)),
%		    io:format("Time in agent: ~w mysec~n", [NT]),
		    gen_udp:send(S#state.usock, Ip, Port, Packet);
		{discarded, _Reason} ->
		    ok
	    end,
	    loop(S);
	{send_pdu, Vsn, Pdu, SecData, To} ->
	    NewS = handle_send_pdu(S, Vsn, Pdu, SecData, To, undefined),
	    loop(NewS);
	{send_pdu_req, Vsn, Pdu, SecData, To, From} ->
	    NewS = handle_send_pdu(S, Vsn, Pdu, SecData, To, From),
	    loop(NewS);
	{discarded_pdu, _Vsn, _ReqId, _ACMData, Variable, _Extra} ->
	    snmp_mpc:discarded_pdu(Variable),
	    loop(S);
	{From, get_log_type} ->
	    From ! {self(), S#state.log},
	    loop(S);
	{set_log_type, Type} ->
	    loop(S#state{log = Type});
	{debug, Flag} ->
	    loop(S#state{debug = Flag});
	{'EXIT', Parent, Reason} when Parent == S#state.parent ->
	    exit(Reason);
	{'EXIT', Pid, Reason} ->
	    NewS = clear_reqs(Pid, S),
	    loop(NewS);
	{system, From, Msg} ->
	    sys:handle_system_msg(Msg, From, S#state.parent, ?MODULE, [], S);
	_ ->
	    loop(S)
    end.

handle_send_pdu(S, Vsn, Pdu, SecData, To, From) ->
    Dbg = S#state.debug,
    case snmp_mpc:generate_msg(Vsn, Pdu, SecData, To) of
	{ok, Packet, Addresses} ->
	    log(S#state.log, Pdu#pdu.type, Packet, Addresses),
	    pdebug(Dbg, "sending pdu ~p, to the following mgs:", [Pdu]),
	    lists:foreach(fun({snmpUDPDomain, {Ip, Port}}) ->
				  ipdebug(Dbg, "    ~w:~w~n",[Ip,Port]),
				  gen_udp:send(S#state.usock, Ip,
					       Port, Packet);
			     (_) -> ok
			  end, Addresses);
	{discarded, Reason} ->
	    pdebug(Dbg, "do not send too big pdu ~p", [Pdu]),
	    ok
    end,
    case From of
	undefined ->
	    S;
	Pid ->
	    link(Pid),
	    NReqs = snmp_misc:keyreplaceadd(Pid, 2, S#state.reqs, 
					    {Pdu#pdu.request_id, From}),
	    S#state{reqs = NReqs}
    end.

handle_response(Vsn, Pdu, From, S) ->
    case lists:keysearch(Pdu#pdu.request_id, 1, S#state.reqs) of
	{value, {_, Pid}} ->
	    Pid ! {snmp_response_received, Vsn, Pdu, From};
	_ ->
	    pdebug(S#state.debug, "No receiver available for response pdu", [])
    end.

clear_reqs(Pid, S) ->
    NReqs = lists:keydelete(Pid, 2, S#state.reqs),
    S#state{reqs = NReqs}.


log(false, _Type, _Packet, _Address) ->
    ok;
log(Log, Type, Packet, Address) ->
    snmp_log:log(Log, Type, Packet, Address).

%%%-----------------------------------------------------------------
%%% System messages
%%%-----------------------------------------------------------------
system_continue(_Parent, _Dbg, S) ->
    loop(S).

system_terminate(Reason, _Parent, _Dbg, S) ->
    exit(Reason).

system_code_change(S, _Module, _OldVsn, _Extra) ->
    {ok, S}.

%%%-----------------------------------------------------------------
%%% DEBUG FUNCTIONS
%%%-----------------------------------------------------------------
pdebug(false, _Format, _X) -> ok;
pdebug(_, Format, X) ->
    Form = lists:concat(["** SNMP Net if debug: ~n   ", Format, "~n"]),
    io:format(Form, X).

ipdebug(false, _Format, _X) -> ok;
ipdebug(_, Format, X) ->
    io:format(Format, X).

%subtr({X1,Y1,Z1}, {X1,Y1,Z2}) ->
%    Z1 - Z2;
%subtr({X1,Y1,Z1}, {X1,Y2,Z2}) ->
%    ((Y1 * 1000000) + Z1) - ((Y2 * 1000000) + Z2).
