%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_note_store).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

%% External exports
-export([start_link/1, get_note/1, set_note/3]).

%% Internal exports
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2]).
-export([timer/2]).

-define(timeout, 30000).  % Perform gc twice in a minute.

-record(state, {notes, timer = start_timer(), timeout = false}).

%%%-----------------------------------------------------------------
%%% Implements a database for notes with a lifetime. Once in a
%%% while, the database will be gc:ed, to get rid of old notes.
%%% This database will not contain much data.
%%%-----------------------------------------------------------------
start_link(Prio) ->
    gen_server:start_link({local, snmp_note_store}, snmp_note_store, [Prio],[]).

%%-----------------------------------------------------------------
%% Interface functions.
%%-----------------------------------------------------------------
get_note(Key) ->
    gen_server:call(snmp_note_store, {get_note, Key}, infinity).
set_note(Lifetime, Key, Value) ->
    gen_server:call(snmp_note_store, {set_note, Lifetime, Key, Value},
		    infinity).

init([Prio]) ->
    process_flag(priority, Prio),
    Notes = ets:new(snmp_note_store, [set, private]),
    {ok, #state{notes = Notes}}.

%%-----------------------------------------------------------------
%% A note is any internal information that has to be
%% stored for some time (the Lifetime).
%% A note is stored in ets as {Key, {BestBefore, Value}},
%% where BestBefore is currentTime + Lifetime. 
%% A GC-op can destroy any notes with CurTime > BestBore.
%% Lifetime is in centiseconds or infinity, in which case
%% the note is eternal.
%%-----------------------------------------------------------------
handle_call({set_note, Lifetime, Key, Value}, _From, State) 
  when integer(Lifetime) ->
    RealUpTime = snmp_misc:now(cs) - snmp:system_start_time(),
    BestBefore = RealUpTime + Lifetime,
    Val = ets:insert(State#state.notes, {Key, {BestBefore, Value}}),
    NState = activate_timer(State),
    {reply, Val, NState};

handle_call({set_note, infinity, Key, Value}, _From, State) ->
    Val = ets:insert(State#state.notes, {Key, {infinity, Value}}),
    {reply, Val, State};

handle_call({get_note, Key}, _From, State) ->
    Val = handle_get_note(Key, State#state.notes),
    {reply, Val, State};

handle_call(stop, _From, State) ->
    {stop, normal, ok, State}.

handle_cast(_, State) ->
    {noreply, State}.
    
%%-----------------------------------------------------------------
%% If there are no possible garbage left, we don't
%% have to wait for timeout, and perform another
%% gc, because we won't do anything. So
%% we switch the timeout off in that case.
%% It will be switched on as soon as we get some
%% other message.
%%-----------------------------------------------------------------
handle_info(timeout, State) ->
    case gc(State#state.notes) of
	nothing_left ->
	    NState = deactivate_timer(State),
	    {noreply, NState};
	work_to_do ->
	    NState = activate_timer(State),
	    {noreply, NState}
    end.

terminate(_Reason, State) ->
    ok.


activate_timer(State) ->
    case State#state.timeout of
	false -> 
	    State#state.timer ! activate,
	    receive
		activated -> ok
	    end;
	true ->
	    ok
    end,
    State#state{timeout = true}.

deactivate_timer(State) ->
    case State#state.timeout of
	true ->
	    State#state.timer ! deactivate,
	    receive
		deactivated -> ok
	    end;
	false ->
	    ok
    end,
    State#state{timeout = false}.

start_timer() ->
    spawn_link(?MODULE, timer, [self(), deactive]).

timer(Pid, deactive) ->
    receive
	activate ->
	    Pid ! activated,
	    timer(Pid, active)
    end;
timer(Pid, active) ->
    receive
	deactivate ->
	    Pid ! deactivated,
	    timer(Pid, deactive)
    after
	?timeout ->
	    Pid ! timeout,
	    timer(Pid, active)
    end.
    

handle_get_note(Key, Notes) ->
    case ets:lookup(Notes, Key) of
	[{Key, {infinity, Val}}] ->
	    Val;
	[{Key, {BestBefore, Val}}] ->
	    StartTime = snmp:system_start_time(),
	    case (snmp_misc:now(cs) - StartTime) of
		Now when BestBefore >= Now ->
		    Val;
		_ ->
		    ets:delete(Notes, Key),
		    undefined
	    end;
	[] -> undefined
    end.

%%-----------------------------------------------------------------
%% Clean up all old notes in the database.
%%-----------------------------------------------------------------
gc(Tab) ->
    RealUpTime = snmp_misc:now(cs) - snmp:system_start_time(),
    gc(nothing_left, ets:tab2list(Tab), Tab, RealUpTime).

gc(Flag, [{Key, {infinity, _}} | T], Tab, Now) -> gc(Flag, T, Tab, Now);
gc(Flag, [{Key, {BestBefore, _}} | T], Tab, Now) 
  when integer(BestBefore), BestBefore < Now ->
    ets:delete(Tab, Key),
    gc(Flag, T, Tab, Now);
gc(Flag, [_ | T], Tab, Now) -> gc(work_to_do, T, Tab, Now);
gc(Flag, [], _Tab, _Now) -> Flag.
    
    
