%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_notification_mib).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-export([init/0, configure/1, invalidate_cache/0,
	 snmpNotifyTable/1, snmpNotifyTable/3,
	 get_targets/0, get_targets/1]).

-include("SNMP-NOTIFICATION-MIB.hrl").
-include("SNMPv2-TC.hrl").


%%-----------------------------------------------------------------
%% Func: init/0
%% Purpose: Creates the tables and variables necessary for the SNMP
%%          target mechanism to work properly.
%%          Note that this function won't destroy any old values.
%%          This function should be called only once.
%%-----------------------------------------------------------------
init() ->
    maybe_create_table(snmpNotifyTable).

%%-----------------------------------------------------------------
%% Func: configure/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads the config-files for the internal tables, and
%%          inserts the data.  Makes sure that all old data in
%%          the tables are deleted, and the new data inserted.
%%          This function makes sure that all (and only) 
%%          config-file-data are in the tables. 
%% Returns: ok
%% Fails: exit(configuration_error)
%% PRE: init/1 has been successfully called
%%-----------------------------------------------------------------
configure(Dir) ->
    Notifs =
	snmp_conf:read_notify_config_files(Dir),
    init_tabs(Notifs),
    invalidate_cache(),
    ok.

maybe_create_table(Name) ->
    case snmp_local_db:table_exists(db(Name)) of
	true -> ok;
	_ -> snmp_local_db:table_create(db(Name))
    end.

init_tabs(Notifs) ->
    snmp_local_db:table_delete(db(snmpNotifyTable)),
    snmp_local_db:table_create(db(snmpNotifyTable)),
    init_notify_table(Notifs).
    
init_notify_table([Row | T]) ->
    Key = element(1, Row),
    snmp_local_db:table_create_row(db(snmpNotifyTable), Key, Row),
    init_notify_table(T);
init_notify_table([]) -> true.


%%-----------------------------------------------------------------
%% Func: get_targets()
%%       get_targets(NotifyName) -> [Target]
%% Types: Target = {NotifyName, Ip, SnmpVsn, NotifyType}
%%        NotifyName = string()  - the INDEX
%%        Ip = {snmpUDPDomain, {[A,B,C,D], UDP}}
%%        SnmpVsn = 
%%        NotifyType = trap | {inform, Timeout. Retry}
%% Purpose: Returns a list of all targets.  Called by snmp_trap
%%          when a trap should be sent.
%%          If a NotifyName is specified, the targets for that
%%          name is returned.
%%-----------------------------------------------------------------
get_targets() ->
    case snmp_local_db:variable_get({snmp_targets_cache, volatile}) of
	{value, Targets} ->
	    Targets;
	 _ ->
	    Targets = find_targets(),
	    update_cache(Targets),
	    Targets
    end.

get_targets(NotifyName) ->
    lists:filter(fun(Entry) when element(1, Entry) == NotifyName -> true;
		    (_) -> false
		 end, get_targets()).
    

%%-----------------------------------------------------------------
%% We use a cache of targets to avoid searching the tables each
%% time a trap is to be sent.  When some of the 3 tables (notify,
%% targetAddr, targetParams) is modified, the cache is invalidated.
%%-----------------------------------------------------------------
update_cache(Targets) ->
    snmp_local_db:variable_set({snmp_targets_cache, volatile}, Targets).

invalidate_cache() ->
    snmp_local_db:variable_delete({snmp_targets_cache, volatile}).

%% Ret: [{NotifyName, Ip, SnmpVsn, NotifyType}]
%%   NotifyType = trap | {inform, Timeout. Retry}
find_targets() ->
    TargAddrs = snmp_target_mib:get_target_addrs(),
    %% TargAddrs = [{TagList, Ip, SnmpVsn, Timeout, Retry}]
    find_targets([], TargAddrs , []).
find_targets(Key, TargAddrs, Res) ->
    case table_next(snmpNotifyTable, Key) of
	endOfTable -> 
	    Res;
	NextKey -> 
	    Elements = [?snmpNotifyTag, ?snmpNotifyType, ?snmpNotifyRowStatus],
	    case snmpNotifyTable(get, NextKey, Elements) of
		[{value, Tag}, {value, Type}, {value, ?'RowStatus_active'}] ->
		    Targs = get_targets(TargAddrs, Tag, Type, NextKey),
		    find_targets(NextKey, TargAddrs, Targs ++ Res);
		_ ->
		    find_targets(NextKey, TargAddrs, Res)
	    end
    end.

get_targets([{TagList, Addr, SnmpVsn, Timeout, Retry} | T], Tag, Type, Name) ->
    case snmp_misc:is_tag_member(Tag, TagList) of
	true -> [{Name, Addr, SnmpVsn, type(Type, Timeout, Retry)} |
		 get_targets(T, Tag, Type, Name)];
	false ->
	    get_targets(T, Tag, Type, Name)
    end;
get_targets([], Tag, Type, Name) ->
    [].

type(trap, _, _) -> trap;
type(inform, Timeout, Retry) -> {inform, Timeout, Retry}.


%%-----------------------------------------------------------------
%% Instrumentation Functions
%%-----------------------------------------------------------------
%% Op == new | delete
snmpNotifyTable(Op) ->
    snmp_generic:table_func(Op, db(snmpNotifyTable)).

%% Op == get | is_set_ok | set | get_next
snmpNotifyTable(get, RowIndex, Cols) ->
    get(snmpNotifyTable, RowIndex, Cols);
snmpNotifyTable(get_next, RowIndex, Cols) ->
    next(snmpNotifyTable, RowIndex, Cols);
snmpNotifyTable(set, RowIndex, Cols) ->
    invalidate_cache(),
    snmp_generic:table_func(set, RowIndex, Cols, db(snmpNotifyTable));
snmpNotifyTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(snmpNotifyTable)).


db(X) -> {X, volatile}.

fa(snmpNotifyTable) -> 2.
 
foi(snmpNotifyTable) -> 1.
 
noc(snmpNotifyTable) -> 5.
 
next(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_next(db(Name), RowIndex, Cols,
                                   fa(Name), foi(Name), noc(Name)).

table_next(Name, RestOid) ->
    snmp_generic:table_next(db(Name), RestOid).

 
get(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_get(db(Name), RowIndex, Cols, foi(Name)).

