%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_target_mib).
-copyright('Copyright (c) 1991-98 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').

-export([init/0, configure/1,
	 snmpTargetSpinLock/1, snmpTargetSpinLock/2,
	 snmpTargetAddrTable/1, snmpTargetAddrTable/3,
	 snmpTargetParamsTable/1, snmpTargetParamsTable/3,
	 table_next/2, get_target_addrs/0]).

-include("SNMP-TARGET-MIB.hrl").
-include("SNMPv2-TC.hrl").
-include("SNMPv2-TM.hrl").

-define(v1_model, 0).
-define(v2c_model, 1).

%%-----------------------------------------------------------------
%% Func: init/0
%% Purpose: Creates the tables and variables necessary for the SNMP
%%          target mechanism to work properly.
%%          Note that this function won't destroy any old values.
%%          This function should be called only once.
%%-----------------------------------------------------------------
init() ->
    maybe_create_table(snmpTargetAddrTable),
    maybe_create_table(snmpTargetParamsTable).

%%-----------------------------------------------------------------
%% Func: configure/1
%% Args: Dir is the directory with trailing dir_separator where
%%       the configuration files can be found.
%% Purpose: Reads the config-files for the internal tables, and
%%          inserts the data.  Makes sure that all old data in
%%          the tables are deleted, and the new data inserted.
%%          This function makes sure that all (and only) 
%%          config-file-data are in the tables. 
%% Returns: ok
%% Fails: exit(configuration_error)
%% PRE: init/1 has been successfully called
%%-----------------------------------------------------------------
configure(Dir) ->
    {Addrs, Params} =
	snmp_conf:read_target_config_files(Dir),
    init_tabs(Addrs, Params),
    snmp_notification_mib:invalidate_cache(),
    ok.

maybe_create_table(Name) ->
    case snmp_local_db:table_exists(db(Name)) of
	true -> ok;
	_ -> snmp_local_db:table_create(db(Name))
    end.

init_tabs(Addrs, Params) ->
    snmp_local_db:table_delete(db(snmpTargetAddrTable)),
    snmp_local_db:table_create(db(snmpTargetAddrTable)),
    init_addr_table(Addrs),
    snmp_local_db:table_delete(db(snmpTargetParamsTable)),
    snmp_local_db:table_create(db(snmpTargetParamsTable)),
    init_params_table(Params).
    
init_addr_table([Row | T]) ->
    Key = element(1, Row),
    snmp_local_db:table_create_row(db(snmpTargetAddrTable), Key, Row),
    init_addr_table(T);
init_addr_table([]) -> true.

init_params_table([Row | T]) ->
    Key = element(1, Row),
    snmp_local_db:table_create_row(db(snmpTargetParamsTable), Key, Row),
    init_params_table(T);
init_params_table([]) -> true.

%% TargAddrs = [{TagList, Ip, SnmpVsn, Timeout, Retry}]
get_target_addrs() ->
    get_target_addrs([], []).

get_target_addrs(Key, Res) ->
    case table_next(snmpTargetAddrTable, Key) of
	endOfTable -> 
	    Res;
	NextKey -> 
	    case snmp_local_db:table_get_row(db(snmpTargetAddrTable),NextKey) of
		{_Key, TDomain, TAddress, Timeout, RetryCount, TagList, Params,
		 _Storage, ?'RowStatus_active'} ->
		    SnmpVsn = get_snmp_vsn(Params),
		    Targ = {TagList, addr(TDomain, TAddress), 
			    SnmpVsn, Timeout, RetryCount},
		    get_target_addrs(NextKey, [Targ | Res]);
		_ ->
		    get_target_addrs(NextKey, Res)
	    end
    end.

addr(?snmpUDPDomain, [A,B,C,D,U1,U2]) ->
    {snmpUDPDomain, {[A,B,C,D], U1 bsl 8 + U2}};
addr(Domain, Addr) ->
    {undefined, Addr}.

get_snmp_vsn(Params) ->
    case snmpTargetParamsTable(get, Params, [?snmpTargetParamsMPModel,
					     ?snmpTargetParamsRowStatus]) of
	[{value, MpModel}, {value, ?'RowStatus_active'}] ->
	    case MpModel of
		?v1_model -> 'version-1';
		?v2c_model -> 'version-2';
		_ -> undefined
	    end;
	_ ->
	    undefined
    end.

%%-----------------------------------------------------------------
%% Instrumentation Functions
%%-----------------------------------------------------------------
snmpTargetSpinLock(new) ->
    snmp_generic:variable_func(new, {snmpTargetSpinLock, volatile}),
    {A1,A2,A3} = erlang:now(),
    random:seed(A1,A2,A3),
    Val = random:uniform(2147483648) - 1,
    snmp_generic:variable_func(set, Val, {snmpTargetSpinLock, volatile});

snmpTargetSpinLock(delete) ->
    ok;

snmpTargetSpinLock(get) ->
    snmp_generic:variable_func(get, {snmpTargetSpinLock, volatile}).

snmpTargetSpinLock(is_set_ok, NewVal) ->
    case snmp_generic:variable_func(get, {snmpTargetSpinLock, volatile}) of
	{value, NewVal} -> noError;
	_ -> inconsistentValue
    end;
snmpTargetSpinLock(set, NewVal) ->
    snmp_generic:variable_func(set, (NewVal + 1) rem 2147483648,
			       {snmpTargetSpinLock, volatile}).


%% Op == new | delete
snmpTargetAddrTable(Op) ->
    snmp_generic:table_func(Op, db(snmpTargetAddrTable)).

%% Op == get | is_set_ok | set | get_next
snmpTargetAddrTable(get, RowIndex, Cols) ->
    get(snmpTargetAddrTable, RowIndex, Cols);
snmpTargetAddrTable(get_next, RowIndex, Cols) ->
    next(snmpTargetAddrTable, RowIndex, Cols);
snmpTargetAddrTable(set, RowIndex, Cols) ->
    snmp_notification_mib:invalidate_cache(),
    snmp_generic:table_func(set, RowIndex, Cols, db(snmpTargetAddrTable));
snmpTargetAddrTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(snmpTargetAddrTable)).

%% Op == new | delete
snmpTargetParamsTable(Op) ->
    snmp_generic:table_func(Op, db(snmpTargetParamsTable)).

%% Op == get | is_set_ok | set | get_next
snmpTargetParamsTable(get, RowIndex, Cols) ->
    get(snmpTargetParamsTable, RowIndex, Cols);
snmpTargetParamsTable(get_next, RowIndex, Cols) ->
    next(snmpTargetParamsTable, RowIndex, Cols);
snmpTargetParamsTable(set, RowIndex, Cols) ->
    snmp_notification_mib:invalidate_cache(),
    snmp_generic:table_func(set, RowIndex, Cols, db(snmpTargetParamsTable));
snmpTargetParamsTable(Op, Arg1, Arg2) ->
    snmp_generic:table_func(Op, Arg1, Arg2, db(snmpTargetParamsTable)).


db(X) -> {X, volatile}.

fa(snmpTargetAddrTable) -> 2;
fa(snmpTargetParamsTable) -> 2.
 
foi(snmpTargetAddrTable) -> 1;
foi(snmpTargetParamsTable) -> 1.
 
noc(snmpTargetAddrTable) -> 9;
noc(snmpTargetParamsTable) -> 7.
 
next(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_next(db(Name), RowIndex, Cols,
                                   fa(Name), foi(Name), noc(Name)).

table_next(Name, RestOid) ->
    snmp_generic:table_next(db(Name), RestOid).

 
get(Name, RowIndex, Cols) ->
    snmp_generic:handle_table_get(db(Name), RowIndex, Cols, foi(Name)).

