%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(snmp_trap).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-author('mbj@erlang.ericsson.se').
-revision('$Revision: 1.11 $').
-date('$Date: 1997/09/11 11:57:11 $').

%%%-----------------------------------------------------------------
%%% This module takes care of all trap handling.
%%%-----------------------------------------------------------------
%% External exports
-export([construct_trap/2, try_initialise_vars/2, send_trap/5]).

%% Internal exports
-export([alias_to_oid/1, make_varbind/1, init_v2_inform/7]).

-include("snmp_types.hrl").
-include("OTP-SNMPEA-MIB.hrl").
-include("STANDARD-MIB.hrl").
-define(enterpriseSpecific, 6).
% TMP
-define(snmpTraps, [1,3,6,1,6,3,1,1,5]).
-define(snmpTrapEnterprise_instance, [1,3,6,1,6,3,1,1,4,3,0]).
-define(snmpTrapOID_instance, [1,3,6,1,6,3,1,1,4,1,0]).


%%-----------------------------------------------------------------
%% Trap mechanism
%% ==============
%% Distributed subagent (dSA) case
%%   The MIB with the TRAP-TYPE macro is loaded in dSA.  This means
%%   that dSA has info on all variables defined in the TRAP-TYPE,
%%   even though some variables may be located in other SA:s (or
%%   in the MA). Other variables that may be sent in the trap, 
%%   must be known by either the dSA, or some of its parent agents
%%   (e.g. the master agent), if the variable should be referred
%%   to by symbolic name. It is however possible to send other
%%   variables as well, but then the entire OID must be provided.
%%   The dSA locates the asn1 type, oid and value for as many
%%   variables as possible. This information, together with the
%%   variables for which the type, value or oid isn't known, is
%%   sent to the dSA's parent. This agent performs the same
%%   operation, and so on, until eventually the MA will receive the
%%   info. The MA then fills in the gaps, and at this point all
%%   oids and types must be known, otherwise an error is signalled,
%%   and the opertaion is aborted. For the unknown values for some 
%%   oids, a get-operation is performed by the MA. This will
%%   retreive the missing values.
%%   At this point, all oid, types and values are known, so the MA
%%   can distribute the traps according to the information in the
%%   internal tables.
%% 
%% Local subagent (lSA) case
%%   This case is similar to the case above.
%%
%% Master agent (MA) case
%%   This case is similar to the case above.
%%
%% NOTE: All trap forwarding between agents is made asynchronously.
%%
%% dSA: Distributed SA  (the #trap is loaded here)
%% nSA: [many] SAs between dSA and MA
%% MA:  Master Agent. (all trap info (destiniations is here))
%% 1) application decides to send a trap.
%% 2) dSA calls send_trap which initialises vars
%% 3) dSA sends all to nSA
%% 4) nSA tries to map symbolic names to oids and find the types
%%    of all variableoids with a value (and no type).
%% 5) nSA sends all to (n-1)SA
%% 6) MA tries to initialise vars
%% 7) MA makes atrappdu, and sends it to all destination.
%%
%% Problems with this implementation
%% =================================
%% It's ok to send {Oid, Value} but not just Oid. (it should be for
%%   any Oid)
%% It's ok to send {Name, Value} but not just Name. (it should be
%%   for Names in the hierarchy)
%% This approach might be too flexible; will people use it?
%% *NOTE*
%% Therefore, in this version we *do not* allow extra variables
%% in traps.
%% *YES* In _this_ version we do.
%%-----------------------------------------------------------------

%%-----------------------------------------------------------------
%% Func: construct_trap/2
%% Args: Trap is an atom
%%       Varbinds is a list of {Variable, Value},
%%         where Variable is an atom or an OID. |
%%         {SymbolicTableCol, RowIndex, Value},
%%         where RowIndex is the indexes for the row.
%%         We don't check the RowIndex.
%% Purpose: This is the initially-called function. It is called
%%          by the agent that found out that a trap should be
%%          sent.
%%          Initialize as many variables as possible.
%% Returns: {ok, TrapRecord, <list of Var>} | error
%%          where Var is returned from initiate_vars.
%% NOTE: Executed at the inital SA
%%-----------------------------------------------------------------
construct_trap(Trap, Varbinds) ->
    case snmp_symbolic_store:get_notification(Trap) of
	undefined -> 
	    snmp_error:user_err("send_trap got undef Trap: ~p" , [Trap]),
	    error;
	{value, TRec} when record(TRec, trap) ->
	    ListOfVars = TRec#trap.oidobjects,
	    OidVbs = lists:map({snmp_trap, alias_to_oid}, [], Varbinds),
	    LV = initiate_vars(ListOfVars, OidVbs),
	    {ok, TRec, try_initialise_vars(get(mibserver), LV)};
	{value, NRec} when record(NRec, notification) ->
	    ListOfVars = NRec#notification.oidobjects,
	    OidVbs = lists:map({snmp_trap, alias_to_oid}, [], Varbinds),
	    LV = initiate_vars(ListOfVars, OidVbs),
	    {ok, NRec, try_initialise_vars(get(mibserver), LV)}
    end.

alias_to_oid({Alias, Val}) when atom(Alias) ->
    case snmp_symbolic_store:aliasname_to_oid(Alias) of
	{value, Oid} -> {lists:append(Oid, [0]), {value, Val}};
	_ ->   	     {Alias, {value, Val}}
    end;
alias_to_oid({Alias, RowIndex, Val}) when atom(Alias) ->
    case snmp_symbolic_store:aliasname_to_oid(Alias) of
	{value, Oid} -> {lists:append(Oid, RowIndex), {value, Val}};
	_ ->   	     {Alias, RowIndex, {value, Val}}
    end;
alias_to_oid({Oid, Val}) -> {Oid, {value, Val}}.

%%-----------------------------------------------------------------
%% Func: initiate_vars/2
%% Args: ListOfVars is a list of {Oid, #asn1_type}
%%       Varbinds is a list of 
%%          {VariableOid, Value} | 
%%          {VariableAtom, Value} |
%%          {TableColAtom, RowIndex, Value}
%% Purpose: For each variable in specified in the TRAP-TYPE macro
%%          (each in ListOfVars), check if it's got a value given
%%          in the Varbinds list.
%%          For each Oid:
%%            1) It has corresponding VariableOid. Use Value.
%%            2) No corresponding VariableOid. No value.
%% Returns: A list of
%%            {VariableOid, #asn1_type, Value} |
%%            {VariableOid, #asn1_type} |
%%            {VariableOid, Value} |
%%            {VariableAtom, Value} |
%%            {TableColAtom, RowIndex, Value}
%% NOTE: Executed at the inital SA
%%-----------------------------------------------------------------
initiate_vars([{Oid, Asn1Type} | T], Varbinds) ->
    case delete_oid_from_varbinds(Oid, Varbinds) of
	{undefined, _, _} ->
	    [{Oid, Asn1Type} | initiate_vars(T, Varbinds)];
	{Value, VarOid, RestOfVarbinds} ->
	    [{VarOid, Asn1Type, Value} | initiate_vars(T, RestOfVarbinds)]
    end;
initiate_vars([], Varbinds) ->
    Varbinds.
    
delete_oid_from_varbinds(Oid, [{VarOid, Value} | T]) ->
    case lists:prefix(Oid, VarOid) of
	true -> 
	    {Value, VarOid, T};
	_ -> 
	    {Value2, VarOid2, T2} = delete_oid_from_varbinds(Oid, T),
	    {Value2, VarOid2, [{VarOid, Value} | T2]}
    end;
delete_oid_from_varbinds(Oid, [H | T]) ->
    {Value, VarOid, T2} = delete_oid_from_varbinds(Oid, T),
    {Value, VarOid, [H | T2]};
delete_oid_from_varbinds(_Oid, []) -> {undefined, undefined, []}.

%%-----------------------------------------------------------------
%% Func: try_initialise_vars(Mib, Varbinds)
%% Args: Mib is the local mib process
%%       Varbinds is a list returned from initiate_vars.
%% Purpose: Try to initialise uninitialised vars.
%% Returns: see initiate_vars
%% NOTE: Executed at the intermediate SAs
%%-----------------------------------------------------------------
try_initialise_vars(Mib, Varbinds) ->
    V = try_map_symbolic(Varbinds),
    try_find_type(V, Mib).

%%-----------------------------------------------------------------
%% Func: try_map_symbolic/1
%% Args: Varbinds is a list returned from initiate_vars.
%% Purpose: Try to map symbolic name to oid for the 
%%          symbolic names left in the Varbinds list.
%% Returns: see initiate_vars.
%% NOTE: Executed at the intermediate SAs
%%-----------------------------------------------------------------
try_map_symbolic([Varbind | Varbinds]) ->
    [localise_oid(Varbind) | try_map_symbolic(Varbinds)];
try_map_symbolic([]) -> [].

localise_oid({VariableName, Value}) when atom(VariableName) ->
    alias_to_oid({VariableName, Value});
localise_oid({VariableName, RowIndex, Value}) when atom(VariableName) ->
    alias_to_oid({VariableName, RowIndex, Value});
localise_oid(X) -> X.

%%-----------------------------------------------------------------
%% Func: try_find_type/2
%% Args: Varbinds is a list returned from initiate_vars.
%%       Mib is a ref to the Mib process corresponding to
%%         this agent.
%% Purpose: Try to find the type for each variableoid with a value
%%          but no type.
%% Returns: see initiate_vars.
%% NOTE: Executed at the intermediate SAs
%%-----------------------------------------------------------------
try_find_type([Varbind | Varbinds], Mib) ->
    [localise_type(Varbind, Mib) | try_find_type(Varbinds, Mib)];
try_find_type([], _) -> [].

localise_type({VariableOid, Type}, Mib) 
  when list(VariableOid), record(Type, asn1_type) ->
    {VariableOid, Type};
localise_type({VariableOid, Value}, Mib) when list(VariableOid) ->
    case snmp_mib:lookup(Mib, VariableOid) of
	{variable, ME} ->
	    {VariableOid, ME#me.asn1_type, Value};
	{table_column, ME, _} ->
	    {VariableOid, ME#me.asn1_type, Value};
	_ ->
	    {VariableOid, Value}
    end;
localise_type(X, _) -> X.

%%-----------------------------------------------------------------
%% Func: make_v1_trap_pdu/4
%% Args: Enterprise = oid()
%%       Specific = integer()
%%       Varbinds is as returned from initiate_vars
%%         (but only {Oid, Type[, Value} permitted)
%%       SysUpTime = integer()
%% Purpose: Make a #trappdu
%%          Checks the Varbinds to see that no symbolic names are
%%          present, and that each var has a type. Performs a get
%%          to find any missing value.
%% Returns: {#trappdu, [byte()] | error
%% Fails: yes
%% NOTE: Executed at the MA
%%-----------------------------------------------------------------
make_v1_trap_pdu(Enterprise, Specific, VarbindList, SysUpTime) ->
    case Enterprise of
	?snmp ->
	    Enterp = sys_object_id(),
	    Generic = Specific,
	    Spec = 0;
	_ ->
	    Enterp = Enterprise,
	    Generic = ?enterpriseSpecific,
	    Spec = Specific
    end,
    {value, AgentIp} = snmp_internal_mib:intAgentIpAddress(get),
    #trappdu{enterprise = Enterp,
	     agent_addr = AgentIp,
	     generic_trap = Generic,
	     specific_trap = Spec,
	     time_stamp = SysUpTime,
	     varbinds = VarbindList}.

make_v2_notif_pdu(Vbs, Type) ->
    #pdu{type = Type,
	 request_id = make_request_id(),
	 error_status = noError,
	 error_index = 0,
	 varbinds = Vbs}.

make_request_id() -> random:uniform(16#FFFFFFF-1).

make_varbind_list(Varbinds) ->
    {VariablesWithValueAndType, VariablesWithType} =
	split_variables(order(Varbinds)),
    V = get_all(VariablesWithType),
    Vars = lists:append([V, VariablesWithValueAndType]),
    lists:map({snmp_trap, make_varbind}, [], unorder(lists:keysort(1, Vars))).

%%-----------------------------------------------------------------
%% Func: send_trap/5
%% Args: TrapRec ::= #trap | #notification
%%       NotifyName ::= string()
%%       Recv = no_receiver | {Ref, Receiver}
%%       Receiver = pid() | atom() | {M,F,A}
%%       Vbs ::= [varbind()]
%%       NetIf ::= pid()
%% Purpose: Default trap sending function.
%%          Sends the trap to the targets pointed out by NotifyName.
%%          If NotifyName is ""; the normal procedure defined in 
%%          SNMP-NOTIFICATION-MIB is used, i.e. the trap is sent to
%%          all managers.
%%          Otherwise, the NotifyName is used to find an entry in the
%%          SnmpNotifyTable which define how to send the notification
%%          (as an Inform or a Trap), and to select targets from
%%          SnmpTargetAddrTable (using the Tag).
%%-----------------------------------------------------------------
send_trap(TrapRec, NotifyName, Recv, Vbs, NetIf) ->
    VarbindList = make_varbind_list(Vbs),
    Dests = find_dests(NotifyName),
    send_trap_pdus(Dests, {TrapRec, VarbindList}, [], [], Recv, NetIf).
	    
get_all(VariablesWithType) ->
    {Order, Varbinds} = extract_order(VariablesWithType, 1),
    case snmp_agent:do_get(snmp_acm:get_root_mib_view(), Varbinds, true) of
	{noError, _, NewVarbinds} ->
	    contract_order(Order, NewVarbinds);
	{ErrorStatus, ErrorIndex, _} ->
	    snmp_error:user_err("snmp_trap: get operation failed "
				"{~p, ~p}~n  in ~p",
				[ErrorStatus, ErrorIndex, Varbinds]),
	    throw(error)
    end.
    
make_varbind(Varbind) when record(Varbind, varbind) ->
    Varbind;
make_varbind({VarOid, ASN1Type, Value}) ->
    case snmp_agent:make_value_a_correct_value(Value, ASN1Type, undef) of
	{value, Type, Val} ->
	    #varbind{oid = VarOid, variabletype = Type, value = Val};
	{error, Reason} -> 
	    snmp_error:user_err("snmp_trap: Invalid value: Oid ~p, Val:~p, ~p",
				[VarOid, Value, Reason]),
	    throw(error)
    end.

order(Varbinds) -> order(Varbinds, 1).
order([H | T], No) -> [{No, H} | order(T, No + 1)];
order([], _) -> [].

unorder([{No, H} | T]) -> [H | unorder(T)];
unorder([]) -> [].

extract_order([{No, {VarOid, _Type}} | T], Index) ->
    {Order, V} = extract_order(T, Index+1),
    {[No | Order], [#varbind{oid = VarOid, org_index = Index} | V]};
extract_order([], _) -> {[], []}.

contract_order([No | Order], [Varbind | T]) ->
    [{No, Varbind} | contract_order(Order, T)];
contract_order([], []) -> [].

split_variables([{No, {VarOid, Type, Val}} | T]) when list(VarOid) ->
    {A, B} = split_variables(T),
    {[{No, {VarOid, Type, Val}} | A], B};
split_variables([{No, {VarOid, Type}} | T]) 
  when list(VarOid), record(Type, asn1_type) ->
    {A, B} = split_variables(T),
    {A, [{No, {VarOid, Type}} | B]};
split_variables([{No, {VarName, Value}} | T]) ->
    snmp_error:user_err("snmp_trap: Undefined variable ~p (~p)",
			[VarName, Value]),
    throw(error);
split_variables([{No, {VarName, RowIndex, Value}} | T]) ->
    snmp_error:user_err("snmp_trap: Undefined variable ~p ~w (~p)",
			[VarName, RowIndex, Value]),
    throw(error);
split_variables([]) -> {[], []}.

%%-----------------------------------------------------------------
%% Func: find_dests/1
%% Returns: [{NotifyName, DestAddr, SnmpVsn, Type}]
%%          A list of all Destination addresses for this community.
%%          Each Destination address is {Domain, Address}
%% NOTE: This function is executed in the master agent's context
%%-----------------------------------------------------------------
find_dests("") ->
    snmp_notification_mib:get_targets();
find_dests(NotifyName) ->
    case snmp_notification_mib:get_targets(NotifyName) of
	[] ->
	    pdebug("snmp_trap: No dests found for snmpNotifyName ~p",
		   [NotifyName]),
	    [];
	Dests ->
	    Dests
    end.


%%-----------------------------------------------------------------
%% NOTE: This function is executed in the master agent's context
%%-----------------------------------------------------------------
send_trap_pdus([{NotifyName, ManagerAddr, Vsn, Type} | T], {TrapRec, Vbs},
	       V1Res, V2Res, Recv, NetIf) ->
    %% Use NotifyName as Community!!!
    Community = NotifyName,
    CommunityKey = [length(Community) | Community],
    Key = lists:append(get_ip(ManagerAddr), CommunityKey),
    case local_get(intCommunityTable, Key, [?intCommunityViewIndex]) of
	[{value, ViewIndex}] ->
	    MibView = snmp_acm:get_mib_view(ViewIndex),
	    case check_all_varbinds(MibView, Vbs) of
		true when Vsn == 'version-1' ->
		    send_trap_pdus(T, {TrapRec, Vbs},
				   [{ManagerAddr, Community} | V1Res],
				   V2Res, Recv, NetIf);
		true when Vsn == 'version-2' ->
		    send_trap_pdus(T, {TrapRec, Vbs}, V1Res,
				   [{ManagerAddr, Community, Type} | V2Res],
				   Recv, NetIf);
		_ ->
		    pdebug("snmp_trap: invalid access for ~p in ~p.~n",
			   [ManagerAddr, Community]),
		    send_trap_pdus(T, {TrapRec, Vbs}, V1Res, V2Res, Recv, NetIf)
	    end;
	undefined ->
	    pdebug("No info for Trap Destination ~p ~p "
		   "in intCommunityTable", [ManagerAddr, Community]),
	    send_trap_pdus(T, {TrapRec, Vbs}, V1Res, V2Res, Recv, NetIf)
    end;
send_trap_pdus([], {TrapRec, Vbs}, V1Res, V2Res, Recv, NetIf) ->
    SysUpTime = snmp_standard_mib:sys_up_time(),
    send_v1_trap(TrapRec, V1Res, Vbs, NetIf, SysUpTime),
    send_v2_trap(TrapRec, V2Res, Vbs, Recv, NetIf, SysUpTime).

send_v1_trap(_TrapRec, [], _Vbs, _NetIf, _SysUpTime) -> ok;
send_v1_trap(#trap{enterpriseoid = Enter, specificcode = Spec},
	     V1Res, Vbs, NetIf, SysUpTime) ->
    TrapPdu = make_v1_trap_pdu(Enter, Spec, Vbs, SysUpTime),
    AddrCommunities = mk_addr_communities(V1Res),
    lists:foreach(fun({Community, Addrs}) ->
			  NetIf ! {send_pdu, 'version-1', TrapPdu,
				   {community, Community}, Addrs}
		  end, AddrCommunities);
send_v1_trap(#notification{oid = Oid}, V1Res, Vbs, NetIf, SysUpTime) ->
    %% Use alg. in rfc2089 to map a v2 trap to a v1 trap
    % delete Counter64 objects from vbs
    NVbs = lists:filter(fun(Vb) when Vb#varbind.variabletype =/= 'Counter64' ->
				true;
			   (_) -> false
			end, Vbs),
    case Oid of
	[1,3,6,1,6,3,1,1,5,Specific] ->
	    Enter = ?snmp,
	    Spec = Specific - 1;
	_ ->
	    case lists:reverse(Oid) of
		[Last, 0 | First] ->
		    Enter = lists:reverse(First),
		    Spec = Last;
		[Last | First] ->
		    Enter = lists:reverse(First),
		    Spec = Last
	    end
    end,
    TrapPdu = make_v1_trap_pdu(Enter, Spec, NVbs, SysUpTime),
    AddrCommunities = mk_addr_communities(V1Res),
    lists:foreach(fun({Community, Addrs}) ->
			  NetIf ! {send_pdu, 'version-1', TrapPdu,
				   {community, Community}, Addrs}
		  end, AddrCommunities).
    
send_v2_trap(_TrapRec, [], _Vbs, _Recv, _NetIf, _SysUpTime) -> ok;
send_v2_trap(#notification{oid = Oid}, V2Res, Vbs, Recv, NetIf, SysUpTime) ->
    send_v2_notif(Oid, V2Res, Vbs, Recv, NetIf, SysUpTime);
send_v2_trap(#trap{enterpriseoid = Enter, specificcode = Spec},
	     V2Res, Vbs, Recv, NetIf, SysUpTime) ->
    %% Use alg. in rfc1908 to map a v1 trap to a v2 trap
    case Enter of
	?snmp ->
	    Oid = ?snmpTraps ++ [Spec + 1],
	    Enterp = sys_object_id();
	_ ->
	    Oid = Enter ++ [0, Spec],
	    Enterp = Enter
    end,
    ExtraVb = #varbind{oid = ?snmpTrapEnterprise_instance,
		       variabletype = 'OBJECT IDENTIFIER',
		       value = Enterp},
    send_v2_notif(Oid, V2Res, Vbs ++ [ExtraVb], Recv, NetIf, SysUpTime).
    
send_v2_notif(Oid, V2Res, Vbs, Recv, NetIf, SysUpTime) ->
    IVbs = [#varbind{oid = ?sysUpTime_instance,
		     variabletype = 'TimeTicks',
		     value = SysUpTime},
	    #varbind{oid = ?snmpTrapOID_instance,
		     variabletype = 'OBJECT IDENTIFIER',
		     value = Oid} | Vbs],
    TrapRecvs = get_trap_recvs(V2Res),
    InformRecvs = get_inform_recvs(V2Res),
    do_send_v2_trap(TrapRecvs, IVbs, NetIf),
    do_send_v2_inform(InformRecvs, IVbs, Recv, NetIf).

%%-----------------------------------------------------------------
%% This one is used to maintain the IP check in intCommunityTable;
%% it will be removed in the future, when v3 is implemented.
%%-----------------------------------------------------------------
get_ip({snmpUDPDomain, {Ip, Udp}}) ->
    Ip;
get_ip(_) ->
    [].

get_trap_recvs([{Addr, Community, trap} | T]) ->
    [{Addr, Community} | get_trap_recvs(T)];
get_trap_recvs([_ | T]) ->
    get_trap_recvs(T);
get_trap_recvs([]) ->
    [].

get_inform_recvs([{Addr, Community, {inform, Timeout, Retry}} | T]) ->
    [{Addr, Community, Timeout, Retry} | get_inform_recvs(T)];
get_inform_recvs([_ | T]) ->
    get_inform_recvs(T);
get_inform_recvs([]) ->
    [].


do_send_v2_trap([], _Vbs, _NetIf) ->
    ok;
do_send_v2_trap(Recvs, Vbs, NetIf) ->
    TrapPdu = make_v2_notif_pdu(Vbs, 'snmpv2-trap'),
    AddrCommunities = mk_addr_communities(Recvs),
    lists:foreach(fun({Community, Addrs}) ->
			  NetIf ! {send_pdu, 'version-2', TrapPdu,
				   {community, Community}, Addrs}
		  end, AddrCommunities),
    ok.

do_send_v2_inform([], _Vbs, Recv, _NetIf) ->
    deliver_recv(Recv, snmp_targets, []);
do_send_v2_inform(Recvs, Vbs, Recv, NetIf) ->
    Targets = lists:map(fun({Addr, _, _, _}) -> Addr end, Recvs),
    deliver_recv(Recv, snmp_targets, Targets),
    lists:foreach(
      fun({Addr, Community, Timeout, Retry}) ->
	      proc_lib:spawn_link(?MODULE, init_v2_inform,
				  [Addr, Timeout, Retry, Vbs,
				   Recv, NetIf, Community])
      end, 
      Recvs).

%% New process
init_v2_inform(Addr, Timeout, Retry, Vbs, Recv, NetIf, Community) ->
    %% Make a new Inform for each recipient; they need unique
    %% request-ids!
    {A1,A2,A3} = erlang:now(),
    random:seed(A1,A2,A3),
    InformPdu = make_v2_notif_pdu(Vbs, 'inform-request'),
    Msg = {send_pdu_req, 'version-2', InformPdu, {community, Community},
	   [Addr], self()},
    send_v2_inform(Addr, Timeout*10, Retry, Msg, Recv, NetIf).
    

send_v2_inform(Addr, Timeout, -1, Msg,  Recv, NetIf) ->
    deliver_recv(Recv, snmp_notification, {no_response, Addr});
send_v2_inform(Addr, Timeout, Retry, Msg, Recv, NetIf) ->
    io:format("using timeout: ~p\n", [Timeout]),
    NetIf ! Msg,
    receive
	{snmp_response_received, _Vsn, _Pdu, _From} ->
	    deliver_recv(Recv, snmp_notification, {got_response, Addr})
    after
	Timeout ->
	    send_v2_inform(Addr, Timeout*2, Retry-1, Msg, Recv, NetIf)
    end.

deliver_recv(no_receiver, MsgId, _Result) ->
    ok;
deliver_recv({Tag, Receiver}, MsgId, Result) ->
    Msg = {MsgId, Tag, Result},
    case Receiver of
	Pid when pid(Pid) ->
	    Pid ! Msg;
	Name when atom(Name) ->
	    catch Name ! Msg;
	{M, F, A} ->
	    catch M:F([Msg | A]);
	Else ->
	    snmp_error:user_err("snmp: bad receiver, ~w\n", [Else])
    end;
deliver_recv(Else, _MsgId, _Result) ->
    snmp_error:user_err("snmp: bad receiver, ~w\n", [Else]).

check_all_varbinds(MibView, [#varbind{oid = Oid} | Vbs]) ->
    case snmp_acm:validate_mib_view(Oid, MibView) of
	true -> check_all_varbinds(MibView, Vbs);
	false -> false
    end;
check_all_varbinds(_MibView, []) -> true.

%%--------------------------------------------------
%% Functions to access the local mib.
%%--------------------------------------------------
sys_object_id() ->
    case snmp_agent:do_get(snmp_acm:get_root_mib_view(),
			   [#varbind{oid = ?sysObjectID_instance}],
			   true) of
	{noError, _, [#varbind{value = Value}]} ->
	    Value;
	X ->
	    snmp_error:user_err("sysObjectID bad return value ~p", [X])
    end.


local_get(Name, Indexes, Elements) ->
    case snmp_internal_mib:check_status(Name, Indexes, status_no(Name)) of
	true -> apply(snmp_internal_mib, Name, [get, Indexes, Elements]);
	false -> undefined
    end.

%%-----------------------------------------------------------------
%% Finds next active row.
%%-----------------------------------------------------------------
local_next(Name, Indexes) ->
    case snmp_internal_mib:table_next(Name, Indexes) of
	endOfTable -> endOfTable;
	NextIndexes -> 
	    case snmp_internal_mib:check_status(Name, NextIndexes, 
						status_no(Name)) of
		true -> NextIndexes;
		false -> local_next(Name, Indexes)
	    end
    end.

status_no(intCommunityTable) -> ?intCommunityStatus;
status_no(intViewTable) -> ?intViewStatus;
status_no(intAddressTable) -> ?intAddressStatus.


%% Collect all ADDRs for each community together.
%% In: [{Addr, Community}]
%% Out: [{Community, [Addr]}]
mk_addr_communities(Recvs) ->
    [{Addr, Comm} | T] = lists:keysort(2, Recvs),
    mic(T, Comm, [Addr], []).

mic([{Addr, Comm} | T], CurComm, AddrList, Res) when Comm == CurComm ->
    mic(T, CurComm, [Addr | AddrList], Res);
mic([{Addr, Comm} | T], CurComm, AddrList, Res) ->
    mic(T, Comm, [Addr], [{CurComm, AddrList} | Res]);
mic([], CurComm, AddrList, Res) ->
    [{CurComm, AddrList} | Res].

pdebug(Format, X) -> 
    case get(debug) of
	false -> ok;
	_ ->
	    Form = lists:concat(["** SNMP Agent debug: ~n   ", Format, "\n"]),
	    io:format(Form, X)
    end.

