%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%%
%% File    : cover.erl
%% Author  : Joe Armstrong
%% Purpose : Coverage anlysis and profiling data
%% Written : 930427
%% Change log: 950420
%%    Re-written to use data base BIS's ets: ...
%%    so we can remove the "bump" stuff from the run-time machine
%% Change log: 961028 : Kenneth Lundin
%% Modifications and bugfixes:
%% quick_anal renamed to quick_analyse
%% -import removed
%% file/1 changed to return {error,Reason} when applicable
%% Change log: 961117 : Kenneth Lundin
%% new file/2, quick_analyse_as_term/1
%% changed printout  format for quick_analyse
%%
%% Bugs asserts a few variables called 'cover$$...'

-module(cover).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

-export([all_covered/0, all_files/1, all_files/2,  analyse/1,  check/1, 
	 file/1, file/2, 
	 most_called/0, most_called/1,
	 quick_analyse/1, quick_analyse_as_term/1,
	 zero/0, zero/1]).

-export([bump/4, internal/0, get_table_name/0]).

%% cover:all_covered()       -- a list of all files available for analysis
%% cover:all_files(Dir)      -- Run cover:file/1 on all files in Dir
%% cover:all_files(Dir,Options)      -- Run cover:file/2 on all files in Dir
%% cover:analyse(Module)     -- Coverage analysis (makes listing file)
%% cover:check(Module)       -> true | false
%% cover:checkv(Module)      -> true | false (verbose version of check)
%% cover:file(Module)        -- prepairs file for analysis
%% cover:most_called()       -- most called of all modules
%% cover:most_called(Module) -- Coverage statistics
%% cover:quick_analyse(Module)  -- quick analysis with printout
%% cover:quick_analyse_as_term(Module)  -- quick analysis returning a term
%% cover:zero()              -- zero coverage data
%% cover:zero(Module)        -- zero coverage data in Module

%% cover:file(Module) -- prepairs file for analysis
%%   1) Assumes <Module> is loaded and runnable
%%   2) Transforms <File>.erl to <File>.COVER.erl
%%   3) Compiles <File>.Cover.erl -> <File>.Cover.erl.(jam|beam) or other 
%%      object file extension returned by code:objfile_extension/0
%%   4) Loads Cover.jam



%%______________________________________________________________________
%% Cover server code:
%%   the process with registered name cover_server_001
%%   "Owns" the coverage data base


% the catch is here because the application might under rare
% circumstances not be started e.g "erl -boot start_old"
start() ->
    case whereis(cover_server_001) of
	undefined ->
	    register(cover_server_001, spawn(cover, internal, []));
	Pid ->
	    true
    end.

cover_server_running() ->
    case whereis(cover_server_001) of
	undefined ->
	    false;
	_ ->
	    true
    end.

get_table_name() ->
    start(),
    cover_server_001 ! {self(), get_table},
    receive
	{cover_server_001, Table} ->
	    Table
    end.

internal() ->
    io:format("~n~n~n"
	      "=====================================~n~n"
	      "Please note that COVER in due time~n"
	      "will be replaced by the COAST module!~n~n"
	      "=====================================~n~n"),
    Table = ets:new('ets_coverage_data', [set, public]),
    loop(Table).

loop(Table) ->
    receive 
	{From, get_table} ->
	    From ! {cover_server_001, Table},
	    loop(Table)
    end.

%%______________________________________________________________________
%%  END of server
%%______________________________________________________________________

%%______________________________________________________________________
%%  bump -- is called by the routines which have been compiled for
%%  coverage analysis.
%%  Note we locally cache the table name for efficiency reasons

bump(Module, Func, Arity, Index) ->
    Table = case get('cover$$tableName') of
		undefined ->
		    Tab = get_table_name(),
		    put('cover$$tableName', Tab),
		    Tab;
		Tab ->
		    Tab
	    end,
    N = lookup(Table, Module, Func, Arity, Index),
    %% io:format('Insert:~w\n',[{Table, Module, Func, Arity, Index}]),
    ets:insert(Table, {{Module, Func, Arity, Index},N+1}).

lookup(Table, Module, Func, Arity, Index) ->
    Val = ets:lookup(Table, {Module,Func,Arity,Index}) ,
    %% io:format('Lookup:~w = ~w\n',[{Table, Module, Func, Arity, Index},Val]),
    case Val of
	[{_,N}] when integer(N) ->
	    N;
	[] ->
	    0
    end.

%%______________________________________________________________________
%%  The transformer

all_files(Dir) ->
    all_files(Dir, false).

all_files(Dir, Options) ->
    case file:list_dir(Dir) of
	{ok, Files} ->
	    Dir2 = case lists:reverse(Dir) of % assure that Dir ends with "/"
		       [$/|Rest] -> Dir;
		       _ -> Dir ++ "/"
		   end,
	    all_files2(Dir2, Files, Options);
	_ ->
	    true
    end.

all_files2(Dir, [H|T], Options) ->
    case lists:reverse(H) of
	[$l,$r,$e,$.,$R,$E,$V,$O,$C,$.|Z] ->
	    all_files2(Dir, T, Options);
	[$l,$r,$e,$.,$r,$e,$v,$o,$c] -> % don't try to cover module cover
	    all_files2(Dir, T, Options);
	[$l,$r,$e,$.|Z] ->
	    case Options of
		false -> file(Dir ++ lists:reverse(Z));
		_ -> file(Dir ++ lists:reverse(Z), Options)
	    end,
	    all_files2(Dir, T, Options);
	_ ->
	    all_files2(Dir, T, Options)
    end;
all_files2(Dir, [], _) -> [].

file(File) ->
    file(File, []).
	
file(File, Options0) ->
    case filename:find_src(File) of
	{error, Reason} ->
	    io:format("Failed to locate file: ~s\n", [File]),
	    {error, Reason};
	{Source, OldOptions} ->
     	    Options = lists:keydelete(outdir, 1, Options0++OldOptions),
	    io:format("Input file:~s\n", [Source]),
	    OutFileRoot = filename:basename(Source),
	    io:format("Output file:~s.COVER.erl\n", [OutFileRoot]),
	    file2(Source, Options, OutFileRoot)
    end.
	    
file2(In, Options, Out) ->
    % io:format("file2(~s,~n~w,~n~s)~n",[In,Options,Out]),
    OutFile = lists:concat([Out,".COVER.erl"]),
    InFile = lists:concat([In,".erl"]),
    IncludePath = lists:reverse(lists:foldl(fun({i,I},Acc) -> [I|Acc];
					       (_,Acc) -> Acc end, [],Options)),
    Defines = lists:reverse(lists:foldl(fun({d,D},Acc) -> [D|Acc];
					   (_,Acc) -> Acc end, [],Options)),
    case catch transform(InFile, OutFile,
			 epp:open(InFile, ["."|IncludePath], Defines),
			 file:open(OutFile,write)) of
	transformed ->
	    io:format("Now compiling module ~s~n", [Out]),
	    ObjExt = code:objfile_extension(),
	    case compile:file(lists:concat([Out,".COVER"]), 
			      [report_errors, report_warnings|Options]) of
		{ok, Module} ->
		    file:rename(lists:concat([Out,".COVER",ObjExt]), 
				lists:concat([Out, ObjExt])),
		    io:format("Module ~p will now be purged and re-loaded~n",
			      [Module]),
		    code:purge(Module),
		    ObjFile = lists:concat([Out,".COVER",ObjExt]),
%		    {ok, Bin} = file:read_file(ObjFile),
%		    code:load_binary(Module,ObjFile,Bin);
		    code:load_abs(lists:concat([Module]));
		_ ->
		    compilation_failed
	    end;
	Error ->
	    transformation_failed,
	    Error
    end.

%%______________________________________________________________________
%% transform file

transform(File, OutFile, {ok, In}, {ok, Out}) ->
    io:format('Transforming ~s for profiling\n',[File]),
    put('cover$$maxBumps',0),
    put('cover$$allFuncs', []),
    Ret = translate(In, Out),
    windup(Out, OutFile), 
    epp:close(In),
    file:close(Out),
    Tab = get_table_name(),
    erase('cover$$maxBumps'),
    erase('cover$$allFuncs'),
    io:format("~n", []),
    Ret;
transform(_, _, _, _) ->
    error_in_opening_some_file.

translate(In,Out) ->
    loop(epp:parse_erl_form(In),In,Out).

loop({ok,{attribute,Line,module,Mod}}, In, Out) ->
    put(module,Mod),
    io:format(Out,'~s',[erl_pp:form({attribute,Line,module,Mod})]),
    io:format(Out,'~s',[erl_pp:form({attribute,Line,export,
				     [{'cover_data_$$_private',0}]})]),
    translate(In,Out);
loop({ok,Form}, In, Out) ->
    %% format('read:~w~n',[Form]),
    tell_user(Form),
    Form1 = munge(Form),
    %% format('wrote:~w~n',[Form1]),
    io:format(Out,'~s',[erl_pp:form(Form1)]),
    translate(In,Out);
loop({eof,_},In,Out) ->
    transformed;
loop({error,{Line,Mod,Error}},In,Out) ->
    io:format('Error Line:~w ~s~n',[Line, apply(Mod,format_error,[Error])]),
    translate(In,Out).

windup(Out, OutFile) ->
    Funcs = get('cover$$allFuncs'),
    Form = {function,0,'cover_data_$$_private',0,
	    [{clause,0,[],[],
	       [erl_parse:abstract({OutFile,Funcs})]}]},
    io:format(Out,'~s',[erl_pp:form(Form)]).

tell_user({function,_,Name,Arity,Clauses2}) ->
    io:format('~w/~w ',[Name, Arity]);
tell_user(_) ->
    true.

munge({attribute,Line,module,Mod}) ->
    put(module,Mod),
    {attribute,Line,module,Mod};
munge({function, Line, Name, Arity, Clauses}) ->
    {function, Line, Name, Arity, munge_clauses(Name, Arity, Clauses)};
munge(X) ->
    X.

munge_clauses(Name, Arity, []) -> [];
munge_clauses(Name, Arity, [{clause, Line, Head,Guard,Body}|T]) ->
    [{clause, Line, Head,Guard,munge_body(Name, Arity, Body)}|
     munge_clauses(Name, Arity, T)].

munge_body(Name, Arity, [H|T]) ->
    [add_bump(Name, Arity),munge_expr(Name, Arity, H)|
     munge_body(Name, Arity, T)];
munge_body(Name, Arity, []) ->
    [].

munge_expr(Name, Arity, {'if', Line, Opts}) -> 
    {'if',Line, munge_option_list(Name, Arity, Opts)};
munge_expr(Name, Arity, {'case', Line, Func, Opts}) -> 
    {'case', Line, Func, munge_option_list(Name, Arity, Opts)};
munge_expr(Name, Arity, {'receive', Line, Opts, Func, TimeOutBody}) -> 
    {'receive', Line,  
     munge_option_list(Name, Arity, Opts), Func, 
     munge_body(Name, Arity, TimeOutBody)};
munge_expr(Name, Arity, {'receive', Line, Opts}) -> 
    {'receive',Line, munge_option_list(Name, Arity, Opts)};
munge_expr(Name, Arity, {match,Line,LS,RS}) ->
    {match,Line,munge_expr(Name, Arity, LS),munge_expr(Name, Arity, RS)};
munge_expr(Name, Arity, {call,Line,Label,Args}) ->
    {call,Line,Label,munge_exprs(Name, Arity, Args)};
munge_expr(Name, Arity, X) ->
    X.

munge_exprs(Name, Arity, [E|Es]) ->
    E1 = munge_expr(Name, Arity, E),
    [E1|munge_exprs(Name, Arity, Es)];
munge_exprs(Name, Arity, []) ->
    [].

munge_option_list(Name, Arity, []) -> [];
munge_option_list(Name, Arity, [{clause, Line, Guard, Body}|T]) ->
    [{clause, Line, Guard,munge_body(Name, Arity, Body)}|
     munge_option_list(Name, Arity, T)];
munge_option_list(Name, Arity, [{clause, Line, Pattern,Guard,Body}|T]) ->
    [{clause, Line, Pattern,Guard,munge_body(Name, Arity, Body)}|
     munge_option_list(Name, Arity, T)];
munge_option_list(Name, Arity, X) ->
    io:format('oops:~w~n',[X]).

add_bump(Name, Arity) ->
    Mod = get(module),
    N = get('cover$$maxBumps'),
    put('cover$$maxBumps', N + 1),
    put('cover$$allFuncs', [{Name,Arity,N}|get('cover$$allFuncs')]),
    {call,0,{remote,0,{atom,0,cover},{atom,0,bump}},
     [{atom,0,Mod},{atom,0,Name},{integer,0,Arity},{integer,0,N}]}.
%%______________________________________________________________________
%% most_called().
%%   list the most called functions in all modules

most_called() ->
    Table = get_table_name(),
    L = ets:match(Table, {{'$1', '$2','$3','$4'}, '$5'}),
    L1 = lists:reverse(lists:sort(merge_calls(lists:sort(L)))),
    show(cum(L1)).

merge_calls([[M,F,A,C,N1],[M,F,A,_,N2]|T]) ->
    merge_calls([[M,F,A,C,N1+N2]|T]);
merge_calls([[M,F,A,_,N]|T]) ->
    [{N,{M,F,A}}|merge_calls(T)];
merge_calls([]) ->
    [].

cum(L) ->
    %% io:format('here:~w\n',[L]),
    Tot = sum_calls(L, 0.0),
    cum(L, Tot, 0.0, []).

cum([{N, W}|T], Tot, Cum, L) ->
    Cum1 = Cum + N,
    cum(T, Tot, Cum1, [{N, (N/Tot)*100, (Cum1/Tot)*100, W}|L]);
cum([], _, _, L) ->
    lists:reverse(L).

sum_calls([{N, _}|T], Sum) -> sum_calls(T, N + Sum);
sum_calls([], Sum)         -> Sum.

%%______________________________________________________________________
%% most_called(Module)
%%   list the most called functions in <Module>

most_called(Module) ->
    Table = get_table_name(),
    L = ets:match(Table, {{Module, '$1','$3','$4'}, '$5'}),
    L1 = add_module_info(L, Module, []),
    L2 = lists:reverse(lists:sort(merge_calls(lists:sort(L1)))),
    show(cum(L2)).

add_module_info([H|T], Mod, L) ->
    add_module_info(T, Mod, [[Mod|H]|L]);
add_module_info([], _, L) ->
    L.

%%______________________________________________________________________
%% zero() 
%%    zero the coverage statistics in all modules

zero() ->
    zero1(all_covered()).

zero1([H|T]) -> zero(H),zero1(T);
zero1([]) -> true.

%%______________________________________________________________________
%% zero(Module) 
%%    zero the coverage statistics in <Module>

zero(Module) ->
    Table = get_table_name(),
    ets:match_delete(Table, {{Module, '_','_','_'}, '_'}).

%%______________________________________________________________________
%% all_covered()

all_covered() ->
    Table = get_table_name(),
    L = ets:match(Table, {{'$1', '_', '_','_'}, '_'}),
    remove_duplicates(L, []).

remove_duplicates([[Mod]|T], L) ->
    case lists:member(Mod, L) of
	true -> remove_duplicates(T, L);
	false -> remove_duplicates(T, [Mod|L])
    end;
remove_duplicates([], L) ->
    L.

show(L) -> show(1, L).

show(Index, [{N, P1, P2, {Mod, Func, Arity}}|T]) ->
    io:format('~8w ~10w ~6.2f ~6.2f ~w:~w/~w \n',
	      [Index, N, P1, P2, Mod, Func,Arity]),
    show(Index+1, T);
show(_, []) ->
    true.

%%______________________________________________________________________
%%  checks if the module is compiled for coverage analysis

check(Module) ->
    erlang:function_exported(Module, 'cover_data_$$_private', 0).



%%______________________________________________________________________
%% quick_analyse(Module)
%%   performs and prints a quick analysis of the coverage data in 
%%   Module

quick_analyse(Module) ->
     case quick_analyse_as_term(Module) of
	 A when atom(A) ->
	     A;
	 R ->
	     io:format("COVERAGE ANALYSIS (QUICK) FOR MODULE ~s~n",[Module]),
	     io:format("*** ~w PERCENT *** of the statements are covered~n",
		       [element(5,R)]),
	     io:format("NOT COVERED:~w COVERED: ~w TOTAL NUMBER OF STATEMENTS: ~w~n",
		       [element(3,R),element(4,R),element(3,R)+element(4,R)]),
	     io:format("~nCOVERAGE FOR FUNCTIONS IN MODULE ~w~n",[Module]),
	     print_func_coverage(keysplit(4,(element(2,R)))),
	     ok
     end.

print_func_coverage([H|T]) ->
    [{_,_,Cov,P}|_] = H,
    case {Cov,P} of 
	{0,0} -> 
	    io:format("*** NOT CALLED *** ~n",[]);
	{_,X} ->
	    io:format("*** ~w PERCENT COVERED *** ~n",[X])
    end,
    print_percent(H),
    print_func_coverage(T);
print_func_coverage([]) -> io:format("~n",[]).

print_percent(L) ->
    print_percent(L,0);
print_percent([]) -> true.
print_percent([{{F,A},_,_,_}|T],Num) when Num < 5 ->
    io:format("~w/~w ",[F,A]),
    print_percent(T,Num + 1);
print_percent([{{F,A},_,_,_}|T],Num)  ->
    io:format("~w/~w~n ",[F,A]),
    print_percent(T,0);
print_percent([],_) -> 
    io:format("~n",[]).

    
%%
%% keysplit(N, TupleList)
%%
%% Return a list of lists in wich
%% Each list has the same element(N, T)
%% 
keysplit(N, TupleList) ->
    keysplit(lists:keysort(N, TupleList), N, [], [], []).

keysplit([T|Ts], N, Key, Acc1, Acc2) when element(N,T) == Key ->
    keysplit(Ts, N, Key, [T|Acc1], Acc2);
keysplit([T|Ts], N, [], Acc1, Acc2) ->
    keysplit(Ts, N, element(N,T), [T|Acc1], Acc2);
keysplit([T|Ts], N, Key, Acc1, Acc2) ->
    keysplit(Ts, N, element(N, T), [T], [Acc1|Acc2]);
keysplit([], _, _, Acc1, Acc2) -> 
    lists:reverse([Acc1|Acc2]).

%%______________________________________________________________________
%% quick_analyse_as_term(Module)
%%   performs a quick analysis of the coverage data in 
%%   Module and returns the result as an Erlang term with the 
%%   following format: 
%%   {Module, [{{Func,Arity}, NumNotCovered, NumCovered, PercentCovered}, ...],
%%     TotalNotCovered,TotalCovered,PercentCovered}
%%   The List with coverage per function is sorted with PercentCovered as key

quick_analyse_as_term(Module) ->

    case ok_to_analyse(Module) of
	ok ->
	    {_,Funcs} = apply(Module, 'cover_data_$$_private', []),
	    put('cover_module_max',length(Funcs)),
	    Table = get_table_name(),
	    qanal2(lists:reverse(Funcs), Module, Table,0);
	Error ->
	    Error
    end.

qanal2(FuncL, Module, Table, Nzeros) ->
    qanal2(FuncL, Module, Table, Nzeros, []).

qanal2([H|T], Module, Table, _Nzeros, Out) ->
    {L1, L2} = split_list(T, [H]),
    qanal2(L2, Module, Table, 0,
	   [func_covered(L1, length(L1), Module, Table)|Out]);

qanal2([], Module, Table, _Nzeros, Out) ->
    Sum = get('cover_module_max'),
    erase('cover_module_max'),
    Covered = lists:foldl(fun(E,Csum) -> element(3,E) + Csum end, 0, Out),
    {Module, 
     lists:keysort(4,Out), 
     Sum - Covered, 
     Covered, round(Covered/Sum*100)}.



split_list([{Func,Arity,I1}|T], [{Func,Arity,I2}|L]) ->
    split_list(T, [{Func,Arity,I1},{Func,Arity,I2}|L]);
split_list(L1, L2) ->
    {L2, L1}.

%%______________________________________________________________________
%%  all_zeros([{Func,Arity,Index}], Module, Table, N) ->  no | {yes, N'}
%%    no       -- if the function *was not* all zeros
%%    {yes, N} -- if the function *was* all zeros. N is the number of zeros

all_zeros([{Func,Arity,I}|T], Module, Table, N) ->
    case lookup(Table, Module, Func, Arity, I) of
	0 ->
	    all_zeros(T, Module, Table, N+1);
	_ ->
	    no
    end;
all_zeros([], _, _, N) ->
    {yes, N}.

func_covered(L, FuncLen, Module, Table) ->
    [{Func,Arity,_} |_] = L, 
    Not_cov = count_zeros(L, Module, Table, 0),
    Cov = FuncLen - Not_cov,
    {{Func,Arity}, Not_cov, Cov, round(Cov/FuncLen*100)}.

%%______________________________________________________________________
%% all_zeros([{Func,Arity,Index}], Module, Table, N) -> N'
%%   Count the number of zero evaluated lines in a function

count_zeros([{Func,Arity,I}|T], Module, Table, N) ->
    case lookup(Table, Module, Func, Arity, I) of
	0 ->
	    count_zeros(T, Module, Table, N+1);
	_ ->
	    count_zeros(T, Module, Table, N)
    end;
count_zeros([], _, _, N) ->
    N.

%%______________________________________________________________________
%% analyse(Module) 
%%   complete coverage analysis

analyse(Module) ->
    case ok_to_analyse(Module) of
	ok ->
	    analyse1(Module);
	Error ->
	    Error
    end.

analyse1(Module) ->
    case code:is_loaded(Module) of
	{file, File} ->
	    io:format('Found Object code in:~s\n',[File]),
	    {ok, Info} = file:file_info(File),
	    ObjLMT = element(5, Info),
	    io:format('Object code last modified:~w\n',[ObjLMT]),
	    %% try to locate the source
	    {Source,_} = apply(Module, 'cover_data_$$_private', []),
	    io:format('Found source file in:~s\n',[Source]),
	    {ok, Info1} = file:file_info(Source),
	    SrcLMT = element(5, Info1),
	    io:format('Source code last modified:~w\n',[SrcLMT]),
	    if 
		ObjLMT > SrcLMT ->
		    %% ok_to_analyse;
		    analyse(Source, Module);
		true ->
		    time_staps_wrong
	    end

	    ;
	false ->
	    module_not_loaded
    end.

analyse(InFile, Module) ->
    OutFile = lists:concat([extract_filename(InFile),".out"]),
    io:format("Analysing Source:~s module:~w\n",[InFile, Module]),
    io:format("Output File:~s\n",[OutFile]),
    {ok, InStream}  = epp:open(InFile, ["."], []),
    {ok, OutStream} = file:open(OutFile, write),
    analyse_loop(InStream, OutStream).

analyse_loop(In, Out) ->
    analyse_loop1(epp:parse_erl_form(In),In,Out).

analyse_loop1({ok,Form},In,Out) ->
    %% format('read:~w~n',[Form]),
    tell_user(Form),
    {Flag, Form1} = merge_bump_data(Form, false),
    case Flag of
	true ->
	    io:format(Out,'~s\n',[erl_pp:form(Form1)]);
	false ->
	    true
    end,
    analyse_loop(In,Out);
analyse_loop1({eof,_},In,Out) ->
    epp:close(In),
    file:close(Out);
analyse_loop1({error,{Line,Mod,What}},In,Out) ->
    io:format('read error in line:~w ~s~n',
	      [Line, apply(format_error,Mod,What)]),
    analyse_loop(In,Out).

merge_bump_data({attribute,Line,module,Mod}, Flag) ->
    put(module,Mod),
    {Flag, {attribute,Line,module,Mod}};
merge_bump_data({function, Line, Name, Arity, Clauses}, Flag) ->
    {Flag1, Clauses1} = merge_bump_data1(Clauses, Flag),
    {Flag1, {function, Line, Name, Arity, Clauses1}};
merge_bump_data(X, Flag) ->
    {Flag, X}.

merge_bump_data1([H|T], Flag) ->
    {Flag1, H1} = merge_bump_data1(H, Flag),
    {Flag2, T1} = merge_bump_data1(T, Flag1),
    {Flag2, [H1|T1]};
merge_bump_data1({call,Line,{remote,_,{atom,_,cover},{atom,_,bump}},Args},
		 Flag) ->
    [{atom,L1,Mod},{atom,L2,Func},{integer,L3,Arity},
     {integer,L3,Index}] = Args,
    case lookup(get_table_name(), Mod, Func,Arity, Index) of
	0 ->
	    %% if their is a single zero change the flag to true
	    {true, {string, Line, "**** Not Covered *****"}};
	N ->
	    {true, {string, Line, 
	     lists:flatten(io_lib:format("**** ~w ****", [N]))}}
    end;
merge_bump_data1(X, Flag) when tuple(X) ->
    {Flag1, L} = merge_bump_data1(tuple_to_list(X), Flag),
    {Flag1, list_to_tuple(L)};
merge_bump_data1(X, Flag) ->
    {Flag, X}.

%%______________________________________________________________________
%% extracts the filename from an atom or string
%%   example: extract_filename('a/b/c/foo') -> "foo"

extract_filename(File) ->
    extract_filename(coerce_to_string(File), []).

extract_filename([], L) ->
    lists:reverse(L);
extract_filename([$/|T], _) ->
    extract_filename(T, []);
extract_filename([H|T], L) ->
    extract_filename(T, [H|L]).

%%______________________________________________________________________
%% shorts

coerce_to_string(X) when atom(X) ->
    atom_to_list(X);
coerce_to_string(X) ->
    X.

%%______________________________________________________________________
%%  ok_to_analyse(Module)
%%    checks if the moudle has been compiled for coverage AND if the 
%     coverage server is running

ok_to_analyse(Module) ->
    case check(Module) of
	true ->
	    case cover_server_running() of
		true ->
		    ok;
		false ->
		    coverage_server_is_not_running
	    end;
	false ->
	    module_is_not_compiled_for_coverage_analysis
    end.

    

























