%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(dbg).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').
-export([p/1, p/2, c/3, c/4, i/0, i/1,stop/0, filter/1]).

%% Local exports
-export([loop0/0, dfilter/1, start/0, collect_tracs/2, un_suspend/1,
	do_relay/1]).


start() ->
    register(dbg, spawn(dbg, loop0, [])).

loop0() ->
    process_flag(trap_exit, true),
    process_flag(priority, max),
    loop([], {dbg, dfilter}).

loop(Ps, Filter) ->
    receive
	Trac when tuple(Trac), element(1, Trac) == trace ->
	    P  = suspend(Trac, []),
	    {L, Procs} = collect_tracs([Trac], [P]),
	    lists:foreach(Filter, [], L),
	    un_suspend(lists:delete(nopid, Procs)),
	    loop(Ps, Filter);
	{From, {filter, F}} ->
	    From ! {dbg, Filter},
	    loop(Ps, F);
	{From, i} ->
	    io:format("~-12s ~-21s Trace ~n",
		      ["Pid", "Initial call"]),
	    display_info(Ps, all()),
	    From ! {dbg, ok},
	    loop(Ps, Filter);
	{From, {i, Pid}} ->
	    From ! {dbg, lists:keysearch(Pid, 1, Ps)},
	    loop(Ps, Filter);
	
	{From, {set, P, Flags}} ->
	    Ps2 = doset(P, Flags, Ps),
	    From ! {dbg, ok},
	    loop(Ps2, Filter);

	{'EXIT', P, _} ->
	    loop(lists:keydelete(P,1, Ps), Filter);

	{From,stop} ->
	    unlink_all(Ps), 
	    From ! {dbg, stopped},
	    exit(normal);
	Other ->
	    io:format(user,"** dbg got ~p~n", [Other]),
	    loop(Ps, Filter)
    end.

collect_tracs(Ack, Procs) ->
    receive
	Trac when tuple(Trac), element(1, Trac) == trace ->
	    P = suspend(Trac, Procs),
	    collect_tracs([Trac | Ack], add(P, Procs));
	Trac when tuple(Trac), element(1, Trac) == print ->
	     collect_tracs([Trac | Ack],Procs)
    after 0 ->
	    {lists:reverse(Ack), Procs}
    end.

add(P, L) ->
    case lists:member(P, L) of
	true -> 
	    L;
	false ->
	    [P|L]
    end.


suspend({trace, From, call, Func}, Procs) when node(From) == node() ->
    case lists:member(From, Procs)  of
	true ->
	    nopid;
	false ->
	    trac(From, true, [suspend]),
	    From
    end;

suspend(X, _) -> 
    nopid.

un_suspend([Pid|Tail]) when node(Pid) == node() ->
    trac(Pid, false, [suspend]), 
    un_suspend(Tail);
un_suspend([]) -> ok.

unlink_all([]) -> ok;
unlink_all([{P,_}|T]) ->
    trac(P, false, all()),
    unlink(P),
    unlink_all(T).

trac(Pid, How, Flag) when node(Pid) == node() ->
    case catch erlang:trace(Pid, How, Flag) of
        true -> 
	    true;
	false ->
	    false;
        _ -> 
	    io:format(user, "** Illegal trace request ** ~p\n",
		      [{erlang, trace, [Pid, How, Flag]}]),
	    false
    end;

trac(Pid, How, Flag) ->
    relay(Pid) ! {self(), erlang, trace, [Pid, How, Flag]},
    true.
        

%% Since we are not allowed to do erlang:trace/3 on remote
%% processe we create a help process at the remote node to
%% do the job for us 
relay(P) when pid(P), node(P) /= node() ->
    N = node(P),
    case get(N) of
	undefined -> 
	    Pid = spawn_link(N, dbg, do_relay, [self()]),
	    put(N, Pid),
	    Pid;
	Pid -> 
	    Pid
    end.

do_relay(P) ->
    process_flag(trap_exit, true),
    do_relay_1(P).

do_relay_1(P) ->
    receive
	{P, M,F, A} ->
	    case catch apply(M,F,A) of
		true -> ok;
		_ ->  P ! {print, "** Illegal trace request **\n", []}
	    end,
	    do_relay_1(P);
	{'EXIT', P, _} ->
	    exit(normal);
	Other ->             %% Here is the normal case for trace i/o
	    P ! Other, 
	    do_relay_1(P)
    end.

dfilter({trace, 'receive', From, {dbg, ok}}) ->
    ignore;
dfilter({trace, 'receive', From, Msg}) ->
    io:format(user, "(~p) << ~p~n", [From, Msg]);
dfilter({trace, send, To, From,{X, {set, X,_}}}) ->
    ignore;
dfilter({trace, send, To, From,Msg}) ->
    io:format(user, "(~p) ~p ! ~p~n", [From,To,Msg]);
dfilter({trace, From, call, Func}) ->
    io:format(user, "(~p) call ~s~n",
	      [From, lists:flatten(ffunc(Func))]);
dfilter({trace, From, Op, Data}) ->
    io:format(user, "(~p) ~p ~p~n", [From, Op, Data]);
dfilter(X) ->
    io:format(user,"~p~n", [X]).


ffunc({M,F, Argl}) ->
    io_lib:format("~p:~p(~s)", [M, F, fargs(Argl)]);
ffunc(X) -> io_lib:format("~p", [X]).
fargs([]) -> [];
fargs([A]) -> io_lib:format("~p", [A]);  %% last arg
fargs([A|Args]) -> [io_lib:format("~p,", [A]) | fargs(Args)].

doset(P, [clear], Ps) ->
    trac(P, false, all()),
    unlink(P),
    lists:keydelete(P,1,Ps);

doset(P, Flags, Ps) ->  % add flags 
    F = transform_flags(Flags),
    case lists:keysearch(P,1,Ps) of
	false ->
	    link(P),
	    case trac(P, true, F) of
		false ->
		    unlink(P),
		    Ps;
		true ->
		    [{P, F} | Ps]
	    end;
	{value, {_, Oldf}} ->
	    case trac(P, true, F) of
		false ->
		    Ps;
		true ->
		    lists:keyreplace(P,1,Ps,{P,merge(Oldf,F)})
	    end
    end.

merge([H|T],S) ->
    case lists:member(H,S) of
	true -> merge(T,S);
	false -> [H|merge(T,S)]
    end;
merge([], S) -> S.
		
	    
transform_flags([]) -> [];
transform_flags([m|Tail]) -> [send, 'receive' | transform_flags(Tail)];
transform_flags([s|Tail]) -> [send |transform_flags(Tail)];
transform_flags([r|Tail]) -> ['receive' | transform_flags(Tail)];
transform_flags([c|Tail]) -> [call | transform_flags(Tail)];
transform_flags([call|Tail]) -> [call|transform_flags(Tail)];
transform_flags([b|Tail]) -> [bifs | transform_flags(Tail)];
transform_flags([p|Tail]) -> [proc|transform_flags(Tail)];
transform_flags([sos|Tail]) -> [set_on_spawn | transform_flags(Tail)];
transform_flags([sol|Tail]) -> [set_on_link | transform_flags(Tail)];
transform_flags([sofs|Tail]) -> [set_on_first_spawn | transform_flags(Tail)];
transform_flags([sofl|Tail]) -> [set_on_first_link|transform_flags(Tail)];
transform_flags([all|_]) -> all();
transform_flags([F|Tail]) -> 
    error_logger:format("** Bad trace flag ~p~n", [F]),
    transform_flags(Tail).


all() -> [send,'receive',call,bifs,procs,set_on_spawn,
	  set_on_first_spawn, set_on_link, set_on_first_link].

display_info([{Pid, What}|T], All) ->
    case pinfo(Pid, initial_call) of
        undefined ->
            display_info(T, All);
	{initial_call, Call} when What == All ->
	    io:format("~-12s ~-21s ~s~n",
		      [io_lib:format("~w",[Pid]),
		       io_lib:format("~p", [Call]),
		       "Trace all"]),
	    display_info(T, All);
        {initial_call, Call} ->
	    io:format("~-12s ~-21s ~s~n",
		      [io_lib:format("~w",[Pid]),
		       io_lib:format("~p", [Call]),
		       format_trace(What)]),
	    display_info(T,All)
    end;
display_info([], _) -> ok.

format_trace(What) ->
    lists:flatten(ft(What)).

ft([]) -> [];
ft([Item]) -> [ts(Item)];
ft([Item|T]) -> [ts(Item) ," | ", ft(T)].

ts(send) -> "s";
ts('receive') -> "r";
ts(call) -> "c";
ts(procs) -> "p";
ts(bifs) -> "b";
ts(set_on_spawn) -> "sos";
ts(set_on_first_spawn) -> "sofs";
ts(set_on_link) -> "sol";
ts(set_on_first_link) -> "sofl".


p(Pid) ->
    p(Pid, [m]).
p(Pid, Flags) when atom(Flags) ->
    p(Pid, [Flags]);
p(Pid, Flags) ->
    P = to_pid(Pid),
    case {P, check_flags(Flags, allflags())} of
	{P, true} when pid(P)  ->
	    req({set, P, Flags});
	_ ->
	    false
    end.

check_flags([], _) -> true;
check_flags([Flag|Tail], All) ->
    case lists:member(Flag, All) of 
	true -> 
	    check_flags(Tail, All);
	false -> 
	    error_logger:format("** Bad trace flag ~p~n", [Flag]),
	    false
    end.

allflags() -> [m,r,s,b,p,c,sol,sofl,sos,sofs,all,clear,call,
	       messages,send,'receive',bifs,proc].

req(R) ->
    ensure() ! {self(), R},
    receive
	{dbg, Reply} -> Reply
    end.

i() ->    req(i).
i(Pid) -> req({i, to_pid(Pid)}).
stop() -> req(stop).
filter(F) -> req({filter, F}).

ensure() ->
    case whereis(dbg) of
	undefined -> 
	    start(),
	    dbg;
	Pid ->
	    Pid
    end.

%% Trace a shell call
c(M,F,A) ->
    c(M,F,A,all).

c(M,F,A,Flags) ->
    p(self(), Flags),
    Res = apply(M,F,A),
    p(self(), clear),
    Res.


%%
%% Turn something into a pid, return 'nopid' on failure 
%%

to_pid(X) when pid(X) ->
    check_pid(X);
to_pid(X) when atom(X) ->
    case whereis(X) of
	undefined -> 
	    error_logger:format("** No such process ~p~n", [X]),
	    false;
	Pid ->
	    check_pid(Pid)
    end;
to_pid(X) when integer(X) ->
    to_pid({0,X,1});

to_pid({X,Y,Z}) ->
    Pid = list_to_pid(lists:concat(["<",integer_to_list(X),".",
				    integer_to_list(Y),".",
				    integer_to_list(Z),">"])),
    check_pid(Pid);

to_pid(X) -> nopid.


check_pid(Pid) ->
    case pinfo(Pid, status) of
	undefined ->
	    error_logger:format("** Can't trace process ~p~n", [Pid]),
	    nopid;
	_ ->
	    Pid
    end.

pinfo(P, X) when node(P) == node() -> erlang:process_info(P, X);
pinfo(P, X) -> check(rpc:call(node(P), erlang, process_info, [P, X])).

check({badrpc, _}) -> undefined;
check(X) -> X.

