%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
%% Copyright (C) 1994, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose:   Manage and trace processes
%% Inspired by: The debug BIF debacle :-)
%% This module is merely a gui to the trace/3 BIF.

-module(pxw_pman).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-vsn('$Revision: /main/release/free/1').

%% The user interface exports ...
-export([proc/1, start/0, start/1, start/2, proc/1, proc/3]).

%% Internal Exports ...
-export([init_full/2, do_relay/1, internal/1, ftm1/1]).
-import(lists, [concat/1, delete/2]).


start() -> start(allmods).
start(Mods) -> 
    Size = get_size(Mods),
    start(Mods, Size).

%% Mods is a list of moduls we want
%% I.e if Mods == [a,b] only processes running code from module
%% a and b will be displayed

start(Mods, Size) ->
    pxw:start(),
    spawn(pxw_pman, init_full, [Size+1, Mods]).

%% In case we are limiting the view to just a number of 
%% different modules AND we have a very large set of processes
%% we don't make grids for all the procs we don't want to see

get_size(allmods) -> length(processes()) + 20;
get_size(Mods) ->
    get_size(processes(),Mods, 20).
get_size([], _, Ack) -> Ack;
get_size([P|Tail], Mods, Ack) ->
    Mod = (catch element(1, func(P))),
    case seemod(Mod, [], Mods) of
	true -> get_size(Tail, Mods, Ack+1);
	false -> get_size(Tail, Mods, Ack)
    end.


proc(undefined) -> undefined;
proc(P) when atom(P) -> proc(whereis(P));
proc({global, N}) -> proc(global:whereis_name(N));
proc(P) when pid(P) -> spawn(pxw_pman, internal, [P]).
proc(X,Y,Z) -> proc(c:pid(X,Y,Z)).

init_full(Size, Seemods) ->
    process_flag(trap_exit, true),
    {P, Shell} = mk_pman_window(Size, Seemods),
    put(node, node()),
    loop(P, ordsets:new_set(), Seemods,Size),
    pxw:cmd([destroy, Shell]),
    init_full(round(Size * 1.4), Seemods).   % grow number of grids 

mk_pman_window(Size, Seemods) ->
    Width = 840,
    put(width, Width),
    pxw:cmd([setdefaults,{list,defaults()}]),
    {widget, Shell} = pxw:cmd([shell,create,name,"",'_','_','_','_',toplevel]),
    pxw:cmd([shell,title,Shell,concat(["pxw_pman on ",node()])]),
    put(shell, Shell),
    pxw:cmd([shell,iconname,Shell,"pxw_pman "]),
    {widget, Pane} = pxw:cmd([paned,create,"erl", Shell,0,0,Width,'_']),
    {widget, Box} = pxw:cmd([box,create, "topbuts", Pane,'_','_','_','_']),
    put(topBox, Box),
    brow(Box, ["Quit", "Refresh", "Nohide", "Grow", "TraceShell","Help"]),
    N = lists:map({erlang, atom_to_list}, [], [node()  | nodes()]),
    {widget, M} = 
        pxw:cmd([menu,create,name,Box,'_','_','_','_',"Nodes",N,N]),
    put(menu, M),

    {widget,View}= pxw:cmd([view,create,"tbox",Pane,'_','_','_',300]),
    put(gridView, View),
    Grid = make_grid(Size),
    New = new_pid_grid(processes(), [], Seemods),    
    update(get(grid), 1, Size, New, mk_list(dummy, length(New))),
    pxw:cmd([shell,popup,Shell]),
    {New, Shell}.

mk_list(_, 0) -> [];
mk_list(X, I) -> [X | mk_list(X, I-1)].

make_grid(Nrows) ->
    View = get(gridView),  %% the view is set up for us
    %% create a new form
    Width = get(width),
    {widget, Form}= pxw:cmd([form,create,"tbox1",View,'_','_',Width-5,'_']),
    put(gridForm, Form),
    Grid = pxw_lib:cmd([grid,create,Form,Nrows,6,[90,300,200,50,100,80],20]),
    put(grid, Grid),
    Grid.


brow(Graph,[H|T]) ->
    pxw:cmd([button,create,name,Graph,'_','_','_','_',H,H]),
    brow(Graph,T);
brow(Graph,[]) ->
    true.

    
loop(Data, Mods,  Seemods,Size) ->
    receive
	{pxw,[button,press,_,"Quit"]} ->
	    exit(topquit);
	{pxw, [menu, select, Menu, N]} ->
	    put(node, list_to_atom(N)),
	    pxw:cmd([shell,title,get(shell), concat(["pxw_pman on ",N])]),
	    self() ! {pxw,[button,press,0,"Refresh"]},
	    loop(Data, Mods, Seemods,Size);
	{pxw,[button,press,Button,"Refresh"]} ->
	    New = new_pid_grid(r_processes(), Mods, Seemods),
	    update(get(grid), 1, Size, New, Data),
	    N = lists:map({erlang, atom_to_list}, [], [node() |nodes()]),
	    pxw:cmd([menu, menu, get(menu), N, N]),
	    loop(New, Mods, Seemods,Size);
	{pxw,[button,press,_,"Grow"]} ->

	    ok;
	{pxw,[button,press,_,"TraceShell"]} ->
	    spawn_link(pxw_pman, internal, [shell]),
	    loop(Data, Mods, Seemods, Size);
	{pxw,[button,press,_,"Help"]} ->

	    spawn_link(pxw_pman, internal, [help]),
	    loop(Data, Mods, Seemods, Size);
	{pxw,[button,press,_,"Nohide"]} ->
	    New = new_pid_grid(r_processes(),M2 = ordsets:new_set(), allmods),
	    update(get(grid), 1, Size, New, Data),
	    loop(New, M2 , allmods,Size);
	{grid,['Btn1Down',Grid, Row, 1,_]} ->
	    Pid = row2pid(Row, Data, Size),
	    spawn_link(pxw_pman, internal, [Pid]),
	    loop(Data, Mods, Seemods, Size);
	{grid,['Btn1Down',Grid, Row, 2,_]} ->
	    Mod = row2mod(Row, Data, Size),
	    spawn_link(pxw_pman, internal, [Mod]),
	    loop(Data, Mods, Seemods, Size);
	{grid,['Btn3Down',Grid, Row, 2]} ->
	    case row2mod(Row, Data, Size) of
		{module, Mod} ->
		    New = new_pid_grid(r_processes(), 
				       M2 = ordsets:add_element(Mod,Mods),
				      Seemods),
		    update(get(grid), 1, Size, New, Data),
		    loop(New, M2, Seemods,Size);
		_ ->
		    loop(Data, Mods, Seemods, Size)
	    end;

	Other ->
	    %%io:format("Got ~w~n", [Other]),
	    loop(Data, Mods, Seemods, Size)
    end.

row2pid(Row, Data, Size) when Row < Size ->
    case catch element(1, element(2, lists:nth(Row, Data))) of
	P when pid(P) -> P;
	_ -> {error, nopid}
    end;
row2pid(_,_,_) -> {error, nopid}.

row2mod(Row, Data, Size) when Row < Size ->
    case catch element(2, element(2, lists:nth(Row, Data))) of
	{Mod, Fun, Args} -> {module, Mod};
	_ -> {error, nopid}
    end;
row2mod(_,_,_) -> {error, nopid}.
    
get_name(P) when pid(P) ->
    case pinfo(P, registered_name) of
	{registered_name, Name} -> Name;
	_ -> []
    end;
get_name(_) -> " ".


msg(P) when pid(P) -> 
    case pinfo(P, messages) of
	undefined -> 0;
	{messages, L} -> length(L)
    end;
msg(_) -> 0.

new_pid_grid(Procs, Mods, Seemods) ->
    new_pid_grid(Procs, Mods,Seemods, []).

new_pid_grid([], Mods, Seemods,Ack) ->
   addrows(1, [{'Pid', 'Func', 'Name', 'Msgs', 'Reds', 'Size'} | Ack]);
new_pid_grid([P |Tail], Mods, Seemods,Ack) ->
    Call = func(P),
    Reds = reds(P),
    Name = get_name(P),
    Msg = msg(P),
    Size = psize(P),
    case seemod(catch(element(1, Call)), Mods, Seemods) of
	true ->
	    new_pid_grid(Tail,Mods, Seemods,
			 [{P,Call, Name, Msg, Reds, Size} |Ack]);
	false ->
	    new_pid_grid(Tail,Mods, Seemods,Ack)
    end.


%% Do we want to display Mod or not ?
seemod(Module, Mods, allmods) ->
    case ordsets:is_element(Module, Mods) of
	false ->  %% At least not hide it
	    true;
	_ ->
	    false
    end;
seemod(Module, Mods, Seemods) ->
    case lists:member(Module, Seemods) of
	true ->
	    seemod(Module, Mods, allmods);
	false ->
	    false
    end.

update(Grid, Row, Size, [H|New], [H|Old]) when Row < Size ->    %% Equal
    update(Grid, Row+1, Size, New, Old);
update(Grid, Row, Size, [H|New], [_|Old]) when Row < Size ->   %% Standard case
    {_, {P,F,Name,Msgs,Reds,Psize}} = H,
    pxw_lib:cmd([grid,remove_text,Grid,Row, 1]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 1,centered,f(P)]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 2]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 2,centered,f(F)]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 3]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 3,centered,f(Name)]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 4]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 4,centered,f(Msgs)]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 5]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 5,centered,f(Reds)]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 6]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 6,centered,f(Psize)]),
    update(Grid,Row+1, Size, New, Old);
update(Grid, Row, Size, [H|New], []) when Row < Size ->   %% add case
    {_, {P,F,Name,Msgs,Reds,Psize}} = H,
    pxw_lib:cmd([grid,insert_text,Grid,Row, 1,centered,f(P)]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 2,centered,f(F)]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 3,centered,f(Name)]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 4,centered,f(Msgs)]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 5,centered,f(Reds)]),
    pxw_lib:cmd([grid,insert_text,Grid,Row, 6,centered,f(Psize)]),
    update(Grid,Row+1, Size, New, []);
    
update(Grid, Row, Size, [], [H|T])  when Row < Size ->
    pxw_lib:cmd([grid,remove_text,Grid,Row, 1]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 2]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 3]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 4]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 5]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 6]),
    update(Grid,Row+1, Size,[], T); 
update(Grid, Row, Size, New, _) when Row == Size ->
    pxw_lib:cmd([grid,remove_text,Grid,Row, 1]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 2]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 3]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 4]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 5]),
    pxw_lib:cmd([grid,remove_text,Grid,Row, 6]),
    case New of
	[] -> ok;
	More ->
	    String = lists:flatten(io_lib:format("~w more procs ....", 
					   [length(More)])),
	    pxw_lib:cmd([grid,insert_text,Grid,Row, 2,centered, String])
    end;

update(_,_,_, _, _) -> done.

addrows(_, []) -> [];
addrows(I, [H|T]) -> [{I, H} | addrows(I+1, T)].

func(Pid) ->
    ee (case pinfo(Pid, current_function) of
	{current_function, 0}  ->
	    pinfo(Pid, initial_call);
	{current_function, current_function} ->
            pinfo(Pid, initial_call);
	X -> X
    end).

reds(Pid) -> ee(pinfo(Pid, reductions)).

psize(P) ->  
    case {pinfo(P, heap_size), pinfo(P, stack_size)} of
	{undefined, undefined} -> 0;
	{undefined, {_,Sz}} -> Sz;
	{{_,Sz}, undefined} -> Sz;
	{{_,Sz0}, {_,Sz1}} -> Sz0 + Sz1
    end.


ee({_, {M, F, A}}) when list(A) -> {M,F, length(A)};
ee({_, X}) -> X;
ee(X) -> X.
		
defaults() ->
    ["*erl*topbuts.skipAdjust:         true",
     "*erl*topoutput.scrollVertical:   always"].

f(X) -> lists:flatten(io_lib:format("~p",[X])).


%% The INFORmation window
internal({error, nopid}) -> 
    ok;  % Just terminate 
internal(shell) ->
    put(is_shell, true),
    Sh = find_shell(),
    Flags = window({shell, Sh}),
    process_flag(trap_exit, true),
    link(Sh),
    display(Sh),
    process_flag(priority, max),
    catch monitor({Sh, Flags});

internal(Pid)  ->
    process_flag(trap_exit, true),
    Flags = window(Pid),
    display(Pid),
    relay(Pid),
    process_flag(priority, max),
    catch monitor({Pid, Flags}).

find_shell() -> find_shell(processes(), []).
find_shell([], []) -> exit(normal);
find_shell([], Ack) -> lists:max(Ack);
find_shell([P|Tail], Ack) ->
    case erlang:process_info(P, initial_call) of
	{initial_call, {shell, evaluator, 3}} -> find_shell(Tail, [P|Ack]);
	_ -> find_shell(Tail, Ack)
    end.
    
window(Pid) ->
    {widget, Shell} =pxw:cmd([shell,create,name,"",'_','_','_','_',toplevel]),
    put(shell, Shell),
    pxw:cmd([shell,title,Shell,title(Pid)]),
    pxw:cmd([shell,iconname,Shell,f(Pid)]),
    {widget, Pane} = pxw:cmd([paned,create,"erl",Shell,0,0,'_','_']),
    {widget,Box} = pxw:cmd([box,create,name,Pane,'_','_','_','_']),
    case is_pid(Pid) of
	true ->
	    brow(Box,["Quit","Kill", "Refresh", "Links", "File"]),
	    N=["Send","Receive","Functioncalls","Bifcalls","Proc",
	       "SetOnSpawn","SetOnFirstSpawn", "SetOnLink", "SetOnFirstLink",
	       "All", "Reset"],
	    {widget, M} = 
		pxw:cmd([menu,create,name,Box,'_','_','_','_',"Trace",N,N]),
	    put(menu, M);
	false ->
	    N = [],
	    brow(Box,["Quit"])
    end,
    {widget, Output}=pxw:cmd([text,create,name,Pane,'_','_',600,300,"txt1"]),
    put(output, Output),
    pxw:cmd([shell,popup,Shell]), N.


title(help) -> "Help";
title({module, Mod}) -> concat(["Module ", Mod]);
title({shell, Sh}) -> concat(["Shell tracer ", pid_to_list(Sh)]);
title(Pid) -> concat(["Process ", pid_to_list(Pid), name(Pid)]).
name(Pid) ->
    case pinfo(Pid, registered_name) of
	{registered_name, Name} -> concat([" named ", Name]);
	_ -> ""
    end.

is_pid({shell,_}) -> true;
is_pid(P) when pid(P) -> true;
is_pid(_) -> false.


%% var is a tuple {Item Flags}
%% Item can be help | Pid 

monitor(Var) ->
    receive
	{pxw, Cmd0} ->  %% Priority messages 
	    monitor(pxw_cmd(Cmd0, Var));
	T0 when tuple(T0), element(1,T0) == 'EXIT' ->
	    monitor(exit_cmd(T0, Var))
    after 0 ->
	    ok
    end,

    receive
	{pxw, Cmd} ->
	    monitor(pxw_cmd(Cmd, Var));
	T when tuple(T) ->
	    case element(1, T) of
		'EXIT' ->
		    monitor(exit_cmd(T, Var));
		trace ->
		    monitor(do_trace_msgs(T, Var));
		print ->
		    monitor(do_trace_msgs(T, Var));
		_ ->
		    monitor(Var)
	    end;
	_ ->
	    monitor(Var)
    end.

exit_cmd({'EXIT', Item, _} , {Item, Flags}) ->
    case is_shell() of 
	true ->
	    %% Shell died 
	    flush(),
	    sleep(100),  %% Give time for new shell to become alive
	    Sh = find_shell(),
	    pxw:cmd([shell,title,get(shell),title({shell, Sh})]),
	    trac_new_shell(Sh, Flags),
	    link(Sh),
	    {Sh, Flags};
	false ->
	    format("Process ~w died ~n", [Item]),
	    {Item, Flags}
    end;
exit_cmd({'EXIT', _, topquit} , {Item, Flags}) ->
    exit(topquit);
exit_cmd({'EXIT', _, quit} , {Item, Flags}) ->
    exit(quit).


pxw_cmd(Cmd, {Item, Flags}) ->
    case Cmd of
	[button,press,_,"Quit"] ->
	    exit(quit);
	[button,press,_,"Refresh"] ->
	    display(Item),
	    {Item, Flags};
	[button,press,_,"Kill"] ->
	    catch exit(Item, kill), 
	    exit(stopped);
	[button,press,_,"File"] ->
	    case get_file(Item) of
		error ->
		    format("Can't open trace file \n", []);
		ok ->
		    ignore;
		{File, Fname} ->
		    format("** Appending trace output to file ~s~n", [Fname]),
		    put(file, File)
	    end,
	    {Item, Flags};
	[button,press,Button,"Links"] ->
	    do_links(element(2,pinfo(Item ,links))),
	    {Item, Flags};
	[menu, select, Menu, "Reset"] ->
	    flush(),
	    trac(Item, false, allflags()),
	    N=["Send","Receive","Functioncalls","Bifcalls","Proc",
	       "SetOnSpawn","SetOnFirstSpawn", 
	       "SetOnLink", "SetOnFirstLink","All", "Reset"],
	    pxw:cmd([menu, menu, get(menu),N,N]),
	    {Item, N};
	[menu, select, Menu, "All"] ->
	    trac(Item, true, allflags()),
	    N= ["Reset"],
	    pxw:cmd([menu, menu, get(menu),N,N]),
	    {Item, N};

	[menu, select, Menu, Object] ->
	    trac(Item, true, [which(Object)]),
	    F2 = delete(Object, Flags),
	    pxw:cmd([menu, menu, get(menu),F2,F2]),
	    {Item, F2};
	_ ->
	    {Item, Flags}
    end.


do_trace_msgs(Tup1, Var) ->
    {L, Suspended} = dbg:collect_tracs([Tup1], []),
    lists:foreach({pxw_pman, ftm1}, [], L),
    dbg:un_suspend(lists:delete(nopid, Suspended)),
    Var.


ftm1({trace, 'receive', From, Msg}) ->
    format("(~w) << ~p~n", [From, Msg]);
ftm1({trace, send, Msg, From, To}) ->
    format("(~w) ~w ! ~p~n", [From,To,Msg]);
ftm1({trace, From, call, Func}) ->
    format("(~w) call ~s~n",[From, ffunc(Func)]);
ftm1({trace, From, Op, Data}) ->
    format("(~w) ~w ~p~n", [From, Op, Data]);
ftm1({print, Format, Args}) ->
    format(Format, Args);
ftm1({trace, send_to_non_existing_process, Msg, From, To}) ->
    format("(~w) ~w(non_existing_process) ! ~p~n", [From,To,Msg]).


%% Set trace flags for the new shell to be the same ones we had on 
%% on the old shell

trac_new_shell(Sh, Flags) ->
    All = ["Send","Receive","Functioncalls","Bifcalls",
	   "Proc", "SetOnSpawn","SetOnFirstSpawn", 
	   "SetOnLink", "SetOnFirstLink"],
    trac_new_shell(Sh, Flags, All).
trac_new_shell(Sh, Flags, []) -> done;
trac_new_shell(Sh, Flags, [Flag|Tail]) ->
    case lists:member(Flag, Flags) of
	false ->
	    erlang:trace(Sh, true, [which(Flag)]);
	true ->
	    ok
    end,
    trac_new_shell(Sh, Flags, Tail).

allflags() -> [send,'receive',call,bifs,procs,set_on_spawn,
	       set_on_first_spawn, set_on_link, set_on_first_link].

sleep(I) -> receive after I -> ok end.

which("Send") -> send;
which("Receive") -> 'receive';
which("Functioncalls") -> call;
which("Bifcalls") -> bifs;
which("Proc") -> procs;
which("SetOnSpawn") -> set_on_spawn;
which("SetOnFirstSpawn") -> set_on_first_spawn;
which("SetOnLink") -> set_on_link;
which("SetOnFirstLink") -> set_on_first_link.

ffunc({M,F, Argl}) ->
    io_lib:format("~w:~w(~s)", [M, F, fargs(Argl)]);
ffunc(X) -> io_lib:format("~p", [X]).
fargs([]) -> [];
fargs([A]) -> io_lib:format("~p", [A]);  %% last arg
fargs([A|Args]) -> [io_lib:format("~p,", [A]) | fargs(Args)].


%% Since we are not allowed to do erlang:trace/3 on remote
%% processe we create a help process at the remote node to
%% do the job for us 

relay(P) when pid(P), node(P) /= node() ->
    put(relay, spawn_link(node(P), pxw_pman, do_relay, [self()]));
relay(_) -> ignore.

do_relay(P) ->
    process_flag(trap_exit, true),
    do_relay_1(P).

do_relay_1(P) ->
    receive
	{P, M,F, A} ->
	    case catch apply(M,F,A) of
		true -> ok;
		_ ->  P ! {print, "** Illegal trace request **\n", []}
	    end,
	    do_relay_1(P);
	{'EXIT', P, _} ->
	    exit(normal);
	Other ->             %% Here is the normal case for trace i/o
	    P ! Other, 
	    do_relay_1(P)
    end.

flush() ->
    receive
	_ -> flush()
    after 0 -> ok
    end.

%% Possibly send a request to do tracing to a remote node.
trac(Pid, How, Flag) when node(Pid) == node() ->
    case catch erlang:trace(Pid, How, Flag) of
	true -> true;
	false -> false;
	_ -> format("** Illegal trace request ** \n", []),error
    end;
	
trac(Pid, How, Flag) ->
    get(relay) ! {self(), erlang, trace, [Pid, How, Flag]}.
    

do_links([]) -> ok;
do_links([Pid | Tail]) when pid(Pid) ->
    spawn_link(pxw_pman, internal, [Pid]),
    do_links(Tail);
do_links([_|T]) -> do_links(T).

format(Format) -> format(Format, []).

format(Format, Args) ->
    Text = io_lib:format(Format, Args),
    pxw:cmd([text,output,get(output),Text]),
    case get(file) of
	undefined -> ok;
	File -> io:format(File,"~s", [Text])
    end.


%% Open a file and send a copy of all trace i/o there as well
get_file(Pid) ->
    case {get(file), is_shell()} of
	{undefined, false} ->
	    L = delete($<, delete($>, pid_to_list(Pid))),
	    get_file2(concat(["pxw_pman.trace.", L]));
	{File, false} ->
	    ok;
	{_, true} ->  %% We're a shell
	    L="pxw_pman.trace.shell",
	    get_file2(L)
    end.
get_file2(L) ->
    case file:open(L, read_write) of
	{error, open} -> 
	    error;
	{ok,File} ->
	    file:position(File, {eof, 0}),
	    io:format(File, 
		      "** New trace round beginning at time:~w, date:~w~n",
		      [time(), date()]),
	    {File, L}
    end.
	    
is_shell() ->
    case get(is_shell) of
	true -> true;
	_ -> false
    end.

display(help) ->
    format(
"To get additonal info about a process click on 
the pid field with the left mouse button.
Then an additional window will be popped up.

This window contains a Trace menu.
The meaning of the different items in the menu are as follows
Send --    trace all messages the process sends 
Receive -- trace all messages the process receives 
Functioncalls -- trace function calls for the process if the 
                 process runs trace compiled code. Code 
                 gets trace compiled if compiled as c:c(F, trace).
Bifcalls -- trace all calls to bifs made by the process 
Proc --     trace events related to concurrency made by the 
            process, such as spawn, link, death etc 
SetOnSpawn -- let the trace flags be inherited by all processes 
            created by the original process. 
SetOnFirstSpawn -- As above but only for the first process which is 
              created by the original process.
All --      Do all of the above. If both SetOnFirstSpawn and SetOnSpawn 
            are set, SetOnFirstSpawn overrides. 
Reset --    Remove all trace for the process.


If the {Mod,Fun,Args} field is clicked on with 
the right button processes running the code from 
module Mod will not be displayed

If the {Mod,Fun,Args} field is clicked on with 
the left button Information about the module Mod 
will be displayed.

If the Nohide button is clicked
all processes will be displayed
(If they fit in the window)

To trace the shell (and processes created from the shell) 
the TraceShell button can be clicked in the topwindow. 
If this button is clicked a trace window will be popped up
that selects the latest shell and traces it.

pxw_pman can be started as pxw_pman:start(ListofModules) and information
about processes running code from any of the modules in 
ListofModules will be displayed only.

pxw_pman can also be started as pxw_pman:proc(Item) where Item 
can be a Pid, Atom (Where atom must be the name of a process) or
{global, Name} where name is the the name of a globaly registered
process.


\n\n\n",[]);

display({module, Mod}) ->
    vformat("", catch apply({Mod, module_info},[]));
display(Pid) ->
    display1(pinfo(Pid)).

display1(undefined) ->
    format('Process is dead~n',[]);
display1(Pi) ->
    proc_format(Pi).

pinfo(P) when node(P) == node() -> process_info(P);
pinfo(P) -> check(rpc:call(node(P), erlang, process_info, [P])).

pinfo(P, X) when node(P) == node() -> erlang:process_info(P, X);
pinfo(P, X) -> check(rpc:call(node(P), erlang, process_info, [P, X])).

check({badrpc, _}) -> undefined;
check(X) -> X.
	    
r_processes() ->
    N = get(node),
    if
	N == node() -> processes();
	true -> rpc:block_call(get(node), erlang, processes, [])
    end.

proc_format(Pi) ->  %% process_info struct
    pformat(lists:keysearch(initial_call, 1, Pi)),
    pformat(lists:keysearch(current_function, 1,Pi)),
    pformat(lists:keysearch(messages, 1,Pi)),
    pformat(lists:keysearch(dictionary,1, Pi)),
    pformat(lists:keysearch(heap_size, 1,Pi)),
    pformat(lists:keysearch(stack_size, 1,Pi)),
    pformat(lists:keysearch(reductions, 1,Pi)).

pformat(false) -> false;
pformat({value,{_, 0}}) -> false;
pformat({value,{_, []}}) -> false;
pformat({value, {Key, Vals}}) ->
    Pad = mkpad(io_lib:format("~p ",[Key])),
    format("~p: ", [Key]), 
    vformat(Pad, Vals).

vformat(Pad, {M,F,A}) when atom(F) ->    
    format("~p:~p", [M,F]), 
    Pad2 = lists:append(Pad, 
			mkpad(io_lib:format("~w:~w",[M,F]))),
    argformat(Pad2, A);

vformat(Pad, [H|T]) ->
    format("["),
    kvformat(Pad, [H|T]);
vformat(Pad, X) -> format("~p~n", [X]).

mkpad([_|T]) -> [32|mkpad(T)];
mkpad([]) -> [].

kvformat(S, [Item]) ->
    format("\n~s~p]\n",[S,Item]);
kvformat(S,[H|T]) ->
    format("\n~s~p,",[S,H]), 
    kvformat(S,T);
kvformat(_,[]) -> 
    format("]\n").

argformat(Pad,A) when integer(A) ->
    format("/~p\n", [A]);
argformat(_,A) ->
    format("/~p\n", [length(A)]),
    format("args: \n"),
    argformat2("      ", A).

argformat2(Pad, Arglist) ->
    Chars = io_lib:format("~p",[Arglist]),
    if
	length(Chars) < (70 - length(Pad)) ->
	    format("~s~s\n", [Pad, Chars]);
	true ->
	    argformat3(Pad, Arglist)
    end.

argformat3(_,[]) -> format("\n");
argformat3(Pad, [H|T]) ->
    Chars = truncate(65,io_lib:format("~s~p",[Pad, H])),
    format("~s,\n", [Chars]),
    argformat3(Pad, T).

truncate(0, Chars) -> ".....";
truncate(I, [H|T]) -> [H|truncate(I-1, T)];
truncate(I, []) -> [].

    





